Package: spada
Title: A 'shiny' Package for Data Analysis
Version: 0.1.8
Authors@R: 
    person("Luís Gustavo", "Schuck", email="luisgustavoschuck@yahoo.com.br",  
    role = c("aut", "cre"))
Maintainer: Luís Gustavo Schuck <luisgustavoschuck@yahoo.com.br>
Description: Provides a 'shiny' application with a user-friendly interface 
  for interactive data analysis. It supports exploratory data analysis 
  through descriptive statistics, data visualization, statistical 
  tests (e.g., normality assessment), linear modeling, data 
  import, transformation and reporting. For more details see Shapiro and Wilk 
  (1965) <doi:10.2307/2333709>.
License: AGPL (>= 3)
URL: https://github.com/lgschuck/spada,
        https://lgschuck.github.io/spada/
BugReports: https://github.com/lgschuck/spada/issues
Depends: R (>= 4.4.0)
Imports: bsicons (>= 0.1.2), bslib (>= 0.10.0), collapse (>= 2.1.6),
        data.table (>= 1.18.2.1), DescTools (>= 0.99.60), ggplot2 (>=
        4.0.1), gt (>= 1.3.0), haven (>= 2.5.5), htmltools (>= 0.5.9),
        mirai (>= 2.5.3), qs2 (>= 0.1.7), rlang (>= 1.1.7), shiny (>=
        1.12.1), shinyWidgets (>= 0.9.0), waiter (>= 0.2.5.1), writexl
        (>= 1.5.4)
Encoding: UTF-8
Suggests: dplyr, knitr, rmarkdown, testthat (>= 3.0.0), tidyselect
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-08-22 21:41:29 UTC; luisg
Author: Luís Gustavo Schuck [aut, cre]
Repository: CRAN
Date/Publication: 2026-08-23 11:20:02 UTC
Built: R 4.5.2; ; 2026-08-23 12:37:26 UTC; unix

Package: spada
Title: A 'shiny' Package for Data Analysis
Version: 0.1.6
Authors@R: 
    person("Luís Gustavo", "Schuck", email="luisgustavoschuck@yahoo.com.br",  
    role = c("aut", "cre"))
Maintainer: Luís Gustavo Schuck <luisgustavoschuck@yahoo.com.br>
Description: Provides a 'shiny' application with a user-friendly interface 
  for interactive data analysis. It supports exploratory data analysis 
  through descriptive statistics, data visualization, statistical 
  tests (e.g., normality assessment), linear modeling, data 
  import, transformation and reporting. For more details see Shapiro and Wilk 
  (1965) <doi:10.2307/2333709>.
License: AGPL (>= 3)
URL: https://github.com/lgschuck/spada,
        https://lgschuck.github.io/spada/
BugReports: https://github.com/lgschuck/spada/issues
Depends: R (>= 4.4.0)
Imports: bsicons (>= 0.1.2), bslib (>= 0.10.0), collapse (>= 2.1.6),
        data.table (>= 1.18.2.1), DescTools (>= 0.99.60), ggplot2 (>=
        4.0.1), gt (>= 1.3.0), haven (>= 2.5.5), htmltools (>= 0.5.9),
        mirai (>= 2.5.3), qs2 (>= 0.1.7), rlang (>= 1.1.7), shiny (>=
        1.12.1), shinyWidgets (>= 0.9.0), waiter (>= 0.2.5.1), writexl
        (>= 1.5.4)
Encoding: UTF-8
RoxygenNote: 7.3.3
Suggests: dplyr, knitr, rmarkdown, testthat (>= 3.0.0), tidyselect
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-06-11 10:38:57 UTC; luisg
Author: Luís Gustavo Schuck [aut, cre]
Repository: CRAN
Date/Publication: 2026-06-11 11:20:02 UTC
Built: R 4.5.2; ; 2026-06-11 15:35:21 UTC; unix

# Generated by roxygen2: do not edit by hand

export(spada)
import(shiny)
importFrom(DescTools,Freq)
importFrom(DescTools,Gmean)
importFrom(DescTools,Hmean)
importFrom(DescTools,Kurt)
importFrom(DescTools,Mode)
importFrom(DescTools,Outlier)
importFrom(DescTools,ShapiroFranciaTest)
importFrom(DescTools,Skew)
importFrom(DescTools,StrCap)
importFrom(DescTools,VIF)
importFrom(DescTools,ZTest)
importFrom(bsicons,bs_icon)
importFrom(bslib,accordion)
importFrom(bslib,accordion_panel)
importFrom(bslib,bind_task_button)
importFrom(bslib,bs_add_rules)
importFrom(bslib,bs_theme)
importFrom(bslib,card)
importFrom(bslib,card_body)
importFrom(bslib,card_footer)
importFrom(bslib,card_header)
importFrom(bslib,font_collection)
importFrom(bslib,input_switch)
importFrom(bslib,input_task_button)
importFrom(bslib,layout_column_wrap)
importFrom(bslib,layout_columns)
importFrom(bslib,layout_sidebar)
importFrom(bslib,nav_item)
importFrom(bslib,nav_menu)
importFrom(bslib,nav_panel)
importFrom(bslib,nav_select)
importFrom(bslib,nav_spacer)
importFrom(bslib,navbar_options)
importFrom(bslib,navset_card_pill)
importFrom(bslib,page_navbar)
importFrom(bslib,popover)
importFrom(bslib,sidebar)
importFrom(bslib,tooltip)
importFrom(bslib,update_switch)
importFrom(bslib,value_box)
importFrom(collapse,"%==%")
importFrom(collapse,allNA)
importFrom(collapse,ffirst)
importFrom(collapse,flast)
importFrom(collapse,fmax)
importFrom(collapse,fmean)
importFrom(collapse,fmedian)
importFrom(collapse,fmin)
importFrom(collapse,fmode)
importFrom(collapse,fncol)
importFrom(collapse,fnobs)
importFrom(collapse,fnrow)
importFrom(collapse,fnunique)
importFrom(collapse,fquantile)
importFrom(collapse,frange)
importFrom(collapse,fsd)
importFrom(collapse,fsum)
importFrom(collapse,fvar)
importFrom(collapse,qtab)
importFrom(collapse,whichNA)
importFrom(data.table,"%between%")
importFrom(data.table,"%notin%")
importFrom(data.table,":=")
importFrom(data.table,.SD)
importFrom(data.table,as.data.table)
importFrom(data.table,between)
importFrom(data.table,copy)
importFrom(data.table,data.table)
importFrom(data.table,fcase)
importFrom(data.table,fcoalesce)
importFrom(data.table,fifelse)
importFrom(data.table,fread)
importFrom(data.table,fwrite)
importFrom(data.table,hour)
importFrom(data.table,is.data.table)
importFrom(data.table,mday)
importFrom(data.table,melt)
importFrom(data.table,minute)
importFrom(data.table,month)
importFrom(data.table,quarter)
importFrom(data.table,rbindlist)
importFrom(data.table,second)
importFrom(data.table,setDT)
importFrom(data.table,setcolorder)
importFrom(data.table,setnames)
importFrom(data.table,setorderv)
importFrom(data.table,shift)
importFrom(data.table,wday)
importFrom(data.table,week)
importFrom(data.table,yday)
importFrom(data.table,year)
importFrom(ggplot2,.data)
importFrom(ggplot2,aes)
importFrom(ggplot2,after_stat)
importFrom(ggplot2,coord_flip)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,element_rect)
importFrom(ggplot2,element_text)
importFrom(ggplot2,geom_bar)
importFrom(ggplot2,geom_boxplot)
importFrom(ggplot2,geom_histogram)
importFrom(ggplot2,geom_hline)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_text)
importFrom(ggplot2,geom_tile)
importFrom(ggplot2,geom_vline)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,labs)
importFrom(ggplot2,scale_fill_gradient2)
importFrom(ggplot2,stat_boxplot)
importFrom(ggplot2,stat_function)
importFrom(ggplot2,stat_qq)
importFrom(ggplot2,stat_qq_line)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_classic)
importFrom(ggplot2,ylim)
importFrom(grDevices,colors)
importFrom(graphics,abline)
importFrom(graphics,barplot)
importFrom(graphics,boxplot)
importFrom(graphics,curve)
importFrom(graphics,hist)
importFrom(graphics,lines)
importFrom(graphics,mtext)
importFrom(graphics,polygon)
importFrom(graphics,text)
importFrom(gt,as_raw_html)
importFrom(gt,cols_align)
importFrom(gt,cols_hide)
importFrom(gt,cols_label)
importFrom(gt,cols_merge)
importFrom(gt,cols_move)
importFrom(gt,cols_width)
importFrom(gt,data_color)
importFrom(gt,fmt_bytes)
importFrom(gt,fmt_icon)
importFrom(gt,fmt_integer)
importFrom(gt,fmt_number)
importFrom(gt,fmt_percent)
importFrom(gt,gt)
importFrom(gt,gt_output)
importFrom(gt,gtsave)
importFrom(gt,opt_interactive)
importFrom(gt,pct)
importFrom(gt,render_gt)
importFrom(gt,sub_missing)
importFrom(gt,sub_values)
importFrom(gt,tab_header)
importFrom(gt,tab_options)
importFrom(gt,tab_spanner)
importFrom(haven,as_factor)
importFrom(haven,is.labelled)
importFrom(haven,read_sav)
importFrom(haven,write_sav)
importFrom(htmltools,plotTag)
importFrom(htmltools,save_html)
importFrom(mirai,daemons)
importFrom(mirai,mirai)
importFrom(qs2,qs_read)
importFrom(qs2,qs_save)
importFrom(rlang,parse_expr)
importFrom(shinyWidgets,colorPickr)
importFrom(shinyWidgets,dropdownButton)
importFrom(shinyWidgets,radioGroupButtons)
importFrom(shinyWidgets,show_toast)
importFrom(shinyWidgets,statiCard)
importFrom(shinyWidgets,updateColorPickr)
importFrom(shinyWidgets,updateRadioGroupButtons)
importFrom(stats,IQR)
importFrom(stats,cor)
importFrom(stats,cor.test)
importFrom(stats,dnorm)
importFrom(stats,dt)
importFrom(stats,formula)
importFrom(stats,ks.test)
importFrom(stats,lm)
importFrom(stats,median)
importFrom(stats,pf)
importFrom(stats,qnorm)
importFrom(stats,qqline)
importFrom(stats,qqnorm)
importFrom(stats,qt)
importFrom(stats,rnorm)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,shapiro.test)
importFrom(stats,t.test)
importFrom(stats,terms)
importFrom(stats,var)
importFrom(tools,R_user_dir)
importFrom(tools,file_ext)
importFrom(tools,toTitleCase)
importFrom(utils,head)
importFrom(utils,object.size)
importFrom(utils,packageDescription)
importFrom(utils,packageVersion)
importFrom(utils,sessionInfo)
importFrom(waiter,useWaiter)
importFrom(waiter,waiterShowOnLoad)
importFrom(waiter,waiter_hide)
importFrom(waiter,waiter_show)
importFrom(writexl,write_xlsx)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# spada (Shiny Package for Data Analysis)

<!-- badges: start -->

[![](https://www.r-pkg.org/badges/version-last-release/spada)](https://cran.r-project.org/package=spada)
[![](https://img.shields.io/badge/R-%3E%3D%204.4-blue)](https://CRAN.R-project.org/package=spada)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-coral.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![](https://cranlogs.r-pkg.org/badges/grand-total/spada)](https://CRAN.R-project.org/package=spada)
[![](https://cranlogs.r-pkg.org/badges/spada)](https://CRAN.R-project.org/package=spada)
[![](https://cranlogs.r-pkg.org/badges/last-week/spada)](https://CRAN.R-project.org/package=spada)
[![](https://cranlogs.r-pkg.org/badges/last-day/spada)](https://CRAN.R-project.org/package=spada)

<!-- badges: end -->

**Spada** provides a ‘shiny’ application with a user-friendly interface
for interactive data analysis. It supports exploratory data analysis
through descriptive statistics, data visualization, statistical tests
(e.g., normality assessment), linear modeling, data import,
transformation and reporting.

This package is inspired in many other tools like:

- IBM SPSS Statistics (<https://www.ibm.com/products/spss-statistics>)

- R Commander package (<https://CRAN.R-project.org/package=Rcmdr>)

- Jamovi (<https://www.jamovi.org/>)

- ydata profiling (<https://docs.profiling.ydata.ai/latest/>)

<div style="border: 2px solid #02517d; padding: 10px; background-color: #a9ccdb;">

<strong>Warning:</strong> Spada is in active development.

</div>

## [Live Demo](https://lgschuck.shinyapps.io/spada)

You may try Spada in shinyapps.io
[Spada](https://lgschuck.shinyapps.io/spada)

## Documentation

You can access Spada Book on this website: [Spada
Book](https://lgschuck.github.io/spada_book/).

## Installation

``` r
install.packages("spada")
```

You can install the development version of spada from
[GitHub](https://github.com/) using the command below.

``` r
install.packages("remotes")
remotes::install_github(
  "lgschuck/spada",
  dependencies = TRUE
)
```

### Loading the package

``` r
library(spada)
```

### Usage

``` r
if(interactive()){
  spada()
}
```

## Tests

Spada is tested via testthat and currently runs over **900 tests**.

## Docker

Spada is distributed as a **Docker Image**.

For instructions in how to download and use Spada go to
[Docker](https://lgschuck.github.io/spada_book/docker.html)

You can use [Docker Hub](https://hub.docker.com/r/lgschuck/spada) to
search for diferent versions.

## Screenshots

<figure>
<img src="man/figures/spada_loading.png" alt="Loading" />
<figcaption aria-hidden="true">Loading</figcaption>
</figure>

### Data

<figure>
<img src="man/figures/spada_home.png" alt="Home" />
<figcaption aria-hidden="true">Home</figcaption>
</figure>

<figure>
<img src="man/figures/spada_home2.png" alt="Home 2" />
<figcaption aria-hidden="true">Home 2</figcaption>
</figure>

#### Data \> Metadata

<figure>
<img src="man/figures/spada_data_metadata.png" alt="Metadata" />
<figcaption aria-hidden="true">Metadata</figcaption>
</figure>

#### Data \> Overview

<figure>
<img src="man/figures/spada_data_overview.png" alt="Overview" />
<figcaption aria-hidden="true">Overview</figcaption>
</figure>

### Edit

<figure>
<img src="man/figures/spada_edit.png" alt="Edit" />
<figcaption aria-hidden="true">Edit</figcaption>
</figure>

### Analysis

<figure>
<img src="man/figures/spada_analysis_exploratory.png" alt="Analysis" />
<figcaption aria-hidden="true">Analysis</figcaption>
</figure>

### Output

<figure>
<img src="man/figures/spada_output.png" alt="Output" />
<figcaption aria-hidden="true">Output</figcaption>
</figure>

#### Exported Output

<figure>
<img src="man/figures/spada_output_exported.png"
alt="Exported Output" />
<figcaption aria-hidden="true">Exported Output</figcaption>
</figure>

### Options

<figure>
<img src="man/figures/spada_config.png" alt="Options" />
<figcaption aria-hidden="true">Options</figcaption>
</figure>

<!-- README.md is generated from README.Rmd. Please edit that file -->

# spada (<u>**S**</u>hiny <u>**Pa**</u>ckage for <u>**D**</u>ata <u>**A**</u>nalysis)

<!-- badges: start -->

[![](https://www.r-pkg.org/badges/version-last-release/spada)](https://cran.r-project.org/package=spada)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-coral.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![](https://cranlogs.r-pkg.org/badges/grand-total/spada)](https://CRAN.R-project.org/package=spada)
[![](https://cranlogs.r-pkg.org/badges/spada)](https://CRAN.R-project.org/package=spada)
[![](https://cranlogs.r-pkg.org/badges/last-week/spada)](https://CRAN.R-project.org/package=spada)
[![](https://cranlogs.r-pkg.org/badges/last-day/spada)](https://CRAN.R-project.org/package=spada)

<!-- badges: end -->

**Spada** provides a ‘shiny’ application with a user-friendly interface
for interactive data analysis. It supports exploratory data analysis
through descriptive statistics, data visualization, statistical tests
(e.g., normality assessment), linear modeling, data import,
transformation and reporting.

This package is inspired in many other tools like:

- IBM SPSS Statistics (<https://www.ibm.com/products/spss-statistics>)

- R Commander package (<https://CRAN.R-project.org/package=Rcmdr>)

- Jamovi (<https://www.jamovi.org/>)

- ydata profiling (<https://docs.profiling.ydata.ai/latest/>)

<div style="border: 2px solid #02517d; padding: 10px; background-color: #a9ccdb;">

<strong>Warning:</strong> Spada is in active development.

</div>

## [Live Demo](https://lgschuck.shinyapps.io/spada)

You may try Spada in shinyapps.io
[Spada](https://lgschuck.shinyapps.io/spada)

## Documentation

You can access Spada Book on this website: [Spada
Book](https://lgschuck.github.io/spada_book/).

## Installation

``` r
install.packages("spada")
```

You can install the development version of spada from
[GitHub](https://github.com/) using the command below.

``` r
install.packages("remotes")
remotes::install_github(
  "lgschuck/spada",
  dependencies = TRUE
)
```

### Loading the package

``` r
library(spada)
```

### Usage

``` r
if(interactive()){
  spada()
}
```

## Docker

Spada is distributed as a **Docker Image**.

For instructions in how to download and use Spada go to
[Docker](https://lgschuck.github.io/spada_book/docker.html)

You can use [Docker Hub](https://hub.docker.com/r/lgschuck/spada) to
search for diferent versions.

## Screenshots

<figure>
<img src="man/figures/spada_loading_0.1.3.png" alt="Loading" />
<figcaption aria-hidden="true">Loading</figcaption>
</figure>

### Data

<figure>
<img src="man/figures/spada_home.png" alt="Home" />
<figcaption aria-hidden="true">Home</figcaption>
</figure>

<figure>
<img src="man/figures/spada_home2.png" alt="Home 2" />
<figcaption aria-hidden="true">Home 2</figcaption>
</figure>

#### Data \> Metadata

<figure>
<img src="man/figures/spada_data_metadata.png" alt="Metadata" />
<figcaption aria-hidden="true">Metadata</figcaption>
</figure>

#### Data \> Overview

<figure>
<img src="man/figures/spada_data_overview.png" alt="Overview" />
<figcaption aria-hidden="true">Overview</figcaption>
</figure>

### Edit

<figure>
<img src="man/figures/spada_edit.png" alt="Edit" />
<figcaption aria-hidden="true">Edit</figcaption>
</figure>

### Analysis

<figure>
<img src="man/figures/spada_analysis_exploratory.png" alt="Analysis" />
<figcaption aria-hidden="true">Analysis</figcaption>
</figure>

### Output

<figure>
<img src="man/figures/spada_output.png" alt="Output" />
<figcaption aria-hidden="true">Output</figcaption>
</figure>

### Options

<figure>
<img src="man/figures/spada_config.png" alt="Options" />
<figcaption aria-hidden="true">Options</figcaption>
</figure>

# ui --------------------------------------------------------------------------
navbar_df_info_ui <- function(id) {
  ns <- NS(id)
  tagList(
    uiOutput(ns('navbar_df_info')),
    fluidRow(
      column(3, mini_btn(ns('df_btn_overview'), '', icon('magnifying-glass'))),
      column(3, mini_btn(ns('df_btn_meta'), '', bs_icon('info-circle'))),
      column(3, mini_btn(ns('df_btn_change'), '', icon('shuffle'))),
      column(3, mini_btn(ns('df_btn_edit'), '', icon('screwdriver-wrench')))
    ),
      tags$br(),
    fluidRow(
      column(3, mini_btn(ns('df_btn_import'), '', icon('upload'))),
      column(3, mini_btn(ns('df_btn_export'), '', icon('download'))),
      column(3, mini_btn(ns('df_btn_output'), '', icon('tablet-screen-button'))),
      column(3, mini_btn(ns('df_btn_explore'), '', icon('chart-simple')))
    )
  )
}

# server ----------------------------------------------------------------------
navbar_df_info_server <- function(id, app_session) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

    output$navbar_df_info <- renderUI({
      req(session$userData$dt$act_mini_meta())

      mini_meta <- session$userData$dt$act_mini_meta()
      tagList(
        div(class = 'spada-navbar-df-info-header',
            h2(session$userData$dt$act_name)),
        p('Rows/Columns:', mini_meta[['row_col']]),
        p("Columns with NA's:", mini_meta[['col_nas']])
      )
    })

    # buttons --------
    observe({
      nav_select('navbar', selected = 'Data', session = app_session)
      nav_select('navset_card_pill_data', selected = 'Overview', session = app_session)
    }) |> bindEvent(input$df_btn_overview)

    observe({
      nav_select('navbar', selected = 'Data', session = app_session)
      nav_select('navset_card_pill_data', selected = 'Metadata', session = app_session)
    }) |> bindEvent(input$df_btn_meta)

    observe({
      nav_select('navbar', selected = 'Data', session = app_session)
      nav_select('navset_card_pill_data', selected = 'Data', session = app_session)
    }) |> bindEvent(input$df_btn_change)

    observe({
      nav_select('navbar', selected = 'Exploratory', session = app_session)
    }) |> bindEvent(input$df_btn_explore)

    observe({
      nav_select('navbar', selected = 'Data', session = app_session)
      nav_select('navset_card_pill_data', selected = 'Import', session = app_session)
    }) |> bindEvent(input$df_btn_import)

    observe({
      nav_select('navbar', selected = 'Data', session = app_session)
      nav_select('navset_card_pill_data', selected = 'Export', session = app_session)
    }) |> bindEvent(input$df_btn_export)

    observe({
      nav_select('navbar', selected = 'Output', session = app_session)
    }) |> bindEvent(input$df_btn_output)

    observe({
      nav_select('navbar', selected = 'Edit', session = app_session)
    }) |> bindEvent(input$df_btn_edit)

  })
}

# ui --------------------------------------------------------------------------
config_ui <- function(id) {
  ns <- NS(id)
  nav_panel(
    value = 'config',
    title = 'Config',
    icon = bs_icon('sliders2'),
    card(
      class = 'big-card',
      layout_columns(
        col_widths = c(8, 4),
        card(
          card_body(
            h4('Colors'),
            layout_columns(
              col_widths = c(3, 3, 3),
              uiOutput(ns('ui_fill_color')),
              uiOutput(ns('ui_line_color')),
              uiOutput(ns('ui_title_color')),
              btn_task(ns('btn_reset_colors'), 'Reset', icon('rotate'),
                       style = 'margin-top: 27px !important;')
            ),
            plotOutput(ns('sample_plot'))
          )
        ),
        card(
          card_body(
            fluidRow(accordion(
              accordion_panel(
                'Theme',
                layout_columns(
                  col_widths = c(7, 5),
                  selectInput(
                    ns('theme_choice'),
                    NULL,
                    choices = c('Spada' = 'spada_theme', 'Dark Spada' = 'spada_dark_theme')
                  ),
                  actionButton(ns('btn_theme'), 'Apply', icon('check'), class = 'btn-task')
                ),
                icon = bs_icon('palette')
              ),
              open = F
            )),
            fluidRow(accordion(
              accordion_panel(
                'Size of input files (MB)',
                layout_columns(
                  col_widths = c(7, 5),
                  numericInput(
                    ns('input_file_size'),
                    NULL,
                    value = 1000,
                    min = 0,
                    step = 500
                  ),
                  actionButton(ns('btn_file_size'), 'Apply', icon('check'), class = 'btn-task')
                ),
                icon = bs_icon('upload')
              ),
              open = F
            )),
            fluidRow(accordion(
              accordion_panel(
                'Config Directories',
                h5('Config'),
                textOutput(ns('conf_dir')),
                h5('Data'),
                textOutput(ns('conf_data_dir')),
                icon = bs_icon('folder-check')
              ),
              open = F
            )),
            fluidRow(accordion(
              accordion_panel(
                'Session',
                radioGroupButtons(
                  ns('radio_restore_session'), 'Restore Session data at startup',
                  c('Always' = 'always', 'Never' = 'never'),
                  size = 'sm', individual = T),
                radioGroupButtons(
                  ns('radio_save_session'), 'Save Session data on exit',
                  c('Always' = 'always', 'Ask' = 'ask', 'Never' = 'never'),
                  size = 'sm', individual = T),
                actionButton(ns('btn_save_session_conf'), 'Apply', icon('check'), class = 'btn-task'),
                icon = bs_icon('sliders')
              ),
              open = F
            )),
            fluidRow(accordion(
              accordion_panel(
                'Plot Limit (thousands of rows)',
                layout_columns(
                  col_widths = c(7, 5),
                  numericInput(
                    ns('input_plot_limit'),
                    NULL,
                    value = 100,
                    min = 0,
                    step = 50
                  ),
                  actionButton(ns('btn_plot_limit'), 'Apply', icon('check'), class = 'btn-task')
                ),
                icon = bs_icon('graph-up')
              ),
              open = F
            )),
            fluidRow(accordion(
              accordion_panel(
                'Current Conf',
                verbatimTextOutput(ns('current_conf')),
                icon = bs_icon('check-square')
              ),
              open = F
            )),
          )
        )
      )
    )
  )
}

# server ----------------------------------------------------------------------
config_server <- function(id, app_session) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  # render colorpickr -----------------------
	  output$ui_fill_color <- renderUI({
	    req(session$userData$conf$plot_fill_color)
	    colorPickr(
  	    inputId = ns('sel_fill'),
  	    label = 'Fill color',
  	    selected = session$userData$conf$plot_fill_color,
  	    update = 'save'
	    )
	  })

	  output$ui_line_color <- renderUI({
	    req(session$userData$conf$plot_line_color)
	    colorPickr(
	      inputId = ns('sel_line'),
	      label = 'Line color',
	      selected = session$userData$conf$plot_line_color,
	      update = 'save'
	    )
	  })

	  output$ui_title_color <- renderUI({
	    req(session$userData$conf$plot_title_color)
	    colorPickr(
	      inputId = ns('sel_title'),
	      label = 'Title color',
	      selected = session$userData$conf$plot_title_color,
	      update = 'save'
	    )
	  })

	  # palette ---------------------------------
	  palette <- reactive({
	    list('fill' = input$sel_fill, 'line' = input$sel_line,
	         'title' = input$sel_title)
	  })

	  observe({
	    req(input$sel_fill, input$sel_line)
	    session$userData$conf$plot_fill_color <- palette()[['fill']]
	    session$userData$conf$plot_line_color <- palette()[['line']]
	    session$userData$conf$plot_title_color <- palette()[['title']]
	  })

	  # reset colors ---------------------------
	  observe({
	    updateColorPickr(session = session,
	                     inputId = 'sel_fill',
	                     value = plot_fill_color)
	    updateColorPickr(session = session,
	                     inputId = 'sel_line',
	                     value = plot_line_color)
	    updateColorPickr(session = session,
	                     inputId = 'sel_title',
	                     value = plot_title_color)
	  }) |> bindEvent(input$btn_reset_colors)

	  # sample plot to show picked colors ------
	  plot_values <- rnorm(1e3)
	  output$sample_plot <- renderPlot({
	    req(palette())
	    hist(
	      plot_values,
	      col = palette()[['fill']],
	      xlab = '',
	      ylab = '',
	      main = 'Title',
	      col.main = palette()[['title']],
	      cex.main = 3
	    )
	    abline(h = 100, col = palette()[['line']], lwd = 3)

	  }) |> bindCache(palette())

	  # directories -----------------------------
	  output$conf_dir <- renderText({
	    req(session$userData$conf$conf_dir)
	    session$userData$conf$conf_dir
    })

	  output$conf_data_dir <- renderText({
	    req(session$userData$conf$data_dir)
	    session$userData$conf$data_dir
	   })

	  # set theme conf value --------------------
	  observe({
	    req(session$userData$conf$theme)
	    updateSelectInput(
	      session = session,
	      inputId = 'theme_choice',
	      selected = session$userData$conf$theme
	    )
	  })

	  # set file size conf value ----------------
	  observe({
	    req(session$userData$conf$file_size)
	    updateSelectInput(
	      session = session,
	      inputId = 'input_file_size',
	      selected = session$userData$conf$file_size
	    )
	  })

	  # set restore sesison conf value ----------
	  observe({
	    req(session$userData$conf$restore_session)
	    updateRadioGroupButtons(
	      session = session,
	      inputId = 'radio_restore_session',
	      selected = session$userData$conf$restore_session
	    )
	  })

	  # set save sesison conf value -------------
	  observe({
	    req(session$userData$conf$save_session)
	    updateRadioGroupButtons(
	      session = session,
	      inputId = 'radio_save_session',
	      selected = session$userData$conf$save_session
	    )
	  })

    # input file size ------------------------
    max_request_size <- reactive(input$input_file_size * 1024 ^ 2)

    observe({
      if(!isTruthy(input$input_file_size) || input$input_file_size < 1) {
        msg('Value must be > 1')
      } else {
        options(shiny.maxRequestSize = max_request_size())
        session$userData$conf$file_size <- input$input_file_size
        msg('New limit applied')
      }
    }) |> bindEvent(input$btn_file_size)

    # change theme ---------------------------
    observe({

      running_modal('Applying theme...')

      session$onFlushed(function(){
        remove_running_modal()
      }, once = TRUE)

      set_spada_theme(
        session = session,
        theme = input$theme_choice
      )

      if(input$theme_choice == 'spada_dark_theme') {
        update_switch(
          id = 'sidebar-dark_mode', value = TRUE, session = app_session
        )
      } else {
        update_switch(
          id = 'sidebar-dark_mode', value = FALSE, session = app_session
        )
      }
    }) |> bindEvent(input$btn_theme)

    # session conf ---------------------------
    observe({
      session$userData$conf$restore_session <- input$radio_restore_session
      session$userData$conf$save_session <- input$radio_save_session

      msg('New settings applied')
    }) |> bindEvent(input$btn_save_session_conf)

    # input plot limit ------------------------
    observe({
      if(!isTruthy(input$input_plot_limit) || input$input_plot_limit <= 0) {
        msg('Value must be > 0')
      } else {
        session$userData$conf$plot_limit <- input$input_plot_limit * 1e3
        msg('New limit applied')
      }
    }) |> bindEvent(input$btn_plot_limit)

    # current conf ------------------------
    output$current_conf <- renderPrint({ session$userData$conf |> reactiveValuesToList() })

  })
}

# ui --------------------------------------------------------------------------
normality_test_ui <- function(id) {
  ns <- NS(id)
  card(
    full_screen = T,
    card_header('Normality Test', class = 'mini-header'),
    layout_sidebar(
      class = 'card-sidebar',
      sidebar = sidebar(uiOutput(ns('parameters'))),
      navset_card_pill(
        nav_panel(
          'Histogram',
          card(
            full_screen = T,
            card_body(plotOutput(ns('hist'))),
            card_footer(
              div(style = "margin-bottom: -8px !important;"),
              layout_columns(
                col_widths = c(1, 2, 2),
                numericInput(ns('bins'), 'Bins', 25, 5, step = 5),
                btn_task(ns('btn_hist'), 'Show Histogram', icon('chart-simple'),
                         style = 'margin-top: 28px'),
                div(insert_output_ui(ns('norm_insert_output_hist')),
                    style = 'margin-top: 28px')
              ),
              div(style = "margin-bottom: -24px !important;"),
            )
          )
        ),
        nav_panel(
          'QQ Plot',
          card(
            full_screen = T,
            card_body(plotOutput(ns('qq_plot'))),
            card_footer(
              btn_task(ns('btn_qq'), 'Show QQ plot', icon('chart-simple')),
              insert_output_ui(ns('norm_insert_output_qq'))
            )
          )
        ),
        nav_panel(
          'KS Test',
          card(
            full_screen = T,
            card_body(
              layout_columns(
                col_widths = c(3, 7, 2),
                uiOutput(ns('conditional_staticard_ks')),
                gt_output(ns('ks_test')),
                div(
                  uiOutput(ns('conditional_save_ks_gt')),
                  br(), br(),
                  uiOutput(ns('conditional_add_output_ks'))
                )
              ),
              uiOutput(ns('ks_test_obs_ui'))
            ),
            card_footer(
              btn_task(ns('btn_ks'), 'Run Test', icon('gear')),
              btn_task(ns('btn_help_ks'), 'Help', icon('question'))
            )
          )
        ),
        nav_panel(
          'Shapiro-Wilk Test',
          card(
            full_screen = T,
            card_body(
              layout_columns(
                col_widths = c(3, 7, 2),
                uiOutput(ns('conditional_staticard_sw')),
                gt_output(ns('sw_test')),
                div(
                  uiOutput(ns('conditional_save_sw_gt')),
                  br(), br(),
                  uiOutput(ns('conditional_add_output_sw'))
                )
              )
            ),
            card_footer(
              btn_task(ns('btn_sw'), 'Run Test', icon('gear')),
              btn_task(ns('btn_help_sw'), 'Help', icon('question'))
            )
          )
        ),
        nav_panel(
          'Shapiro-Francia Test',
          card(
            full_screen = T,
            card_body(
              layout_columns(
                col_widths = c(3, 7, 2),
                uiOutput(ns('conditional_staticard_sf')),
                gt_output(ns('sf_test')),
                div(
                  uiOutput(ns('conditional_save_sf_gt')),
                  br(), br(),
                  uiOutput(ns('conditional_add_output_sf'))
                )
              )
            ),
            card_footer(
              btn_task(ns('btn_sf'), 'Run Test', icon('gear')),
              btn_task(ns('btn_help_sf'), 'Help', icon('question'))
            )
          )
        )
      )
    )
  )
}

# server ----------------------------------------------------------------------
normality_test_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  df <- reactive(get_act_dt(session))

    df_active <- reactive({
      req(df())
      df()[, lapply(df(), is.numeric) == T, with = F]
    })

    var_analysis <- reactive({
      req(df_active())

      df_names <- df_active() |> names()

      var_analysis <- session$userData$dt$act_meta()[perc_nas != 1, var]

      df_names[df_names %in% var_analysis]
    })

    var <- reactive({
      req(df_active())
      req(input$sel_var)

      temp <- df_active()[[input$sel_var]]
      temp[!is.na(temp)]
    })

    var_len <- reactive(var() |> length())

    output$parameters <- renderUI({
      req(var_analysis())
      tagList(
        selectInput(ns('sel_var'), 'Variable 1', var_analysis()),
        p('* Showing only numeric variables')
      )
    })

    # histogram ---------------------------------------------------------------
    output$hist <- renderPlot({
      validate(need(input$bins > 0, 'Bins must be 1 or higher'))

      norm_hist()
    }, res = 96)

    task_hist <- ExtendedTask$new(function(spada_plot_fun,
                                           df,
                                           title,
                                           bins,
                                           plot_fill_color,
                                           plot_line_color,
                                           sample_limit,
                                           mean_value,
                                           sd_value) {
      mirai({
        spada_plot_fun(
          type = 'hist_density',
          df = df,
          xvar = 'x',
          xlab = 'Values',
          ylab = 'Density',
          title = title,
          bins = bins,
          fill_color = plot_fill_color,
          line_color = plot_line_color,
          point_shape = if (sample_limit > 1e4 && nrow(df) > 1e4) '.' else 20,
          sample_limit = sample_limit,
          mean_value = mean_value,
          sd_value = sd_value
        )
      },
      spada_plot_fun = spada_plot_fun,
      df = df,
      title = title,
      bins = bins,
      plot_fill_color = plot_fill_color,
      plot_line_color = plot_line_color,
      sample_limit = sample_limit,
      mean_value = mean_value,
      sd_value = sd_value)
    }) |> bind_task_button('btn_hist')

    observe({
      req(input$sel_var, var())
      task_hist$invoke(
        spada_plot_fun = spada_plot,
        df = data.frame(x = var()),
        bins = input$bins,
        title = paste('Histogram -', input$sel_var),
        plot_fill_color = session$userData$conf$plot_fill_color,
        plot_line_color = session$userData$conf$plot_line_color,
        sample_limit = session$userData$conf$plot_limit,
        mean_value = fmean(var(), na.rm = TRUE),
        sd_value = fsd(var(), na.rm = TRUE)
      )
    }) |> bindEvent(input$btn_hist)

    norm_hist <- reactive({ task_hist$result() })

    # insert histogram to output ----------------------------------------------
    insert_output_server(
      'norm_insert_output_hist',
      reactive(plot_tag(norm_hist())),
      'Histogram'
    )

    # qq plot -----------------------------------------------------------------
    output$qq_plot <- renderPlot({
      norm_qq_plot()
    }, res = 96)

    task_qq <- ExtendedTask$new(function(spada_plot_fun,
                                         df,
                                         title,
                                         plot_fill_color,
                                         plot_line_color,
                                         sample_limit) {
      mirai({
        spada_plot_fun(
          type = 'qq_plot',
          df = df,
          xvar = 'x',
          xlab = 'Theoretical Quantiles',
          ylab = 'Sample Quantiles',
          title = title,
          fill_color = plot_fill_color,
          line_color = plot_line_color,
          sample_limit = sample_limit
        )
      },
      spada_plot_fun = spada_plot_fun,
      df = df,
      title = title,
      plot_fill_color = plot_fill_color,
      plot_line_color = plot_line_color,
      sample_limit = sample_limit)
    }) |> bind_task_button('btn_qq')

    observe({
      req(input$sel_var, var())
      task_qq$invoke(
        spada_plot_fun = spada_plot,
        df = data.frame(x = var()),
        title = paste('Normal QQ Plot:', input$sel_var),
        plot_fill_color = session$userData$conf$plot_fill_color,
        plot_line_color = session$userData$conf$plot_line_color,
        sample_limit = session$userData$conf$plot_limit
      )
    }) |> bindEvent(input$btn_qq)

    norm_qq_plot <- reactive({ task_qq$result() })

    # insert to output --------------------------------------------------------
    insert_output_server(
      'norm_insert_output_qq',
      reactive(plot_tag(norm_qq_plot())),
      'QQ Plot'
    )

    output$conditional_add_output_qq <- renderUI({
      req(norm_qq_plot())
      insert_output_ui(ns('norm_insert_output_qq'))
    })

    # ks test -----------------------------------------------------------------
    ks_results <- reactive({
      req(input$sel_var)
      req(var())

      if(anyDuplicated(var()) > 0){
        # small error to avoid ties
        inputed_error <- reactive(abs(min_nona(var())/1e6))

        test_value <- reactive(var() + rnorm(length(var()),
                                             mean = 0,
                                             sd = inputed_error()))

        df <- ks.test(test_value(), 'pnorm') |> unlist() |> as.data.frame()

        df$results <- rownames(df)
        names(df) <- c('values', 'results')

        df[df$results == 'data.name', ]$values <- paste(input$sel_var)

        results = list(
          'results' = df,
          'observations' = paste0('There are tied values, normally distributed',
                                  ' random noise was added (mean = 0 and sd = ',
                                  inputed_error(), ')')
        )
      } else {
        df <- ks.test(var(), 'pnorm') |> unlist() |> as.data.frame()

        df$results <- rownames(df)
        names(df) <- c('values', 'results')

        df[df$results == 'data.name', ]$values <- paste(input$sel_var)

        results = list('results' = df)
      }

    }) |> bindEvent(input$btn_ks)

    ks_results_gt <- reactive({
      req(ks_results())

      ks_results()$results |>
        gt() |>
        cols_move(columns = 'values', after = 'results') |>
        gt::cols_label('values' = 'Values', 'results' = 'Results') |>
        tab_header('Kolmogorov-Smirnov Normality Test')
    })

    # ks staticards -----------------------------------------------------------
    output$conditional_staticard_ks <- renderUI({
      req(ks_results())

      stat_d <- ks_results()$results$values[which(ks_results()$results$results == 'statistic.D')]
      p_value <- ks_results()$results$values[which(ks_results()$results$results == 'p.value')]

      tagList(
        stati_card(stat_d |> as.numeric() |> f_num(dig = 5), 'Statistic D (test value)'),
        stati_card(p_value |> as.numeric() |> f_num(dig = 5), 'p value')
      )
    })

    output$ks_test <- render_gt({
      req(ks_results_gt())
      ks_results_gt()
    })

    output$ks_test_obs <- renderPrint({
      req(ks_results()$observations)
      ks_results()$observations
    })

    output$ks_test_obs_ui <- renderUI({
      req(ks_results()$observations)
      verbatimTextOutput(ns('ks_test_obs'))
    })

    save_gt_server('ks_save_gt', ks_results_gt)

    output$conditional_save_ks_gt <- renderUI({
      req(ks_results_gt())
      save_gt_ui(ns('ks_save_gt'))
    })

    # insert to output --------------------------------------------------------
    insert_output_server(
      'norm_insert_output_ks',
      reactive(ks_results_gt()),
      'Kolmogorov-Smirnov Normality Test'
    )

    output$conditional_add_output_ks <- renderUI({
      req(ks_results_gt())
      insert_output_ui(ns('norm_insert_output_ks'))
    })

    # help ks function --------------------------------------------------------
    observe({
      fun_help_modal('stats', 'ks.test')
    }) |> bindEvent(input$btn_help_ks)

    # sw test -----------------------------------------------------------------
    sw_results <- reactive({
      req(input$sel_var)
      req(var())
      req(var_len())

      var_len <- var_len()
      if (var_len < 3 || var_len > 5000) {
        msg(paste0('Sample size must be between 3 and 5000 (actual: ', var_len, ')'), 3)

        return()
      }

      if (test_all_equal(var())) {
        msg('Shapiro-Wilk test: the values can not be all equal')
        return()
      }

      df <- shapiro.test(var()) |> unlist() |> as.data.frame()

      df$results <- rownames(df)
      names(df) <- c('values', 'results')

      df[df$results == 'data.name', ]$values <- paste(input$sel_var)

      df
    }) |> bindEvent(input$btn_sw)

    sw_results_gt <- reactive({
      req(sw_results())

      sw_results() |>
        gt() |>
        cols_move(columns = 'values', after = 'results') |>
        cols_label('values' = 'Values', 'results' = 'Results') |>
        tab_header('Shapiro-Wilk Normality Test')
    })

    # sharpiro-wilk staticards ------------------------------------------------
    output$conditional_staticard_sw <- renderUI({
      req(sw_results())

      stat_w <- sw_results()$values[which(sw_results()$results == 'statistic.W')]
      p_value <- sw_results()$values[which(sw_results()$results == 'p.value')]

      tagList(
        stati_card(stat_w |> as.numeric() |> f_num(dig = 5), 'Statistic W (test value)'),
        stati_card(p_value |> as.numeric() |> f_num(dig = 5), 'p value')
      )
    })

    output$sw_test <- render_gt({
      req(sw_results_gt())
      sw_results_gt()
    })

    save_gt_server('sw_save_gt', sw_results_gt)

    output$conditional_save_sw_gt <- renderUI({
      req(sw_results_gt())
      save_gt_ui(ns('sw_save_gt'))
    })

    # insert to output --------------------------------------------------------
    insert_output_server(
      'norm_insert_output_sw',
      reactive(sw_results_gt()),
      'Shapiro-Wilk Normality Test'
    )

    output$conditional_add_output_sw <- renderUI({
      req(sw_results_gt())
      insert_output_ui(ns('norm_insert_output_sw'))
    })

    # help file of shapiro.test
    observe({
      fun_help_modal('stats', 'shapiro.test')
    }) |> bindEvent(input$btn_help_sw)

    # sf test -----------------------------------------------------------------
    sf_results <- reactive({
      req(input$sel_var)
      req(var())
      req(var_len())
      var_len <- var_len()
      if (var_len < 5 || var_len > 5000) {
        msg(paste0('Sample size must be between 5 and 5000 (actual: ', var_len, ')'), 3)

        return()
      }

      if (test_all_equal(var())) {
        msg('Shapiro-Francia test: the values can not be all equal')
        return()
      }

      df <- ShapiroFranciaTest(var()) |> unlist() |> as.data.frame()

      df$results <- rownames(df)
      names(df) <- c('values', 'results')

      df[df$results == 'data.name', ]$values <- paste(input$sel_var)

      df
    }) |> bindEvent(input$btn_sf)

    sf_results_gt <- reactive({
      req(sf_results())

      sf_results() |>
        gt() |>
        cols_move(columns = 'values', after = 'results') |>
        cols_label('values' = 'Values', 'results' = 'Results') |>
        tab_header('Shapiro-Francia Normality Test')
    })


    # sharpiro-francia staticards ---------------------------------------------
    output$conditional_staticard_sf <- renderUI({
      req(sf_results())

      stat_w <- sf_results()$values[which(sw_results()$results == 'statistic.W')]
      p_value <- sf_results()$values[which(sw_results()$results == 'p.value')]

      tagList(
        stati_card(stat_w |> as.numeric() |> f_num(dig = 5), 'Statistic W (test value)'),
        stati_card(p_value |> as.numeric() |> f_num(dig = 5), 'p value')
      )
    })

    output$sf_test <- render_gt({
      req(sf_results_gt())
      sf_results_gt()
    })

    save_gt_server('sf_save_gt', sf_results_gt)

    output$conditional_save_sf_gt <- renderUI({
      req(sf_results_gt())
      save_gt_ui(ns('sf_save_gt'))
    })

    # insert to output --------------------------------------------------------
    insert_output_server(
      'norm_insert_output_sf',
      reactive(sf_results_gt()),
      'Shapiro-Francia Normality Test'
    )

    output$conditional_add_output_sf <- renderUI({
      req(sf_results_gt())
      insert_output_ui(ns('norm_insert_output_sf'))
    })

    # help file of ShapiroFranciaTest
    observe({
      fun_help_modal('DescTools', 'ShapiroFranciaTest')
    }) |> bindEvent(input$btn_help_sf)

  })
}

# ui --------------------------------------------------------------------------
data_ui <- function(id) {
  ns <- NS(id)

  card(
    card_body(
      selectInput(
        ns('sel_dt'),
        'Select a dataset',
        choices = NULL
      ),
      textInput(ns('txt_new_name'), 'New name'),
      layout_column_wrap(
        btn_task(ns('btn_new_name'), 'Rename dataset', icon('file-signature')),
        btn_task(ns('btn_active'), 'Make dataset Active', icon('check')),
        btn_task(ns('btn_copy_dt'), 'Copy dataset', icon('copy')),
        btn_task(ns('btn_del_dt'), 'Delete dataset', icon('trash-can'))
      )
    )
  )
}

# server ----------------------------------------------------------------------
data_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  # update datasets input ---------------------------------------------------
	  observe({
	    req(session$userData$dt_names(), session$userData$dt$act_name)

	    choices <- c(
	      session$userData$dt$act_name,
	      setdiff(session$userData$dt_names(), session$userData$dt$act_name)
	    )

	    updateSelectInput(
	      session,
	      'sel_dt',
	      choices  = choices,
	      selected = session$userData$dt$act_name
	    )
	  })

	  # make active dataset event --------------
	  observe({
	    req(input$sel_dt)
	    session$userData$dt$act_name <- input$sel_dt
	    session$userData$dt$bkp0 <- copy(get_act_dt(session))
	    session$userData$dt$bkp <- NULL
	    msg(paste('Dataset', session$userData$dt$act_name, 'is the active one'))
	    updateTextInput(session, 'txt_new_name', value = '')
	  }) |> bindEvent(input$btn_active)

	  # rename dataset event -------------------
	  observe({
	    if(!is_name_available(input$txt_new_name, session$userData$dt_names())){
	      msg_error('New name is not valid or already in use')
	    } else {
	      names(session$userData$dt$dt)[session$userData$dt_names() == input$sel_dt] <- input$txt_new_name
	      # update active dataset if necessary
	      if(session$userData$dt$act_name == input$sel_dt){
	        session$userData$dt$act_name <- input$txt_new_name
	      }
	      # update metadata names
	      names(session$userData$dt$meta)[names(session$userData$dt$meta) == input$sel_dt] <- input$txt_new_name

	      msg('New name applied')
	      updateTextInput(session, 'txt_new_name', value = '')
	    }
	  }) |> bindEvent(input$btn_new_name)

	  # copy dataset event ---------------------
	  observe({
	    if(!is_name_available(input$txt_new_name, session$userData$dt_names())){
	      msg_error('New name is not valid or already in use')
	    } else {
	      session$userData$dt$dt[[input$txt_new_name]] <- session$userData$dt$dt[[input$sel_dt]]
	      # update metadata
	      session$userData$dt$meta[[input$txt_new_name]] <- session$userData$dt$meta[[input$sel_dt]]
	      gc()

	      msg(paste('Dataset', input$txt_new_name, 'created'))

	      updateTextInput(session, 'txt_new_name', value = '')
	    }
	  }) |> bindEvent(input$btn_copy_dt)

	  # delete dataset event -------------------
	  observe({
	    if(session$userData$dt$act_name == input$sel_dt){
	      msg_error('You can not delete the active dataset')
	    } else {
	      session$userData$dt$dt[[input$sel_dt]] <- NULL

	      # delete metadata
	      session$userData$dt$meta[[input$sel_dt]] <- NULL
	      gc()
	      msg(paste('Dataset', input$sel_dt, 'deleted'))

	      updateTextInput(session, 'txt_new_name', value = '')
	    }
	  }) |> bindEvent(input$btn_del_dt)

  })
}

# Function with the ui of spada.R
spada_ui <- function(conf){
  tagList(

    # favicon
    tags$head(tags$link(rel = 'icon', href = 'spada/favicon.ico')),

    busyIndicatorOptions(
      spinner_type = 'bars2',
      spinner_color = main_color,
      fade_opacity = '0.1'
    ),

    useBusyIndicators(),

    # css style for waiter screen ---------------------------------------------
    tag_js_waiter_text,

    useWaiter(),

    waiter_screen,

    show_startup_screen(),

    # close the app ---------
    tag_js_exit,

    page_navbar(
      id = 'navbar',
      theme = spada_bs_theme(conf$theme),
      title = 'Spada',
      navbar_options = navbar_options(
        bg = navbar_bg,
        underline = F
      ),

      # page sidebar ----------------------------------------------------------
      sidebar = sidebar_ui('sidebar'),

      # page data -------------------------------------------------------------
      nav_panel(
        'Data',
        icon = bs_icon('info-square-fill'),

        card(
          full_screen = T,
          card_body(
            class = 'big-card',
            navset_card_pill(
              id = 'navset_card_pill_data',
              nav_panel('Highlights', data_highlights_ui('pD_highlights')),
              nav_panel('Metadata', metadata_ui('pD_metadata')),
              nav_panel('Overview', data_overview_ui('pD_overview')),
              nav_panel('Data', data_ui('pD_data')),
              nav_panel('Import', import_file_ui('pD_import')),
              nav_panel('Export', export_file_ui('pD_export'))
            )
          )
        )
      ),

      # page edit -------------------------------------------------------------
      nav_panel(
        'Edit',
        icon = icon('screwdriver-wrench'),

        card(
          full_screen = T,
           card_body(
             class = 'big-card',
             layout_columns(
               navset_card_pill(
                 nav_panel(
                   'Calculate',
                   calculate_cols_ui('pE_calculate_cols')
                 ),
                 nav_panel('Convert', convert_cols_ui('pE_convert_cols')),
                 nav_panel('Duplicates', duplicates_ui('pE_duplicates')),
                 nav_panel(
                   'Filter',
                   layout_column_wrap(
                     filter_rows_ui('pE_filter_rows'),
                     select_cols_ui('pE_filter_sel_cols'),
                   )
                 ),
                 nav_panel('Join', join_ui('pE_join')),
                 nav_panel('Missing', missing_ui('pE_missing')),
                 nav_panel(
                   'Order',
                   layout_column_wrap(
                     order_rows_ui('pE_order_rows'),
                     order_cols_ui('pE_order_cols')
                   )
                 ),
                 nav_panel('Rename', rename_cols_ui('pE_rename_cols')),
                 nav_panel(
                   'Summarise',
                   layout_column_wrap(
                     groupby_ui('pE_groupby'),
                     summarise_ui('pE_summarise')
                   )
                 )
               )
             )
           ),
           card_footer(data_bkp_ui('pE_bkp'), class = 'big-card-footer')
        )
      ),

      # page analysis ---------------------------------------------------------
      nav_menu(
        title = 'Analysis',
        nav_panel('Exploratory',
                  icon = bs_icon('bar-chart-line'),
                  exploratory_ui('pA_exploratory')),
        nav_panel('Descriptive Stats',
                  icon = bs_icon('list-columns-reverse'),
                  descriptive_stats_ui('pA_desc_stats')),
        nav_panel('Correlation Matrix',
                  icon = bs_icon('table'),
                  cor_matrix_ui('pA_cor_matrix')),
        nav_panel('Correlation Test',
                  icon = bs_icon('magnet'),
                  correlation_ui('pA_correlation')
        ),
        nav_panel('Normality Test',
                  icon = bs_icon('bell'),
                  normality_test_ui('pA_normality_test')
        ),
        nav_panel('One-sample t Test', icon = icon('t'), one_t_test_ui('pA_t_test')),
        nav_panel('Z Test', icon = icon('z'), z_test_ui('pA_z_test')),
        nav_panel('Linear Model', icon = bs_icon('graph-up'), lm_ui('pA_lm'))
      ), # end of analysis menu

      # output ----------------------------------------------------------------
      nav_panel(
        title = 'Output',
        icon = icon('tablet-screen-button'),
        card(
          full_screen = T,
          card_body(class = 'big-card', output_ui('pO_output'))
        )
      ),

      # menu options ----------------------------------------------------------
      nav_menu(
        title = 'Options',
        config_ui('pC'),
        nav_item(
          tags$a(
            icon('code'),
            'Reference',
            href = 'https://lgschuck.github.io/spada/',
            target = '_blank'
          )
        ),
        nav_item(
          tags$a(
            icon('book'),
            'Spada Book',
            href = 'https://lgschuck.github.io/spada_book/',
            target = '_blank'
          )
        ),
        about_spada_ui('about_spada'),
        nav_panel(
          value = 'exit',
          title = 'Exit',
          icon = icon('arrow-right-from-bracket')
        )),

      nav_item(
        tags$a(
          icon('github'),
          href = 'https://github.com/lgschuck/spada',
          target = '_blank'
        )),

      nav_spacer(),

      # active dataset --------------------------------------------------------
      nav_item(h4(textOutput('dt_act_name')) |>
                 popover(navbar_df_info_ui('navbar_df_info'),
                         placement = 'bottom',
                         options = list(customClass = 'preview-dt-popup-mini')))
    )
  )
}

# ui --------------------------------------------------------------------------
duplicates_ui <- function(id) {
  ns <- NS(id)

  layout_column_wrap(
    card(
      card_header('Identify Duplicates', class = 'mini-header'),
      card_body(
        selectizeInput(
          ns('vars_sel_dup'),
          'Variables',
          '',
          multiple = T,
          options = list(plugins = list('remove_button', 'clear_button')),
          width = '80%'
        ),
        textInput(ns('txt_new_dup_name'), 'Duplicaeted Variable Name', 'Duplicated'),
        radioButtons(ns('radio_last'), 'Which Duplicated',
                     c('First case' = TRUE, 'Last case' = FALSE), inline = T),
        radioButtons(ns('radio_primary'), 'Mark Primary Cases',
                     c('Yes' = TRUE, 'No' = FALSE), inline = T),
        conditionalPanel(
          condition = sprintf("input['%s'] == 'TRUE'", ns('radio_primary')),
          textInput(ns('txt_new_primary_name'), 'Primary Variable Name', 'Primary')
        )
      ),
      card_footer(btn_task(ns('btn_identify'), 'Run', icon('check')))
    ),
    card(
      card_header('Remove Duplicates', class = 'mini-header'),
      card_body(
        selectInput(ns('vars_sel_drop'), 'Variable', ''),
      ),
      card_footer(btn_task(ns('btn_remove'), 'Remove', icon('trash-can')))
    )
  )
}

# server ----------------------------------------------------------------------
duplicates_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  df_names <- reactive(get_act_dt(session) |> names())

	  # update inputs ------------
	  observe({
	    req(df_names())
	    updateSelectizeInput(
	      session,
	      'vars_sel_dup',
	      choices = c('', df_names())
	    )

	    updateSelectInput(
	      session,
	      'vars_sel_drop',
	      choices = c('', df_names())
	    )
	  }) |> bindEvent(df_names())

	  # detect duplicates -----------
	  observe({

	    selected_vars <- input$vars_sel_dup
	    dup_name <- input$txt_new_dup_name
	    last_dup <- as.logical(input$radio_last)
	    primary <- as.logical(input$radio_primary)
	    primary_name <- input$txt_new_primary_name

	    new_vars <- if(primary){
	      c(dup_name, primary_name)
	    } else {
	      dup_name
	    }

	    if(selected_vars |> length() == 0){
	      msg('Select at least one variable')
	      return()
	    } else {
	      if(is_name_available(dup_name, df_names()) &&
	         (!primary || is_name_available(primary_name, df_names()))
	       ){

  	      running_modal()
  	      temp <- copy(get_act_dt(session))

  	      dup_vector <- duplicated(temp, fromLast = last_dup, by = selected_vars)

  	      dup_factor <- factor(
  	        dup_vector,
  	        levels = c(TRUE, FALSE),
  	        labels = c('Duplicated', 'Not Duplicated')
  	      )

  	      temp[, new_var := dup, env = list(
  	        new_var = dup_name,
  	        dup = dup_factor
  	      )]

  	      if(primary){
  	        primary_factor <- factor(
  	          !dup_vector,
  	          levels = c(TRUE, FALSE),
  	          labels = c('Primary', 'Not Primary')
  	        )

  	        temp[, new_var := p, env = list(
  	          new_var = primary_name,
  	          p = primary_factor
  	        )]
  	      }

  	      update_act_dt(session, copy(temp), updated_cols = new_vars)
  	      rm(temp)

  	      remove_running_modal()

  	      updateTextInput(session, 'txt_new_dup_name', value = 'Duplicated')
  	      updateTextInput(session, 'txt_new_primary_name', value = 'Primary')

  	    } else {
  	      msg_error('New names are not valid or already in use')
  	      return()
  	    }
	    }
	  }) |> bindEvent(input$btn_identify)

	  # drop duplicates -----------
	  observe({

	    selected_vars <- input$vars_sel_drop

	    not_dup_factor <- !(is.factor(get_act_dt(session)[[selected_vars]]) &&
	      all(levels(get_act_dt(session)[[selected_vars]]) %in% c('Duplicated', 'Not Duplicated')))

	    if(!isTruthy(selected_vars)){
	      msg('Select at least one variable')
	      return()
	    } else if(not_dup_factor){
	      msg('The variable is not a factor of Duplicated and Not Duplicated Levels', 4)
	      return()
	    } else {

        temp <- copy(get_act_dt(session))
        temp <- temp[var == 'Not Duplicated', , env = list(var = selected_vars)]

        if(!is_spada_df(temp)){
          abort_filter_modal()
        } else {
          running_modal()
          update_act_dt(session, copy(temp))
          updateSelectInput(session, 'vars_sel_drop', selected = character(0))
          remove_running_modal()
        }
      }
	  }) |> bindEvent(input$btn_remove)

  })
}

# ui --------------------------------------------------------------------------
sidebar_ui <- function(id) {
  ns <- NS(id)

  sidebar(
    class = 'spada-main-sidebar',
    open = T,
    accordion(
      open = T,
      accordion_panel(
        class = 'accordion-sidebar',
        'Active Dataset',
        icon = bs_icon('check2-square', size = '1.75em'),
        selectInput(ns('sel_act_dt'), NULL, choices = character(0)),
        uiOutput(ns('df_info')),
        fluidRow(
          column(3, mini_btn(ns('df_btn_overview'), '', icon('magnifying-glass')) |>
                   tooltip('Overview', placement = 'bottom')),
          column(3, mini_btn(ns('df_btn_meta'), '', bs_icon('info-circle')) |>
                   tooltip('Metadata', placement = 'bottom')),
          column(3, mini_btn(ns('df_btn_change'), '', icon('shuffle')) |>
                   tooltip('Change dataset', placement = 'bottom')),
          column(3, mini_btn(ns('df_btn_edit'), '', icon('screwdriver-wrench')) |>
                   tooltip('Edit', placement = 'bottom'))

        ),
        tags$br(),
        fluidRow(
          column(3, mini_btn(ns('df_btn_import'), '', icon('upload')) |>
                   tooltip('Import', placement = 'bottom')),
          column(3, mini_btn(ns('df_btn_export'), '', bs_icon('download')) |>
                   tooltip('Export', placement = 'bottom')),
          column(3, mini_btn(ns('df_btn_output'), '', icon('tablet-screen-button')) |>
                   tooltip('Output', placement = 'bottom')),
          column(3, mini_btn(ns('df_btn_explore'), '', icon('chart-simple')) |>
                   tooltip('Exploratory Analysis', placement = 'bottom'))
        )
      )
    ),
    accordion(
      open = T,
      accordion_panel(
        class = 'accordion-sidebar',
        'Datasets',
        icon = bs_icon('stack', size = '1.75em'),
        selectInput(ns('sel_datasets_names'), '', choices = character(0)),
        actionButton(ns('btn_preview_dt'), 'Preview', icon('magnifying-glass'),
                     class = 'btn-task') |>
          popover(htmlOutput(ns('df_preview')),
                  options = list(customClass = 'preview-dt-popup'))
      )
    ),
    btn_task(ns('btn_save_session'), 'Save Session', icon('save')),
    input_switch(id = ns('dark_mode'), label = 'Dark mode', value = FALSE)
  )
}

# server ----------------------------------------------------------------------
sidebar_server <- function(id, app_session) {
  moduleServer(id, function(input, output, session) {
  	ns <- session$ns

  	# active dataset info -----------------------------------------------------
    output$df_info <- renderUI({
      div(
        style = 'font-size: 0.9rem;',
        p(
          'Rows/Columns: ',
          textOutput(ns('row_col'), inline = TRUE)
        ),
        p(
          "Columns with NA's: ",
          textOutput(ns('col_nas'), inline = TRUE)
        )
      )
    })

    output$row_col <- renderText({
      session$userData$dt$act_mini_meta()[['row_col']]
    })

    output$col_nas <- renderText({
      session$userData$dt$act_mini_meta()[['col_nas']]
    })

    # update active dataset list --------
    observe({
      req(session$userData$dt_names(), session$userData$dt$act_name)

      updateSelectInput(
        session,
        'sel_act_dt',
        choices = c(
          session$userData$dt$act_name,
          setdiff(session$userData$dt_names(), session$userData$dt$act_name)
        ),
        selected = session$userData$dt$act_name
      )
    })

    # update active dataset --------
    observe({
      req(input$sel_act_dt)
      session$userData$dt$act_name <- input$sel_act_dt
      session$userData$dt$bkp0 <- copy(get_act_dt(session))
      session$userData$dt$bkp <- NULL
    }) |> bindEvent(input$sel_act_dt, ignoreInit = T)

    # mini buttons --------
    observe({
      nav_select('navbar', selected = 'Data', session = app_session)
      nav_select('navset_card_pill_data', selected = 'Overview', session = app_session)
    }) |> bindEvent(input$df_btn_overview)

    observe({
      nav_select('navbar', selected = 'Data', session = app_session)
      nav_select('navset_card_pill_data', selected = 'Metadata', session = app_session)
    }) |> bindEvent(input$df_btn_meta)

    observe({
      nav_select('navbar', selected = 'Data', session = app_session)
      nav_select('navset_card_pill_data', selected = 'Data', session = app_session)
    }) |> bindEvent(input$df_btn_change)

    observe({
      nav_select('navbar', selected = 'Exploratory', session = app_session)
    }) |> bindEvent(input$df_btn_explore)

    observe({
      nav_select('navbar', selected = 'Data', session = app_session)
      nav_select('navset_card_pill_data', selected = 'Import', session = app_session)
    }) |> bindEvent(input$df_btn_import)

    observe({
      nav_select('navbar', selected = 'Data', session = app_session)
      nav_select('navset_card_pill_data', selected = 'Export', session = app_session)
    }) |> bindEvent(input$df_btn_export)

    observe({
      nav_select('navbar', selected = 'Output', session = app_session)
    }) |> bindEvent(input$df_btn_output)

    observe({
      nav_select('navbar', selected = 'Edit', session = app_session)
    }) |> bindEvent(input$df_btn_edit)

    # list of datasets --------------------------------------------------------
    observe({
      req(session$userData$dt_names(), session$userData$dt$act_name)

      choices <- c(
        session$userData$dt$act_name,
        setdiff(session$userData$dt_names(), session$userData$dt$act_name)
      )

      updateSelectInput(session, 'sel_datasets_names', choices = choices)
    })

    # preview dataset -------------------------
    output$df_preview <- renderUI({
      req(session$userData$dt$dt, input$sel_datasets_names)

      head(session$userData$dt$dt[[input$sel_datasets_names]], 5) |>
        gt() |>
        tab_header(input$sel_datasets_names) |>
        tab_options(heading.align = 'left') |>
        as_raw_html() |>
        HTML()
    })

    # save session ------------------------------------------------------------
    observe({

      if(!session$userData$run_local){
        abort_run_local_modal('Spada must run locally to be able to save the session.')
      } else {
        spada_save(session$userData$conf$data_dir,
                   session$userData$dt$dt,
                   'data.qs2')

        spada_save(session$userData$conf$data_dir,
                   session$userData$out$elements,
                   'output.qs2')

        spada_save(session$userData$conf$conf_dir,
                   reactiveValuesToList(session$userData$conf),
                   'conf.qs2')

        show_toast(
          title = 'Session saved',
          type = 'info',
          position = 'bottom-start',
          timer = 2000,
          timerProgressBar = F,
          width = '250px'
        )
      }
    }) |> bindEvent(input$btn_save_session)

    # apply dark mode ------------
    observe({
      new_theme <- if(isTRUE(input$dark_mode)) {
        'spada_dark_theme'
      } else {
        'spada_theme'
      }

      if(identical(session$userData$conf$theme, new_theme)) return()

      running_modal('Applying theme...')

      session$onFlushed(function(){
        remove_running_modal()
      }, once = TRUE)

      set_spada_theme(session = session, theme = new_theme)

    }) |> bindEvent(input$dark_mode, ignoreInit = TRUE)

  })
}

# ui --------------------------------------------------------------------------
groupby_ui <- function(id) {
  ns <- NS(id)

  card(
    card_header('Group By', class = 'mini-header'),
    card_body(
      selectizeInput(ns('vars_groupby'), 'Group by', '',
                     multiple = T,
                     width = '80%',
                     options = list(plugins = list('remove_button', 'clear_button'))),
      div(
        layout_column_wrap(
          selectInput(ns('vars_sel'), 'Variables', ''),
          selectInput(ns('fun'), 'Select a function', ''),
          textInput(ns('txt_new_name'), 'New variable name')
        ),
        layout_column_wrap(
          btn_task(ns('btn_add_var'), 'Add Variable', icon('plus')),
          btn_task(ns('btn_rm_var'), 'Remove Variable', icon('xmark')),
          btn_task(ns('btn_view_var'), 'View New Variables',
                   icon('magnifying-glass'))
        ),
        radioButtons(ns('radio_overwrite'), 'Dataset',
                       c('New' = 'new', 'Overwrite' = 'overwrite'), inline = T
                       ),
        conditionalPanel(
            condition = "input.radio_overwrite == 'new'", ns = ns,
            textInput(ns('txt_new_dt_name'), 'New Dataset Name', placeholder = 'new_dataset')
        )
      )
    ),
    card_footer(btn_task(ns('btn_groupby'), 'Group By', icon('check')))
  )
}

# server ----------------------------------------------------------------------
groupby_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  df_names <- reactive(get_act_dt(session) |> names())

	  group <- reactiveValues(
      new_vars = NULL,
      funs = NULL,
      vars = NULL
	  )

	  # variables ---------------------------------------------------------------
	  observe({
	    req(df_names(), input$vars_groupby, session$userData$dt$act_meta())
	    updateSelectInput(
	      session,
	      'vars_sel',
	      choices = df_names()[df_names() %notin% input$vars_groupby]
	    )
	  }) |> bindEvent(input$vars_groupby, session$userData$dt$act_meta())

    # render variables group --------------------------------------------------
	  observe({
	    req(df_names())
	    updateSelectizeInput(
	      session,
	      'vars_groupby',
	      choices = c('', df_names())
	    )
	  }) |> bindEvent(df_names())

    # suggest name for new variable -------------------------------------------
    observe({
      req(input$vars_sel, input$fun)
      updateTextInput(session, 'txt_new_name',
                      value = paste0(input$vars_sel, '_', input$fun))
    }) |> bindEvent(input$vars_sel, input$fun)

    # render functions choices ------------------------------------------------
    selected_var_type <- reactive({
      req(input$vars_sel)
      obj_type(get_act_dt(session)[[input$vars_sel]])
    })

    observe({
      req(selected_var_type())

      choices = switch(
        selected_var_type(),
        'double'   = groupby_math_funs,
        'integer'  = groupby_math_funs,
        'char'     = groupby_char_funs,
        'date'     = groupby_date_funs,
        'logical'  = groupby_logical_funs,
        ' '
      )

      updateSelectInput(
        session,
        'fun',
        choices = choices,
        selected = choices[1]
      )
    }) |> bindEvent(selected_var_type())

    # add and remove buttons --------------------------------------------------
    # add new vars --------
    observe({

      if(!is_name_available(input$txt_new_name, group$newvars)){
        msg('New name is not valid or already inserted')
        return()
      } else if(!(input$fun %in% allowed_operations)) {
        msg_error('Function are not allowed')
        return()
      } else {
        group$newvars <- append(group$newvars, input$txt_new_name)
        group$funs <- append(group$funs, input$fun)
        group$vars <- append(group$vars, input$vars_sel)

        msg('New variable inserted')
      }

    }) |> bindEvent(input$btn_add_var)

    # remove inserted vars -------
    observe({

      if(input$txt_new_name %notin% group$newvars){
        msg('Variable not previously inserted')
      } else {

        index <- which(group$newvars == input$txt_new_name)

        group$newvars <- group$newvars[-index]
        group$funs <- group$funs[-index]
        group$vars <- group$vars[-index]

        msg('Variable removed')
      }

    }) |> bindEvent(input$btn_rm_var)

    # view inserted --------
    observe({
      req(group$newvars, group$funs, group$vars)

      inserted <- build_calls(group$newvars, group$funs, group$vars)

      df <- data.frame(
        new_var = group$newvars,
        operator = rep('=', length(group$newvars)),
        operation = as.character(inserted)[-1]
      )

      df <- df |>
        gt::gt() |>
        cols_label(
          new_var = '',
          operator = '',
          operation = ''
        ) |>
        opt_interactive(
          page_size_default = 5,
          page_size_values = c(5, 10, 25, 50, 100),
          use_resizers = T,
          use_highlight = T,
          use_compact_mode = T,
          use_text_wrapping = F,
          use_page_size_select = T
        ) |>
        tab_options(column_labels.hidden = T)

      showModal(modalDialog(
        title = div(icon('layer-group'), 'New Variables'),
        div(df),
        div(p(paste('Group by:', paste0(input$vars_groupby, collapse = ', ')))),
        size = 'l',
        easyClose = T
      ))

    }) |> bindEvent(input$btn_view_var)

    # clean inserted values if change active dataset
    observe({
      req(input$vars_groupby, group$newvars, group$funs)
      req(session$userData$dt$act_meta())

      group$newvars <- NULL
      group$funs <- NULL
      group$vars <- NULL

    }) |> bindEvent(session$userData$dt$act_meta())

    # apply function events ---------------------------------------------------
    observe({
      if(!isTruthy(group$newvars) || !isTruthy(group$vars)) {
        msg('Select at least one variable')
        return()
      } else if(!isTruthy(group$funs)){
        msg('Select a function')
        return()
      } else if(!(any(group$funs %in% allowed_operations))) {
        msg_error('Function are not allowed')
        return()
      }  else if (input$radio_overwrite == 'new' &&
                  (!is_name_available(input$txt_new_dt_name,
                                      session$userData$dt_names()))
                  )
      {
        msg_error('New name is not valid or already in use')
        return()
      } else {
        running_modal()

        j_calls <- build_calls(group$newvars, group$funs, group$vars)

        temp <- copy(get_act_dt(session))

        temp <- temp[, j, by = groupby ,
                     env = list(j = j_calls, groupby = input$vars_groupby |> as.list())]

        # overwrite -------
        if(input$radio_overwrite == 'overwrite'){

          update_act_dt(session, temp)

        } else if(input$radio_overwrite == 'new'){

          append_dt(session, temp, input$txt_new_dt_name)
          append_meta(session, temp |> df_info(), input$txt_new_dt_name)

        }
        rm(temp)

        # clean values ---------
        group$newvars <- NULL
        group$funs <- NULL
        group$vars <- NULL

        remove_running_modal()
      }

    }) |> bindEvent(input$btn_groupby)
  })
}

# tests/testthat/test-groupby_module.R

dt <- mtcars |> as.data.table()

# test group by - no groupby var ----------------------------------------------
test_that('Test group by - no group by var - overwrite', {
  testServer(groupby_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = dt),
      act_name = 'mtcars',
      meta = list('mtcars' = mtcars |> df_info())
    )

    session$userData$dt_names <- reactive('mtcars')
    session$userData$dt$act_meta <- reactive(dt |> df_info())
    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      vars_groupby = NULL,
      vars_sel = 'mpg',
      fun = 'min_na',
      txt_new_name = 'min_mpg',
      radio_overwrite = 'overwrite'
    )

    session$setInputs(btn_add_var = 1)

    session$setInputs(btn_groupby = 1)

    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      dt[, .(min_mpg = min(mpg)), by = NULL |> as.list()]
    )

    session$setInputs(vars_groupby = c())

    session$setInputs(btn_groupby = 2)

    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      dt[, .(min_mpg = min(mpg)), by = c() |> as.list()]
    )
    expect_equal(last_msg, 'Remove modal')
  })
})

# test group by - add and group -----------------------------------------------
test_that('Test group by - 1x1 var - overwrite', {
  testServer(groupby_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = dt),
      act_name = 'mtcars',
      meta = list('mtcars' = mtcars |> df_info())
    )

    session$userData$dt_names <- reactive('mtcars')
    session$userData$dt$act_meta <- reactive(dt |> df_info())
    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      vars_groupby = 'cyl',
      vars_sel = 'mpg',
      fun = 'min_na',
      txt_new_name = 'min_mpg',
      radio_overwrite = 'overwrite'
    )

    session$setInputs(btn_add_var = 1)

    session$setInputs(btn_groupby = 1)

    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      dt[, .(min_mpg = min(mpg)), by = 'cyl']
    )
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test group by - 2x1 var - overwrite', {
  testServer(groupby_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = dt),
      act_name = 'mtcars',
      meta = list('mtcars' = mtcars |> df_info())
    )

    session$userData$dt_names <- reactive('mtcars')
    session$userData$dt$act_meta <- reactive(dt |> df_info())
    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      vars_groupby = c('cyl', 'am'),
      vars_sel = 'hp',
      fun = 'max_na',
      txt_new_name = 'max_hp',
      radio_overwrite = 'overwrite'
    )
    session$setInputs(btn_add_var = 1)

    session$setInputs(btn_groupby = 1)

    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      dt[, .(max_hp = max(hp)), by = c('cyl', 'am')]
    )
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test group by - 2x2 var - overwrite', {
  testServer(groupby_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = dt),
      act_name = 'mtcars',
      meta = list('mtcars' = mtcars |> df_info())
    )

    session$userData$dt_names <- reactive('mtcars')
    session$userData$dt$act_meta <- reactive(dt |> df_info())
    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      vars_groupby = c('cyl', 'am'),
      vars_sel = 'hp',
      fun = 'sum_na',
      txt_new_name = 'sum_hp',
      radio_overwrite = 'overwrite'
    )
    session$setInputs(btn_add_var = 1)

    session$setInputs(
      vars_sel = 'mpg',
      fun = 'mean_na',
      txt_new_name = 'mean_mpg',
      radio_overwrite = 'overwrite'
    )
    session$setInputs(btn_add_var = 2)

    session$setInputs(btn_groupby = 1)

    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      dt[, .(sum_hp = sum(hp),
             mean_mpg = mean(mpg)), by = c('cyl', 'am')]
    )
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test group by - 2x2 var - new dataset', {
  testServer(groupby_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = dt),
      act_name = 'mtcars',
      meta = list('mtcars' = mtcars |> df_info())
    )

    session$userData$dt_names <- reactive('mtcars')
    session$userData$dt$act_meta <- reactive(dt |> df_info())
    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      vars_groupby = c('cyl', 'am'),
      vars_sel = 'hp',
      fun = 'sum_na',
      txt_new_name = 'sum_hp',
      radio_overwrite = 'new',
      txt_new_dt_name = 'new_dt'
    )
    session$setInputs(btn_add_var = 1)

    session$setInputs(
      vars_sel = 'mpg',
      fun = 'mean_na',
      txt_new_name = 'mean_mpg'
    )
    session$setInputs(btn_add_var = 2)

    session$setInputs(btn_groupby = 1)

    expect_equal(
      session$userData$dt$dt[['new_dt']],
      dt[, .(sum_hp = sum(hp),
             mean_mpg = mean(mpg)), by = c('cyl', 'am')]
    )

    expect_equal(
      session$userData$dt$meta[['new_dt']],
      dt[, .(sum_hp = sum(hp),
             mean_mpg = mean(mpg)), by = c('cyl', 'am')] |> df_info()
    )
    expect_equal(last_msg, 'Remove modal')
  })
})

# test group by - overwrite - metadata update ---------------------------------
# test using spada_server for run the metadata update
# only with he groupby module the metadata is not updated
dfs <- list('df_iris' = iris |> as.data.table(),
            'df_mtcars' = mtcars |> as.data.table())

temp_dir <- tempdir()

start_conf <- list(
  'conf_dir' = file.path(temp_dir, 'conf'),
  'data_dir' = file.path(temp_dir, 'data'),
  'theme' = 'spada_theme',
  'file_size' = 1000,
  'restore_session' = 'never',
  'save_session' = 'ask',
  'restore_data_status' = 0,
  'restore_output_status' = 0,
  'restore_status' = NULL,
  'plot_fill_color' = plot_fill_color,
  'plot_line_color' = plot_line_color
)

test_that('Test group by - overwrite - metadata update', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt$act_name <- 'df_mtcars'

    groupby_server('groupby')

    session$setInputs(
      'groupby-vars_groupby' = c('cyl', 'am'),
      'groupby-vars_sel' = 'hp',
      'groupby-fun' = 'sum_na',
      'groupby-txt_new_name' = 'sum_hp',
      'groupby-radio_overwrite' = 'overwrite',
      'groupby-txt_new_dt_name' = 'new_dt'
    )
    session$setInputs('groupby-btn_add_var' = 1)

    session$setInputs(
      'groupby-vars_sel' = 'mpg',
      'groupby-fun' = 'mean_na',
      'groupby-txt_new_name'= 'mean_mpg'
    )
    session$setInputs('groupby-btn_add_var' = 2)

    session$setInputs('groupby-btn_groupby' = 1)

    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      dt[, .(sum_hp = sum(hp),
             mean_mpg = mean(mpg)), by = c('cyl', 'am')]
    )

    expect_equal(
      session$userData$dt$meta[[session$userData$dt$act_name]],
      dt[, .(sum_hp = sum(hp),
             mean_mpg = mean(mpg)), by = c('cyl', 'am')] |> df_info()
    )
    expect_equal(last_msg, 'Remove modal')
  })
})

# test group by - remove and group --------------------------------------------
test_that('Test group by - 1x1 var - overwrite - remove', {
  testServer(groupby_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = dt),
      act_name = 'mtcars',
      meta = list('mtcars' = mtcars |> df_info())
    )

    session$userData$dt_names <- reactive('mtcars')
    session$userData$dt$act_meta <- reactive(dt |> df_info())
    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      vars_groupby = 'cyl',
      vars_sel = 'mpg',
      fun = 'min_na',
      txt_new_name = 'min_mpg',
      radio_overwrite = 'overwrite'
    )

    session$setInputs(btn_add_var = 1)

    session$setInputs(
      vars_sel = 'hp',
      fun = 'mean_na',
      txt_new_name = 'mean_hp'
    )
    session$setInputs(btn_add_var = 2)

    expect_equal(group$newvars, c('min_mpg', 'mean_hp'))
    expect_equal(group$funs, c('min_na', 'mean_na'))
    expect_equal(group$vars, c('mpg', 'hp'))

    session$setInputs(
      txt_new_name = 'min_mpg',
      btn_rm_var = 2
    )

    expect_equal(group$newvars, 'mean_hp')
    expect_equal(group$funs, 'mean_na')
    expect_equal(group$vars, 'hp')

    session$setInputs(btn_groupby = 1)

    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      dt[, .(mean_hp = mean(hp)), by = 'cyl']
    )
    expect_equal(last_msg, 'Remove modal')
  })
})

utils::globalVariables(
  c('.N',
    'cols',
    'density',
    'dt_names_react',
    'dup',
    'f_icon',
    'fun',
    'groupby',
    'i',
    'j',
    'label',
    'measure',
    'Measures',
    'new_var',
    'n_nas',
    'n_unique',
    'n_valid',
    'n_zero',
    'perc_nas',
    'perc_unique',
    'perc_valid',
    'perc_zero',
    'preview',
    'Pr(>|t|)',
    'q1',
    'q3',
    'results',
    'rows',
    'sel_var',
    'Sig Levels',
    'size',
    'terms',
    'value',
    'values',
    'var1',
    'Var1',
    'Var2',
    'x',
    'y',
    'integer_var',
    'numeric_var',
    'char_var',
    'char_long_var',
    'char_colors_var',
    'date_var',
    'datetime_var',
    'factor_var',
    'logical_var',
    'posixct_var',
    'waiter_right_foot'
    )
  )

# ui --------------------------------------------------------------------------
exploratory_ui <- function(id) {
  ns <- NS(id)

  card(
    full_screen = T,
    card_body(
      class = 'big-card',
      layout_columns(
        col_widths = c(2, 7, 3),
        navset_card_pill(
          full_screen = T,
          nav_panel('Parameters',
                    selectInput(ns('sel_vars'),
                                list('Main Variable', bs_icon('info-circle')) |>
                                  ttip('Dependent Variable'),
                                choices = NULL),
                    selectInput(ns('sel_vars2'),
                                list('Variable 2', bs_icon('info-circle')) |>
                                  ttip('Independent Variable'),
                                choices = NULL)
                    ),
          nav_panel('Filters',
                    checkboxInput(
                      ns('outliers'),
                      list('Remove Outliers',
                      bs_icon('info-circle')) |>
                        ttip('Only for numeric vars')
                    )
          )
        ),
        navset_card_pill(
          full_screen = T,
          nav_panel(
            'Distribution',
            full_screen = T,
            card_body(plotOutput(ns('dist_plot'))),
            card_footer(
              fluidRow(
                column(6,
                       radioGroupButtons(
                         ns('radio_dist_plot'),
                         'Plot type:',
                         c('Histogram' = 'hist',
                           'Boxplot' = 'boxplot',
                           'Boxplot by Groups' = 'boxplot_group',
                           'Dots' = 'dots',
                           'Barplot' = 'barplot'), size = 'sm', individual = T)),
                column(2, numericInput(ns('var_percentile'), 'Percentile', 50, 0, 100, 5)),
                column(1, conditionalPanel(
                  condition = "input.radio_dist_plot == 'hist'", ns = ns,
                  numericInput(ns('bins'), 'Bins', 25, 5, step = 5))
                ),
                column(3, div(insert_output_ui(ns('insert_dist_plot'))),
                    style = 'margin-top: 28px')
              ),
              div(style = 'margin-bottom: -8px !important;'),
            )
          ),
          nav_panel(
            'Scatter',
            full_screen = T,
            card_body(plotOutput(ns('scatter_plot'))),
            card_footer(
              layout_column_wrap(
                checkboxInput(
                  ns('scatter_lm'),
                  list('Plot Linear Model', bs_icon('info-circle')) |>
                    ttip('Show the line only if LM model was created')),
                btn_task(ns('btn_scatter'), 'Generate Plot', icon('chart-simple')),
                insert_output_ui(ns('insert_scatter'))
              ),
              div(style = 'margin-bottom: -18px !important;'),
            )
          ),
          nav_panel('Table', full_screen = T, table_values_ui(ns('table_values'))),
          nav_panel('Frequencies', full_screen = T, frequencies_ui(ns('frequencies'))),
          nav_panel(
            'Linear Model',
            full_screen = T,
            navset_card_pill(
              nav_panel(
                'Parameters',
                sliderInput(ns('sample_size'), 'Sample Size (%)', 0, 100, 100) |>
                  tooltip('Applied only if valid values are greater than 10.000'),
                layout_column_wrap(
                  btn_task(ns('btn_scatter_lm_run'), 'Run Linear Model', icon('gear')),
                  btn_task(ns('btn_scatter_lm_clear'), 'Clear Linear Model', icon('trash-can'))
                )
              ),
              nav_panel(
                'Output',
                card_body(gt_output(ns('lm_var_table')), gt_output(ns('lm_metrics'))),
                card_footer(insert_output_ui(ns('insert_lm_model_output')))
              ),
              nav_panel(
                'Residuals',
                plotOutput(ns('lm_resid_plot')),
                card_footer(
                  layout_column_wrap(
                    radioGroupButtons(
                      ns('radio_lm_resid'),
                      'Plot type:',
                      c(
                        'Histogram' = 'hist',
                        'Boxplot' = 'boxplot',
                        'Dots' = 'dots',
                        'QQ Plot' = 'qq_plot'
                      ),
                      size = 'sm',
                      individual = T
                    ),
                    btn_task(
                      ns('btn_lm_resid'),
                      'Plot residuals',
                      icon('chart-simple'),
                      style = 'margin-top: 28px'
                    ),
                    div(insert_output_ui(ns(
                      'insert_lm_resid_plot'
                    )), style = 'margin-top: 28px')
                  ),
                  div(style = 'margin-bottom: -24px !important;'),
                )
              ),
            )
          ),
        ),
        navset_card_pill(
          nav_panel('Stats', full_screen = T, stats_table_ui(ns('pA_stats')))
        )
      )
    )
  )
}

# server ----------------------------------------------------------------------
exploratory_server <- function(id) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns

    # df active ---------------------------------------------------------------
    df <- reactiveValues()
    observe({
      df$df_active <- get_act_dt(session)
    })

    var_analysis <- reactive({
      session$userData$dt$act_meta()[perc_nas != 1, var]
    })

    observe({
      vars <- var_analysis()
      req(vars)

      updateSelectInput(
        session,
        "sel_vars",
        choices = vars,
        selected = vars[1]
      )

      updateSelectInput(
        session,
        "sel_vars2",
        choices = vars,
        selected = vars[2]
      )
    })


    # outlier filter ----------------------------------------------------------
    outliers_index <- reactive({
      v <- df$df_active[[input$sel_vars]]
      if(input$outliers & is.numeric(v)) {
        q1 <- fquantile(v, 0.25)
        q3 <- fquantile(v, 0.75)
        dist_interquatile <- q3 - q1
        v >= (q1 - 1.5 * dist_interquatile) & v <= (q3 + 1.5 * dist_interquatile)
      } else {
        rep(T, length(v))
      }
    })

    # values to analysis page -------------------------------------------------
    var <- reactive({
      req(input$sel_vars)
      df$df_active[[input$sel_vars]][outliers_index()]
    })

    var2 <- reactive({
      req(input$sel_vars2)
      df$df_active[[input$sel_vars2]][outliers_index()]
    })

    var_percentile <- reactive(
      if(isTruthy(input$var_percentile) && is.numeric(var()) &&
         between(input$var_percentile, 0, 100)){
        fquantile(var(), input$var_percentile / 100)
      } else { NA }
    )

    # render plots ------------------------------------------------------------
    dist_plot <- reactive({
      req(var())
      req(var2())

      if (input$radio_dist_plot == 'barplot'){
        validate(need(!is.numeric(var()), 'Var can not be numeric'))

        spada_plot(type = 'barplot',
                   df = data.frame(x = var()),
                   xvar = 'x',
                   ylab = 'Count',
                   fill_color = session$userData$conf$plot_fill_color,
                   sample_limit = session$userData$conf$plot_limit
        )

      } else {
        validate(need(is.numeric(var()), 'Var must be numeric'))

        if (input$radio_dist_plot == 'boxplot_group'){
          validate(
            need(!is.numeric(var2()) | (is.numeric(var2()) & is.integer(var2())),
                 'Variable 2 can not be float'),
            need(!is.complex(var2()), 'Variable 2 can not be complex')
          )

          spada_plot(
            type = 'boxplot_group',
            df = data.frame(x = {
              if (var2() |> is.numeric())
                as.factor(var2())
              else
                var2()
            }, y = var()),
            xvar = 'x',
            yvar = 'y',
            fill_color = session$userData$conf$plot_fill_color,
            line_color = session$userData$conf$plot_line_color,
            vertical_line = var_percentile(),
            sample_limit = session$userData$conf$plot_limit
          )

        } else {
          validate(
            need(isTruthy(input$var_percentile)
                 && between(input$var_percentile, 0, 100), 'Percentile must be between 0 and 100')
          )

          if(input$radio_dist_plot == 'hist'){
            validate(need(input$bins > 0, 'Bins must be 1 or higher'))

            spada_plot(type = 'hist',
                       df = data.frame(x = var()),
                       xvar = 'x',
                       ylab = 'Count',
                       fill_color = session$userData$conf$plot_fill_color,
                       line_color = session$userData$conf$plot_line_color,
                       title_color = session$userData$conf$plot_title_color,
                       title = paste('Histogram -', input$sel_vars),
                       bins = input$bins,
                       vertical_line = var_percentile(),
                       sample_limit = session$userData$conf$plot_limit
            )

          } else if (input$radio_dist_plot == 'boxplot'){

            spada_plot(type = 'boxplot',
                       df = data.frame(x = var()),
                       xvar = 'x',
                       fill_color = session$userData$conf$plot_fill_color,
                       line_color = session$userData$conf$plot_line_color,
                       vertical_line = var_percentile(),
                       sample_limit = session$userData$conf$plot_limit
            )

          } else if (input$radio_dist_plot == 'dots'){

            spada_plot(type = 'dots',
                       df = data.frame(x = seq_along(var()), y = var()),
                       xvar = 'x',
                       yvar = 'y',
                       xlab = 'Index',
                       ylab = 'Values',
                       fill_color = session$userData$conf$plot_fill_color,
                       line_color = session$userData$conf$plot_line_color,
                       vertical_line = var_percentile(),
                       point_shape = if(session$userData$conf$plot_limit > 1e4 &&
                                        length(var()) > 1e4) '.' else 20,
                       sample_limit = session$userData$conf$plot_limit
            )
          }
        }
      }
    })

    output$dist_plot <- renderPlot({
      req(dist_plot())
      dist_plot()
    }, res = 96)
    # render scatter plot -----------------------------------------------------
    scatter_plot <- reactive({

      model_plot <- (input$scatter_lm &&
                       linear_model$y_name == input$sel_vars &&
                       linear_model$x_name == input$sel_vars2)

      s_plot <- spada_plot(type = 'scatter',
                 df = data.frame(x = var2(), y = var()),
                 xvar = 'x',
                 yvar = 'y',
                 xlab = input$sel_vars2,
                 ylab = input$sel_vars,
                 title = if(model_plot){
                   paste(
                    'Adjusted R Squared:',
                    summary(linear_model$model)$r.squared |> round(4)
                    )
                 } else {
                   paste('Pearson Correlation:', stats_correlation() |> round(4))
                 },
                 fill_color = session$userData$conf$plot_fill_color,
                 title_color = session$userData$conf$plot_title_color,
                 point_shape = if(session$userData$conf$plot_limit > 1e4 &&
                                  length(var()) > 1e4) '.' else 20,
                 sample_limit = session$userData$conf$plot_limit
        )
        # insert model line
        if(model_plot){
          s_plot <- s_plot +
            geom_line(
              data = data.frame(x = linear_model$x, y = linear_model$y),
              aes(x = x, y = y),
              color = session$userData$conf$plot_line_color,
              linewidth = 1
            )
        }

      s_plot

    })|> bindEvent(input$btn_scatter)

    output$scatter_plot <- renderPlot({
      validate(
        need(is.numeric(var()) && is.numeric(var2()), 'Variables must be numeric')
      )
      scatter_plot()
    }, res = 96)

    # tables ------------------------------------------------------------------
    table_values_server('table_values',
                        var,
                        var2,
                        reactive(input$sel_vars),
                        reactive(input$sel_vars2))

    # frequencies -------------------------------------------------------------
    frequencies_server('frequencies', var, reactive(input$sel_vars))

    # linear model ------------------------------------------------------------
    linear_model <- reactiveValues(
      model = NULL,
      x = NULL,
      y = NULL,
      x_name = '',
      y_name = ''
    )

    observe({
      if(!is.numeric(var())){
        msg('The Dependent variable must be numeric', 2.5)
      } else if (input$sel_vars == input$sel_vars2) {
        msg('Select diferent variables for X and Y.', 2.5)
      } else {
        linear_model$y_name <- input$sel_vars
        linear_model$x_name <- input$sel_vars2

        var_size <- length(var())

        if(var_size < 10e3) {
          var_y <- var()
          var_x <- var2()
        } else {
          sample_size <- min(var_size,
                             floor(var_size * min(1, max(0, input$sample_size/100))))
          lm_sample <- sample.int(var_size, sample_size, replace = F) |>
            sort()
          var_y <- var()[lm_sample]
          var_x <- var2()[lm_sample]
        }

        linear_model$model <- lm(var_y ~ var_x, model = F)
        linear_model$x <- var_x
        linear_model$y <- linear_model$model$fitted.values
        msg('Lm model completed.')
      }
    }) |> bindEvent(input$btn_scatter_lm_run)

    observe({
      linear_model$model <- NULL
      linear_model$model$residuals <- NULL
      linear_model$x <- NULL
      linear_model$y <- NULL
      linear_model$x_name <- ''
      linear_model$y_name <- ''

      # clear the residual plot to avoid incorrect output element
      update_lm_resid_plot(update_lm_resid_plot() + 1)

      msg('Lm model cleared.')
    }) |> bindEvent(input$btn_scatter_lm_clear)

    # print linear model ------------------------------------------------------
    lm_var_table <- reactive({
      req(linear_model$model)
      output <- linear_model_df_output(linear_model$model |> summary())
      output$Variable <- gsub('var_x', linear_model$x_name, output$Variable)

      output |>
        gt() |>
        tab_header(title = 'Linear Model',
                   subtitle = paste('Independent Variable:',
                                    linear_model$y_name))
    })

    lm_metrics <- reactive({
      req(linear_model$model)
      linear_model_df_metrics(linear_model$model |> summary()) |>
        gt() |> tab_header('Model metrics')
    })

    output$lm_var_table <- render_gt({
      req(lm_var_table())
      lm_var_table()
    })

    output$lm_metrics <- render_gt({
      req(lm_metrics())
      lm_metrics()
    })

    # plot linear model residuals ---------------------------------------------
    update_lm_resid_plot <- reactiveVal(0)

    observe({
      update_lm_resid_plot(update_lm_resid_plot() + 1)
    }) |> bindEvent(input$btn_lm_resid)

    lm_resid_plot <- reactive({
      req(linear_model$model$residuals)
      req(update_lm_resid_plot() > 0)

      if(input$radio_lm_resid == 'hist'){

        spada_plot(type = 'hist',
                   df = data.frame(x = linear_model$model$residuals),
                   xvar = 'x',
                   ylab = 'Count',
                   fill_color = session$userData$conf$plot_fill_color,
                   line_color = session$userData$conf$plot_line_color,
                   title_color = session$userData$conf$plot_title_color,
                   sample_limit = session$userData$conf$plot_limit
        )

      } else if (input$radio_lm_resid == 'boxplot'){

        spada_plot(type = 'boxplot',
                   df = data.frame(x = linear_model$model$residuals),
                   xvar = 'x',
                   fill_color = session$userData$conf$plot_fill_color,
                   line_color = session$userData$conf$plot_line_color,
                   sample_limit = session$userData$conf$plot_limit
        )

      } else if (input$radio_lm_resid == 'dots'){

        spada_plot(type = 'dots',
                   df = data.frame(x = seq_along(linear_model$model$residuals),
                                   y = linear_model$model$residuals),
                   xvar = 'x',
                   yvar = 'y',
                   xlab = 'Index',
                   ylab = 'Values',
                   fill_color = session$userData$conf$plot_fill_color,
                   line_color = session$userData$conf$plot_line_color,
                   vertical_line = 0,
                   point_shape = if(session$userData$conf$plot_limit > 1e4 &&
                                    length(linear_model$model$residuals) > 1e4) '.' else 20,
                   sample_limit = session$userData$conf$plot_limit,
                   line_type = 2
        )
      } else if (input$radio_lm_resid == 'qq_plot'){
        spada_plot(
          type = 'qq_plot',
          df = data.frame(x = linear_model$model$residuals),
          xvar = 'x',
          xlab = 'Theoretical Quantiles',
          ylab = 'Sample Quantiles',
          title = 'Linear Model - Residuals',
          fill_color = session$userData$conf$plot_fill_color,
          line_color = session$userData$conf$plot_line_color,
          sample_limit = session$userData$conf$plot_limit
        )
      }
    }) |> bindEvent(update_lm_resid_plot())

    output$lm_resid_plot <- renderPlot({
      validate(need(isTruthy(linear_model$model), 'No residuals to plot'))

      lm_resid_plot()
    }, res = 96)

    # metrics -----------------------------------------------------------------
    stats_sd <- reactive(if(is.numeric(var())) fsd(var(), na.rm = T) else NA)

    stats_correlation <- reactive(
      if(is.numeric(var()) && is.numeric(var2()) && stats_sd() != 0 &&
         !is.na(stats_sd())){
        sd_var2 <- fsd(var2(), na.rm = T)
        if(sd_var2 == 0 || sd_var2 |> is.na()) {
          NA
        } else {
          cor(var(), var2(), method = 'p', use = 'na.or.complete')
        }
      } else { NA }
    )
    # stats table -------------------------------------------------------------
    mod_stats_table <- stats_table_server(
      'pA_stats',
      var,
      reactive(input$sel_vars),
      reactive(input$var_percentile),
      var_percentile,
      stats_sd,
      stats_correlation
    )

    # insert dist plot to output ----------------------------------------------
    insert_output_server(
      'insert_dist_plot',
      reactive(plot_tag(dist_plot())),
      'Exploratory Plot'
    )

    # insert scatter to output ------------------------------------------------
    insert_output_server(
      'insert_scatter',
      reactive(plot_tag(scatter_plot())),
      'Scatter Plot'
    )

    # insert lm model output --------------------------------------------------
    insert_output_server(
      'insert_lm_model_output',
      reactive(gen_table2(lm_var_table(), lm_metrics())),
      'Linear Model'
    )

    # insert lm residual plot to output ---------------------------------------
    insert_output_server(
      'insert_lm_resid_plot',
      reactive(plot_tag(lm_resid_plot())),
      'Linear Model - Residuals Plot'
    )

  })
}

# ui --------------------------------------------------------------------------
table_values_ui <- function(id) {
  ns <- NS(id)

  tagList(
    card_body(
      fluidRow(
        column(4,
               radioGroupButtons(
                 ns('table_var'),
                 'Table variables:',
                 c('1 Variable' = '1v', '2 Variables' = '2v'),
                 size = 'sm',
                 individual = T
               )
        ),
        column(4,
               radioGroupButtons(
                 ns('table_type'),
                 'Table type:',
                 c('Absolute Values' = 'abs_table', 'Percent Values' = 'perc_table'),
                 size = 'sm',
                 individual = T
               )
        ),
      ),
      gt_output(ns('table')),
    ),
    card_footer(insert_output_ui(ns('insert_table_values')))
  )
}

# server ----------------------------------------------------------------------
table_values_server <- function(id, var, var2, var_name, var2_name) {
  moduleServer(id, function(input, output, session) {

    ns <- session$ns

    table_values <- reactive({
      req(var(), input$table_var, input$table_type)

      var <- var()

      if(input$table_var == '1v') {

        validate(need(is.character(var) || is.factor(var) || is.logical(var),
                      'Var must be character, factor or logical'))

        tab <- var |> qtab()
        if(input$table_type == 'perc_table') tab <- tab |> prop.table()

        tab <- tab |> as.data.frame()
        names(tab) <- c('Var1', 'Freq')
        tab

      } else if (input$table_var == '2v'){
        req(var2())
        var2 <- var2()

        validate(need(
          var_name() != var2_name() && is_categorical(var) && is_categorical(var2),
          'Select two different variables of type character, factor or logical')
        )

        tab <- qtab(var, var2)

        if(input$table_type == 'perc_table') tab <- tab |> prop.table()

        tab <- tab |> as.data.frame.matrix()

        cbind(Var1 = rownames(tab), tab)
      }
    })

    table_values_gt <- reactive({
      req(table_values())
      req(input$table_var, input$table_type, var_name())

      var_name <- var_name()
      tab <- table_values()

      if(input$table_var == '1v') {

        y_label <- if(input$table_type == 'abs_table') 'Frequency' else 'Relative Frequency'

        tab <- tab |>
          gt() |>
          cols_label(
            Var1 = var_name,
            Freq = y_label
          )

        if(input$table_type == 'abs_table') {
          tab |> fmt_number(columns = Freq)
        } else {
          tab |> fmt_percent(columns = Freq, decimals = 4)
        }

      } else if(input$table_var == '2v'){
        req(var2_name())
        var2_name <- var2_name()

        cols <- names(tab)[-1]

        tab <- tab |>
          gt() |>
          cols_label(Var1 = '') |>
          tab_spanner(
            label = var_name,
            columns = Var1
          ) |>
          tab_spanner(
            label = var2_name,
            columns = cols
          )

        if(input$table_type == 'abs_table') {
          tab |> fmt_number()
        } else {
          tab |> fmt_percent(decimals = 4)
        }
      }
    })

    output$table <- render_gt({
      req(table_values_gt())
      table_values_gt() |>
        opt_interactive(
          page_size_default = 10,
          use_filters = T,
          use_resizers = T,
          use_highlight = T,
          use_compact_mode = T,
          use_text_wrapping = F,
          use_page_size_select = T
        )
    })

    # insert table of values to output ----------------------------------------
    insert_output_server('insert_table_values', table_values_gt, 'Table')

  })
}

# ui --------------------------------------------------------------------------
calculate_cols_ui <- function(id) {
  ns <- NS(id)

  layout_column_wrap(
    card(
      card_header('Apply Function', class = 'mini-header'),
      card_body(
        selectInput(ns('vars_sel'), 'Variable', ''),

        selectInput(ns('fun'), 'Select a function', ''),
        textInput(ns('txt_new_name_fun'), 'New variable name'),
        selectizeInput(ns('vars_groupby_fun'), 'Group by', '',
                       multiple = T,
                       options = list(plugins = list('remove_button', 'clear_button')))

      ),
      card_footer(btn_task(
        ns('btn_apply_fun'), 'Apply Function', icon('check')
      ))
    ),
    card(
      card_header('Freehand', class = 'mini-header'),
      card_body(
        textInput(ns('txt_new_name_free'), 'New var name', placeholder = 'new_var'),
        textAreaInput(ns('txt_code_input'), 'Input code', width = '800px', height = '200px'),
        fluidRow(
          column(4, btn_task(ns('btn_allowed_operations'), 'Show Allowed Operations')),
          column(4, btn_task(ns('btn_help'), 'Help', icon('question')))
        ),
        selectizeInput(ns('vars_groupby_free'),
                       'Group By',
                       choices = '',
                       multiple = T,
                       options = list(plugins = list('remove_button', 'clear_button'))
        )
      ),
      card_footer(btn_task(
        ns('btn_apply_free'), 'Run Code', icon('gears')
      ))
    )
  )
}

# server ----------------------------------------------------------------------
calculate_cols_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  df_names <- reactive(get_act_dt(session) |> names())

	  observe({
	    req(df_names())
	    updateSelectInput(
	      session,
	      'vars_sel',
	      choices = c('', df_names())
	    )
	  }) |> bindEvent(df_names())

    # render variables group -------------------------------------------------
	  observe({
	    req(df_names())
	    updateSelectizeInput(
	      session,
	      'vars_groupby_fun',
	      choices = c('', df_names())
	    )
	  }) |> bindEvent(df_names())


    # suggest name for new variable -------------------------------------------
    observe({
      req(input$vars_sel, input$fun)
      updateTextInput(session, 'txt_new_name_fun',
                      value = paste0(input$vars_sel, '_', input$fun))
    }) |> bindEvent(input$vars_sel, input$fun)

    # render functions choices ------------------------------------------------
    selected_var_type <- reactive({
      req(input$vars_sel)
      obj_type(get_act_dt(session)[[input$vars_sel]])
    })

    observe({
      req(selected_var_type())

      updateSelectInput(
        session,
        'fun',
        choices = switch(
          selected_var_type(),
          'double'   = math_funs,
          'integer'  = math_funs,
          'char'     = char_funs,
          'date'     = date_funs,
          'logical'  = logical_funs,
          'factor'   = factor_funs,
          'complex'  = complex_funs,
          character(0)
        )
      )
    }) |> bindEvent(selected_var_type())

    # apply function events ---------------------------------------------------
    observe({
      if(!isTruthy(input$vars_sel)) {
        msg('Select a variable')
      } else if(!isTruthy(input$fun)){
        msg('Select a function')
      } else if(input$fun %in% dangerous_operations ||
                input$fun %notin% allowed_operations){
        msg('Function is not allowed')
      } else {
        if(is_name_available(input$txt_new_name_fun, df_names())) {

          running_modal()

          temp <- copy(get_act_dt(session))

          temp[, new_var := fun(var1), by = groupby, env = list(
            new_var = input$txt_new_name_fun,
            fun = input$fun,
            var1 = input$vars_sel,
            groupby = input$vars_groupby_fun |> as.list()
          )]

          update_act_dt(session, copy(temp), updated_cols = input$txt_new_name_fun)
          rm(temp)

          remove_running_modal()

          updateTextInput(session, 'txt_new_name_fun', value = '')

        } else {
          msg_error('New name is not valid or already in use')
        }

      }
    }) |> bindEvent(input$btn_apply_fun)

    # apply freehand events ---------------------------------------------------
    #  groupby vars -----------------------------------------------------------
    observe({
      req(df_names())
      updateSelectizeInput(
        session,
        'vars_groupby_free',
        choices = c('', df_names())
      )
    }) |> bindEvent(df_names())

    # calculate var -----------------------------------------------------------
    observe({

      if(is_name_available(input$txt_new_name_free, df_names())){

        # parse code ----------------------------------------------------------
        parsed_code <- try(parse_expr(input$txt_code_input), silent = T)

        if(inherits(parsed_code, 'try-error')){
          return(msg_error('Error to validate expression. Check code'))
        }

        # get operations and variables ----------------------------------------
        code_operations <- all.names(parsed_code, unique = T)
        code_vars <- all.vars(parsed_code)

        code_operations <- code_operations[!code_operations %in% code_vars]

        #  check variables and operations -------------------------------------
        if (!all(code_operations %in% allowed_operations)) {
          msg_error('Some operations are not allowed')
        } else if (!all(code_vars %in% c(df_names(), 'T', 'F'))) {
          msg_error('Some variables are not present in the dataset')
        } else {
          # create safe env for evaluation ------------------------------------
          running_modal()

          e1 <- safe_env(allowed_operations)

          # run code ----------------------------------------------------------
          e1$temp <- copy(get_act_dt(session))

          # start new variable
          e1$new_var_name <- input$txt_new_name_free
          e1$vars_groupby <- input$vars_groupby_free |> as.list()
          e1$parsed_code <- parsed_code

          e1$temp <- eval(expression(

            try(temp[, new_var_name := parsed_code,
                     by = groupby,
                     env = list(new_var_name = new_var_name,
                                parsed_code = parsed_code,
                                groupby = vars_groupby)],
                silent = TRUE)
            ),
            envir = e1
          )
          remove_running_modal()

          if(inherits(e1$temp, 'try-error')){
            return(msg_error('Error in expression. Check code'))
          } else{

            update_act_dt(session, copy(e1$temp), updated_cols = input$txt_new_name_free)

            rm(e1)

            updateTextInput(session, 'txt_new_name_free',
                            value = '', placeholder = 'new_var')

            updateTextAreaInput(session, 'txt_code_input', value = '')

          }
        }

      } else {
        msg_error('New name is not valid or already in use')
      }

    }) |> bindEvent(input$btn_apply_free)

    # show allowed operations -------------------------------------------------
    observe({
      show_allowed_op()
    }) |> bindEvent(input$btn_allowed_operations)

    # show help ---------------------------------------------------------------
    observe({
      operations_help(
        div(
          style = '
              border: 1px solid #d9e3f0;
              border-radius: 10px;
              padding: 20px;
              box-shadow: 0 2px 8px rgba(0,0,0,0.08);
            ',

          h4(icon('wand-magic-sparkles'), 'Examples of operations'),

          tags$p('Spada allows you to perform a variety of data manipulation tasks.'),

          tags$ul(
            tags$li(strong('Create variables:'), "as.character('Spada')"),
            tags$li(strong('Math operations:'), 'Sepal.Length + Petal.Width'),
            tags$li(strong('If else:'), 'fifelse(Petal.Length > 6, 99, -99)'),
            tags$li(strong('Apply functions:'), 'sum(Petal.Width)')
          )
        )
      )
    }) |> bindEvent(input$btn_help)

  })
}

# ui --------------------------------------------------------------------------
order_rows_ui <- function(id) {
  ns <- NS(id)

  card(
    card_header('Order Rows', class = 'mini-header'),
    card_body(
      selectizeInput(
        ns('vars_rows'),
        'Order by',
        NULL,
        multiple = T,
        options = list(plugins = list('remove_button', 'clear_button')),
        width = '80%'
      ),
      selectizeInput(
        ns('vars_descending'),
        list('Vars in descending order', bs_icon('info-circle')) |>
          ttip('If not informed, the order will be Ascending', PLACE = 'right'),
        '',
        multiple = T,
        options = list(plugins = list("remove_button", "clear_button")),
        width = '80%'
      ),
      radioButtons(ns('radio_nas'),
                   "NA's placement",
                   c('Last' = 'last', 'First' = 'first'),
                   inline = T),
    ),
    card_footer(
      btn_task(ns('btn_order_rows'), 'Order Rows', icon('shuffle')),
    )
  )
}

# server ----------------------------------------------------------------------
order_rows_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

    df_names <- reactive(get_act_dt(session) |> names())

    observe({
      req(df_names())
      updateSelectizeInput(
        session,
        'vars_rows',
        choices = c('', df_names())
      )
    }) |> bindEvent(df_names())

    # vars in descending order
    observe({
      updateSelectizeInput(
        session,
        'vars_descending',
        label = 'Descending Order',
        choices = input$vars_rows
      )
    }) |> bindEvent(input$vars_rows)

    # btn order rows ---------------------------
    observe({
      if(!isTruthy(input$vars_rows)){
        msg('Select at least one variable')
      } else {
        running_modal()
        rows_position <- rep(1, input$vars_rows |> length())

        rows_position[which(input$vars_rows %in% input$vars_descending)] <- -1

        temp <- copy(get_act_dt(session))

        setorderv(temp, cols = input$vars_rows,
                  order = rows_position,
                  na.last = if(input$radio_nas == 'last'){
                    TRUE
                  } else if (input$radio_nas == 'first'){
                    FALSE
                  } else { FALSE }
        )

        update_act_dt(session, copy(temp), FALSE)
        rm(temp)

        remove_running_modal()
      }
    }) |> bindEvent(input$btn_order_rows)

  })
}

# ui --------------------------------------------------------------------------
filter_rows_ui <- function(id) {
  ns <- NS(id)

  card(
    card_header('Filter Rows', class = 'mini-header'),
    card_body(
      radioGroupButtons(
        ns('filter_type'),
        'Filter Type',
        c(
          'One Variable' = 'one',
          'Two Variables' = 'two',
          'Sample' = 'sample',
          'Freehand' = 'free',
          'From Dataset' = 'dataset'
        ),
        selected = character(0),
        size = 'sm',
        individual = T
      ),

      # panel for one var filter ----------------------------------------------
      conditionalPanel(
        condition = "input.filter_type == 'one'",
        ns = ns,
        selectInput(ns('one_var_sel'), 'Variable', ''),
        selectInput(ns('one_var_operator'), 'Operator', c('', filter_operators)),
        uiOutput(ns('ui_one_var_value'))
      ),

      # panel for two var filter ----------------------------------------------
      conditionalPanel(
        condition = "input.filter_type == 'two'",
        ns = ns,
        selectInput(ns('two_var_sel1'), 'Variable 1', ''),
        selectInput(
          ns('two_var_operator'),
          'Operator',
          c('', equal_operators, compare_operators)
        ),
        selectInput(ns('two_var_sel2'), 'Variable 2', '')
      ),

      # panel for sample filter -----------------------------------------------
      conditionalPanel(
        condition = "input.filter_type == 'sample'",
        ns = ns,
        radioButtons(
          ns('sample_type'),
          NULL,
          c('Number of Rows' = 'rows', '% of Rows' = 'percent'),
          inline = T
        ),
        uiOutput(ns('ui_sample'))
      ),

      # panel for free filter -------------------------------------------------
      conditionalPanel(
        condition = "input.filter_type == 'free'",
        ns = ns,
        textAreaInput(
          ns('txt_code_input'),
          'Input Code',
          width = '800px',
          height = '200px',
          resize = 'both'
        ),
        fluidRow(
          column(4, btn_task(ns('btn_allowed_operations'), 'Show Allowed Operations')),
          column(4, btn_task(ns('btn_help'), 'Help', icon('question')))
        )
      ),

      # panel for other dataset -----------------------------------------------
      conditionalPanel(
        condition = "input.filter_type == 'dataset'",
        ns = ns,
        selectInput(ns('dt_1_var_sel'), 'Variable (active dataset)', ''),
        selectInput(
          ns('dt_var_operator'),
          'Operator',
          c('', in_operators)
        ),
        selectInput(ns('dt_dt_sel'), 'Dataset (other non active)', ''),
        selectInput(ns('dt_2_var_sel'), 'Variable (from other dataset)', '')
      ),
    ),
    card_footer(btn_task(ns('btn_filter'), 'Apply filters', icon('check')))
  )
}

# server ----------------------------------------------------------------------
filter_rows_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  df_names <- reactive(get_act_dt(session) |> names())

	  dts <- reactive({
	    req(session$userData$dt$act_name)
	    setdiff(session$userData$dt$dt |> names(), session$userData$dt$act_name)
	  })

    nrow_df_active <- reactive(get_act_dt(session) |> nrow())

    # variable inputs ---------------------------------------------------------
    observe({
      req(df_names())
      updateSelectInput(
        session,
        'one_var_sel',
        choices = c('', df_names())
      )
    }) |> bindEvent(df_names())

    observe({
      req(df_names())
      updateSelectInput(
        session,
        'two_var_sel1',
        choices = c('', df_names())
      )
    }) |> bindEvent(df_names())

    observe({
      req(df_names())
      updateSelectInput(
        session,
        'two_var_sel2',
        choices = c('', df_names())
      )
    }) |> bindEvent(df_names())

    # dataset filter var 1 ----------
    observe({
      req(df_names())
      updateSelectInput(
        session,
        'dt_1_var_sel',
        choices = c('', df_names())
      )
    }) |> bindEvent(df_names())

    # dataset filter dt names ----------
    observe({
      req(dts())
      updateSelectInput(
        session,
        'dt_dt_sel',
        choices = c('', dts())
      )
    }) |> bindEvent(dts())

    # dataset filter var 2 ----------
    observe({
      req(input$dt_dt_sel)
      updateSelectInput(
        session,
        'dt_2_var_sel',
        choices = c('', session$userData$dt$dt[[ input$dt_dt_sel ]] |> names())
      )
    }) |> bindEvent(input$dt_dt_sel)


    # selected column type ----------------------------------------------------
    col_type_one_var <- reactive({
      req(input$one_var_sel)
      get_act_dt(session)[[input$one_var_sel]] |> obj_type()
    })

    col_type_two_var1 <- reactive({
      req(input$two_var_sel1)
      get_act_dt(session)[[input$two_var_sel1]] |> obj_type()
    })

    # render UI for sample type -----------------------------------------------
    output$ui_sample <- renderUI({
      req(input$sample_type)
      tagList(
        if (input$sample_type == 'rows') {
          numericInput(
            ns('n_rows'), 'Number of Rows',
            value = (nrow_df_active()/2) |> ceiling(),
            min = 1, max = nrow_df_active()
          )
        } else if (input$sample_type == 'percent') {
          sliderInput(ns('sample_size'), 'Sample Size (%)', 1, 100, 50)
        },
        checkboxInput(ns('x_sample_replace'), 'Replace') |>
          ttip('Allow sample with replacement')
      )
    })

    # render UI for value input dynamically -----------------------------------
    observe({
      req(input$one_var_sel, input$one_var_operator, col_type_one_var())
      output$ui_one_var_value <- renderUI({
        if (col_type_one_var() == 'date'){
          if(input$one_var_operator %in% c(equal_operators, compare_operators)){
            dateInput(ns('one_var_value'), 'Date')
          } else if (input$one_var_operator %in% between_operators){
            dateRangeInput(ns('one_var_value'), 'Date', end = Sys.Date() + 30)
          } else if (input$one_var_operator %in% in_operators){
            tagList(
              layout_column_wrap(
                dateInput(ns('one_var_value'), 'Date'),
                textInput(ns('preview_value'), 'Preview Values')
              ),
              layout_column_wrap(
                btn_task(ns('btn_insert_value'), 'Insert Value', icon('plus')),
                btn_task(ns('btn_clear_value'), 'Clear All Values', icon('xmark'))
              )
            )
          }
        } else if (col_type_one_var() %in% c('double', 'integer')){
          if(input$one_var_operator %in% c(equal_operators, compare_operators)){
            numericInput(ns('one_var_value'), 'Value', value = 0)
          } else if (input$one_var_operator %in% between_operators){
            tagList(
              numericInput(ns('one_var_value'), 'Inicial Value', value = 0),
              numericInput(ns('one_var_value2'), 'Final Value', value = 0)
            )
          } else if (input$one_var_operator %in% in_operators){
            tagList(
              layout_column_wrap(
                numericInput(ns('one_var_value'), 'Value', value = 0),
                textInput(ns('preview_value'), 'Preview Values')
              ),
              layout_column_wrap(
                btn_task(ns('btn_insert_value'), 'Insert Value', icon('plus')),
                btn_task(ns('btn_clear_value'), 'Clear All Values', icon('xmark'))
              )
            )
          }
        } else if (col_type_one_var() == 'factor'){
          if(input$one_var_operator %in% c(equal_operators, in_operators)){
            selectizeInput(
              ns('one_var_value'),
              'Value',
              choices = get_act_dt(session)[[input$one_var_sel]] |> levels(),
              multiple = T,
              options = list(create = T)
            )
          }
        } else if (col_type_one_var() %in% c('char', 'complex')){
          if(input$one_var_operator %in% c(equal_operators, in_operators)){
            selectizeInput(
              ns('one_var_value'),
              list('Value', bs_icon("info-circle")) |>
                ttip(PLACE = 'right', 'Text should not be in quotes'),
              choices = NULL,
              multiple = T,
              options = list(create = T)
            )
          }
        } else if (col_type_one_var() == 'logical'){
          div()
        } else if(isTruthy(input$one_var_sel) && isTruthy(input$one_var_operator)){
          selectizeInput(
            ns('one_var_value'),
            list('Value', bs_icon("info-circle")) |>
              ttip(PLACE = 'right', 'Text should not be in quotes'),
            choices = NULL,
            multiple = T,
            options = list(create = T)
          )
        }
      })
    }) |> bindEvent(input$one_var_sel, input$one_var_operator)

    # update selectinput to show pertinent operators --------------------------
    observe({
      req(input$one_var_sel, col_type_one_var())
      updateSelectInput(
        session, 'one_var_operator',
        choices =
          if(col_type_one_var() %in% c('factor', 'char', 'complex')){
            c('',
              equal_operators,
              na_operators,
              in_operators)
          } else if (col_type_one_var() == 'logical'){
            c('', logical_operators)
          } else if (col_type_one_var() == 'date'){
            c('', filter_operators[filter_operators %notin% logical_operators])
          } else if (col_type_one_var() %in% c('double', 'integer')){
            c('', filter_operators[filter_operators %notin% logical_operators])
          } else { c('', filter_operators) }
      )
    }) |> bindEvent(input$one_var_sel)

    # update selectinput two vars filter to show pertinent operators ----------
    observe({
      req(input$two_var_sel1)
      updateSelectInput(
        session, 'two_var_operator',
        choices =
          if(col_type_two_var1() %in% c('factor', 'char', 'complex', 'logical')){
            c('', equal_operators)
          } else if (col_type_two_var1() %in% c('date', 'double', 'integer')){
            c('', equal_operators, compare_operators) }
      )
    }) |> bindEvent(input$two_var_sel1)

    # temporary values for multi-inputs ---------------------------------------
    value_temp <- reactiveValues(value_temp_inserted = NULL)

    # insert values -----------------------------------------------------------
    observe({
      req(input$one_var_value)
      if(is.null(value_temp$value_temp_inserted)){
        value_temp$value_temp_inserted <- input$one_var_value
      } else {
        value_temp$value_temp_inserted <- c(value_temp$value_temp_inserted,
                                            input$one_var_value)
      }

      updateTextInput(session, 'preview_value',
                      value = value_temp$value_temp_inserted)
    }) |> bindEvent(input$btn_insert_value)

    # clear inserted values
    observe({
      value_temp$value_temp_inserted <- NULL

      updateTextInput(session, 'preview_value', value = '')
    }) |> bindEvent(input$btn_clear_value)

    # apply btn filter rows ---------------------------------------------------
    observe({
      if(!isTruthy(input$filter_type)){
        msg('Select the Filter type')
        return()
      } else {
        # filter events for one variable --------------------------------------
        if (input$filter_type == 'one') {
          # test if var and operator were informed
          if (!isTruthy(input$one_var_sel)) {
            msg('Select a variable')
            return()
          } else if (!isTruthy(input$one_var_operator)) {
            msg('Select an operator')
            return()
          } else if (!isTruthy(input$one_var_value) &
                     input$one_var_operator %notin%
                     c(na_operators, logical_operators, outlier_operators)) {
            msg('Insert a value')
            return()
          } else if (input$one_var_operator %in% between_operators) {
            if (col_type_one_var() %in% c('double', 'integer') &
                !isTruthy(input$one_var_value2)) {
              msg('Inform inicial and final values')
              return()
            } else if (col_type_one_var() == 'date' &
                       (!isTruthy(input$one_var_value[1]) |
                        !isTruthy(input$one_var_value[2]))) {
              msg('Inform inicial and final dates')
              return()
            }
          } else if (input$one_var_operator %in% in_operators &
                     col_type_one_var() %in% c('date', 'numeric') &
                     is.null(value_temp$value_temp_inserted)) {
            msg('Insert values')
            return()
          }

          # use inserted values
          if (input$one_var_operator %in% between_operators) {
            if (col_type_one_var() %in% c('double', 'integer')) {
              value_temp$value_temp <- c(input$one_var_value, input$one_var_value2)
            } else if (col_type_one_var() == 'date') {
              value_temp$value_temp <- input$one_var_value
            }
          } else if (input$one_var_operator %in% in_operators &
                     col_type_one_var() %notin% c('char', 'complex', 'factor')) {
            value_temp$value_temp <- value_temp$value_temp_inserted
          } else {
            value_temp$value_temp <- input$one_var_value
          }
          value_temp$len <- value_temp$value_temp |> length()

          # pass values to filter function
          if (input$one_var_operator %in%
              c(na_operators, logical_operators, outlier_operators)) {
            temp <- copy(get_act_dt(session))
            temp <- filter_rows(temp, input$one_var_sel, input$one_var_operator, NULL)
          } else if (value_temp$len > 1 & input$one_var_operator %in%
                     c(equal_operators, compare_operators)) {
            msg('Operator requires value of length 1')
            return()
          } else {

            temp <- copy(get_act_dt(session))
            temp <- filter_rows(temp,
                                input$one_var_sel,
                                input$one_var_operator,
                                value_temp$value_temp)
          }

          # reset value_temp
          value_temp$value_temp_inserted <- NULL
          value_temp$value_temp <- NULL
          value_temp$len <- NULL

          # clear value after click in button
          updateSelectInput(session, 'var', selected = '')
          updateSelectInput(session, 'operator', selected = '')

          # filter events for 2 variables -------------------------------------
        } else if (input$filter_type == 'two') {
          temp <- copy(get_act_dt(session))
          if (!isTruthy(input$two_var_sel1) || !isTruthy(input$two_var_sel2)) {
            msg('Select 2 variables')
            return()
          } else if (!isTruthy(input$two_var_operator)) {
            msg('Select an operator')
            return()
          } else if (temp[[input$two_var_sel1]] |> obj_type() !=
                     temp[[input$two_var_sel2]] |> obj_type()) {
            msg('Variables must be of the same type')
            return()
          } else {

            temp <- filter_rows_2vars(temp,
                                      input$two_var_sel1,
                                      input$two_var_sel2,
                                      input$two_var_operator)
          }

          # filter events for sample ------------------------------------------
        } else if (input$filter_type == 'sample') {
          if (input$sample_type == 'rows') {
            if (!isTruthy(input$n_rows) ||
                !between(input$n_rows, 1, nrow_df_active())) {
              msg(paste('Number of rows must be between 1 and', nrow_df_active()))
              return()
            } else {

              temp <- copy(get_act_dt(session))
              temp <- temp[sample(1:nrow_df_active(),
                                  input$n_rows,
                                  replace = input$x_sample_replace), ]
            }

          } else if (input$sample_type == 'percent') {

            temp <- copy(get_act_dt(session))
            temp <- temp[sample(
              1:nrow_df_active(),
              input$sample_size / 100 * nrow_df_active(),
              replace = input$x_sample_replace
            ), ]
          }

          # filter events for freehand ----------------------------------------
        } else if (input$filter_type == 'free') {

          # parse code --------------------------------------------------------
          parsed_code <- try(parse_expr(input$txt_code_input), silent = T)

          if(inherits(parsed_code, "try-error")){
            return(msg_error('Error to validate expression. Check code'))
          }

          # get operations and variables --------------------------------------
          code_operations <- all.names(parsed_code, unique = T)
          code_vars <- all.vars(parsed_code)

          code_operations <- code_operations[!code_operations %in% code_vars]

          #  check variables and operations -----------------------------------
          if (!all(code_operations %in% allowed_operations)) {
            msg_error('Some operations are not allowed')
            return()
          } else if (!all(code_vars %in% c(df_names(), 'T', 'F'))) {
            msg_error('Some variables are not present in the dataset')
            return()
          } else {

            # create safe env for evaluation ----------------------------------
            e1 <- safe_env(allowed_operations)

            # run code --------------------------------------------------------
            temp <- copy(get_act_dt(session))
            e1$temp <- copy(temp)
            e1$parsed_code <- parsed_code

            e1$temp <- eval(expression(

              try(temp[parsed_code, , env = list(parsed_code = parsed_code)],
                  silent = TRUE)
              ),
              envir = e1
            )

            if(inherits(e1$temp, 'try-error')){
              return(msg_error('Error in expression. Check code'))
            } else{
              temp <- copy(e1$temp)
              rm(e1)

              updateTextAreaInput(session, 'txt_code_input', value = '')
            }
          }

          # filter dataset -------------------------------------
        } else if (input$filter_type == 'dataset') {
          if (!isTruthy(input$dt_1_var_sel) || !isTruthy(input$dt_2_var_sel)) {
          msg('Select 2 variables')
          return()
          } else if (!isTruthy(input$dt_dt_sel)) {
            msg('Select a dataset')
            return()
          } else if (!isTruthy(input$dt_var_operator)) {
            msg('Select an operator')
            return()
          }
          # apply filter -----
          temp <- copy(get_act_dt(session))
          temp <- filter_rows(
            temp,
            input$dt_1_var_sel,
            input$dt_var_operator,
            session$userData$dt$dt[[ input$dt_dt_sel ]][[ input$dt_2_var_sel ]] |>
              unique()
          )
        }

        if(!is_spada_df(temp)){
          abort_filter_modal()
        } else {
          running_modal()
          update_act_dt(session, copy(temp))
          remove_running_modal()
        }

        rm(temp)

      }
    }) |> bindEvent(input$btn_filter)

    # show allowed operations -------------------------------------------------
    observe({
      show_allowed_op()
    }) |> bindEvent(input$btn_allowed_operations)

    # show help ---------------------------------------------------------------
    observe({
      operations_help(
        div(
          style = '
              border: 1px solid #d9e3f0;
              border-radius: 10px;
              padding: 20px;
              box-shadow: 0 2px 8px rgba(0,0,0,0.08);
            ',

          h4(icon('wand-magic-sparkles'), 'Examples of operations'),

          tags$p('Spada allows you to perform a variety of filter tasks.'),

          tags$ul(
            tags$li(strong('Compare Values:'), 'Sepal.Length > 6'),
            tags$li(strong('Compare Variables:'), 'Sepal.Length == Sepal.Width * 2'),
            tags$li(strong('Apply functions:'), 'is.na(Sepal.Length)')
          )
        )
      )
    }) |> bindEvent(input$btn_help)

  })
}

# Function with the server of spada.R
spada_server <- function(datasets, conf, run_local = TRUE){
  function(input, output, session) {

    previous_shiny.maxRequestSize <- getOption('shiny.maxRequestSize')
    options(shiny.maxRequestSize = conf$file_size * 1024 ^ 2)

    onStop(function(){
      options(shiny.maxRequestSize = previous_shiny.maxRequestSize)
      daemons(0)
    })

    session$onFlushed(function(){
      waiter_hide()
    })

    session$userData$run_local <- run_local
    session$onSessionEnded(function(){
      if(session$userData$run_local) stopApp()
    })

    # conf values -------------------------------------------------------------
    session$userData$conf <- reactiveValues(
      no_input_data = conf$no_input_data,
      conf_dir = conf$conf_dir,
      data_dir = conf$data_dir,
      theme = conf$theme,
      file_size = conf$file_size,
      restore_session = conf$restore_session,
      save_session = conf$save_session,
      restore_data_status = conf$restore_data_status,
      restore_output_status = conf$restore_output_status,
      restore_status = conf$restore_status,
      plot_fill_color = conf$plot_fill_color,
      plot_line_color = conf$plot_line_color,
      plot_title_color = conf$plot_title_color,
      plot_limit = conf$plot_limit,
      run_local = run_local
    )

    # restore session module --------------------------------------------------
    restore_session_server('restore_session')

    # data --------------------------------------------------------------------
    session$userData$dt <- reactiveValues(
      dt = datasets,
      act_name = names(datasets[1]),
      bkp0 = NULL,
      bkp = NULL,
      meta = lapply(datasets, df_info),
      updated_cols = NULL,
      data_changed_type = 'new_data'
    )
    session$userData$out <- reactiveValues(elements = list())

    session$userData$data_changed <- reactiveVal(0)
    session$userData$out_edit_trigger <- reactiveVal(NULL)

    session$userData$dt$bkp0 <- isolate(get_act_dt(session))

    session$userData$dt_names <- reactive({
      req(session$userData$dt$dt)
      names(session$userData$dt$dt)
    })

    output$dt_act_name <- renderText(short_name(session$userData$dt$act_name))

    # datasets metadata -------------------------------------------------------
    observe({
      req(session$userData$dt$dt, session$userData$data_changed() > 0)

      if(is.null(session$userData$dt$updated_cols) &&
         session$userData$dt$data_changed_type == 'new_data'){

        # recalculate meta for all cols
        session$userData$dt$meta[[session$userData$dt$act_name]] <- get_act_dt(session) |>
          df_info()
      } else {

        # update name of vars
        if(session$userData$dt$data_changed_type == 'rename_cols'){

          session$userData$dt$meta[[session$userData$dt$act_name]] <- update_meta(
            previous_meta = session$userData$dt$meta[[session$userData$dt$act_name]],
            col_names = get_act_dt(session) |> names(),
            change_type = session$userData$dt$data_changed_type
          )
          # keep meta of selected vars
        } else if (session$userData$dt$data_changed_type == 'select_cols'){

          session$userData$dt$meta[[session$userData$dt$act_name]] <- update_meta(
            previous_meta = session$userData$dt$meta[[session$userData$dt$act_name]],
            col_names = get_act_dt(session) |> names(),
            ncols = get_act_dt(session) |> fncol(),
            change_type = session$userData$dt$data_changed_type
          )

          # update order of vars in meta df
        } else if (session$userData$dt$data_changed_type == 'order_cols'){

          session$userData$dt$meta[[session$userData$dt$act_name]] <- update_meta(
            previous_meta = session$userData$dt$meta[[session$userData$dt$act_name]],
            col_names = get_act_dt(session) |> names(),
            change_type = session$userData$dt$data_changed_type
          )

        } else if (session$userData$dt$data_changed_type %in% c('convert_cols', 'new_data')){

          # update meta of changed variables
          session$userData$dt$meta[[session$userData$dt$act_name]] <- update_meta(
            dt = get_act_dt(session)[, .SD, .SDcols = session$userData$dt$updated_cols],
            previous_meta = session$userData$dt$meta[[session$userData$dt$act_name]],
            col_names = get_act_dt(session) |> names(),
            updated_cols = session$userData$dt$updated_cols,
            ncols = get_act_dt(session) |> fncol(),
            change_type = session$userData$dt$data_changed_type
          )
        }
        session$userData$dt$updated_cols <- NULL
        session$userData$dt$data_changed_type = 'new_data'
      }

    }) |> bindEvent(session$userData$data_changed())

    session$userData$dt$gt_info <- reactive({
      req(session$userData$dt$meta)

      Map(gt_info, session$userData$dt$meta,
          df_name = names(session$userData$dt$meta))
    })

    # df active metadata ---------------------
    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$meta)

      session$userData$dt$meta[[session$userData$dt$act_name]]
    })

    # info to use in sidebar and navbar modules --------
    session$userData$dt$act_mini_meta <- reactive({
      req(session$userData$dt$act_meta())

      dt_act_meta <- session$userData$dt$act_meta()

      list(
        'row_col' = paste(
          f_int(dt_act_meta[['rows']][1]),
          '/',
          f_int(dt_act_meta[['cols']][1])
        ),
        'col_nas' = dt_act_meta[n_nas > 0, ] |> fnrow()
      )
    })

    # navbar ------------------------------------------------------------------
    navbar_df_info_server('navbar_df_info', app_session = session)

    # side bar ----------------------------------------------------------------
    sidebar_server('sidebar', app_session = session)

    # metadata ----------------------------------------------------------------
    metadata_server('pD_metadata')

    # overview -------------------------------
    data_overview_server('pD_overview')

    # values for boxes -----------------------
    data_highlights_server('pD_highlights')

    # manage datasets ------------------------
    data_server('pD_data')

    # export file ----------------------------
    export_file_server('pD_export')

    # import file ----------------------------
    import_file_server('pD_import')

    # edit page events --------------------------------------------------------

    # calculate cols events ------------------
    calculate_cols_server('pE_calculate_cols')

    # convert events -------------------------
    convert_cols_server('pE_convert_cols')

    # duplicates events -------------------------
    duplicates_server('pE_duplicates')

    # filter events --------------------------
    filter_rows_server('pE_filter_rows')

    # select cols ----------------------------
    select_cols_server('pE_filter_sel_cols')

    # join datasets --------------------------
    join_server('pE_join')

    # missing data events --------------------
    missing_server('pE_missing')

    # order rows events ----------------------
    order_rows_server('pE_order_rows')

    # order cols events ----------------------
    order_cols_server('pE_order_cols')

    # rename cols events ---------------------
    rename_cols_server('pE_rename_cols')

    # summarise events -----------------------
    groupby_server('pE_groupby')

    summarise_server('pE_summarise')

    # edit backup events ---------------------
    data_bkp_server('pE_bkp')

    # analysis page events ----------------------------------------------------
    exploratory_server('pA_exploratory')

    # descriptive stats ----------------------
    descriptive_stats_server('pA_desc_stats')

    # correlation ----------------------------
    cor_matrix_server('pA_cor_matrix')
    correlation_server('pA_correlation')

    # normality test -------------------------
    normality_test_server('pA_normality_test')

    # stats test -----------------------------
    one_t_test_server('pA_t_test')
    z_test_server('pA_z_test')

    # linear model ---------------------------
    lm_server('pA_lm')

    # output events -----------------------------------------------------------
    output_server('pO_output')

    # config events -----------------------------------------------------------
    config_server('pC', app_session = session)

    # about events ------------------------------------------------------------
    about_spada_server('about_spada')

    # exit app event ----------------------------------------------------------
    observe({
      req(input$navbar)
      if (input$navbar == 'exit') {

        if (!session$userData$run_local ||
            session$userData$conf$save_session == 'never'){
          exit_without_save(session)
        } else if(session$userData$conf$save_session == 'ask'){
          # ask to save data and ouput
          showModal(modalDialog(
            title = div(icon('save'), 'Save Session'),
            'Do you want to save Data and Output objects?',
            easyClose = FALSE,
            size = 'm',
            footer = tagList(
              actionButton('btn_cancel_exit', 'Cancel', icon = icon('xmark'), class = 'btn-task btn-task-neutral'),
              actionButton('btn_cancel_save_session', 'No', icon = icon('xmark'), class = 'btn-task btn-task-cancel'),
              actionButton('btn_confirm_save_session', 'Yes', icon = icon('check'), class = 'btn-task')
            )
          ))
        } else if (session$userData$conf$save_session == 'always'){
          exit_with_save(session)
        }
      }
    })

    # cancel exit ----------------------------
    observe({
      removeModal()
      nav_select('navbar', selected = 'Data', session = session)
      nav_select('navset_card_pill_data', selected = 'Highlights', session = session)
    }) |> bindEvent(input$btn_cancel_exit)

    # cancel save session --------------------
    observe({
      removeModal()
      exit_without_save(session)
    }) |> bindEvent(input$btn_cancel_save_session)

    # confirm save session -------------------
    observe({
      removeModal()
      exit_with_save(session)
    }) |> bindEvent(input$btn_confirm_save_session)

  }
}

# tests/testthat/test-restore_session_module.R

dfs <- list('df1' = data.table(x = 1:5),
            'df2' = data.table(y = letters[1:20]))

temp_dir <- tempdir()
dir.create(file.path(temp_dir, 'data'))

start_conf <- list(
  'no_input_data' = TRUE,
  'conf_dir' = file.path(temp_dir, 'conf'),
  'data_dir' = file.path(temp_dir, 'data'),
  'theme' = 'spada_theme',
  'file_size' = 1000,
  'restore_session' = 'always',
  'save_session' = 'ask',
  'restore_data_status' = 0,
  'restore_output_status' = 0,
  'restore_status' = NULL,
  'plot_fill_color' = plot_fill_color,
  'plot_line_color' = plot_line_color
)

# test restore session - no data no output ------------------------------------
test_that('Test restore session no data and no output ', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    restore_session_server('restore_session')

    session$flushReact()
    expect_equal(session$userData$conf$restore_data_status, 2)
    expect_equal(session$userData$conf$restore_output_status, 2)
    expect_equal(session$userData$conf$restore_status, '2.2')

  })
})

# test restore session - load data and output ---------------------------------
output_for_test <- list('id_1' = gen_output())

qs_save(dfs, file.path(temp_dir, 'data', 'data.qs2'))
qs_save(output_for_test, file.path(temp_dir, 'data', 'output.qs2'))

test_that('Test restore session load data and output ', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    restore_session_server('restore_session')

    session$flushReact()
    expect_equal(session$userData$conf$restore_data_status, 1)
    expect_equal(session$userData$conf$restore_output_status, 1)
    expect_equal(session$userData$conf$restore_status, '1.1')

    expect_equal(session$userData$dt$dt |> names(), c('df1', 'df2'))
    expect_equal(session$userData$dt$dt, dfs)
    expect_equal(session$userData$out$elements[[1]]$title,
                 output_for_test[[1]]$title)
    expect_equal(session$userData$out$elements[[1]]$annotation,
                 output_for_test[[1]]$annotation)
    expect_equal(session$userData$out$elements[[1]]$element,
                 output_for_test[[1]]$element)
    expect_equal(session$userData$out$elements[[1]]$card,
                 output_for_test[[1]]$card)
    expect_equal(session$userData$out$elements[[1]]$btn_x |> class(),
                 output_for_test[[1]]$btn_x |> class())
    expect_s3_class(session$userData$out$elements[[1]]$btn_x, 'shiny.tag')
    expect_equal(session$userData$out$elements[[1]]$btn_e |> class(),
                 output_for_test[[1]]$btn_e |> class())
    expect_s3_class(session$userData$out$elements[[1]]$btn_e, 'shiny.tag')
  })
})

# test restore session - invalid format data and output -----------------------
qs_save(list(1:5), file.path(temp_dir, 'data', 'data.qs2'))
qs_save(list(letters), file.path(temp_dir, 'data', 'output.qs2'))

test_that('Test restore session load data and output ', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    restore_session_server('restore_session')

    session$flushReact()
    expect_equal(session$userData$conf$restore_data_status, 3)
    expect_equal(session$userData$conf$restore_output_status, 3)
    expect_equal(session$userData$conf$restore_status, '3.3')
  })
})

# tests/testthat/test-output_module.R

out_el <- list(
  'id_1' = gen_output(element = div(h2('Test'))),
  'id_2' = gen_output(element = iris |> gt())
)

# test elements ---------------------------------------------------------------
test_that('Test output structure', {
  testServer(output_server, {

    session$userData$out <- reactiveValues(elements = out_el)

    expect_length(session$userData$out$elements, 2)
    expect_true(inherits(session$userData$out$elements[[1]], 'list'))
    expect_true(inherits(session$userData$out$elements[[2]], 'list'))

    # test id
    expect_true(inherits(session$userData$out$elements[[1]]$id, 'character'))
    expect_true(inherits(session$userData$out$elements[[2]]$id, 'character'))

    # test title
    expect_true(inherits(session$userData$out$elements[[1]]$title, 'character'))
    expect_true(inherits(session$userData$out$elements[[2]]$title, 'character'))

    # test annotation
    expect_true(inherits(session$userData$out$elements[[1]]$annotation, 'character'))
    expect_true(inherits(session$userData$out$elements[[2]]$annotation, 'character'))

    # test element
    expect_true(inherits(session$userData$out$elements[[1]]$element, c('shiny.tag', 'gt_tbl')))
    expect_true(inherits(session$userData$out$elements[[2]]$element, c('shiny.tag', 'gt_tbl')))

    # test card
    expect_true(inherits(session$userData$out$elements[[1]]$card, 'shiny.tag'))
    expect_true(inherits(session$userData$out$elements[[2]]$card, 'shiny.tag'))

    # test btn_x
    expect_true(inherits(session$userData$out$elements[[1]]$btn_x, 'shiny.tag'))
    expect_true(inherits(session$userData$out$elements[[2]]$btn_x, 'shiny.tag'))

    # test btn_e
    expect_true(inherits(session$userData$out$elements[[1]]$btn_e, 'shiny.tag'))
    expect_true(inherits(session$userData$out$elements[[2]]$btn_e, 'shiny.tag'))

  })
})

# test btn reset --------------------------------------------------------------
test_that('Reset button clears output', {
  testServer(output_server, {

    session$userData$out <- reactiveValues(elements = out_el)
    session$userData$out_edit_trigger <- reactiveVal(NULL)

    expect_length(session$userData$out$elements, 2)
    session$setInputs(btn_reset = 1)

    expect_length(session$userData$out$elements, 2)
    session$setInputs(btn_confirm_reset = 1)

    expect_length(session$userData$out$elements, 0)
    expect_true(inherits(session$userData$out$elements, 'list'))
  })
})

# test save output ------------------------------------------------------------
test_that('Save output stores qs2 file', {
  testServer(output_server, {

    tmpdir <- tempdir()

    session$userData$conf <- reactiveValues(data_dir = tmpdir)
    session$userData$out <- reactiveValues(elements = out_el)
    session$userData$run_local <- TRUE
    session$userData$out_edit_trigger <- reactiveVal(NULL)

    session$setInputs(btn_save_output_session = 1)
    session$setInputs(btn_confirm_save_output = 1)

    saved_file <- file.path(tmpdir, 'output.qs2')
    expect_true(file.exists(saved_file))

    saved_content <- qs_read(saved_file)
    expect_true(is.list(saved_content))
    expect_length(saved_content, 2)
    expect_true(inherits(saved_content[[1]], 'list'))
    expect_true(inherits(saved_content[[2]], 'list'))

    # test id
    expect_true(inherits(saved_content[[1]]$id, 'character'))
    expect_true(inherits(saved_content[[2]]$id, 'character'))

    # test title
    expect_true(inherits(saved_content[[1]]$title, 'character'))
    expect_true(inherits(saved_content[[2]]$title, 'character'))

    # test annotation
    expect_true(inherits(saved_content[[1]]$annotation, 'character'))
    expect_true(inherits(saved_content[[2]]$annotation, 'character'))

    # test element
    expect_true(inherits(saved_content[[1]]$element, c('shiny.tag', 'gt_tbl')))
    expect_true(inherits(saved_content[[2]]$element, c('shiny.tag', 'gt_tbl')))

    # test card
    expect_true(inherits(saved_content[[1]]$card, 'shiny.tag'))
    expect_true(inherits(saved_content[[2]]$card, 'shiny.tag'))

    # test btn_x
    expect_true(inherits(saved_content[[1]]$btn_x, 'shiny.tag'))
    expect_true(inherits(saved_content[[2]]$btn_x, 'shiny.tag'))

    # test btn_e
    expect_true(inherits(saved_content[[1]]$btn_e, 'shiny.tag'))
    expect_true(inherits(saved_content[[2]]$btn_e, 'shiny.tag'))

  })
})

# test import output ----------------------------------------------------------
test_that('Import output loads qs2 file correctly', {
  testServer(output_server, {

    tmpdir <- tempdir()
    qs_save(out_el, file.path(tmpdir, 'output.qs2'))

    session$userData$conf <- reactiveValues(data_dir = tmpdir)
    session$userData$out <- reactiveValues(elements = list())
    session$userData$run_local <- TRUE
    session$userData$out_edit_trigger <- reactiveVal(NULL)

    session$setInputs(btn_import_output_session = 1)
    session$setInputs(btn_confirm_import_output = 1)

    expect_length(session$userData$out$elements, 2)
    expect_equal(
      grepl(pattern = 'id_', session$userData$out$elements),
      grepl(pattern = 'id_', out_el)
    )

    expect_true(inherits(session$userData$out$elements[[1]], 'list'))
    expect_true(inherits(session$userData$out$elements[[2]], 'list'))

    # test id
    expect_true(inherits(session$userData$out$elements[[1]]$id, 'character'))
    expect_true(inherits(session$userData$out$elements[[2]]$id, 'character'))

    # test title
    expect_true(inherits(session$userData$out$elements[[1]]$title, 'character'))
    expect_true(inherits(session$userData$out$elements[[2]]$title, 'character'))

    # test annotation
    expect_true(inherits(session$userData$out$elements[[1]]$annotation, 'character'))
    expect_true(inherits(session$userData$out$elements[[2]]$annotation, 'character'))

    # test element
    expect_true(inherits(session$userData$out$elements[[1]]$element, c('shiny.tag', 'gt_tbl')))
    expect_true(inherits(session$userData$out$elements[[2]]$element, c('shiny.tag', 'gt_tbl')))

    # test card
    expect_true(inherits(session$userData$out$elements[[1]]$card, 'shiny.tag'))
    expect_true(inherits(session$userData$out$elements[[2]]$card, 'shiny.tag'))

    # test btn_x
    expect_true(inherits(session$userData$out$elements[[1]]$btn_x, 'shiny.tag'))
    expect_true(inherits(session$userData$out$elements[[2]]$btn_x, 'shiny.tag'))

    # test btn_e
    expect_true(inherits(session$userData$out$elements[[1]]$btn_e, 'shiny.tag'))
    expect_true(inherits(session$userData$out$elements[[2]]$btn_e, 'shiny.tag'))
  })
})

# test download html ----------------------------------------------------------
test_that('Download HTML ', {
  testServer(output_server, {

    session$userData$out <- reactiveValues(elements = out_el)
    session$userData$out_edit_trigger <- reactiveVal(NULL)

    expect_equal(
      output$btn_save_html |> basename() |> substr(1, 15),
      paste0(
        'spada_output_',
        format(Sys.time(), format = '%Y%m%d%H%M%S'),
        '.html') |> substr(1, 15)
    )
  })
})

# test printable output -------------------------------------------------------
test_that('Test output - printable output', {
  testServer(output_server, {

    session$userData$out <- reactiveValues(elements = out_el)
    session$userData$out_edit_trigger <- reactiveVal(NULL)

    session$setInputs(sel_show_output = Inf)

    # wait for the extended_task
    while (task_printable_out$status() == 'running') {
      session$flushReact()
    }

    expect_equal(task_printable_out$status(), 'success')
    expect_s3_class(printable_output()[[1]], c('shiny.tag'))

    expect_true(
      all(
        sapply(
          printable_output(),
          \(x){ class(x) == 'shiny.tag' }
        )
      )
    )
  })
})

# test edit output element ----------------------------------------------------
dfs <- list('df_iris' = iris |> as.data.table(),
            'df_mtcars' = mtcars |> as.data.table())

temp_dir <- tempdir()

start_conf <- list(
  'conf_dir' = file.path(temp_dir, 'conf'),
  'data_dir' = file.path(temp_dir, 'data'),
  'theme' = 'spada_theme',
  'file_size' = 1000,
  'restore_session' = 'never',
  'save_session' = 'ask',
  'restore_data_status' = 0,
  'restore_output_status' = 0,
  'restore_status' = NULL,
  'plot_fill_color' = plot_fill_color,
  'plot_line_color' = plot_line_color
)

test_that('Test output - edit element', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    session$userData$out <- reactiveValues(elements = list())
    session$userData$out_edit_trigger <- reactiveVal(NULL)

    out_el <- div(h2('Element'))

    insert_output_server('insert_output', reactive(out_el))
    output_server('output')

    # insert output
    session$setInputs(
      'insert_output-output_title' = 'Title 1',
      'insert_output-output_annot' = 'Annotation 1',
      'insert_output-btn_add_output' = 1,
      'insert_output-btn_confirm_add_output' = 1
    )

    expect_length(session$userData$out$elements, 1)
    expect_equal(session$userData$out$elements[[ 1 ]]$title, 'Title 1')
    expect_equal(session$userData$out$elements[[ 1 ]]$annotation, 'Annotation 1')

    # trigger btn edit
    id <- session$userData$out$elements[[ 1 ]]$id
    edit_btn <- list()
    edit_btn[[paste0('insert_output-btn_eout_', id)]] = 1
    do.call(session$setInputs, edit_btn)

    # edit output
    session$setInputs(
      'output-output_edit_title' = 'New title',
      'output-output_edit_annot' = 'New annotation',
      'output-btn_confirm_edit_output' = 1
    )

    expect_type(session$userData$out$elements, 'list')
    expect_length(session$userData$out$elements, 1)
    expect_equal(session$userData$out$elements[[ 1 ]]$title, 'New title')
    expect_equal(session$userData$out$elements[[ 1 ]]$annotation, 'New annotation')
  })
})


# test delete output element --------------------------------------------------
test_that('Test output - edit element', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    session$userData$out <- reactiveValues(elements = list())
    session$userData$out_edit_trigger <- reactiveVal(NULL)

    out_el <- div(h2('Element'))

    insert_output_server('insert_output', reactive(out_el))
    output_server('output')

    # insert output
    session$setInputs(
      'insert_output-output_title' = 'Title 1',
      'insert_output-output_annot' = 'Annotation 1',
      'insert_output-btn_add_output' = 1,
      'insert_output-btn_confirm_add_output' = 1
    )

    expect_length(session$userData$out$elements, 1)
    expect_equal(session$userData$out$elements[[ 1 ]]$title, 'Title 1')
    expect_equal(session$userData$out$elements[[ 1 ]]$annotation, 'Annotation 1')

    # trigger btn edit
    id <- session$userData$out$elements[[ 1 ]]$id
    delete_btn <- list()
    delete_btn[[paste0('insert_output-btn_xout_', id)]] = 1
    do.call(session$setInputs, delete_btn)

    expect_type(session$userData$out$elements, 'list')
    expect_length(session$userData$out$elements, 0)
  })
})

# ui --------------------------------------------------------------------------
convert_cols_ui <- function(id) {
  ns <- NS(id)

  layout_column_wrap(
    card(
      card_header('Conversions', class = 'mini-header'),
      card_body(
        layout_column_wrap(
          height = '200px',
          selectInput(ns('vars_sel'), 'Variable', ''),
          textInput(ns('txt_current_format'), 'Current Type / Class')
        ),
        layout_column_wrap(
          height = '800px',
          selectInput(ns('sel_format'), 'Select the new format',
                      c('', 'as.numeric', 'as.integer',
                        'as.character', 'as.Date', 'as.factor',
                        'as.double')),
          conditionalPanel(
            condition = sprintf("input['%s'] == 'as.Date'", ns('sel_format')),
            selectInput(
              ns('sel_date_formats'),
              list('Select the entry format', bs_icon('info-circle')) |>
                tooltip('For character and factor inputs'),
              date_formats),
            dateInput(ns('sel_date_origin'),
                      list('Select a start date', bs_icon('info-circle')) |>
                        tooltip('For numeric inputs'),
                      value = '1970-01-01')
          )
        )
      ),
      card_footer(btn_task(ns('btn_apply'), 'Apply conversion', icon('check')))
    ),
    card(
      card_header('Preview', class = 'mini-header'),
      card_body(gt_output(ns('preview_gt'))),
      card_footer(btn_task(ns('btn_preview_sample'),
                           'Show new sample', icon('rotate-right')))
    )
  )
}

# server ----------------------------------------------------------------------
convert_cols_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

    df_names <- reactive(get_act_dt(session) |> names())

    # update select vars input ---------------
    observe({
      req(df_names())
      updateSelectInput(
        session,
        'vars_sel',
        choices = c('', df_names())
      )
    }) |> bindEvent(df_names())

    # update current format txt --------------
    observe({
      if(!isTruthy(input$vars_sel)){
        format <- ''
      } else {
        format <- paste(
          'Type: [', get_act_dt(session)[[input$vars_sel]] |> typeof(),
          '] | Class: [',
          paste(get_act_dt(session)[[input$vars_sel]] |> class(), collapse = '/'),
          ']'
        )
      }
      updateTextInput(session, 'txt_current_format', value = format)
    }) |> bindEvent(input$vars_sel)

    # sample to preview conversion -----------
    preview_sample_trigger <- reactiveVal(1)
    preview_sample <- reactive({
      nrow_act_dt <- get_act_dt(session) |> nrow()

      if(nrow_act_dt < 8) {
        rep(TRUE, nrow_act_dt)
      } else {
        sample(nrow_act_dt, 8, replace = F)
      }
    }) |> bindEvent(preview_sample_trigger(), session$userData$dt$act_meta())

    # update sample in button click ----------
    observe({
      preview_sample_trigger(preview_sample_trigger() + 1)
    }) |> bindEvent(input$btn_preview_sample)

    preview_df <- reactive({
      req(input$vars_sel)
      req(input$sel_format)

      if(input$sel_format == 'as.Date'){
        req(input$sel_date_formats)
      }

      preview_df_temp <- subset(
        get_act_dt(session)[preview_sample(), ], select = input$vars_sel)

      preview_df_temp[
        , preview := convert(var1,
                             type = input$sel_format,
                             date_format = input$sel_date_formats,
                             date_origin = input$sel_date_origin),
        env = list(var1 = input$vars_sel)]
    })

    # render preview gt ----------------------
    output$preview_gt <- render_gt({
      req(input$vars_sel, input$sel_format, preview_df())
      if(input$vars_sel %in% df_names()){
        preview_df() |>
          lapply(\(x) if(is.complex(x)) as.character(x) else x) |>
          as.data.frame() |>
          gt() |>
          cols_align(align = 'right') |>
          opt_interactive(
            use_sorting = F,
            use_pagination = F,
            use_highlight = T,
            use_compact_mode = T) |>
          tab_options(table.background.color = bg_color)
      }
    })

    # apply conversions -------------------------------------------------------
    observe({
      if(!isTruthy(input$vars_sel) || !isTruthy(input$sel_format)){
        msg('Select a variable and a new format')
        return()
      } else {
        running_modal()
        temp <- copy(get_act_dt(session))

        temp[, input$vars_sel :=
                       convert(var1,
                               type = input$sel_format,
                               date_format = input$sel_date_formats,
                               date_origin = input$sel_date_origin),
                     env = list(var1 = input$vars_sel)]

        update_act_dt(session, copy(temp), updated_cols = input$vars_sel,
                      change_type = 'convert_cols')
        rm(temp)
        remove_running_modal()

        updateSelectInput(session, 'vars_sel', selected = character(0))
        updateSelectInput(session, 'sel_format', selected = character(0))
      }
    }) |> bindEvent(input$btn_apply)
  })
}

# ui --------------------------------------------------------------------------
one_t_test_ui <- function(id) {
  ns <- NS(id)
  card(
    full_screen = T,
    card_header('One-sample t Test', class = 'mini-header'),
    layout_sidebar(
      class = 'card-sidebar',
      sidebar = sidebar(uiOutput(ns('parameters'))),
      navset_card_pill(
        nav_panel(
          'Test',
          card(
            layout_sidebar(
              sidebar = sidebar(
                h5('Sample Values', style = 'margin-bottom: -18px;'),
                textOutput(ns('sample_mean')),
                h5('Parameters', style = 'margin-bottom: -18px;'),
                layout_columns(
                  numericInput(
                    ns('mu'),
                    list('Mean', bs_icon('info-circle') |>
                           ttip('True Value of Mean')), 0, width = '180px')
                ),
                radioButtons(ns('radio_alternative'), 'Alternative',
                             c('Two sided' = 'two.sided',
                               'Less' = 'less',
                               'Greater' = 'greater')),
                numericInput(ns('confidence'), 'Confidence Interval - %',
                             value = 95, 0, 100, 5, width = '180px'),
                layout_columns(
                  col_widths = c(8),
                  btn_task(ns('btn_run_test'), 'Run Test', icon('gear'))
                ),
                layout_columns(
                  col_widths = c(8),
                  btn_task(ns('btn_help_ttest'), 'Help', icon('question'))
                ),
              ),
              card_body(
                layout_columns(
                  col_widths = c(8, 4),
                  div(
                    uiOutput(ns('conditional_plot')),

                    br(),

                    uiOutput(ns('conditional_staticard_ttest'))
                  ),
                  div(
                    gt_output(ns('ttest_gt')),
                    br(),
                    uiOutput(ns('conditional_save_gt'))
                  )
                )
              ),
              card_footer(
                uiOutput(ns('conditional_add_output'))
              )
            )
          )
        ),
        nav_panel(
          'Histogram',
          card(
            full_screen = T,
            card_body(plotOutput(ns('hist'))),
            card_footer(
              div(style = 'margin-bottom: -8px !important;'),
              layout_columns(
                col_widths = c(1, 3, 3),
                numericInput(ns('bins'), 'Bins', 25, 5, step = 5),
                btn_task(ns('btn_hist'), 'Show Histogram', icon('chart-simple'),
                         style = 'margin-top: 28px'),
                div(insert_output_ui(ns('ttest_insert_output_hist')),
                    style = 'margin-top: 28px')
              ),
              div(style = 'margin-bottom: -24px !important;'),
            )
          )
        )
      )
    )
  )
}

# server ----------------------------------------------------------------------
one_t_test_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  df <- reactive(get_act_dt(session))

    ttest <- reactiveValues(results = NULL)

    df_active <- reactive(df()[, lapply(df(), is.numeric) == T, with = F])

    var_analysis <- reactive({
      df_names <- df_active() |> names()

      var_analysis <- session$userData$dt$act_meta()[perc_nas != 1, var]

      df_names[df_names %in% var_analysis]
    })

    var <- reactive({
      req(input$sel_var)
      df_active()[[input$sel_var]]
    })

    # calculate values --------------------------------------------------------
    sample_mean <- reactive(var() |> fmean(na.rm = T))

    output$sample_mean <- renderText({
      paste('Mean:', sample_mean() |> f_num(dig = 2))
    })

    # input vars --------------------------------------------------------------
    output$parameters <- renderUI({
      tagList(
        selectInput(ns('sel_var'), 'Variable', var_analysis()),
        p('* Showing only numeric variables')
      )
    })

    # run t test --------------------------------------------------------------
    observe({
      req(input$sel_var)
      req(input$radio_alternative)

      if(!isTruthy(input$mu)){
        msg_error('Inform a value for the Mean')
      } else if(!isTruthy(input$confidence) ||
         !between(input$confidence, 0, 100)) {
        msg_error('Confidence interval must be between 0 and 100%', 2)
      } else if (fnobs(var()) < 2) {
        msg_error('Inform at least 2 valid values')
      } else {
        df <- t.test(var(),
                    alternative = input$radio_alternative,
                    mu = input$mu,
                    conf.level = input$confidence/100)

        df <- df |>
          unlist() |> as.data.frame()

        df$results <- rownames(df)
        names(df) <- c('values', 'results')

        df[df$results == 'data.name', ]$values <- input$sel_var

        ttest$results <- df

        ttest$confidence <- input$confidence/100
      }
    }) |> bindEvent(input$btn_run_test)

    # results gt table --------------------------------------------------------
    ttest_results_gt <- reactive({
      req(ttest$results)
      ttest$results |>
        gt() |>
        cols_move(columns = 'values', after = 'results') |>
        cols_label('values' = 'Values', 'results' = 'Results') |>
        tab_header('One-sample t Test')
    })

    output$ttest_gt <- render_gt({
      req(ttest_results_gt())
      ttest_results_gt()
    })

    # save gt module ----------------------------------------------------------
    save_gt_server('ttest_save_gt', ttest_results_gt)

    output$conditional_save_gt <- renderUI({
      req(ttest_results_gt())
      save_gt_ui(ns('ttest_save_gt'))
    })

    # insert to output --------------------------------------------------------
    insert_output_server(
      'ttest_insert_output',
      reactive(
        gen_table2(
          plot_tag(ttest_plot()(), w = 600, h = 300),
          ttest_results_gt(),
          '75%',
          '25%'
        )
      ),
      'One-sample t Test'
    )

    output$conditional_add_output <- renderUI({
      req(ttest_results_gt())
      insert_output_ui(ns('ttest_insert_output'))
    })

    # test plot ---------------------------------------------------------------
    output$conditional_plot <- renderUI({
      req(ttest_plot())
      plotOutput(ns('ttest_plot'))
    })

    ttest_plot <- reactive({
      req(ttest$results)
      req(ttest$confidence)

      alt <- ttest$results$values[which(ttest$results$results == 'alternative')]

      t_value <- ttest$results$values[which(ttest$results$results == 'statistic.t')]

      df <- ttest$results$values[which(ttest$results$results == 'parameter.df')]

      function(){
        plot_t_test(
          confidence = ttest$confidence,
          df = df |> as.numeric(),
          test_type = alt,
          t_value = t_value |> as.numeric()
        )
      }
    })

    output$ttest_plot <- renderPlot({
      req(ttest$results)
      req(ttest$confidence)
      req(ttest_plot())

      ttest_plot()()
    })

    # staticards --------------------------------------------------------------
    output$conditional_staticard_ttest <- renderUI({
      req(ttest_results_gt())

      t_value <- ttest$results$values[which(ttest$results$results == 'statistic.t')]
      p_value <- ttest$results$values[which(ttest$results$results == 'p.value')]

      layout_columns(
        col_widths = c(6, 6),
        stati_card(f_num(as.numeric(t_value), dig = 3), 't value'),
        stati_card(f_num(as.numeric(p_value), dig = 3), 'p value')
      )
    })

    # help modal --------------------------------------------------------------
    observe({
      fun_help_modal('stats', 't.test')
    }) |> bindEvent(input$btn_help_ttest)

    # histogram ---------------------------------------------------------------
    output$hist <- renderPlot({
      req(ttest_hist())
      validate(need(input$bins > 0, 'Bins must be 1 or higher'))

      ttest_hist()
    }, res = 96)

    task_hist <- ExtendedTask$new(function(spada_plot_fun,
                                           df,
                                           title,
                                           bins,
                                           plot_fill_color,
                                           plot_line_color,
                                           sample_limit,
                                           mean_value,
                                           sd_value) {
      mirai({
        spada_plot_fun(
          type = 'hist_density',
          df = df,
          xvar = 'x',
          xlab = 'Values',
          ylab = 'Density',
          title = title,
          bins = bins,
          fill_color = plot_fill_color,
          line_color = plot_line_color,
          point_shape = if (sample_limit > 1e4 && nrow(df) > 1e4) '.' else 20,
          sample_limit = sample_limit,
          mean_value = mean_value,
          sd_value = sd_value
        )
      },
      spada_plot_fun = spada_plot_fun,
      df = df,
      title = title,
      bins = bins,
      plot_fill_color = plot_fill_color,
      plot_line_color = plot_line_color,
      sample_limit = sample_limit,
      mean_value = mean_value,
      sd_value = sd_value)
    }) |> bind_task_button('btn_hist')

    observe({
      req(input$sel_var, var())
      task_hist$invoke(
        spada_plot_fun = spada_plot,
        df = data.frame(x = var()),
        bins = input$bins,
        title = paste('Histogram -', input$sel_var),
        plot_fill_color = session$userData$conf$plot_fill_color,
        plot_line_color = session$userData$conf$plot_line_color,
        sample_limit = session$userData$conf$plot_limit,
        mean_value = fmean(var(), na.rm = TRUE),
        sd_value = fsd(var(), na.rm = TRUE)
      )
    }) |> bindEvent(input$btn_hist)

    ttest_hist <- reactive({ task_hist$result() })

    # insert histogram to output ----------------------------------------------
    insert_output_server(
      'ttest_insert_output_hist',
      reactive(plot_tag(ttest_hist())),
      'Histogram'
    )

  })
}

# ui --------------------------------------------------------------------------
lm_ui <- function(id) {

  ns <- NS(id)
  card(
    full_screen = T,
    card_header('Linear Model', class = 'mini-header'),
    layout_sidebar(
      class = 'card-sidebar',
      sidebar = sidebar(
        selectInput(ns('sel_yvar'), 'Dependent Variable', character(0)),
        selectizeInput(ns('sel_xvar'), 'Independent Variables', character(0),
                       multiple = T,
                       options = list(plugins = list('remove_button', 'clear_button'))),
        layout_columns(
          col_widths = c(8),
          btn_task(ns('btn_run_lm'), 'Run Model', icon('gear'))
        ),
        layout_columns(
          col_widths = c(8),
          btn_task(ns('btn_help_lm'), 'Help', icon('question'))
        )
      ),
      navset_card_pill(
        nav_panel(
          'Output',
          card(
            card_body(
              gt_output(ns('lm_var_table')),
              gt_output(ns('lm_metrics'))
            ),
            card_footer(
              uiOutput(ns('conditional_add_output')),
              uiOutput(ns('conditional_save_model'))
            )
          )
        ),
        nav_panel(
          'Residuals',
          plotOutput(ns('lm_resid_plot')),
          card_footer(
            fluidRow(
              column(
                4,
                radioGroupButtons(
                  ns('radio_lm_resid'),
                  'Plot type:',
                  c(
                    'Histogram' = 'hist',
                    'Boxplot' = 'boxplot',
                    'Dots' = 'dots',
                    'QQ Plot' = 'qq_plot'
                  ),
                  size = 'sm',
                  individual = T
                )
              ),
              column(
                2,
                btn_task(ns('btn_lm_resid'),
                         'Plot residuals',
                         icon('chart-simple'),
                         style = 'margin-top: 28px')
              ),
              column(
                2,
                div(insert_output_ui(ns('insert_lm_resid_plot')), style = 'margin-top: 28px')
              )
            )
          )
        ),
        nav_panel(
          'VIF',
          gt_output(ns('vif_gt')),
          card_footer(
            layout_columns(
              col_widths = c(2, 2, 2),
              btn_task(ns('btn_vif'), 'Run VIF', icon('gear')),
              insert_output_ui(ns('insert_lm_vif')),
              btn_task(ns('btn_help_vif'), 'Help', icon('question'))
            )
          )
        )
      )
    )
  )
}

# server ----------------------------------------------------------------------
lm_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  df <- reactive(get_act_dt(session))

    var_analysis <- reactive({
      session$userData$dt$act_meta()[perc_nas != 1, var]
    })

    yvar <- reactive({
      req(var_analysis())
      intersect(var_analysis(), names(df())[sapply(df(), is.numeric)])
    })

    xvar <- reactive({
      req(input$sel_yvar)
      if(length(yvar()) > 0){
        var_analysis()[var_analysis() %notin% input$sel_yvar]
      } else {
        character(0)
      }
    })

    observe({
      updateSelectInput(session, 'sel_yvar', choices = yvar(), selected = yvar()[1])
    })

    observe({
      updateSelectizeInput(session, 'sel_xvar', choices = xvar())
    })

    # linear model ------------------------------------------------------------
    linear_model <- reactiveValues(
      model = NULL,
      summary = NULL,
      x = NULL,
      y = NULL,
      x_name = '',
      y_name = ''
    )

    observe({
      if (!isTruthy(input$sel_yvar) || !isTruthy(input$sel_xvar)) {
        msg('Select dependent and independent variables')
        return()
      }

      linear_model$y_name <- input$sel_yvar
      linear_model$x_name <- paste(input$sel_xvar, collapse = '+')

      form <- formula(paste(linear_model$y_name, '~', linear_model$x_name))

      linear_model$model <- lm(form, data = df(), model = F)

      linear_model$summary <- summary(linear_model$model)

      msg('Lm model completed.', DURATION = 0.5)

    }) |> bindEvent(input$btn_run_lm)

    # linear model output -----------------------------------------------------
    lm_var_table <- reactive({
      req(linear_model$model)
      linear_model_df_output(linear_model$summary) |>
        gt() |>
        tab_header(title = 'Linear Model',
                   subtitle = paste('Dependent Variable:', linear_model$y_name))
    })

    lm_metrics <- reactive({
      req(linear_model$model)
      linear_model_df_metrics(linear_model$summary) |>
        gt() |> tab_header('Model metrics')
    })

    output$lm_var_table <- render_gt({
      req(lm_var_table())
      lm_var_table()
    })

    output$lm_metrics <- render_gt({
      req(lm_metrics())
      lm_metrics()
    })

    # help events -------------------------------------------------------------
    observe({
      fun_help_modal('stats', 'lm')
    }) |> bindEvent(input$btn_help_lm)

    # insert model to output --------------------------------------------------
    insert_output_server(
      'lm_insert_output',
      reactive(gen_table2(lm_var_table(), lm_metrics())),
      'Linear Model'
    )

    output$conditional_add_output <- renderUI({
      req(linear_model$model)
      insert_output_ui(ns('lm_insert_output'))
    })

    # save model object -------------------------------------------------------
    output$conditional_save_model <- renderUI({
      req(linear_model$model)
      actionButton(ns('btn_save_model'), 'Save Model', icon('download'), class = 'btn-task')
    })

    observe({
      showModal(
        modalDialog(
          title = div(icon('download'), 'Save Model'),
          h5('This will save the Model Output and the Metrics tables.'),
          textInput(ns('model_filename'), 'File name', value = 'model'),
          footer = tagList(
            actionButton(ns('btn_close_save_model'), 'Close', icon('xmark'),
                     class = 'btn-task btn-task-cancel'),
            downloadButton(ns('down_handler'), 'Save model',
                           class = 'btn-task', icon = icon('download'))
          ),
          size = 'l'
        )
      )
    }) |> bindEvent(input$btn_save_model)

    observe({
      removeModal()
    }) |> bindEvent(input$btn_close_save_model)

    # download handler for the model file ----
    output$down_handler <- downloadHandler(

      filename = function() {
        paste0(input$model_filename, '.RDS')
      },
      content = function(file) {
        saveRDS(
          list(
            'model' = linear_model_df_output(linear_model$summary),
            'metrics' = linear_model_df_metrics(linear_model$summary)
          ),
          file,
          compress = F
        )
      }
    )

    # plot linear model residuals ---------------------------------------------
    update_lm_resid_plot <- reactiveVal(0)

    observe({
      update_lm_resid_plot(update_lm_resid_plot() + 1)
    }) |> bindEvent(input$btn_lm_resid)

    lm_resid_plot <- reactive({
      req(linear_model$model$residuals)
      req(update_lm_resid_plot() > 0)

      if(input$radio_lm_resid == 'hist'){

        spada_plot(type = 'hist',
                   df = data.frame(x = linear_model$model$residuals),
                   xvar = 'x',
                   ylab = 'Count',
                   fill_color = session$userData$conf$plot_fill_color,
                   line_color = session$userData$conf$plot_line_color,
                   title_color = session$userData$conf$plot_title_color,
                   title = 'Linear Model - Residuals',
                   sample_limit = session$userData$conf$plot_limit
        )

      } else if (input$radio_lm_resid == 'boxplot'){

        spada_plot(type = 'boxplot',
                   df = data.frame(x = linear_model$model$residuals),
                   xvar = 'x',
                   fill_color = session$userData$conf$plot_fill_color,
                   line_color = session$userData$conf$plot_line_color,
                   title_color = session$userData$conf$plot_title_color,
                   title = 'Linear Model - Residuals',
                   sample_limit = session$userData$conf$plot_limit
        )

      } else if (input$radio_lm_resid == 'dots'){

        spada_plot(type = 'dots',
                   df = data.frame(x = seq_along(linear_model$model$residuals),
                                   y = linear_model$model$residuals),
                   xvar = 'x',
                   yvar = 'y',
                   xlab = 'Index',
                   ylab = 'Values',
                   fill_color = session$userData$conf$plot_fill_color,
                   line_color = session$userData$conf$plot_line_color,
                   title_color = session$userData$conf$plot_title_color,
                   title = 'Linear Model - Residuals',
                   vertical_line = 0,
                   point_shape = if(session$userData$conf$plot_limit > 1e4 &&
                                    length(linear_model$model$residuals) > 1e4) '.' else 20,
                   sample_limit = session$userData$conf$plot_limit,
                   line_type = 2
        )
      } else if (input$radio_lm_resid == 'qq_plot'){
        spada_plot(
          type = 'qq_plot',
          df = data.frame(x = linear_model$model$residuals),
          xvar = 'x',
          xlab = 'Theoretical Quantiles',
          ylab = 'Sample Quantiles',
          title = 'Linear Model - Residuals',
          fill_color = session$userData$conf$plot_fill_color,
          line_color = session$userData$conf$plot_line_color,
          sample_limit = session$userData$conf$plot_limit
        )
      }
    }) |> bindEvent(update_lm_resid_plot())

    output$lm_resid_plot <- renderPlot({
      validate(need(isTruthy(linear_model$model), 'No residuals to plot'))

      lm_resid_plot()
    }, res = 96)

    # insert lm residual plot to output ---------------------------------------
    insert_output_server(
      'insert_lm_resid_plot',
      reactive(plot_tag(lm_resid_plot())),
      'Linear Model - Residuals Plot'
    )

    # VIF ---------------------------------------------------------------------

    vif <- reactive({
      req(linear_model$model)

      mod <- linear_model$model
      if(length(labels(terms(mod))) < 2){
        msg('Model contains fewer than 2 terms')
        return()
      }

      df <- VIF(mod) |> as.data.frame()

      if (ncol(df) == 1) {
        colnames(df) <- 'VIF'
      } else {
        colnames(df) <- c('GVIF', 'Df', 'Adjusted GVIF')
      }

      cbind(data.frame(Variable = rownames(df)), df)

    }) |> bindEvent(input$btn_vif)

    vif_gt <- reactive({
      req(vif())

      vif() |>
        gt() |>
        tab_header(title = 'Linear Model - VIF')
    })

    output$vif_gt <- render_gt({
      req(vif_gt())
      vif_gt()
    })

    # insert vif to output ----------------------------------------------------
    insert_output_server('insert_lm_vif', vif_gt, 'Linear Model - VIF')

    # help events -------------------------------------------------------------
    observe({
      fun_help_modal('DescTools', 'VIF')
    }) |> bindEvent(input$btn_help_vif)
  })
}

#' Launch a 'shiny' application for data analysis
#'
#' @description Generates a 'shiny' application for interactive data analysis.
#'
#' @param ... Objects of data.frame class
#' @param run_local Logical. Whether to run the application locally. If TRUE the app will close at session end.
#'
#' @examples
#' if(interactive()) spada(datasets::mtcars)
#'
#' @return An object of class 'shiny.appobj' representing the 'shiny' application.
#'   Printing the object launches the interactive app in a web browser.
#'
#' @export
#'
#' @import shiny
#'
#' @importFrom bsicons bs_icon
#'
#' @importFrom bslib accordion accordion_panel bind_task_button bs_add_rules
#'             bs_theme card card_body card_footer card_header font_collection
#'             input_switch input_task_button layout_column_wrap layout_columns
#'             layout_sidebar navbar_options nav_item nav_menu nav_panel
#'             nav_select nav_spacer navset_card_pill page_navbar popover
#'             sidebar tooltip update_switch value_box
#'
#' @importFrom collapse allNA ffirst flast fmean fmedian fmax fmin fmode fncol
#'             fnobs fnrow fnunique fquantile fsd fsum fvar frange qtab whichNA
#'             %==%
#'
#' @importFrom data.table %between% %notin% .SD := as.data.table between copy
#'             data.table fcase fcoalesce fifelse fread fwrite hour is.data.table
#'             mday melt minute month quarter rbindlist second setcolorder setDT
#'             setnames setorderv shift wday week yday year
#'
#' @importFrom DescTools Freq Gmean Hmean Kurt Mode Outlier ShapiroFranciaTest
#'             Skew StrCap VIF ZTest
#'
#' @importFrom ggplot2 .data aes after_stat coord_flip element_blank
#'             element_rect element_text geom_bar geom_boxplot geom_histogram
#'             geom_hline geom_line geom_point geom_text geom_tile geom_vline
#'             ggplot labs scale_fill_gradient2 stat_boxplot stat_function
#'             stat_qq stat_qq_line theme theme_classic ylim
#'
#' @importFrom graphics abline barplot boxplot curve hist lines mtext polygon text
#'
#' @importFrom grDevices colors
#'
#' @importFrom gt as_raw_html cols_align cols_hide cols_label cols_merge cols_move
#'             cols_width data_color fmt_bytes fmt_icon fmt_integer
#'             fmt_number fmt_percent gt gt_output gtsave opt_interactive pct
#'             render_gt sub_missing sub_values tab_header tab_options tab_spanner
#'
#' @importFrom haven as_factor is.labelled read_sav write_sav
#'
#' @importFrom htmltools plotTag save_html
#'
#' @importFrom mirai daemons mirai
#'
#' @importFrom qs2 qs_read qs_save
#'
#' @importFrom rlang parse_expr
#'
#' @importFrom shinyWidgets colorPickr dropdownButton show_toast radioGroupButtons
#'             statiCard updateColorPickr updateRadioGroupButtons
#'
#' @importFrom stats cor cor.test dnorm dt formula IQR dnorm ks.test median lm
#'             pf qnorm qqline qqnorm qt rnorm sd setNames shapiro.test t.test
#'             terms var
#'
#' @importFrom tools file_ext R_user_dir toTitleCase
#'
#' @importFrom utils object.size head packageDescription packageVersion sessionInfo
#'
#' @importFrom waiter useWaiter waiter_hide waiter_show waiterShowOnLoad
#'
#' @importFrom writexl write_xlsx

spada <- function(..., run_local = TRUE) {

  run_local <- !isFALSE(run_local)
  daemons(1)

  datasets <- list(...)
  if(length(datasets) == 0){
    datasets <- list('iris' = datasets::iris, 'mtcars' = datasets::mtcars)
    no_input_data <- TRUE
  } else{
    no_input_data <- FALSE
  }

  stopifnot('Objects must be data.frame and have at least 1 row and 1 col each' =
              sapply(datasets, is_spada_df) |> all())

  # set datasets names
  if(is.null(names(datasets))){
    names(datasets) <- lapply(substitute(list(...))[-1], deparse) |>
      unlist() |>
      make.names(unique = T)
  } else {
    new_names <- paste0('dataset_', 1:length(datasets))
    empty_names <- which(names(datasets) == '')
    names(datasets)[empty_names] <- new_names[empty_names]

    names(datasets) <- make.names(names(datasets), unique = T)
  }

  # make sure all datasets variables are valid names
  datasets <- lapply(datasets, make_var_names)

  # make valid cols --------------------------
  datasets <- lapply(
    datasets,
    \(df) {
      lapply(df, make_valid_cols) |> as.data.table()
    }
  )

  # read conf values ----------------------------------------------------------
  spada_conf_dir <- spada_user_dir('config')
  spada_data_dir <- spada_user_dir('data')

  start_conf <- c(
    'no_input_data' = no_input_data,
    'conf_dir' = spada_conf_dir,
    'data_dir' = spada_data_dir,
    default_conf
  )

  start_conf <- load_conf(start_conf, spada_conf_dir, themes_names)

  # resources -----------------------------------------------------------------
  addResourcePath('spada', system.file('www', package = 'spada'))

  ### Run App -----------------------------------------------------------------
  shinyApp(spada_ui(start_conf),
           spada_server(datasets, start_conf, run_local),
           options = list(launch.browser = T))
}

# tests/testthat/test-spada.R

# test spada - input data.frames ----------------------------------------------
test_that('spada() rejects empty data.frame', {
  expect_error(spada(data.frame()),
               'Objects must be data.frame and have at least 1 row and 1 col each')
})

test_that('spada() rejects valid data.frame plus invalid object', {
  expect_error(spada(iris, list()),
               'Objects must be data.frame and have at least 1 row and 1 col each')
})

test_that('spada() rejects valid data.frame plus empty data.frame', {
  expect_error(spada(iris, data.frame()),
               'Objects must be data.frame and have at least 1 row and 1 col each')
})

test_that('spada() rejects valid data.frame plus a vector', {
  expect_error(spada(iris, x = 1:10),
               'Objects must be data.frame and have at least 1 row and 1 col each')
})

test_that('spada() rejects zero row data.frame', {
  expect_error(spada(data.frame(Var = character(0))),
               'Objects must be data.frame and have at least 1 row and 1 col each')
})

# ui --------------------------------------------------------------------------
select_cols_ui <- function(id) {
  ns <- NS(id)

  card(
    card_header('Select Columns', class = 'mini-header'),
    card_body(
      selectizeInput(
        ns('vars_sel'),
        'Variable',
        NULL,
        multiple = T,
        options = list(plugins = list('remove_button', 'clear_button')),
        width = '80%'
      ),

      radioButtons(ns('radio_var_sel'), NULL,
                   c('Drop' = 'drop', 'Keep' = 'keep'), inline = T)
    ),
    card_footer(btn_task(ns('btn_sel'), 'Apply selection', icon('check')))
  )
}

# server ----------------------------------------------------------------------
select_cols_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

    df_names <- reactive(get_act_dt(session) |> names())

    observe({
      req(df_names())
      updateSelectizeInput(
        session,
        'vars_sel',
        choices = c('', df_names())
      )
    }) |> bindEvent(df_names())

    observe({
      if(input$vars_sel |> length() == 0){
        msg('Select at least one variable')
        return()
      } else if(input$radio_var_sel == 'drop' && (all(df_names() %in% input$vars_sel))){
        msg('Leave at least 1 variable')
        return()
      } else {

        running_modal()
        temp <- copy(get_act_dt(session))

        if(input$radio_var_sel == 'keep') {

          temp <- temp[, .SD, .SDcols = input$vars_sel]

        } else if (input$radio_var_sel == 'drop'){
          temp <- temp[, .SD, .SDcols = setdiff(df_names(), input$vars_sel)]
        }

        update_act_dt(session, copy(temp), updated_cols = temp |> names(),
                      change_type = 'select_cols')
        rm(temp)
        remove_running_modal()
      }
    }) |> bindEvent(input$btn_sel)

  })
}

# ui --------------------------------------------------------------------------
correlation_ui <- function(id) {
  ns <- NS(id)
  card(
    full_screen = T,
    card_header('Correlation Test', class = 'mini-header'),
    layout_sidebar(
      class = 'card-sidebar',
      sidebar = sidebar(uiOutput(ns('parameters'))),
      navset_card_pill(
        nav_panel(
          'Test',
          card(
            layout_sidebar(
              sidebar = sidebar(
                h5('Parameters', style = 'margin-bottom: -18px;'),
                radioButtons(ns('radio_method'), 'Method',
                             c('Pearson' = 'pearson',
                               'Kendall' = 'kendall',
                               'Spearman' = 'spearman')),
                radioButtons(ns('radio_alternative'), 'Alternative',
                             c('Two sided' = 'two.sided',
                               'Less' = 'less',
                               'Greater' = 'greater')),
                numericInput(ns('confidence'), 'Confidence Interval - %',
                             value = 95, 0, 100, 5, width = '180px'),
                layout_columns(
                  col_widths = c(8),
                  btn_task(ns('btn_run_test'), 'Run Test', icon('gear'))
                ),
                layout_columns(
                  col_widths = c(8),
                  btn_task(ns('btn_help_cor'), 'Help', icon('question'))
                ),
              ),
              layout_columns(
                col_widths = c(3, 6, 3),
                uiOutput(ns('conditional_staticard_cor')),
                gt_output(ns('cor_gt')),
                div(
                  uiOutput(ns('cond_add_output')),
                  br(), br(),
                  uiOutput(ns('conditional_save_gt'))
                ),
              )
            )
          )
        ),
        nav_panel(
          'Scatter',
          card(
            full_screen = T,
            card_body(plotOutput(ns('scatter_plot'))),
            card_footer(
              layout_columns(
                col_widths = c(3, 3),
                btn_task(ns('btn_scatter'), 'Show Scatter Plot', icon('chart-simple')),
                insert_output_ui(ns('insert_scatter'))
              )
            )
          )
        )
      )
    )
  )
}

# server ----------------------------------------------------------------------
correlation_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  df <- reactive(get_act_dt(session))

    # outupt objects ----------------------------------------------------------
    cor_test <- reactiveValues(results = NULL)

    df_active <- reactive({
      req(df())
      df()[, lapply(df(), is.numeric) == T, with = F]
    })

    var_analysis <- reactive({
      df_names <- df_active() |> names()

      var_analysis <- session$userData$dt$act_meta()[perc_nas != 1, var]

      df_names[df_names %in% var_analysis]
    })

    output$parameters <- renderUI({
      tagList(
        selectInput(ns('sel_var1'), 'Variable 1', var_analysis()),
        selectInput(ns('sel_var2'), 'Variable 2', var_analysis(),
                    var_analysis()[2]),
        p('* Showing only numeric variables')
      )
    })

    # run test events ---------------------------------------------------------
    observe({
      req(input$sel_var1)
      req(input$sel_var2)
      req(input$radio_alternative)
      req(input$radio_method)

      if(!isTruthy(input$confidence) ||
         !between(input$confidence, 0, 100)) {
        msg_error('Confidence interval must be between 0 and 100%', 2)
      } else if (sum(!is.na(df_active()[[input$sel_var1]])) < 3 ||
                 sum(!is.na(df_active()[[input$sel_var2]])) < 3) {
        msg_error('Inform at least 3 valid values for each variable')
      } else if (sd(df_active()[[input$sel_var1]], na.rm = T) == 0 ||
                 sd(df_active()[[input$sel_var2]], na.rm = T) == 0) {
        msg_error('The Standard deviation of any of the variables can not be zero', 3)
      } else {
        df <- cor.test(df_active()[[input$sel_var1]],
                       df_active()[[input$sel_var2]],
                       alternative = input$radio_alternative,
                       method = input$radio_method,
                       conf.level = input$confidence/100,
                       exact = F)

        df <- df |>
          unlist() |> as.data.frame()

        df$results <- rownames(df)
        names(df) <- c('values', 'results')

        df[df$results == 'data.name', ]$values <-
          paste(input$sel_var1, '/', input$sel_var2)

        cor_test$results <- df
      }
    }) |> bindEvent(input$btn_run_test)

    # results -----------------------------------------------------------------
    cor_results_gt <- reactive({
      req(cor_test$results)
      cor_test$results |>
        gt() |>
        cols_move(columns = 'values', after = 'results') |>
        cols_label('values' = 'Values', 'results' = 'Results') |>
        tab_header('Correlation Test')
    })

    output$cor_gt <- render_gt({
      req(cor_results_gt())
      cor_results_gt()
    })

    # save gt -----------------------------------------------------------------
    save_gt_server('cor_save_gt', cor_results_gt)

    output$conditional_save_gt <- renderUI({
      req(cor_results_gt())
      save_gt_ui(ns('cor_save_gt'))
    })

    # insert to output --------------------------------------------------------
    insert_output_server('insert_gt', cor_results_gt, 'Correlation Test')

    output$cond_add_output <- renderUI({
      req(cor_results_gt())
      insert_output_ui(ns('insert_gt'))
    })

    # staticards --------------------------------------------------------------
    cor_staticards <- reactive({
      req(cor_results_gt())

      idx_cor <- cor_test$results$results %in% c(
        'estimate.cor', 'estimate.tau', 'estimate.rho'
      )
      est_cor <- cor_test$results$values[idx_cor]

      indx_p <- cor_test$results$results %in% c('p.value')
      p_value <-  cor_test$results$values[indx_p]

      tagList(
        stati_card(f_num(as.numeric(est_cor), dig = 3), 'Correlation'),
        stati_card(f_num(as.numeric(p_value), dig = 3), 'p value')
      )
    })

    output$conditional_staticard_cor <- renderUI({
      req(cor_staticards())
      cor_staticards()
    })

    # scatter plot ------------------------------------------------------------
    task_scatter <- ExtendedTask$new(function(spada_plot_fun,
                                              df,
                                              sel_var1,
                                              sel_var2,
                                              plot_fill_color,
                                              sample_limit) {
      mirai({
        spada_plot_fun(
          type = 'scatter',
          df = df,
          xvar = sel_var1,
          yvar = sel_var2,
          xlab = sel_var1,
          ylab = sel_var2,
          fill_color = plot_fill_color,
          point_shape = if(sample_limit > 1e4 && nrow(df) > 1e4) '.' else 20,
          sample_limit = sample_limit
        )
      },
      spada_plot_fun = spada_plot_fun,
      df = df,
      sel_var1 = sel_var1,
      sel_var2 = sel_var2,
      plot_fill_color = plot_fill_color,
      sample_limit = sample_limit)
    }) |> bind_task_button('btn_scatter')

    observe({
      req(input$sel_var1, input$sel_var2, df_active())
      task_scatter$invoke(
        spada_plot_fun = spada_plot,
        df = df_active(),
        sel_var1 = input$sel_var1,
        sel_var2 = input$sel_var2,
        plot_fill_color = session$userData$conf$plot_fill_color,
        sample_limit = session$userData$conf$plot_limit
      )
    }) |> bindEvent(input$btn_scatter)

    scatter_plot <- reactive({ task_scatter$result() })

    output$scatter_plot <- renderPlot({
      req(scatter_plot())
      scatter_plot()
    }, res = 96)

    # insert scatter to output ------------------------------------------------
    insert_output_server(
      'insert_scatter',
      reactive(plot_tag(scatter_plot())),
      'Scatter Plot'
    )

    # help events -------------------------------------------------------------
    observe({
      fun_help_modal('stats', 'cor.test')
    }) |> bindEvent(input$btn_help_cor)

  })
}

# ui --------------------------------------------------------------------------
output_ui <- function(id) {
  ns <- NS(id)

  navset_card_pill(
    nav_panel(
      'Output',
      card(
        card_body(fillable = F, uiOutput(ns('panel'))),
        card_footer(
          fluidRow(
            column(1, checkboxInput(ns('x_flip'), 'Flip', T) |>
                     ttip('Flip upside down'),
                   style = 'margin-top: 28px'),
            column(
              3,
              radioGroupButtons(
                ns('sel_show_output'),
                'Show',
                c('All' = Inf,
                  'Last 1' = 1,
                  'Last 3' = 3,
                  'Last 5' = 5,
                  'None' = -Inf
                ),
                selected = 3,
                size = 'sm',
                individual = T
              )
            ),
            column(
              8,
              layout_columns(
                col_widths = c(3, 3, 3, 3),
                actionButton(ns('btn_reset'), 'Reset', icon('rotate-right'),
                             class = 'btn-task'),
                downloadButton(
                  ns('btn_save_html'),
                  'Save HTML',
                  class = 'btn-task',
                  icon = icon('download')
                ),
                actionButton(
                  ns('btn_save_output_session'),
                  'Save Output',
                  icon('download'),
                  class = 'btn-task'
                ),
                actionButton(
                  ns('btn_import_output_session'),
                  'Import Output',
                  icon('upload'),
                  class = 'btn-task'
                )
              ),
              style = 'margin-top: 28px'
            )
          ),
          div(style = 'margin-bottom: -8px !important;')
        )
      )
    ),
    nav_panel('Output Meta', gt_output(ns('elements_meta'))))
}

# server ----------------------------------------------------------------------
output_server <- function(id) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns

    # edit output events ------------------------------------------------------
    observe({
      id <- session$userData$out_edit_trigger()

      el_title <- session$userData$out$elements[[ id ]]$title
      el_annot <- session$userData$out$elements[[ id ]]$annotation
      el_element <- session$userData$out$elements[[ id ]]$element

      showModal(
        modalDialog(
          title = div(icon('pen-to-square'), 'Edit Output element'),
          size = 'l',
          tagList(
            textInput(ns('output_edit_title'), 'Title', value = el_title),
            textAreaInput(
              ns('output_edit_annot'),
              'Annotation',
              value = el_annot,
              width = '90%',
              rows = 5,
              resize = 'both'
            )
          ),
          footer = tagList(
            actionButton(ns('btn_cancel_edit_output'), 'Cancel',
                         icon = icon('xmark'), class = 'btn-task btn-task-cancel'),
            actionButton(ns('btn_confirm_edit_output'), 'Submit',
                         icon = icon('check'), class = 'btn-task')
          )
        )
      )

    }) |> bindEvent(session$userData$out_edit_trigger())

    # cancel_edit_output ----------------------
    observe({
      session$userData$out_edit_trigger(NULL)
      removeModal()
    }) |> bindEvent(input$btn_cancel_edit_output)

    # cancel_confirm_edit_output ----------------------
    observe({
      id <- session$userData$out_edit_trigger()
      new_title <- input$output_edit_title
      new_annot <- input$output_edit_annot
      new_element <- session$userData$out$elements[[ id ]]$element

      # update values ----------
      session$userData$out$elements[[ id ]]$title <- new_title
      session$userData$out$elements[[ id ]]$annotation <- new_annot
      session$userData$out$elements[[ id ]]$card <- report_card(
        new_title, new_annot, new_element)

      session$userData$out_edit_trigger(NULL)
      removeModal()
    }) |> bindEvent(input$btn_confirm_edit_output)

    # render panel ------------------------------------------------------------
    task_printable_out <- ExtendedTask$new(function(accordion,
                                                accordion_panel,
                                                div_fun = div,
                                                printable_report_card,
                                                elements,
                                                sel_show_output) {

      mirai({
        if (is.null(elements) || length(elements) == 0) {
          list()
        } else {

          len <- length(elements)

          lapply(seq_along(elements), function(i) {

            x <- elements[[i]]

            div(
              style = 'margin-bottom: 8px;',
              accordion(
                open = if (i > len - as.numeric(sel_show_output)) TRUE else FALSE,
                multiple = TRUE,
                accordion_panel(
                  paste(i, '-', x$title),
                  printable_report_card(list(x$btn_e, x$btn_x), x$card)
                )
              )
            )
          })
        }
      },
      accordion = accordion,
      accordion_panel = accordion_panel,
      div = div,
      printable_report_card = printable_report_card,
      elements = elements,
      sel_show_output = sel_show_output
      )
    })

    observe({
      req(session$userData$out$elements)
      req(input$sel_show_output)

        task_printable_out$invoke(
          accordion = accordion,
          accordion_panel = accordion_panel,
          div = div,
          printable_report_card = printable_report_card,
          elements = session$userData$out$elements,
          sel_show_output = as.numeric(input$sel_show_output)
        )
    }) |> bindEvent(session$userData$out$elements, input$sel_show_output)

    printable_output <- reactive({
      req(task_printable_out$result())
      task_printable_out$result()
    })

    output$panel <- renderUI({
      req(printable_output())

      if (input$x_flip) {
        tagList(printable_output() |> rev())
      } else {
        tagList(printable_output())
      }
    })

    # render output metadata --------------------------------------------------
    output$elements_meta <- render_gt({
      data.frame(
        position = seq_along(session$userData$out$elements),
        id = sapply(session$userData$out$elements, function(x) x$id),
        title = sapply(session$userData$out$elements, function(x) x$title)
        ) |> gt()
    })

    # reset output ------------------------------------------------------------
    observe({
      showModal(modalDialog(
        title = div(icon('rotate-right'), 'Reset Output'),
        'Do you want to remove all the elements?',
        easyClose = FALSE,
        size = 'l',
        footer = tagList(
          actionButton(ns('btn_cancel_reset'), 'No', icon('xmark'),
                       class = 'btn-task btn-task-cancel'),
          btn_task(ns('btn_confirm_reset'), 'Yes', icon('check'))
        )
      ))
    }) |> bindEvent(input$btn_reset)

    observe({
      session$userData$out$elements <- list()
      removeModal()
    }) |> bindEvent(input$btn_confirm_reset)

    observe({
      removeModal()
    }) |> bindEvent(input$btn_cancel_reset)

    # save html ---------------------------------------------------------------
    output$btn_save_html <- downloadHandler(
      filename = function() {
        paste0('spada_output_', format(Sys.time(), format = '%Y%m%d%H%M%S'), '.html')
      },
      content = function(file) {
        save_html(spada_output(session$userData$out$elements), file)
      }
    )

    # save session output -----------------------------------------------------
    observe({
      if(!session$userData$run_local){
        abort_run_local_modal('Spada must run locally to be able to save the output.')
      } else {
        showModal(modalDialog(
          title = div(icon('download'), 'Save Output'),
          'Do you want to save the current Output? This will overwrite the previous one.',
          easyClose = FALSE,
          size = 'l',
          footer = tagList(
            actionButton(ns('btn_cancel_save_output'), 'No', icon('xmark'),
                         class = 'btn-task btn-task-cancel'),
            btn_task(ns('btn_confirm_save_output'), 'Yes', icon('check'))
          )
        ))
      }
    }) |> bindEvent(input$btn_save_output_session)

    # cancel save output
    observe({ removeModal() }) |> bindEvent(input$btn_cancel_save_output)

    # confirm save output
    observe({
      removeModal()

      spada_save(session$userData$conf$data_dir,
                 session$userData$out$elements,
                 'output.qs2')

      msg('Output saved', 2.5)
    }) |> bindEvent(input$btn_confirm_save_output)

    # import session output ---------------------------------------------------
    observe({
      if(!session$userData$run_local){
        abort_run_local_modal('Spada must run locally to be able to import the output.')
      } else {
        showModal(modalDialog(
          title = div(icon('upload'), 'Import Output'),
          'Do you want to import the previous saved Output? This will erase
          the current Output elements.',
          easyClose = FALSE,
          size = 'l',
          footer = tagList(
            actionButton(ns('btn_cancel_import_output'), 'No', icon('xmark'),
                     class = 'btn-task btn-task-cancel'),
            btn_task(ns('btn_confirm_import_output'), 'Yes', icon('check'))
          )
        ))
      }
    }) |> bindEvent(input$btn_import_output_session)

    # cancel import output
    observe({ removeModal() }) |> bindEvent(input$btn_cancel_import_output)

    # confirm import output
    observe({
      removeModal()

      if(!file.exists(paste0(session$userData$conf$data_dir, '/output.qs2'))){
        msg_error('Output file not found')
      } else {
        check_dir(session$userData$conf$data_dir)
        temp_output <- qs_read(paste0(session$userData$conf$data_dir, '/output.qs2'))

        if(test_output_format(temp_output)){

          temp_output <- lapply(temp_output, function(x) {

            id <- gen_element_id()
            x$id <- id
            btn_xid <- paste0('btn_xout_', id)
            btn_eid <- paste0('btn_eout_', id)

            register_output_events(id, session, input, btn_xid, btn_eid)

            add_output_element(
              id,
              x$title,
              x$annotation,
              x$element,
              ns(btn_xid),
              ns(btn_eid)
            )
          })
          # rename items of list with new ids
          names(temp_output) <- vapply(temp_output, function(x) x$id, character(1))

          session$userData$out$elements <- temp_output
          msg(paste('Output imported from', session$userData$conf$data_dir), 2.5)
        } else {
          msg_error('Output in invalid format')
        }
      }
    }) |> bindEvent(input$btn_confirm_import_output)

  })
}

# ui --------------------------------------------------------------------------
data_overview_ui <- function(id) {
  ns <- NS(id)
  card(
    card_body(
      selectInput(
        inputId = ns('sel_dataset'),
        label = 'Dataset',
        choices = NULL
      ),
      gt_output(ns('gt'))
    ),
    card_footer(
      fluidRow(
        column(2, numericInput(ns('size_sample'), 'Number of rows', 500, 0, 1e4, 500)),
        column(2, radioGroupButtons(
          ns('radio_sample'), 'Show',
          c('First rows' = 'first', 'Last rows' = 'last', 'Sample' = 'sample'),
          size = 'sm', individual = T)),
        column(2, div(
          insert_output_ui(ns('data_overview_insert_output'))),
          style = 'margin-top: 28px'
        )
      )
    )
  )
}

# server ----------------------------------------------------------------------
data_overview_server <- function(id) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns

    observe({
      req(session$userData$dt_names())

      updateSelectInput(
        session,
        inputId = 'sel_dataset',
        choices = c(
          session$userData$dt$act_name,
          setdiff(session$userData$dt_names(), session$userData$dt$act_name)
        )
      )
    })

    df <- reactive({
      req(session$userData$dt$dt, input$sel_dataset)

      session$userData$dt$dt[[input$sel_dataset]]
    })

    idx <- reactive({
      req(df(), input$size_sample)

      validate(need(input$size_sample > 0, 'Number of rows must be > 0'))

      nrow_df <- nrow(df())
      n_show <- max(1, input$size_sample)
      n_show <- min(n_show, nrow_df)

      if(input$radio_sample == 'first'){
        idx <- 1:n_show
      } else if(input$radio_sample == 'last'){
        start_idx <- max(1, nrow_df - n_show + 1)
        idx <- start_idx:nrow_df
      } else if (input$radio_sample == 'sample'){
        idx <- sample.int(nrow_df, n_show, replace = F)
      }
    })

    data_filtered <- reactive({
      req(idx())

      df()[idx(), ]
    })

    data_gt <- reactive({
      req(data_filtered())
      data_filtered() |>
        gt() |>
        tab_header(input$sel_dataset)
    })

    output$gt <- render_gt({
      req(data_gt())

      data_gt() |>
        cols_align(align = 'right') |>
        opt_interactive(
          page_size_default = 10,
          use_filters = T,
          use_resizers = T,
          use_highlight = T,
          use_compact_mode = T,
          use_text_wrapping = F,
          use_page_size_select = T
        )
    })

    # insert to output module -------------------------------------------------
    insert_output_server(
      'data_overview_insert_output',
      reactive(data_gt() |> tab_options(table.width = pct(90))),
      'Data Overview'
    )

  })
}

# ui --------------------------------------------------------------------------
missing_ui <- function(id) {
  ns <- NS(id)

  layout_column_wrap(
    card(
      card_header('Identify Missing Values', class = 'mini-header'),
      card_body(
        selectInput(ns('var_sel_missing'), 'Variable', ''),
        textInput(ns('txt_missing_name'), 'Missing Variable Name', 'Missing')
      ),
      card_footer(btn_task(ns('btn_identify'), 'Run', icon('check')))
    ),
    card(
      card_header('Replace Missing Values', class = 'mini-header'),
      card_body(
        selectInput(ns('var_sel_replace'), 'Variable', ''),
        uiOutput(ns('ui_replace_method')),
        uiOutput(ns('ui_replace_value'))
      ),
      card_footer(btn_task(ns('btn_replace'), 'Replace', icon('check')))
    )
  )
}

# server ----------------------------------------------------------------------
missing_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  df_names <- reactive(get_act_dt(session) |> names())

	  # update inputs ------------
	  observe({
	    req(df_names())
	    updateSelectInput(
	      session,
	      'var_sel_missing',
	      choices = c('', df_names())
	    )

	    updateSelectInput(
	      session,
	      'var_sel_replace',
	      choices = c('', df_names())
	    )
	  }) |> bindEvent(df_names())

	  # detect missing -----------
	  observe({

	    selected_var <- input$var_sel_missing
	    missing_name <- input$txt_missing_name

	    if(!isTruthy(selected_var)){
	      msg('Select at least one variable')
	      return()
	    } else {
	      if(is_name_available(missing_name, df_names())){

  	      running_modal()
  	      temp <- copy(get_act_dt(session))

  	      temp[, ]

  	      temp[, new_var := factor(is.na(sel_var), levels = c(TRUE, FALSE), labels = c('Missing', 'Not Missing' )), env = list(
  	        new_var = missing_name,
  	        sel_var = selected_var
  	      )]

  	      update_act_dt(session, copy(temp), updated_cols = missing_name)
  	      rm(temp)

  	      remove_running_modal()

  	      updateTextInput(session, 'txt_missing_name', value = 'Missing')

  	    } else {
  	      msg_error('New names are not valid or already in use')
  	      return()
  	    }
	    }
	  }) |> bindEvent(input$btn_identify)

	  # replace missing -----------
	  output$ui_replace_method <- renderUI({
	    req(input$var_sel_replace)
	    type <- obj_type(get_act_dt(session)[[input$var_sel_replace]])

	    methods <- c('Constant' = 'constant')

	    if (type %in% c('double', 'integer')) {
	      methods <- c(methods, 'Mean' = 'mean', 'Median' = 'median')
	    }

	    selectInput(
	      ns('sel_replace_method'),
	      'Replacement',
	      choices = methods
	    )
	  })

	  output$ui_replace_value <- renderUI({
	    req(input$var_sel_replace, input$sel_replace_method)

	    type <- obj_type(get_act_dt(session)[[input$var_sel_replace]])

      if(type %in% c('double', 'integer')) {

        if (input$sel_replace_method != 'constant') {
          NULL
        } else {
          numericInput(ns('num_value'), 'Value', 0)
        }
      } else if(type == 'logical') {
        selectInput(ns('logical_value'), 'Value', c(TRUE, FALSE))
      } else if(type == 'factor') {
        selectInput(
          ns('factor_value'),
          'Value',
          choices = levels(get_act_dt(session)[[input$var_sel_replace]])
        )
      } else if(type %in% c('date', 'posix')) {
        dateInput(ns('date_value'), 'Date')
      } else {
        textInput(ns('txt_replace_value'), 'Value')
      }
	  })

    # replace value -----------
	  observe({
	    selected_var <- input$var_sel_replace

	    if(!isTruthy(selected_var)){
	      msg('Select at least one variable')
	      return()
	    }

	    method <- input$sel_replace_method
	    sel_var_type <- obj_type(get_act_dt(session)[[selected_var]])

	    replace_value <- switch(
	      sel_var_type,
	      'double' = input$num_value |> as.double(),
	      'integer' = input$num_value |> as.integer(),
	      'char' = input$txt_replace_value,
	      'logical' = input$logical_value |> as.character(),
	      'factor' = input$factor_value |> as.factor(),
	      'date' = input$date_value |> as.Date(),
	      'posix' = input$date_value |> as.Date()
	    )

	    if(method %in% c('constant') && !isTruthy(replace_value)){
	      msg('Inform a value')
	      return()
	    }

	    temp <- copy(get_act_dt(session))

      new_value <- switch(
        method,
        'mean' = mean_nona(temp[[selected_var]]),
        'median' = median_nona(temp[[selected_var]]),
        'constant' = replace_value
      )

      # new value must keep the type of actual variable
      actual_type <- temp[[selected_var]] |> obj_type()
      new_value <- convert(new_value, actual_type)

      temp <- try(
        temp[is.na(var), var := new_value, env = list(var = selected_var)],
        silent = TRUE
      )

      if (inherits(temp, 'try-error')) {
        msg_error('Unable to replace missing values')
        return()
      }

      if(!is_spada_df(temp)){
        abort_filter_modal()
      } else {
        running_modal()
        update_act_dt(session, copy(temp))

        updateSelectInput(session, 'var_sel_replace', selected = character(0))
        remove_running_modal()
      }
	  }) |> bindEvent(input$btn_replace)

  })
}

# ui --------------------------------------------------------------------------
rename_cols_ui <- function(id) {
  ns <- NS(id)

  layout_column_wrap(
    card(
      card_header('Rename Column', class = 'mini-header'),
      card_body(
        selectInput(ns('vars_sel'), 'Variable', ''),

        textInput(ns('txt_new_name'), 'New name'),
      ),
      card_footer(btn_task(ns('btn_rename'), 'Rename Variable', icon('check')))
    ),
    card(
      card_header('Rename Multiple Columns', class = 'mini-header'),
      card_body(
        selectizeInput(ns('vars_sel_multi'), 'Variables', '',
                       multiple = T,
                       options = list(plugins = list('remove_button', 'clear_button')),
                       width = '80%'),

        radioButtons(ns('rename_method'), 'Rename Method',
                     choices = c('Add Prefix/Suffix' = 'prefix_suffix',
                                 'Apply Function' = 'function',
                                 'Replace Part' = 'replace',
                                 'Remove Part' = 'remove')),

        conditionalPanel(
          condition = "input.rename_method == 'prefix_suffix'", ns = ns,
          textInput(ns('txt_new_name_multi'), 'New part to insert'),
          radioButtons(ns('part_position'), 'Part Position',
                       choices = c('Prefix' = 'prefix', 'Suffix' = 'suffix')),
          radioButtons(ns('name_separator'), 'Separator',
                       choices = c('Underscore (_)' = '_', 'Dot (.)' = '.', 'None' = ''))
        ),

        conditionalPanel(
          condition = "input.rename_method == 'function'", ns = ns,
          selectInput(ns('name_function'), 'Function to Apply',
                      choices = c('Upper Case' = 'toupper',
                                  'Lower Case' = 'tolower',
                                  'Title Case' = 'toTitleCase'))
        ),

        conditionalPanel(
          condition = "input.rename_method == 'replace'", ns = ns,
          textInput(ns('txt_replace_part'), 'Part to Replace'),
          textInput(ns('txt_replace_new_part'), 'New Part'),
        ),

        conditionalPanel(
          condition = "input.rename_method == 'remove'", ns = ns,
          textInput(ns('txt_remove_part'), 'Part to Remove')
        ),

      ),
      card_footer(btn_task(ns('btn_rename_multi'), 'Rename Variables', icon('check')))
    )
  )
}

# server ----------------------------------------------------------------------
rename_cols_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

    df_names <- reactive(get_act_dt(session) |> names())

    # rename 1 variable ------------------------------------------------------
    observe({
      req(df_names())
      updateSelectInput(
        session,
        'vars_sel',
        choices = c('', df_names())
      )
    }) |> bindEvent(df_names())


    observe({
      if(!isTruthy(input$vars_sel)){
        msg('Select at least one variable')
      } else {
        if(is_name_available(input$txt_new_name, df_names())){
          running_modal()
          temp <- copy(get_act_dt(session))

          setnames(temp, input$vars_sel, input$txt_new_name)

          update_act_dt(session, copy(temp), updated_cols = input$txt_new_name,
                        change_type = 'rename_cols')

          rm(temp)

          remove_running_modal()
        } else {
          msg_error('New name is not valid or already in use')
        }
      }
    }) |> bindEvent(input$btn_rename)

    # rename multiple veriables -----------------------------------------------
    observe({
      req(df_names())
      updateSelectizeInput(
        session,
        'vars_sel_multi',
        choices = c('', df_names())
      )
    }) |> bindEvent(df_names())

    observe({
      if(!isTruthy(input$vars_sel_multi)){
        msg('Select at least one variable')
      } else {

        selected_names <- input$vars_sel_multi

        if (input$rename_method == 'prefix_suffix') {

          if(!isTruthy(input$txt_new_name_multi)){
            msg('Inform a new part to insert')
            return()
          } else {
            if(input$part_position == 'prefix') {
              new_names <- paste0(input$txt_new_name_multi,
                                  input$name_separator, selected_names)
            } else if(input$part_position == 'suffix'){
              new_names <- paste0(selected_names, input$name_separator,
                                  input$txt_new_name_multi)
            }
          }
        } else if(input$rename_method == 'function') {
          new_names <- sapply(selected_names, match.fun(input$name_function))
        } else if(input$rename_method == 'replace') {

          if(!isTruthy(input$txt_replace_part) || !isTruthy(input$txt_replace_new_part)){
            msg('Inform the text to replace')
            return()
          } else {
            new_names <- gsub(input$txt_replace_part, input$txt_replace_new_part,
                              selected_names, fixed = TRUE)
          }

        } else if(input$rename_method == 'remove') {

          if(!isTruthy(input$txt_remove_part)){
            msg('Inform the text to be removed')
            return()
          } else {
            new_names <- gsub(input$txt_remove_part, '', selected_names, fixed = TRUE)
          }
        }

        # validate unique and valid names -------------------------------------
        if (is_valid_name(new_names) |> all()) {

          if(all(new_names == selected_names)){
            msg('No changes to apply')
          } else if(anyDuplicated(new_names)){
            msg_error('There is duplication in the new names')
          } else if(any(new_names %in% setdiff(df_names(), selected_names))){
            msg_error('Some new names already in use')
          } else {
            running_modal()
            temp <- copy(get_act_dt(session))

            setnames(temp, input$vars_sel_multi, new_names)

            update_act_dt(session, copy(temp), updated_cols = input$new_names,
                          change_type = 'rename_cols')

            rm(temp)
            remove_running_modal()
          }
        } else {
          msg_error('New names are not valid')
        }
      }
    }) |> bindEvent(input$btn_rename_multi)

  })
}

# ui --------------------------------------------------------------------------
data_bkp_ui <- function(id) {
  ns <- NS(id)

  layout_column_wrap(
    btn_task(ns('btn_reset'), 'Reset Dataset', icon('arrow-rotate-right')) |>
      ttip('Restore to previous state (before been set as the Active dataset)'),
    btn_task(ns('btn_bkp'), 'Create Backup', icon('cloud-arrow-up')) |>
      ttip('Create a copy of the Active dataset'),
    btn_task(ns('btn_restore'), 'Restore Backup', icon('cloud-arrow-down')) |>
      ttip('Restore a previously created backup'),
    btn_task(ns('btn_clear_bkp'), 'Clear Backup', icon('trash-can')) |>
      ttip('Erase the backup'),
  )
}

# server ----------------------------------------------------------------------
data_bkp_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  # reset df active ------------------------
	  observe({
	    update_act_dt(session, copy(session$userData$dt$bkp0))
	    msg('Active Dataset Reseted')
	  }) |> bindEvent(input$btn_reset)

	  # create backup --------------------------
	  observe({
	    session$userData$dt$bkp <- copy(get_act_dt(session))
	    msg('Backup created')
	  }) |> bindEvent(input$btn_bkp)

	  # restore backup -------------------------
	  observe({
	    if(is.null(session$userData$dt$bkp)){
	      msg('No backup to restore')
	    } else {
	      update_act_dt(session, copy(session$userData$dt$bkp))
	      msg('Backup restored')
	    }
	  }) |> bindEvent(input$btn_restore)

	  # clear backup ---------------------------
	  observe({
	    if(is.null(session$userData$dt$bkp)){
	      msg('No backup to clear')
	    } else {
	      session$userData$dt$bkp <- NULL
	      msg('Backup cleared')
	    }
	  }) |> bindEvent(input$btn_clear_bkp)

  })
}

# ============================================================================.
# ---------------------------- OBJECTS ---------------------------------------
# ============================================================================.

# default conf ----------------------------------------------------------------
default_conf <- list(
  'theme' = 'spada_theme',
  'file_size' = 1000,
  'restore_session' = 'never',
  'save_session' = 'ask',
  'restore_data_status' = 0,
  'restore_output_status' = 0,
  'restore_status' = NULL,
  'plot_fill_color' = plot_fill_color,
  'plot_line_color' = plot_line_color,
  'plot_title_color' = plot_title_color,
  'plot_limit' = 1e5
)

# math functions --------------------------------------------------------------
math_funs <- c(
  'Mean' = 'mean_na',
  'Mean no NA' = 'mean_nona',
  'Geometric Mean' = 'Gmean',
  'Harmonic Mean' = 'Hmean',
  # 'Mode' = 'Mode',
  'Standard Deviation' = 'fsd',
  'Variance' = 'fvar',
  'Min' = 'min_na',
  'Min no NA' = 'min_nona',
  'Max' = 'max_na',
  'Max no NA' = 'max_nona',
  'First' = 'first_na',
  'First no NA' = 'first_nona',
  'Last' = 'last_na',
  'Last no NA' = 'last_nona',
  'Lag' = 'shift',
  # 'Range' = 'range',
  'IQR' = 'IQR',
  'Skewness' = 'Skew',
  'Kurtosis' = 'Kurt',
  'Sum' = 'sum_na',
  'Sum no NA' = 'sum_nona',
  'Cum Sum' = 'cumsum',
  'Prod' = 'prod',
  'Cum Prod' = 'cumprod',
  'Square Root' = 'sqrt',
  'Exponential' = 'exp',
  'Log' = 'log',
  'Log2' = 'log2',
  'Log10' = 'log10',
  'Ceiling' = 'ceiling',
  'Floor' = 'floor',
  'Trunc' = 'trunc',
  'Signif' = 'signif',
  'Round' = 'round',
  'Abs' = 'abs',
  'Sine' = 'sin',
  'Cosine' = 'cos',
  'Tangent' = 'tan',
  'Order' = 'order',
  'As Numeric' = 'as.numeric',
  'Is Numeric' = 'is.numeric',
  'As Double' = 'as.double',
  'Is Double' = 'is.double',
  'As Integer' = 'as.integer',
  'Is Integer' = 'is.integer',
  'Is Integer' = 'is.integer',
  'As Numeric' = 'as.numeric',
  'Is Numeric' = 'is.numeric'
  )

groupby_math_funs <- c(
  'Mean' = 'mean_na',
  'Mean no NA' = 'mean_nona',
  'Geometric Mean' = 'Gmean',
  'Harmonic Mean' = 'Hmean',
  'Standard Deviation' = 'sd',
  'Variance' = 'var',
  'Min' = 'min_na',
  'Min no NA' = 'min_nona',
  'Max' = 'max_na',
  'Max no NA' = 'max_nona',
  'First' = 'fina',
  'Last' = 'lana',
  'Skewness' = 'Skew',
  'Kurtosis' = 'Kurt',
  'Sum' = 'sum_na',
  'Sum no NA' = 'sum_nona',
  'Prod' = 'prod'
)

# char functions --------------------------------------------------------------
char_funs <- c(
  'To Upper' = 'toupper',
  'To Lower' = 'tolower',
  'N Char' = 'nchar',
  'As Char' = 'as.character',
  'Is Char' = 'is.character'
)

groupby_char_funs <- c(
  'First' = 'fina',
  'Last' = 'lana'
)

# date functions --------------------------------------------------------------
date_funs <- c(
  # data.table package
  'Year' = 'year',
  'Month' = 'month',
  'Month Day' = 'mday',
  'Week Day' = 'wday',
  'Year Day' = 'yday',
  'Week' = 'week',
  'Quarter' = 'quarter',
  'Hour' = 'hour',
  'Minute' = 'minute',
  'Second' = 'second',
  # spada package
  'Is Date' = 'is_date'
)

groupby_date_funs <- c(
  'Min' = 'min_na',
  'Min no NA' = 'min_nona',
  'Max' = 'max_na',
  'Max no NA' = 'max_nona',
  'First' = 'fina',
  'Last' = 'lana'
)

# factor functions ------------------------------------------------------------
factor_funs <- c(
  'Number of Levels' = 'nlevels',
  'As factor' = 'as.factor',
  'Is Factor' = 'is.factor',
  'Is Ordered' = 'is.ordered'
)

# logical functions -----------------------------------------------------------
logical_funs <- c(
  'All True' = 'all',
  'Any True' = 'any',
  'As Logical' = 'as.logical',
  'Is Logical' = 'is.logical',
  'Number of True' = 'sum_na',
  'Proportion of True' = 'mean_na'
)

groupby_logical_funs <- c(
  'Min' = 'min_na',
  'Min no NA' = 'min_nona',
  'Max' = 'max_na',
  'Max no NA' = 'max_nona',
  'First' = 'fina',
  'Last' = 'lana',
  'Sum' = 'sum_na',
  'Sum no NA' = 'sum_nona'
)

# complex functions -----------------------------------------------------------
complex_funs <- c(
  'Real Part' = 'Re',
  'Imaginary Part' = 'Im',
  'Is Complex' = 'is.complex'
)

# basic operations ------------------------------------------------------------
basic_operations <- c(
  # base package
  '+', '-', '+', '-', '/', '*', '^',
  '==', '!=', '!', '>', '>=', '<', '<=',
  '&', '|', 'xor',
  '(', '[', '<-', '$',
  '%in%', 'is.na', 'is.null', 'is.nan', 'na.omit',
  '::', ':', 'T', 'F',
  'list', 'try',
  'paste', 'paste0', 'substr',
  'isTRUE', 'isFALSE',
  'c',
  'as.Date', 'as.POSIXct', 'as.POSIXlt',
  'ifelse',
  # base math
  'mean', 'median', 'sd', 'var', 'min', 'max', 'sum',

  # data.table package
  '%notin%', 'between', '%between%', 'fifelse', 'fcase',

  # DescTools package
  'Outlier',

  #stats package
  'quantile',

  #spada utils functions
  'first_na', 'first_nona', 'last_na', 'last_nona', 'max_na', 'max_nona',
  'mean_na', 'mean_nona', 'min_na', 'min_nona', 'sum_na', 'sum_nona'
)

# dangerous operations --------------------------------------------------------
dangerous_operations <- c(
  # R base - code
  'assign', 'attach',
  'do.call',
  'eval', 'eval.parent', 'evalq',
  'get', 'globalenv', 'new.env', 'parent.env', 'parse',
  'remove', 'rm', 'setwd', 'substitute', 'Sys.setenv',

  # R base - files system
  'dir.create', 'dir.exists', 'dir.remove',
  'file.copy', 'file.create', 'file.remove', 'file.rename',
  'load', 'save', 'saveRDS', 'unlink',

  # R base - network
  'curl', 'download.file', 'httr::DELETE', 'httr::GET', 'httr::POST', 'httr::PUT',
  'RCurl::getURL', 'RCurl::postForm', 'socketConnection', 'url',

  # R base - IO
  'read.csv', 'readLines', 'read.table', 'readRDS',
  'write.csv', 'writeLines', 'write.table',

  # R base system
  'shell', 'shell.exec', 'system', 'system2', 'gc',

  # data.table
  ':=',
  'set', 'setattr', 'setcolorder', 'setnames', 'setDT',
  'setDF', 'setkey', 'setorder', 'substitute2'
)

# allowed operations ----------------------------------------------------------
allowed_operations <- c(
  basic_operations,
  math_funs,
  char_funs,
  date_funs,
  factor_funs,
  logical_funs
) |> unique() |> sort()

# list of filters -------------------------------------------------------------
equal_operators <- c('== (Equal)' = '==',
                     '!= (Not Equal)' = '!=')

compare_operators <- c('> (Greater)' = '>',
                       '>= (Greater or Equal)' = '>=',
                       '< (Less)' = '<',
                       '<= (Less or Equal)' = '<=')

na_operators <- c('Is NA (is.na)' = 'is_na',
                  'Not NA (! is.na)' = 'not_na')

in_operators <- c('In (%in%)' = 'in',
                  'Not In (! %in%)' = 'not_in')

between_operators <- c('Between' = 'between',
                       'Not Between' = 'not_between')

outlier_operators <- c('Outlier' = 'outlier',
                       'Not Outlier' = 'not_outlier')

logical_operators <- c('TRUE' = 'is_true',
                       'FALSE' = 'is_false')

filter_operators <- c(
  equal_operators,
  compare_operators,
  na_operators,
  in_operators,
  between_operators,
  outlier_operators,
  logical_operators
)

# lits of date formats --------------------------------------------------------
date_formats <- c(
  # hifen
  'YYYY-MM-DD' = '%Y-%m-%d',
  'DD-MM-YYY' = '%d-%m-%Y',
  'MM-DD-YY' = '%m-%d-%y',
  'YYYY-MMM-DD' = '%Y-%b-%d',
  'YY-MMM-DD' = '%y-%b-%d',
  'YY-MM-DD' = '%y-%m-%d',
  # bars
  'YYYY/MM/DD' = '%Y/%m/%d',
  'DD/MM/YYYY' = '%d/%m/%Y',
  'MM/DD/YY' = '%m/%d/%y',
  'YYYY/MMM/DD' = '%Y/%b/%d',
  'YY/MMM/DD' = '%y/%b/%d',
  'YY/MM/DD' = '%y/%m/%d',
  # dots
  'YYYY.MM.DD' = '%Y.%m.%d',
  'DD.MM.YYYY' = '%d.%m.%Y',
  'MM.DD.YY' = '%m.%d.%y',
  'YYYY.MMM.DD' = '%Y.%b.%d',
  'YY.MMM.DD' = '%y.%b.%d',
  'YY.MM.DD' = '%y.%m.%d',
  # no separator
  'YYYYMMDD' = '%Y%m%d',
  'DDMMYYYY' = '%d%m%Y',
  'MMDDYY' = '%m%d%y',
  'YYYYMMMDD' = '%Y%b%d',
  'YYMMMDD' = '%y%b%d',
  'YYMMDD' = '%y%m%d'
)

# list of summarise functions -------------------------------------------------
summarise_functions <- c(
  'Distinct' = 'distinct',
  'Count' = 'count'
)

# close browser tab -----------------------------------------------------------
js_exit <- "Shiny.addCustomMessageHandler('closeWindow', function(m) {window.close();});"

tag_js_exit <- tags$head(tags$script(HTML(js_exit)))

# waiter screen ----------------------------------------------------------------
waiter_right_foot <- function(){
  div(paste('Version', packageVersion('spada')), class = 'spada-waiter-right-footer')
}
waiter_progress_bar <- div(class = 'spada-waiter-progress',  div(class = 'spada-waiter-bar'))

waiter_screen <- tags$style(
  HTML(
    "

    .waiter-overlay-content{
      width: 100% !important;
      max-width: none !important;
      display: flex !important;
      justify-content: center;
      align-items: center;
    }

    .waiter-overlay {
      font-family: 'Segoe UI', sans-serif;

      display: flex;
      justify-content: center;
      align-items: center;
      overflow:hidden;

      width: 100vw;
      height: 100vh;

      background:

          radial-gradient(circle at top,
              rgba(255,255,255,.28),
              transparent 45%),

          radial-gradient(circle,
              rgba(95,110,125,.18) 2px,
              transparent 2px),

          linear-gradient(
              180deg,
              #0A6CA4 0%,
              #02517D 35%,
              #014A72 70%,
              #013D5E 100%
          );

      background-size:
          cover,
          25px 25px,
          cover;

      position:relative;
    }

    .waiter-overlay .spada-waiter-card{
        position :relative;
        z-index: 1;
        width: 90%;
        max-width: 800px;
        padding: 60px 60px 60px 60px;
        background: #EEF2F5;
        border-radius: 16px;
        box-shadow:
            0 30px 70px rgba(0,0,0,.30),
            0 8px 20px rgba(0,0,0,.18);
        text-align:left;
        animation:fade .8s ease;
    }

    @keyframes fade{
        from{
            opacity: 0;
            transform: translateY(18px);
        }
        to{
            opacity: 1;
            transform: translateY(0);
        }
    }

    .waiter-overlay .spada-waiter-title{
        font-size: 60px;
        font-weight: 300;
        letter-spacing: 12px;
        color: #1C4B71;
    }

    .waiter-overlay .spada-waiter-subtitle{
        margin-top: 10px;
        font-size: 32px;
        opacity: 0.95;
        letter-spacing: 4px;
        color: #607B95;
    }

    .waiter-overlay .spada-waiter-subtitle2{
        margin-top: 10px;
        font-size: 24px;
        font-weight: 100;
        opacity: 0.9;
        letter-spacing: 4px;
        color: #607B95;
    }

    .waiter-overlay .spada-waiter-subtitle3{
        margin-top: 15px;
        font-size: 18px;
        font-weight: 100;
        opacity: 0.85;
        letter-spacing: 4px;
        color: #607B95;
    }

    .waiter-overlay .spada-waiter-progress{
        margin: 42px auto 24px;
        width: 100%;
        height: 8px;
        background: #D9E2EB;
        border-radius: 999px;
        position: relative;
        overflow: hidden;
    }

    .waiter-overlay .spada-waiter-bar{
        position: absolute;
        top: 0;
        left: 0;
        width: 20%;
        height: 100%;
        border-radius: 999px;
        background:
            linear-gradient(
                90deg,
                #2F80ED,
                #72C2FF,
                #2F80ED
            );

        box-shadow:
            0 0 10px rgba(47,128,237,.35);

        animation: loading 3.5s ease-in-out infinite alternate;
    }

    @keyframes loading{
        from{ left: 0; }
        to{ left: 80%; }
    }

    .waiter-overlay .spada-waiter-card-footer{
        margin-top: 40px;
        color: #8FA4B8;
        font-size: 13px;
    }

    .waiter-overlay .spada-waiter-right-footer{
        position: fixed;
        right: 24px;
        bottom: 20px;
        color: #6F89A1;
        font-size: 16px;
        letter-spacing: 1px;
    }

    .waiter-overlay .spada-waiter-log{
      color: #607B95;
      font-size: 16px;
      opacity: 0.85;
    }
  "
  )
)

# spada_output_css ------------------------------------------------------------
spada_output_css <- tags$head(tags$style(HTML(
  paste(
    "
      body {
        font-family: 'Segoe UI', Ubuntu sans-serif;
      }

      h2 {
        font-size: 1.5rem !important;
        margin-bottom: 12px ! important;
        color: ", main_color, " !important;
        font-weight: 400 !important;
      }

      .toc-title { color: ", main_color, ";}

      .toc ul{
        list-style:none;
        padding-left:0;
        margin:0;
      }

      .toc li{
        margin-bottom:4px;
      }

      .toc a{
        display:block;
        padding:8px 12px;

        color:#444444;
        text-decoration:none;

        border-radius:6px;

        transition:
          background-color .25s ease,
          color .25s ease,
          transform .25s ease,
          box-shadow .25s ease;
      }

      .toc a:hover{
        background:", main_color, ";
        color:white;

        transform:translateX(5px);

        box-shadow:0 2px 8px rgba(13,110,253,.25);
      }
    "
  )
)))

# ============================================================================.
# ---------------------------- FUNCTIONS -------------------------------------
# ============================================================================.
# short name ------------------------------------------------------------------
short_name <- function(name, max_size = 20){
  stopifnot(is.character(name))
  if(nchar(name) <= max_size) name else paste0(substr(name, 1, max_size - 3) , '...')
}
# spada_user_dir --------------------------------------------------------------
spada_user_dir <- function(which){
  dir <- normalizePath(R_user_dir('spada', which), winslash = '/', mustWork = F)

  check_dir(dir)

  return(dir)
}

# set theme -------------------------------------------------------------------
set_spada_theme <- function(session, theme){

  session$setCurrentTheme(spada_bs_theme(theme))
  session$userData$conf$theme <- theme
}

# load conf -------------------------------------------------------------------
load_conf <- function(start_conf,
                      r_user_conf_dir,
                      themes_names) {

  conf_path <- file.path(r_user_conf_dir, 'conf.qs2')

  if (file.exists(conf_path)) {
    conf_saved <- qs_read(conf_path)

    required_fields <- c(
      'theme',
      'file_size',
      'restore_session',
      'save_session',
      'plot_fill_color',
      'plot_line_color',
      'plot_title_color',
      'plot_limit'
    )

    # if all TRUE copy saved conf
    if (
      # test required items
      is.list(conf_saved) &&
        all(required_fields %in% names(conf_saved)) &&

      # test theme
      length(conf_saved$theme) == 1 &&
        isTRUE(conf_saved$theme %in% themes_names) &&

      # file size
      length(conf_saved$file_size) == 1 &&
        is.numeric(conf_saved$file_size) && conf_saved$file_size > 0 &&

      # restore session
      length(conf_saved$restore_session) == 1 &&
        isTRUE(conf_saved$restore_session %in% c('always', 'ask', 'never')) &&

      # save session
      length(conf_saved$save_session) == 1 &&
        isTRUE(conf_saved$save_session %in% c('always', 'ask', 'never')) &&

      # fill color
      length(conf_saved$plot_fill_color) == 1 &&
        isTRUE(is_hex_color(conf_saved$plot_fill_color)) &&

      # line color
      length(conf_saved$plot_line_color) == 1 &&
        isTRUE(is_hex_color(conf_saved$plot_line_color)) &&

      # title color
      length(conf_saved$plot_title_color) == 1 &&
        isTRUE(is_hex_color(conf_saved$plot_title_color)) &&

      # plot limit
      length(conf_saved$plot_limit) == 1 &&
        is.numeric(conf_saved$plot_limit)
      ){
      # copy if test is passed
      start_conf$theme <- conf_saved$theme
      start_conf$file_size <- conf_saved$file_size
      start_conf$restore_session <- conf_saved$restore_session
      start_conf$save_session <- conf_saved$save_session
      start_conf$plot_fill_color <- conf_saved$plot_fill_color
      start_conf$plot_line_color <- conf_saved$plot_line_color
      start_conf$plot_title_color <- conf_saved$plot_title_color
      start_conf$plot_limit <- conf_saved$plot_limit
    }
  }
  if(test_write_access(conf_path)) qs_save(start_conf, conf_path)

  return(start_conf)
}

# generate 2 column table in html ---------------------------------------------
gen_table2 <- function(element1, element2, w1 = '50%', w2 = '50%') {
  div(
    tags$table(
      style = 'width: 95%',
      tags$tr(
        tags$td(style = paste('padding: 10px; width:', w1, ';'), element1),
        tags$td(style = paste('padding: 10px; width:', w2, ';'), element2)
      )
    )
  )
}

# card to insert in output ----------------------------------------------------
report_card <- function(title = 'Spada - Output', annotation = NULL,
                        content = NULL){
  div(
      h2(title,
         style = paste0(
           " font-size: 1.5rem;",
           " margin-bottom: 12px;",
           " color:", main_color ,";",
           " text-align: center;"
           )
        ),
      p(annotation),
      content
  )
}

printable_report_card <- function(btns, card, id = NULL){
  div(
    style = paste0(
      "border: 2px solid", main_color, "; border-radius: 4px; ",
      "padding: 16px; margin-bottom: 8px; "
    ),
    div(style = "display: flex; width: 100%;",
        div(card, style = "width: 98%;"),
        tagList(
          lapply(btns,
               function(x){ div(x, style = 'width: 2%; text-align: right;') }
          )
        )
    ),
    br(),
    p(id)
  )
}

# output header ---------------------------------------------------------------
spada_output_info_card <- function(content){
  div(
    style = '
    padding: 20px 24px;

    background: linear-gradient(
      135deg,
      #02517d 0%,
      #0A6CA4 100%
    );

    color: #ffffff;
    margin-bottom: 25px;
  ',
    content
  )
}

spada_output <- function(output){
  toc <- tags$div(
    class = 'toc',
    style = '
    width:15vw;
    min-width:220px;
    height:100vh;
    padding:16px 16px 16px 24px;
    border-left:1px solid #ddd;
    box-sizing:border-box;
    overflow-y:auto;
    position:sticky;
    top:0;
  ',

    tags$h2(class = 'toc-title', 'Table of contents'),

    tags$ul(
      tagList(
        tags$li(tags$a(href = '#home', 'Home')),

        lapply(
          output,
          function(x) { tags$li(tags$a(href = paste0('#', x$id), x$title)) }
        )
      )
    )
  )

  cards <- tags$div(
    style = 'width:85vw; padding:24px; box-sizing:border-box;',

    lapply(
      output,
      function(x) {
        div(
          id = x$id,
          style = 'margin-bottom:32px;',
          printable_report_card(NULL, x$card, NULL)
        )

      }
    )
  )

  body_content <- tags$div(
    style = 'display:flex; align-items:flex-start;',
    cards,
    toc
  )

  header <- spada_output_info_card(
    div(
      style = 'display: flex; flex-direction: column;',
      tags$div(
        'Spada Output',
        style = ' font-size: 32px; font-weight: 450; letter-spacing: 4px;')
    )
  )

  footer <- spada_output_info_card(
    div(
      style = '
      display: flex;
      justify-content: space-between;
      align-items: center;
      flex-wrap: wrap;
      gap: 15px;
    ',

      # left
      div(
        paste('Spada:', utils::packageVersion('spada')),
        tags$br(),

        tags$a(
          href = 'https://cran.r-project.org/package=spada',
          target = '_blank',
          style = 'color: white;',
          'CRAN'
        ),

        ' | ',

        tags$a(
          href = 'https://github.com/lgschuck/spada',
          target = '_blank',
          style = 'color: white;',
          'GitHub'
        ),

        ' | ',

        tags$a(
          href = 'https://lgschuck.github.io/spada/',
          target = '_blank',
          style = 'color: white;',
          'Documentation'
        ),

        ' | ',

        tags$a(
          href = 'https://lgschuck.github.io/spada_book/',
          target = '_blank',
          style = 'color: white;',
          'Book'
        )
      ),

      # right
      div(
        style = 'text-align: right;',
        format(Sys.time(), '%Y-%m-%d %H:%M:%S %Z')
      )
    )
  )

  tags$html(
    spada_output_css,
    tags$body(div(id = 'home', header), body_content, footer)
  )
}

# add output element ----------------------------------------------------------
add_output_element <- function(id, title, annotation, element, btn_xid, btn_eid){

  btn_x <- actionButton(btn_xid, '', icon('x'), class = 'micro-btn-cancel')
  btn_e <- actionButton(btn_eid, '', icon('pen-to-square'), class = 'micro-btn-cancel')

  list(
    'id' = id,
    'title' = title,
    'annotation' = annotation,
    'element' = element,
    'card' = report_card(title, annotation, element),
    'btn_x' = btn_x,
    'btn_e' = btn_e
  )
}

# register outputs events -----------------------------------------------------
register_output_events <- function(id, session, input, btn_xid, btn_eid) {
  observe({
    session$userData$out$elements[[id]] <- NULL
  }) |> bindEvent(input[[btn_xid]], once = TRUE)

  observe({
    session$userData$out_edit_trigger(id)
  }) |> bindEvent(input[[btn_eid]])
}

# generate output object ------------------------------------------------------
gen_output <- function(element = div(h2('Element'))){

  x <- sample(1e3, 1)
  title <- paste('Test title', x)
  annotation <- paste('Test annotation', x)
  id <- gen_element_id()
  btn_xid <- paste0('btn_xout_', id)
  btn_eid <- paste0('btn_eout_', id)

  element <- if(is.null(element)) data.frame(x = letters) |> gt() else element

  add_output_element(
    id,
    title,
    annotation,
    element,
    btn_xid,
    btn_eid
  )
}

# generate element id for outputs ---------------------------------------------
gen_element_id <- function(id = 'id', time_only = FALSE){
  if(time_only){
    gsub('.', '', format(Sys.time(), '%Y%m%d%H%M%OS8'), fixed = T)
  } else {
    paste0(id, '_', gsub('.', '', format(Sys.time(), '%Y%m%d%H%M%OS8'), fixed = T))
  }
}

# allowed operations function -------------------------------------------------
show_allowed_op <- function(){
  showModal(
    modalDialog(
      title = div(icon('check'), 'Allowed Operations'),
      selectInput(
        inputId = 'allowed_op_selected',
        label   = 'List of operations',
        choices = allowed_operations,
        selected = allowed_operations[1]
      ),
      easyClose = TRUE
    )
  )
}

# safe env function -----------------------------------------------------------
safe_env <- function(operations = NULL){

  e <- new.env(parent = emptyenv())

  lapply(operations, function(x) {
    assign(x, get(x), envir = e)
  })

  return(e)
}

# test dataset ----------------------------------------------------------------
add_na <- function(x, prop = 0.05) {
  idx <- sample.int(length(x), ceiling(length(x) * prop))
  x[idx] <- NA
  x
}

test_dataset <- function(n_row = 1e3, n_col = 18) {

  test_data <- data.table(
    integer_var = sample(1:100, n_row, replace = TRUE),
    numeric_var = rnorm(n_row),
    char_var = sample(letters, n_row, replace = TRUE),
    char_long_var = rep(paste(letters, collapse = ''), n_row),
    char_colors_var = sample(colors(), n_row, replace = TRUE),
    date_var = Sys.Date() + sample(-49:50, n_row, replace = TRUE),
    posixct_var = as.POSIXct(Sys.time()) +
      sample(0:(365 * 24 * 60 * 60), n_row, replace = TRUE),
    factor_var = factor(sample(paste0('factor_', 1:10), n_row, replace = TRUE)),
    logical_var = sample(c(TRUE, FALSE), n_row, replace = TRUE)
  )

  test_data[, `:=`(
    integer_var_na = add_na(integer_var),
    numeric_var_na = add_na(numeric_var),
    char_var_na = add_na(char_var),
    char_long_var_na = add_na(char_long_var),
    char_colors_var_na = add_na(char_colors_var),
    date_var_na = add_na(date_var),
    posixct_var_na = add_na(posixct_var),
    factor_var_na = add_na(factor_var),
    logical_var_na = add_na(logical_var)
  )]

  test_data <- test_data[, sort(names(test_data)), with = FALSE]

  extra_cols <- n_col - ncol(test_data)

  if (extra_cols > 0) {
    for (i in seq_len(extra_cols)) {
      test_data[[paste0("extra_col_", i)]] <- rnorm(n_row)
    }
  }

  test_data
}

# empty plot function ---------------------------------------------------------
empty_plot <- function(msg = 'No plot', c = 2){
  plot(1:10, 1:10, type = 'n', xlab = '', ylab = '')
  text(5, 5, msg, cex = c)
}

# bslib btn task --------------------------------------------------------------
btn_task <- function(ID,
                     LABEL,
                     ICON = NULL,
                     LABEL_BUSY = 'Running...',
                     ...){
  input_task_button(
    id = ID,
    label = LABEL,
    icon = ICON,
    label_busy = LABEL_BUSY,
    class = 'btn-task',
    ...
  )
}

# mini_btn --------------------------------------------------------------------
mini_btn <- function(ID, LABEL, ICON){
  actionButton(
    inputId = ID,
    label = LABEL,
    icon = ICON,
    class = 'mini-btn'
  )
}

# messages - shownotification -------------------------------------------------
msg <- function(TEXT, DURATION = 2.3){
  show_toast(
    title = TEXT,
    type = 'info',
    position = 'center',
    timer = DURATION * 1000,
    timerProgressBar = F,
    width = '650px'
  )
}

msg_error <- function(TEXT, DURATION = 2.3){
  show_toast(
    title = TEXT,
    type = 'error',
    position = 'center',
    timer = DURATION * 1000,
    timerProgressBar = F,
    width = '650px'
  )
}

bmsg <- function(TEXT, DURATION = 2.3){
  bslib::show_toast(

    bslib::toast(
      TEXT,
      header = bslib::toast_header(
        title = TEXT,
        status = "Spada"
      ),
      type = NULL,
      duration_s = DURATION,
      position = c('middle', 'center')
    )
  )
}

# try convert -----------------------------------------------------------------
try_convert <- function(x, fun){
  tryCatch(fun(x),
           error = function(e) rep(NA, x |> length()),
           warning = function(w) rep(NA, x |> length()))
}

# convert function ------------------------------------------------------------
convert <- function(x, type, date_format = '%Y-%m-%d',
                    date_origin = '1970-01-01'){
  if(x |> is.raw()) x <- as.numeric(x)

  if(type %in% c('as.numeric', 'numeric')){
    suppressWarnings(as.numeric(x))
  } else if(type%in% c('as.integer', 'integer')){
    suppressWarnings(as.integer(x))
  } else if(type %in% c('as.character', 'char')){
    as.character(x)
  } else if(type %in% c('as.Date', 'date')){
    if(x |> inherits('Date')) {
      x
    } else if(is.numeric(x)){
      as.Date(x, origin = date_origin)
    } else if(is.raw(x) || is.complex(x)){
      suppressWarnings(
        as.Date(x |> as.numeric(), origin = date_origin)
      )
    } else as.Date(x, format = date_format)
  } else if(type %in% c('as.factor', 'factor')){
    as.factor(x)
  } else if(type %in% c('as.double', 'double')){
    suppressWarnings(as.double(x))
  } else if(type %in% c('as.complex', 'complex')){
    if(is.complex(x)){
      x
    } else {
      x1 <- suppressWarnings(as.numeric(x))
      as.complex(x1)
    }
  } else if (type %in% c('as.logical', 'logical')){
    as.logical(x)
  } else if (type %in% c('as.posix', 'posix')){
    as.POSIXct(x)
  }
}

# filter rows function --------------------------------------------------------
filter_rows <- function(dt, var, operator, filter_value){

  stopifnot(is.data.table(dt))

  if(operator == '=='){
    dt[var1 == filter_value, , env = list(var1 = var)]
  } else if(operator == '!='){
    dt[var1 != filter_value, , env = list(var1 = var)]
  } else if(operator == '>'){
    dt[var1 > filter_value, , env = list(var1 = var) ]
  } else if(operator == '>='){
    dt[var1 >= filter_value, , env = list(var1 = var) ]
  } else if(operator == '<'){
    dt[var1 < filter_value, , env = list(var1 = var) ]
  } else if(operator == '<='){
    dt[var1 <= filter_value, , env = list(var1 = var) ]
  } else if(operator == 'is_na'){
    dt[is.na(var1), , env = list(var1 = var) ]
  } else if(operator == 'not_na'){
    dt[!is.na(var1), , env = list(var1 = var) ]
  } else if(operator == 'in'){
    dt[var1 %in% filter_value,  , env = list(var1 = var)]
  } else if(operator == 'not_in'){
    dt[!var1 %in% filter_value, , env = list(var1 = var) ]
  } else if(operator == 'between'){
    dt[var1 %between% filter_value, , env = list(var1 = var) ]
  } else if(operator == 'not_between'){
    dt[!(var1 %between% filter_value), , env = list(var1 = var) ]
  } else if(operator == 'outlier'){
    dt[Outlier(var1, value = F, na.rm = T), , env = list(var1 = var)]
  } else if(operator == 'not_outlier'){
    dt[!Outlier(var1, value = F, na.rm = T), , env = list(var1 = var)]
  } else if(operator == 'is_true'){
    dt[var1 == TRUE,  , env = list(var1 = var)]
  } else if(operator == 'is_false'){
    dt[var1 == FALSE,  , env = list(var1 = var)]
  }
}

# filter rows 2 vars ----------------------------------------------------------
filter_rows_2vars <- function(dt, var1, var2, operator){

  stopifnot(is.data.table(dt))

  v1 <- dt[[var1]]
  v2 <- dt[[var2]]

  # Convert to char if factor
  if (is.factor(v1) && is.factor(v2)) {
    v1 <- as.character(v1)
    v2 <- as.character(v2)
  }

  if(operator == '=='){
    dt[v1 == v2, ]
  } else if(operator == '!='){
    dt[v1 != v2, ]
  } else if(operator == '>'){
    dt[v1 > v2, ]
  } else if(operator == '>='){
    dt[v1 >= v2, ]
  } else if(operator == '<'){
    dt[v1 < v2, ]
  } else if(operator == '<='){
    dt[v1 <= v2, ]
  } else {
    stop('Invalid Operator')
  }
}

# tip place top by default ----------------------------------------------------
ttip <- function(TRIGGER, ..., ID = NULL, PLACE = 'top'){
  tooltip(trigger = TRIGGER, ... = ..., id = ID, placement = PLACE)
}

# get function help -----------------------------------------------------------
get_help_file <- function(pak, fun){
  paste(
    utils::capture.output(
      tools::Rd2HTML(tools::Rd_db(pak)[[paste0(fun, '.Rd')]])
    ),
    collapse = '\n'
  )
}

# fun help modal --------------------------------------------------------------
fun_help_modal <- function(pak, fun) {
  showModal(modalDialog(
    HTML(get_help_file(pak, fun)),
    title = div(icon('circle-question'), paste('Help:', pak, '-', fun)),
    easyClose = TRUE,
    size = 'xl'
  ))
}

# operations help -------------------------------------------------------------
operations_help <- function(div_content){
  showModal(
    modalDialog(
      title = div(icon('circle-question'), ' Help'),
      size = 'l',
      easyClose = TRUE,
      div_content
    )
  )
}

# format decimals -------------------------------------------------------------
f_dec <- function(x, dig = 0){
  if(is.numeric(x) && !is.na(x) |> all()){
    format(round(x, dig), nsmall = dig, scientific = F)
  } else {
    NA
  }
}

# make var names --------------------------------------------------------------
make_var_names <- function(df){
  stopifnot(df |> is.data.frame())
  names(df) <- names(df) |> make.names(unique = T)
  return(df)
}

# test all equal --------------------------------------------------------------
test_all_equal <- function(x){
  all(x == x[1])
}

# obj type --------------------------------------------------------------------
obj_type <- function(x){
  if(x |> is.numeric() & x |> is.integer()) 'integer'
  else if (x |> is.numeric() & x |> is.double()) 'double'
  else if (x |> inherits('Date')) 'date'
  else if (x |> inherits(c('POSIXct', 'POSIXlt', 'POSIXt'))) 'posix'
  else if (x |> is.factor()) 'factor'
  else if (x |> is.character()) 'char'
  else if (x |> is.logical()) 'logical'
  else if (x |> is.complex()) 'complex'
  else 'other'
}

# stati_card ------------------------------------------------------------------
stati_card <- function(VALUE, SUBTITLE, ICON = NULL, LEFT = T,
                       COLOR = '#FFFFFF', BACKGROUND = main_color,
                       ANIMATE = T, DURATION = 30, ID = NULL){
  statiCard(value = VALUE,
            subtitle = SUBTITLE,
            icon = ICON,
            left = LEFT,
            color = COLOR,
            background = BACKGROUND,
            animate = ANIMATE,
            duration = DURATION,
            id = ID)
}

# make valid cols -------------------------------------------------------------
make_valid_cols <- function(x){
  if(is.raw(x) || is.complex(x)){
    as.character(x)
  } else {
    x
  }
}

# is spada data frame ---------------------------------------------------------
is_spada_df <- function(df){
  is.data.frame(df) && all(sapply(df, is.atomic)) && ncol(df) > 0 && nrow(df) > 0
}

# is spada data table ---------------------------------------------------------
is_spada_dt <- function(dt){
  is.data.table(dt) && is_spada_df(dt)
}

# test data format-------------------------------------------------------------
test_data_format <- function(data){
  is.list(data) &&
    length(data) > 0 &&
    (sapply(data, is_spada_df) |> all())
}


# running modal ---------------------------------------------------------------
running_modal <- function(text = 'Running...'){
  showModal(
    modalDialog(
      div(
        icon('spinner', class='fa fa-spinner fa-spin fa-1x'),
        span(text, style = 'color:#585c5e;')
      ),
      footer = NULL,
      easyClose = FALSE,
      size = 'm',
      fade = T
    )
  )
}

remove_running_modal <- function(time = 0.7){
  Sys.sleep(time)
  removeModal()
}

# not spada df modal ----------------------------------------------------------
abort_filter_modal <- function(){
  showModal(modalDialog(
    'The resultant dataset must be valid (nrow and ncol > 0)',
    title = div(icon('triangle-exclamation'), 'Operation Aborted')
  ))
}

# abort save session/output ---------------------------------------------------
abort_run_local_modal <- function(msg = 'Spada must run locally'){
  showModal(modalDialog(
    msg,
    title = div(icon('triangle-exclamation'), 'Operation Aborted'),
    easyClose = T
  ))
}

# show startup function -------------------------------------------------------
show_startup_screen <- function() {
  waiterShowOnLoad(
    html = tagList(
      div(
        class = 'spada-waiter-card',
        div(class = 'spada-waiter-title', 'Spada'),
        div(class = 'spada-waiter-subtitle2', 'a Shiny Package for Data Analysis'),
        div(class = 'spada-waiter-progress',  div(class = 'spada-waiter-bar')),
        div(id = 'spada_startup_text', class = 'spada-waiter-log', 'Building UI'),
        div(class = 'spada-waiter-card-footer', HTML('Powered by <strong>R and Shiny</strong>')),
      ),
      waiter_right_foot()
    )
  )
}

# exit screen function --------------------------------------------------------
show_exit_screen <- function(save = TRUE) {
  waiter_show(
    html = {
      if(save){
        tagList(
          div(class = 'spada-waiter-card',
            div('Spada is saving your work', class = 'spada-waiter-subtitle2'),
            div('Please do not close this window', class = 'spada-waiter-subtitle3'),
            div(class = 'spada-waiter-progress',  div(class = 'spada-waiter-bar')),
            div(id = 'spada_exit_text', class = 'spada-waiter-log', '')
          ),
          waiter_right_foot()
        )
      } else {
        html = tagList(
          div(
            class = 'spada-waiter-card',
            div('Closing Spada', class = 'spada-waiter-subtitle2'),
            waiter_progress_bar,
            div(id = 'spada_exit_text', class = 'spada-waiter-log', 'Saving conf')
          ),
          waiter_right_foot()
        )
      }
    }
  )
}

# update startup text ---------------------------------------------------------
update_waiter_text <- function(session, id, text) {
  session$sendCustomMessage('update-text', list(id = id, text = text))
}

# update start up text-----------------------------------------------------------
waiter_text <- "
  Shiny.addCustomMessageHandler(
    'update-text',
    function(data){

      const el = document.getElementById(data.id);

      if(el){
        el.innerText = data.text;
      }
    }
  )
"

tag_js_waiter_text <- tags$head(tags$script(HTML(waiter_text)))

# spada save function ---------------------------------------------------------
spada_save <- function(dir, object, file_name){
  out_file <- file.path(dir, file_name)

  check_dir(dir)

  try_save <- try(qs_save(object, out_file), silent = T)

  if(inherits(try_save, 'try-error')){
    tmp_file <- temp_file(file_name)
    qs_save(object, tmp_file)
    no_write_msg(out_file, tmp_file)
  }
}

# exit spada with saving session ----------------------------------------------
exit_sleep_time <- 3

exit_with_save <- function(session){

  show_exit_screen()
  update_waiter_text(session, 'spada_exit_text', 'Saving data')
  spada_save(session$userData$conf$data_dir,
             session$userData$dt$dt,
             'data.qs2')
  Sys.sleep(exit_sleep_time/3)
  update_waiter_text(session, 'spada_exit_text', 'Saving output')
  spada_save(session$userData$conf$data_dir,
             session$userData$out$elements,
             'output.qs2')
  Sys.sleep(exit_sleep_time/3)
  update_waiter_text(session, 'spada_exit_text', 'Saving conf')
  spada_save(session$userData$conf$conf_dir,
             reactiveValuesToList(session$userData$conf),
             'conf.qs2')
  Sys.sleep(exit_sleep_time/3)
  update_waiter_text(session, 'spada_exit_text', 'Done')
  session$sendCustomMessage(type = 'closeWindow', message = 'message')
  if(session$userData$run_local) stopApp()
}

# exit spada without saving session -------------------------------------------
exit_without_save <- function(session){

  show_exit_screen(F)

  spada_save(session$userData$conf$conf_dir,
             reactiveValuesToList(session$userData$conf),
             'conf.qs2')
  Sys.sleep(exit_sleep_time)
  update_waiter_text(session, 'spada_exit_text', 'Done')
  session$sendCustomMessage(type = 'closeWindow', message = 'message')
  if(session$userData$run_local) stopApp()
}

# test write access -----------------------------------------------------------
test_write_access <- function(file){
  all(!file.exists(file) || isTRUE(file.access(file, 2) == 0))
}

# temp normalized file path ---------------------------------------------------
temp_file <- function(file_name = ''){
  file.path(normalizePath(tempdir(), winslash = '/'), file_name)
}

# no access to write message --------------------------------------------------
no_write_msg <- function(file1, file2){
  message(paste(
    Sys.time(),
    '- Spada could not write on:\n',
    file1,
    '\n Data saved in',
    file2
    )
  )
}

# make names append lists -----------------------------------------------------
make_names_append_list <- function(new_list, actual_names,
                                   suffix = '_previous', index = 1) {

    names(new_list) <- make.names(names(new_list), unique = TRUE)

    while (any(grepl(paste0("^.*", suffix, "$"), actual_names))) {
      suffix <- paste0(suffix, "_", i)
      index <- index + 1
    }

    conflict_names <- names(new_list) %in% actual_names
    names(new_list)[conflict_names] <- paste0(names(new_list)[conflict_names], suffix)

    return(new_list)
}

# plot z test -----------------------------------------------------------------
plot_z_test <- function(confidence = 0.95, test_type = 'two.sided',
                        z_value = qnorm(confidence),
                        color_fill = 'brown3', color_line = 'steelblue') {
  # Define the standardized x-axis (Z-scores)
  x <- seq(-4, 4, length.out = 1000)  # Standardized Z-values
  y <- dnorm(x, mean = 0, sd = 1)  # Standard normal distribution

  alpha <- 1 - confidence

  # Determine critical values in standardized scale
  if (test_type == 'two.sided') {
    z_critical <- qnorm(1 - alpha / 2)
    critical_left <- -z_critical
    critical_right <- z_critical
  } else if (test_type == 'greater') {
    critical_right <- qnorm(1 - alpha)
  } else if (test_type == 'less') {
    critical_left <- qnorm(alpha)
  } else {
    stop("Invalid 'test_type'. Use 'two.sided', 'greater', or 'less'.")
  }

  # Create the plot
  plot(x, y, type = 'l', lwd = 2, col = color_line,
       xlab = 'Standardized Values (Z)', ylab = 'Density',
       main = paste('Z Test\n Type:', test_type, '- Conf.',
                    confidence * 100, '% | Z value:', z_value |> f_num(dig = 3)))

  # Highlight the critical regions
  if (test_type == 'two.sided') {
    polygon(c(x[x <= critical_left], critical_left),
            c(y[x <= critical_left], 0), col = color_fill, border = NA)
    polygon(c(x[x >= critical_right], critical_right),
            c(y[x >= critical_right], 0), col = color_fill, border = NA)

  } else if (test_type == 'less') {
    polygon(c(x[x <= critical_left], critical_left),
            c(y[x <= critical_left], 0), col = color_fill, border = NA)

  } else if (test_type == 'greater') {
    polygon(c(x[x >= critical_right], critical_right),
            c(y[x >= critical_right], 0), col = color_fill, border = NA)
  }

  if(abs(z_value) <= 4){
    abline(v = z_value, col = 'black', lwd = 2, lty = 2)
  }

}

# plot t test -----------------------------------------------------------------
plot_t_test <- function(confidence = 0.95, test_type = 'two.sided',
                        df = 10, t_value = qt(confidence, df = 10),
                        color_fill = 'brown3', color_line = 'steelblue') {
    x <- seq(-4, 4, length.out = 1000)
  y <- dt(x, df = df)

  alpha <- 1 - confidence

  # Determine critical values in standardized scale
  if (test_type == 'two.sided') {
    t_critical <- qt(1 - alpha / 2, df = df)
    critical_left <- -t_critical
    critical_right <- t_critical
  } else if (test_type == 'greater') {
    critical_right <- qt(1 - alpha, df = df)
  } else if (test_type == 'less') {
    critical_left <- qt(alpha, df = df)
  } else {
    stop("Invalid 'test_type'. Use 'two.sided', 'greater', or 'less'.")
  }

  # Create the plot
  plot(x, y, type = 'l', lwd = 2, col = color_line,
       xlab = 't-value', ylab = 'Density',
       main = paste(
         'One-sample t Test\n',
         'Type:', test_type,
         '- Conf.', confidence * 100,
         '% | t value =', t_value |> f_num(dig = 3),
        '(df =', df |> f_int(), ')'
       )
  )

  # Highlight the critical regions
  if (test_type == 'two.sided') {
    polygon(c(x[x <= critical_left], critical_left),
            c(y[x <= critical_left], 0), col = color_fill, border = NA)
    polygon(c(x[x >= critical_right], critical_right),
            c(y[x >= critical_right], 0), col = color_fill, border = NA)

  } else if (test_type == 'less') {
    polygon(c(x[x <= critical_left], critical_left),
            c(y[x <= critical_left], 0), col = color_fill, border = NA)

  } else if (test_type == 'greater') {
    polygon(c(x[x >= critical_right], critical_right),
            c(y[x >= critical_right], 0), col = color_fill, border = NA)
  }

  if(abs(t_value) <= 4){
    abline(v = t_value, col = 'black', lwd = 2, lty = 2)
  }

}

# test if is hex color --------------------------------------------------------
is_hex_color <- function(x) {
  grepl("^#[A-Fa-f0-9]{6}$", x)
}

# test output format-----------------------------------------------------------
test_output_format <- function(output){

  # if list of len 0 is OK
  (is.list(output) && length(output) == 0) ||

    # if len > 0 must pass
    (is.list(output) && length(output) > 0 &&

       # all inside elements must be lists
       all(sapply(output, class) == 'list') &&

       # all names of inside lists must match
       all(
         sapply(output, \(x) {
           all(names(x) %in% c('id', 'title', 'annotation', 'element', 'card', 'btn_x', 'btn_e'))
         })
       ) &&
       # test class of each element in the inside lists
       (
         all(sapply(output, \(x){ 'character' %in% class(x$id)})) &&
           all(sapply(output, \(x){ 'character' %in% class(x$title) })) &&
           all(sapply(output, \(x){ 'character' %in% class(x$annotation) })) &&
           all(sapply(output, \(x){ any(c('gt_tbl', 'html', 'shiny.tag') %in% class(x$element)) })) &&
           all(sapply(output, \(x){ 'shiny.tag' %in% class(x$card) })) &&
           all(sapply(output, \(x){ 'shiny.tag' %in% class(x$btn_x) })) &&
           all(sapply(output, \(x){ 'shiny.tag' %in% class(x$btn_e) }))
       )
    )
}

# check existence of directory ------------------------------------------------
check_dir <- function(dir){
  if(!dir.exists(dir)) dir.create(dir, recursive = T)
}

# linear model df output ------------------------------------------------------
linear_model_df_output <- function(model_summary){
  stopifnot(model_summary |> class() == 'summary.lm')

  table_summary <- as.data.frame(model_summary$coefficients)

  table_summary$Variable <- rownames(table_summary)
  rownames(table_summary) <- NULL

  table_summary <- table_summary[, c("Variable", "Estimate", "Std. Error",
                                     "t value", "Pr(>|t|)")]

  table_summary <- table_summary |> as.data.table()
  table_summary[, `Sig Levels` := fcase(
    `Pr(>|t|)` < 0.001, '***',
    `Pr(>|t|)` < 0.01, '**',
    `Pr(>|t|)` < 0.05, '*',
    `Pr(>|t|)` < 0.1, '.',
    default = ''
  )]
}

# linear model gt metrics -----------------------------------------------------
linear_model_df_metrics <- function(model_summary){
  stopifnot(model_summary |> class() == 'summary.lm')

  f <- model_summary$fstatistic
  f_p_value <- pf(f['value'], f['numdf'], f['dendf'], lower.tail = FALSE)
  data.frame(
    Metric = c('Residual Std Error', 'R-squared', 'Adjusted R-squared',
               'F-statistic', 'F p-value'),
    Value = c(
      model_summary$sigma,
      model_summary$r.squared,
      model_summary$adj.r.squared,
      model_summary$fstatistic[1],
      f_p_value
    )
  )
}

# get active dt ---------------------------------------------------------------
get_act_dt <- function(session){
  session$userData$dt$dt[[ session$userData$dt$act_name ]]
}

# update active dt ------------------------------------------------------------
update_act_dt <- function(session, new_dt, data_changed = TRUE,
                          updated_cols = NULL,
                          change_type = c('new_data',
                                          'convert_cols',
                                          'rename_cols',
                                          'order_cols',
                                          'select_cols')){

  stopifnot(new_dt |> is.data.table())

  change_type <- match.arg(change_type)

  new_dt <- lapply(new_dt, make_valid_cols) |> as.data.table()

  session$userData$dt$dt[[ session$userData$dt$act_name ]] <- new_dt

  if(data_changed) {

    session$userData$dt$updated_cols <- updated_cols
    session$userData$dt$data_changed_type <- change_type
    session$userData$data_changed(session$userData$data_changed() + 1)
  }
}

# append dt -------------------------------------------------------------------
append_dt <- function(session, new_dt, new_dt_name) {

  stopifnot(new_dt |> is.data.frame())
  stopifnot(new_dt_name |> is_valid_name() &&
              new_dt_name %notin% names(session$userData$dt$dt))

  new_dt <- lapply(new_dt, make_valid_cols) |> as.data.table() |> make_var_names()

  new_list <- list(new_dt)
  names(new_list) <- new_dt_name

  session$userData$dt$dt <- c(session$userData$dt$dt, new_list)
}

# append meta data ------------------------------------------------------------
append_meta <- function(session, new_meta, new_dt_name){
  stopifnot(new_meta |> is.data.frame())

  meta_list <- list(new_meta)
  names(meta_list) <- new_dt_name

  session$userData$dt$meta <- c(
    session$userData$dt$meta,
    meta_list
  )
}

# update meta data ------------------------------------------------------------
update_meta <- function(dt = data.table(),
                        previous_meta, col_names, updated_cols, ncols,
                        change_type = c('new_data',
                                        'convert_cols',
                                        'rename_cols',
                                        'order_cols',
                                        'select_cols')){

  stopifnot(dt |> is.data.table())
  stopifnot(previous_meta |> is.data.table())

  change_type <- match.arg(change_type)

  if(change_type %in% c('new_data', 'convert_cols')){
    new_meta <- dt |> df_info()

    # delete updated vars from previous meta
    previous_meta <- previous_meta[var %notin% updated_cols, ]

    new_meta <- rbindlist(list(previous_meta, new_meta), fill = T)

    order_df <- data.table(var = col_names, index = 1:length(col_names))

    new_meta <- merge(new_meta, order_df, by = 'var', all.x = TRUE)

    setorderv(new_meta, c('index'))

    new_meta$index <- NULL
    new_meta$cols <- ncols

    return(new_meta)
  } else if (change_type == 'rename_cols'){
    previous_meta$var <- col_names
    return(previous_meta)
  } else if (change_type == 'order_cols'){

    order_df <- data.table(var = col_names, index = 1:length(col_names))

    meta_reordered <- merge(previous_meta, order_df, by = 'var', all.x = TRUE)

    setorderv(meta_reordered, c('index'))

    meta_reordered$index <- NULL

    return(meta_reordered)

  } else if (change_type == 'select_cols'){
    previous_meta <- previous_meta[var %in% col_names, ]

    previous_meta$cols <- ncols
    return(previous_meta)
  }
}

# descriptive stats -----------------------------------------------------------
desc_stats <- function(df = NULL,
                       fmt_digits = 9,
                       central_tendency = c('mean', 'gmean', 'hmean', 'median', 'mode'),
                       dispersion = c('min', 'max', 'IQR', 'range', 'var', 'sd'),
                       shape = c('skew', 'kurt')) {

  desc_stats <- list()

  # central tendency
  if('mean' %in% central_tendency){
    desc_stats$Mean <- sapply(
      df,
      \(x) {if(x |> is.numeric()) mean_nona(x) |> f_num(dig = fmt_digits) else NA })
  }

  if('gmean' %in% central_tendency){
    desc_stats$Gmean <- sapply(
      df,
      \(x) {if(x |> is.numeric()) suppressWarnings(Gmean(x, na.rm = T)) |> f_num(dig = fmt_digits) else NA })
  }

  if('hmean' %in% central_tendency){
    desc_stats$Hmean <- sapply(
      df,
      \(x) {if(x |> is.numeric()) Hmean(x, na.rm = T) |> f_num(dig = fmt_digits) else NA })
  }

  if('median' %in% central_tendency){
    desc_stats$Median <- sapply(
      df,
      \(x) {if(x |> is.numeric()) fmedian(x, na.rm = T) |> f_num(dig = fmt_digits)  else NA })
  }

  if('mode' %in% central_tendency){
    desc_stats$Mode <- sapply(
      df,
      \(x) {if(x |> is.numeric() ||
               x |> is.character() ||
               x |> is.factor()){
        x_mode <- fmode(x, na.rm = T) |> f_num(dig = fmt_digits)
        if(is.na(x_mode) |> all()) NA else paste(x_mode, collapse = ' | ')
      } else { NA }
      })
  }

  # dispersion
  if('min' %in% dispersion){
    desc_stats$Min <- sapply(
      df,
      \(x) {if(x |> is.numeric()) min_nona(x) |> f_num(dig = fmt_digits) else NA })
  }

  if('max' %in% dispersion){
    desc_stats$Max <- sapply(
      df,
      \(x) {if(x |> is.numeric()) max_nona(x) |> f_num(dig = fmt_digits) else NA })
  }

  if('IQR' %in% dispersion){
    desc_stats$IQR <- sapply(
      df,
      \(x) {if(x |> is.numeric()) IQR(x, na.rm = T) |> f_num(dig = fmt_digits) else NA })
  }

  if('range' %in% dispersion){
    desc_stats$Range <- sapply(
      df,
      \(x) {
        if(x |> is.numeric()){
          paste('[', frange(x, na.rm = T) |> f_num(dig = fmt_digits) , ']', collapse = '--->')
        } else { NA }
      }
    )
  }

  if('var' %in% dispersion){
    desc_stats$Variance <- sapply(
      df,
      \(x) {if(x |> is.numeric()) fvar(x, na.rm = T) |> f_num(dig = fmt_digits) else NA })
  }

  if('sd' %in% dispersion){
    desc_stats[['Standard Deviation']] <- sapply(
      df,
      \(x) {if(x |> is.numeric()) fsd(x, na.rm = T) |> f_num(dig = fmt_digits) else NA })
  }

  if('skew' %in% shape){
    desc_stats[['Skewness']] <- sapply(
      df,
      \(x) {if(x |> is.numeric()) Skew(x, na.rm = T) |> f_num(dig = fmt_digits) else NA })
  }

  if('kurt' %in% shape){
    desc_stats[['Kurtosis']] <- sapply(
      df,
      \(x) {if(x |> is.numeric()) Kurt(x, na.rm = T) |> f_num(dig = fmt_digits) else NA })
  }

  desc_stats
}

# plots -----------------------------------------------------------------------
spada_plot <- function(
    type = 'hist',
    df = NULL,
    xvar = NULL,
    yvar = NULL,
    zvar = NULL,
    xlab = '',
    ylab = '',
    fill_color = '#0099F8',
    line_color = '#000000',
    title_color = '#02517d',
    title = NULL,
    bins = 25,
    vertical_line = NULL,
    point_shape = '.',
    line_type = 1,
    mean_value = NULL,
    sd_value = NULL,
    sample_limit = 1e5,
    na_rm = TRUE
  ){

  if(nrow(df) > sample_limit){
    df <- df[sample.int(nrow(df), sample_limit), , drop = FALSE]
  }

  stopifnot(is.data.frame(df))

  if(type == 'hist'){

    ggplot(data = df, aes(x = .data[[xvar]])) +
      geom_histogram(
        bins = bins,
        fill = fill_color,
        color = '#000000',
        na.rm = na_rm
      ) +
      geom_vline(xintercept = vertical_line, color = line_color, linetype = line_type) +
      labs(title = title, x = xlab, y = ylab) +
      theme_classic() +
      theme(axis.text.x = element_text(size = 14),
            axis.text.y = element_text(size = 14),
            axis.title.y = element_text(size = 16),
            plot.title = element_text(color = title_color, size = 16, face = 'bold')
      )
  } else if (type == 'hist_density'){

    ggplot(data = df, aes(x = .data[[xvar]])) +
      geom_histogram(aes(y = after_stat(density)),
                     bins = bins,
                     fill = fill_color,
                     color = 'black',
                     na.rm = na_rm) +
      stat_function(fun = dnorm,
                    args = list(mean = mean_value, sd = sd_value),
                    color = line_color,
                    linewidth = 1) +
      labs(title = title, x = xlab, y = ylab) +
      theme_classic() +
      theme(axis.text.x = element_text(size = 14),
            axis.text.y = element_text(size = 14),
            axis.title.x = element_text(size = 16),
            axis.title.y = element_text(size = 16),
            plot.title = element_text(color = title_color, size = 16, face = 'bold')
      )

  } else if (type == 'boxplot'){

    ggplot(data = df, aes(x = .data[[xvar]])) +
      stat_boxplot(geom = 'errorbar', width = 0.3, na.rm = na_rm) +
      geom_boxplot(fill = fill_color, na.rm = na_rm) +
      ylim(-1.2, 1.2) +
      geom_vline(xintercept = vertical_line, color = line_color, linetype = line_type) +
      labs(title = title, x = xlab, y = ylab) +
      theme_classic() +
      theme(
        axis.ticks.y = element_blank(),
        axis.text.y  = element_blank(),
        axis.line.y  = element_blank(),
        panel.border = element_rect(color = '#000000', fill = NA),
        axis.text.x = element_text(size = 14),
        plot.title = element_text(color = title_color, size = 16, face = 'bold')
      )
  } else if(type == 'dots'){

    ggplot(data = df, aes(x = .data[[xvar]], y = .data[[yvar]])) +
      geom_point(shape = point_shape, color = fill_color, na.rm = na_rm) +
      geom_hline(yintercept = vertical_line, color = line_color, linetype = line_type) +
      labs(title = title, x = xlab, y = ylab) +
      theme_classic() +
      theme(axis.text.x = element_text(size = 14),
            axis.text.y = element_text(size = 14),
            axis.title.x = element_text(size = 16),
            axis.title.y = element_text(size = 16),
            plot.title = element_text(color = title_color, size = 16, face = 'bold')
      )
  } else if (type == 'barplot'){

    ggplot(data = df, aes(x = factor(.data[[xvar]]))) +
      geom_bar(fill = fill_color, na.rm = na_rm) +
      labs(title = title, x = xlab, y = ylab) +
      theme_classic() +
      theme(axis.text.x = element_text(size = 14),
            axis.text.y = element_text(size = 14),
            axis.title.y = element_text(size = 16),
            plot.title = element_text(color = title_color, size = 16, face = 'bold')
      )

  } else if (type == 'boxplot_group'){

    ggplot(data = df, aes(x = .data[[xvar]], y = .data[[yvar]], fill = .data[[xvar]])) +
      stat_boxplot(geom = 'errorbar', width = 0.3, na.rm = na_rm) +
      geom_boxplot(orientation = 'x', na.rm = na_rm) +
      geom_hline(yintercept = vertical_line,
                 color = line_color) +
      coord_flip() +
      labs(title = title, x = xlab, y = ylab) +
      theme_classic() +
      theme(
        legend.position = 'none',
        axis.ticks.y = element_blank(),
        axis.line.y  = element_blank(),
        panel.border = element_rect(color = 'black', fill = NA),
        axis.text.x = element_text(size = 14),
        axis.text.y = element_text(size = 14),
        plot.title = element_text(color = title_color, size = 16, face = 'bold')
      )
  } else if (type == 'scatter'){

    ggplot(data = df, aes(x = .data[[xvar]], y = .data[[yvar]])) +
      geom_point(color = fill_color, shape = point_shape, na.rm = na_rm) +
      labs(title = title, x = xlab, y = ylab) +
      theme_classic() +
      theme(axis.text.x = element_text(size = 14),
            axis.text.y = element_text(size = 14),
            axis.title.x = element_text(size = 16),
            axis.title.y = element_text(size = 16),
            plot.title = element_text(color = title_color, size = 16, face = 'bold')
      )
  } else if (type == 'qq_plot'){
    ggplot(data = df, aes(sample = .data[[xvar]])) +
      stat_qq(color = fill_color, na.rm = na_rm) +
      stat_qq_line(color = line_color, na.rm = na_rm) +
      labs(title = title, x = xlab, y = ylab) +
      theme_classic() +
      theme(axis.text.x = element_text(size = 14),
            axis.text.y = element_text(size = 14),
            axis.title.x = element_text(size = 16),
            axis.title.y = element_text(size = 16),
            plot.title = element_text(color = title_color, size = 16, face = 'bold')
      )
  }
}

# plotTag with attributes -----------------------------------------------------
plot_tag <- function(plot, w = 900, h = 300){
  plotTag(
    plot,
    '',
    width = w,
    height = h,
    suppressSize = 'xy',
    attribs = list(style = 'max-width:90%; height:auto;')
  )
}

# status row for restore session ----------------------------------------------
status_row <- function(icon_name, color, title, type_icon){
  div(
    style = paste0(
      'display:flex;
       align-items:center;
       gap:16px;
       padding:18px 20px;
       border-radius:16px;
       background:#FFFFFF;
       border-left:5px solid ', color, ';
       box-shadow:0 2px 10px rgba(0,0,0,.05);'
    ),

    div(
      style = paste0(
        'width:44px;
         height:44px;
         border-radius:12px;
         background:', color, '18;
         display:flex;
         align-items:center;
         justify-content:center;
         flex-shrink:0;'
      ),
      icon(icon_name, style = paste0('font-size:22px;color:', color))
    ),

    div(
      style = '
        flex:1;
        font-size:20px;
        font-weight:400;
        color:#1C4B71;
      ',
      title
    ),

    div(
      style = '
        color:#607B95;
        font-size:26px;
        opacity:.75;
      ',
      icon(type_icon)
    )
  )
}

# display restore status ------------------------------------------------------
display_restore_status <- function(session_restore_status, btn_ok){
  list_check_restore <- div(
    style = 'display:flex; flex-direction:column; gap:12px;',

    switch(
      substr(session_restore_status, 1, 1),
      '1' = status_row('check', '#2e7d32', 'Data restored successfully', 'database'),
      '2' = status_row('times', '#c62828', 'Data not found', 'database'),
      '3' = status_row('circle-question', '#ed6c02', 'Data in invalid format', 'database')
    ),

    switch(
      substr(session_restore_status, 3, 3),
      '1' = status_row('check', '#2e7d32', 'Output restored successfully', 'tablet-screen-button'),
      '2' = status_row('times', '#c62828', 'Output not found', 'tablet-screen-button'),
      '3' = status_row('circle-question', '#ed6c02', 'Output in invalid format', 'tablet-screen-button')
    )
  )

  # show modal
  showModal(modalDialog(
    title = div(h2(
      bs_icon('database-up', size = '40px',
              style = 'margin-right: 8px; color:#FFFFFF'),
      'Session Status'
    )),
    div(style = 'padding:12px; border-radius:0px;', list_check_restore),
    size = 'l',
    easyClose = T,
    footer = div(style = 'text-align:right;', btn_ok)

  ))
}

# summarise dataset -----------------------------------------------------------
summarise_dt <- function(dt, fun = c('count', 'distinct'), vars){

  stopifnot(is.data.table(dt))
  stopifnot(all(vars %in% (dt |> names())))

  fun <- match.arg(fun)
  if(fun == 'distinct'){
    dt[, .SD, .SDcols = vars] |> unique()
  } else if(fun == 'count'){
    dt[, .N, by = vars]
  }
}

# build calls -----------------------------------------------------------------
build_calls <- function(new, funs, vars){

  vars_as_symbols <- lapply(vars, as.name)

  calls <- Map(
    \(v, f) call(f, v),
    vars_as_symbols,
    funs
  )

  as.call(c(as.name('.'), setNames(calls, new)))
}

# dt join ---------------------------------------------------------------------
dt_join <- function(dt1, dt2, vars1, vars2,
                     join_type = c('left', 'right', 'full', 'inner')){

  stopifnot(is_spada_dt(dt1))
  stopifnot(is_spada_dt(dt2))
  stopifnot(length(vars1) > 0)
  stopifnot(length(vars2) > 0)

  join_type <- match.arg(join_type)

  types_x <- sapply(dt1[, .SD, .SDcols = vars1], obj_type)
  types_y <- sapply(dt2[, .SD, .SDcols = vars2], obj_type)

  stopifnot('The variables types must match' = all(types_x == types_y))

  temp <- merge(
    dt1,
    dt2,
    by.x = vars1,
    by.y = vars2,
    all = join_type == 'full',
    all.x = any(c('left', 'full') %in% join_type),
    all.y = any(c('right', 'full') %in% join_type),
    allow.cartesian = T,
    suffixes = c('_x', '_y')
  )
  return(temp)
}

# math and other functions ----------------------------------------------------

df_info <- function(df, type = c('basic', 'complete')){
  stopifnot(is.data.frame(df))

  type = match.arg(type)

  if (fncol(df) == 0) {
    dt <- data.table(
      var = 'v1',
      type = NA,
      class = NA,
      min = NA,
      max = NA,
      n_valid = NA,
      perc_valid = NA,
      n_zero = NA,
      perc_zero = NA,
      n_nas = NA,
      perc_nas = NA,
      rows = NA,
      cols = NA
    )
    if(type == 'complete'){
      dt[, `:=`(n_unique = NA, perc_unique = NA, size = 0)]
    }

  } else {
    rows <- fnrow(df)
    cols <- fncol(df)

    res <- lapply(seq_len(cols), function(j) {
      x <- df[[j]]

      nas <- whichNA(x) |> NROW()
      valid <- rows - nas

      if(type == 'complete'){ uniq <- fnunique(x) }

      if (is.numeric(x)){
        zeros <- NROW(x %==% 0)

        rangev <- frange(x, na.rm = T)
        minv <- rangev[1]
        maxv <- rangev[2]
      } else {
        zeros <- 0
        minv <- NA
        maxv <- NA
      }

      if(type == 'basic'){
        list(
          type = typeof(x),
          class = paste(class(x), collapse = '/'),
          min = minv,
          max = maxv,
          n_valid = valid,
          perc_valid = valid / rows,
          n_zero = zeros,
          perc_zero = zeros / rows,
          n_nas = nas,
          perc_nas = nas / rows
        )
      } else if (type == 'complete'){
        list(
          type = typeof(x),
          class = paste(class(x), collapse = '/'),
          size = as.numeric(object.size(x)),
          min = minv,
          max = maxv,
          n_valid = valid,
          perc_valid = valid / rows,
          n_unique = uniq,
          perc_unique = uniq / rows,
          n_zero = zeros,
          perc_zero = zeros / rows,
          n_nas = nas,
          perc_nas = nas / rows
        )
      }
    })

    dt <- data.table::rbindlist(res)
    dt$var <- names(df)
    dt[, `:=`(rows = rows, cols = cols)]

  }

  if(type == 'basic'){
    setcolorder(
      dt,
      neworder = c(
        'var',
        'type',
        'class',
        'min',
        'max',
        'n_valid',
        'perc_valid',
        'n_zero',
        'perc_zero',
        'n_nas',
        'perc_nas',
        'rows',
        'cols')
    )
  } else if(type =='complete'){
    setcolorder(
      dt,
      neworder = c(
        'var',
        'type',
        'class',
        'size',
        'min',
        'max',
        'n_valid',
        'perc_valid',
        'n_unique',
        'perc_unique',
        'n_zero',
        'perc_zero',
        'n_nas',
        'perc_nas',
        'rows',
        'cols')
    )
  }
  return(dt)
}

f_num <- function(x, big = ',', dec = '.', thousand = 'K',
                  million = 'M', billion = 'B', dig = 0){

  if(x |> is.numeric()){
    fcase(
      is.infinite(x), paste(x),
      x > 1e9,
      paste(format(round(x/1e9, digits = dig), nsmall = dig,
                   decimal.mark = dec, big.mark = big, scientific = F), billion),
      x > 1e6,
      paste(format(round(x/1e6, digits = dig), nsmall = dig,
                   decimal.mark = dec, big.mark = big, scientific = F), million),
      x > 1e3,
      paste(format(round(x/1e3, digits = dig), nsmall = dig,
                   decimal.mark = dec, big.mark = big, scientific = F), thousand),
      default = format(round(x, dig), nsmall = dig, scientific = F)
    )
  } else {
    x
  }
}

f_int <- function(int){
  formatC(int, format = 'd', big.mark = ',', digits = 0)
}

gt_info <- function(df, df_name){
  stopifnot(is.data.frame(df))

  dt <- df |> as.data.table()

  dt[, f_icon := fcase(
    class == 'integer', '1',
    class == 'character', 'a',
    class == 'numeric', 'calculator',
    class %in% c('Date', 'IDate/Date', 'POSIXt', 'POSIXct', 'POSIXlt',
                 'POSIXct/POSIXt', 'POSIXlt/POSIXt'), 'calendar',
    class == 'factor', 'sitemap',
    class == 'raw', 'sd-card',
    class == 'complex', 'info',
    class == 'logical', 'puzzle-piece')]

  dt_gt <- dt[, `:=` (n_valid = f_num(n_valid, dig = 3),
                      n_zero = f_num(n_zero, dig = 3),
                      n_nas = f_num(n_nas, dig = 3))] |>
    gt()

  # if all NA do nothing
  if(!allNA(dt$min)){
    dt_gt <- dt_gt |>
      data_color(columns = 'min', palette = yl_palette)
  }

  # if all NA do nothing
  if(!allNA(dt$max)){
    dt_gt <- dt_gt |>
      data_color(columns = 'max', palette = pk_palette)
  }

  dt_gt |>
    tab_header(
      title = df_name,
      subtitle = paste(f_int(dt$rows[1]), 'rows -', f_int(dt$cols[1]), 'variables')
    ) |>
    cols_hide(columns = c('rows', 'cols')) |>
    fmt_percent(columns = c('perc_valid', 'perc_zero', 'perc_nas')) |>
    data_color(columns = 'n_valid', palette = lg_palette) |>
    data_color(columns = 'n_zero', palette = gray_palette) |>
    data_color(columns = 'n_nas', palette = red_palette) |>
    fmt_integer(columns = c('n_valid', 'n_nas')) |>
    fmt_number(columns = c('min', 'max', 'n_valid', 'n_zero', 'n_nas')) |>
    sub_missing() |>
    cols_move(columns = 'f_icon', after = 'var') |>
    fmt_icon(columns = 'f_icon') |>
    cols_label(
      var = 'Variable',
      type = 'Type',
      class = 'Class',
      min = 'Min',
      max = 'Max',
      n_valid = 'Valid',
      n_zero = 'Zeros',
      n_nas = "NA's",
      f_icon = ''
    ) |>
    cols_width(f_icon ~ px(30)) |>
    cols_merge(columns = c('class', 'f_icon'), pattern = "{2} {1}") |>
    cols_merge(columns = c('n_valid', 'perc_valid'), pattern = "{1} / {2}") |>
    cols_merge(columns = c('n_zero', 'perc_zero'), pattern = "{1} / {2}") |>
    cols_merge(columns = c('n_nas', 'perc_nas'), pattern = "{1} / {2}") |>
    tab_options(table.background.color = '#ffffff', heading.title.font.size = 20)
}

is_date <- function(x){
  inherits(x, c('Date', 'POSIXt', 'POSIXct', 'POSIXlt'))
}

is_categorical <- function(x) {
  is.character(x) || is.factor(x) || is.logical(x)
}

is_valid_name <- function(x){
  x == make.names(x)
}

is_name_available <- function(n, names){
  is_valid_name(n) && n %notin% names
}

is_var_available <- function(var_name, df){
  is_valid_name(var_name) && var_name %notin% names(df)
}

# wrapper functions from collapse package for speed ---------------------------
first_na <- function(x){
  ffirst(x, na.rm = F)
}

first_nona <- function(x){
  ffirst(x, na.rm = T)
}

last_na <- function(x){
  flast(x, na.rm = F)
}

last_nona <- function(x){
  flast(x, na.rm = T)
}

max_na <- function(x){
  fmax(x, na.rm = F)
}

max_nona <- function(x){
  fmax(x, na.rm = T)
}

mean_na <- function(x){
  fmean(x, na.rm = F)
}

mean_nona <- function(x){
  fmean(x, na.rm = T)
}

median_na <- function(x){
  fmedian(x, na.rm = F)
}

median_nona <- function(x){
  fmedian(x, na.rm = T)
}

min_na <- function(x){
  fmin(x, na.rm = F)
}

min_nona <- function(x){
  fmin(x, na.rm = T)
}

sum_na <- function(x){
  fsum(x, na.rm = F)
}

sum_nona <- function(x){
  fsum(x, na.rm = T)
}

# tests/testthat/test-lm_module.R

# test lm module - variable selection -----------------------------------------
test_that('lm_server test variable selection', {
  testServer(lm_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$setInputs(
      sel_yvar = 'Petal.Length',
      sel_xvar = c('Sepal.Length')
    )

    expect_equal(var_analysis(), names(iris))
    expect_equal(yvar(), setdiff(names(iris), 'Species'))
    expect_equal(xvar(), setdiff(names(iris), c('Petal.Length')))
  })
})

# test lm module - run model --------------------------------------------------
test_that('lm_server - test run model', {
  testServer(lm_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$setInputs(
      sel_yvar = 'Petal.Length',
      sel_xvar = c('Sepal.Length'),
      btn_run_lm = 1
    )

    expect_equal(linear_model$y_name, 'Petal.Length')
    expect_equal(linear_model$x_name, 'Sepal.Length')

    m1 <- lm(Petal.Length ~ Sepal.Length, data = iris, model = F)
    s1 <- summary(m1)

    expect_equal(linear_model$summary$coefficients, s1$coefficients)
    expect_equal(linear_model$summary$r.squared, s1$r.squared)
    expect_equal(linear_model$summary$adj.r.squared, s1$adj.r.squared)
    expect_equal(linear_model$summary$fstatistic, s1$fstatistic)
  })
})

# test lm module - gt tables --------------------------------------------------
test_that('lm_server - test gt tables', {
  testServer(lm_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$setInputs(
      sel_yvar = 'Petal.Length',
      sel_xvar = c('Sepal.Length'),
      btn_run_lm = 1
    )

    expect_s3_class(lm_var_table(), 'gt_tbl')
    expect_s3_class(lm_metrics(), 'gt_tbl')
  })
})


# test lm module - save model -------------------------------------------------
test_that('lm_server - save model', {
  testServer(lm_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$setInputs(
      sel_yvar = 'Petal.Length',
      sel_xvar = c('Sepal.Length'),
      btn_run_lm = 1,
      btn_save_model = 1,
      model_filename = 'my_model'
    )

    expect_equal(output$down_handler |> basename(), 'my_model.RDS')
  })
})

# test lm module - plot residuals ---------------------------------------------
test_that('lm_server - plot residuals', {
  testServer(lm_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$userData$conf <- reactiveValues(
      plot_fill_color = '#229999',
      plot_line_color = '#44aa44',
      plot_limit = 1e5
    )

    session$setInputs(
      sel_yvar = 'Petal.Width',
      sel_xvar = 'Petal.Length',
      btn_run_lm = 1
    )

    expect_equal(update_lm_resid_plot(), 0)

    session$setInputs(
      radio_lm_resid = 'hist',
      btn_lm_resid = 1
    )

    expect_equal(update_lm_resid_plot(), 1)
    expect_s3_class(lm_resid_plot(), 'ggplot')
    expect_true(is.reactive(lm_resid_plot))
  })
})

# tests/testthat/test-cor_matrix_module.R

# test correlation matrix -----------------------------------------------------
test_that('test correlation matrix - pearson method', {
  testServer(cor_matrix_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    # inputs ----------------------------------------------------------------
    session$setInputs(
      sel_vars = c('Sepal.Length', 'Sepal.Width'),
      radio_method = 'pearson',
      btn_cor_matrix = 1
    )

    while(task_cor_matrix$status() == 'running'){
      session$flushReact()
    }

    expect_s3_class(df_cor(), 'data.table')

    expect_equal(c('Var1', 'Var2', 'value', 'label', 'method'), names(df_cor()))
    expect_equal(nrow(df_cor()), 4)

    expect_equal(
      df_cor()[Var1 == 'Sepal.Length' & Var2 == 'Sepal.Width', round(value, 3)],
      round(cor(iris$Sepal.Length, iris$Sepal.Width), 3)
    )
  })
})

# tests/testthat/test-utils.R

# test safe env function ------------------------------------------------------
test_that('test safe env', {
  e1 <- safe_env()
  expect_identical(parent.env(e1), emptyenv())
  expect_identical(e1 |> length(), 0L)

  e1 <- safe_env('sum')
  expect_identical(parent.env(e1), emptyenv())
  expect_identical(e1 |> length(), 1L)
  expect_identical(e1 |> ls(), 'sum')
  expect_identical(e1$sum, base::sum)
  expect_identical(e1$sum |> class(), 'function')
  expect_identical(e1$sum(1:3), 6L)
})

# test convert function -------------------------------------------------------
df <- data.frame(
  num = as.numeric(1:3),
  int = 1:3L,
  char = c('1', 'a', '1990-01-01'),
  date = as.Date(1:3),
  posix = as.POSIXct(as.Date(1:3)),
  fact = c('a', '3', '1950/01/01'),
  doub = as.double(1:3),
  comp = complex(real = 1:3, imaginary = 2:4)
)

# test convert function - numeric ---------------
test_that('convert function - numeric', {
  # to numeric
  expect_identical(convert(df$num, 'as.numeric'), c(1, 2, 3))

  # to integer
  expect_identical(convert(df$num, 'as.integer'), c(1L, 2L, 3L))

  # to char
  expect_identical(convert(df$num, 'as.character'), c('1', '2', '3'))

  # to date
  expect_identical(convert(df$num, 'as.Date'),
               as.Date(c('1970-01-02', '1970-01-03', '1970-01-04')))
  expect_s3_class(convert(df$num, 'as.Date'), 'Date')

  # to factor
  expect_identical(convert(df$num, 'as.factor'), df$num |> as.factor())

  # to double
  expect_identical(convert(df$num, 'as.double'), c(1, 2, 3))

  # to complex
  expect_identical(convert(df$num, 'as.complex'), c(1+0i, 2+0i, 3+0i))
})

# test convert function - integer ---------------
test_that('convert function - integer', {
  # to numeric
  expect_identical(convert(df$int, 'as.numeric'), c(1, 2, 3))

  # to integer
  expect_identical(convert(df$int, 'as.integer'), c(1L, 2L, 3L))

  # to char
  expect_identical(convert(df$int, 'as.character'), c('1', '2', '3'))

  # to date
  expect_identical(convert(df$int, 'as.Date'),
               as.Date(c('1970-01-02', '1970-01-03', '1970-01-04')))
  expect_s3_class(convert(df$int, 'as.Date'), 'Date')

  # to factor
  expect_identical(convert(df$int, 'as.factor'), df$num |> as.factor())

  # to double
  expect_identical(convert(df$int, 'as.double'), c(1, 2, 3))

  # to complex
  expect_identical(convert(df$int, 'as.complex'), c(1+0i, 2+0i, 3+0i))
})

# test convert function - char ---------------
test_that('convert function - char', {
  # to numeric
  expect_identical(convert(df$char, 'as.numeric'), c(1, NA, NA))

  # to integer
  expect_identical(convert(df$char, 'as.integer'), c(1L, NA, NA))

  # to char
  expect_identical(convert(df$char, 'as.character'), df$char)

  # to date
  expect_identical(convert(df$char, 'as.Date'), as.Date(c(NA, NA, '1990-01-01')))
  expect_s3_class(convert(df$char, 'as.Date'), 'Date')

  # to factor
  expect_identical(convert(df$char, 'as.factor'), df$char |> as.factor())

  # to double
  expect_identical(convert(df$char, 'as.double'), c(1, NA, NA))

  # to complex
  expect_identical(convert(df$char, 'as.complex'), c(1+0i, NA, NA))
})

# test convert function - date ---------------
test_that('convert function - date', {
  # to numeric
  expect_identical(convert(df$date, 'as.numeric'), c(1, 2, 3))

  # to integer
  expect_identical(convert(df$date, 'as.integer'), c(1L, 2L, 3L))

  # to char
  expect_identical(convert(df$date, 'as.character'), c('1970-01-02', '1970-01-03', '1970-01-04'))

  # to date
  expect_identical(convert(df$date, 'as.Date'), as.Date(c(1:3)))
  expect_s3_class(convert(df$date, 'as.Date'), 'Date')

  # to factor
  expect_identical(convert(df$date, 'as.factor'), df$date |> as.factor())

  # to double
  expect_identical(convert(df$date, 'as.double'), c(1, 2, 3))

  # to complex
  expect_identical(convert(df$date, 'as.complex'), c(1+0i, 2+0i, 3+0i))
})

# test convert function - posixct ---------------
test_that('convert function - posixct', {
  # to numeric
  expect_identical(convert(df$posix, 'as.numeric'), c(86400, 172800, 259200))
  # to integer
  expect_identical(convert(df$posix, 'as.integer'), c(86400L, 172800L, 259200L))

  # to char
  expect_identical(convert(df$posix, 'as.character'),
               c('1970-01-02', '1970-01-03', '1970-01-04'))

  # to date
  expect_identical(convert(df$posix, 'as.Date'),
               as.Date(c('1970-01-02', '1970-01-03', '1970-01-04')))
  expect_s3_class(convert(df$posix, 'as.Date'), 'Date')

  # to factor
  expect_identical(convert(df$posix, 'as.factor'), df$posix |> as.factor())

  # to double
  expect_identical(convert(df$posix, 'as.double'), c(86400, 172800, 259200))

  # to complex
  expect_identical(convert(df$posix, 'as.complex'), c(86400+0i, 172800+0i, 259200+0i))
})

# test convert function - factor ---------------
test_that('convert function - factor', {
  # to numeric
  expect_identical(convert(df$fact, 'as.numeric'), c(NA, 3, NA))

  # to integer
  expect_identical(convert(df$fact, 'as.integer'), c(NA, 3L, NA))

  # to char
  expect_identical(convert(df$fact, 'as.character'), c('a', '3', '1950/01/01'))

  # to date
  expect_identical(convert(df$fact, 'as.Date'), as.Date(c(NA, NA, NA)))
  expect_identical(convert(df$fact, 'as.Date', date_format = '%Y/%m/%d'),
               as.Date(c(NA, NA, '1950-01-01')))
  expect_s3_class(convert(df$fact, 'as.Date'), 'Date')

  # to factor
  expect_identical(convert(df$fact, 'as.factor'), df$fact |> as.factor())

  # to double
  expect_identical(convert(df$fact, 'as.double'), c(NA, 3, NA))

  # to complex
  expect_identical(convert(df$fact, 'as.complex'), c(NA, 3+0i, NA))
})

# test convert function - double ---------------
test_that('convert function - double', {
  # to numeric
  expect_identical(convert(df$doub, 'as.numeric'), c(1, 2, 3))

  # to integer
  expect_identical(convert(df$doub, 'as.integer'), c(1L, 2L, 3L))

  # to char
  expect_identical(convert(df$doub, 'as.character'), c('1', '2', '3'))

  # to date
  expect_identical(convert(df$doub, 'as.Date'), as.Date(c('1970-01-02', '1970-01-03', '1970-01-04')))
  expect_s3_class(convert(df$doub, 'as.Date'), 'Date')

  # to factor
  expect_identical(convert(df$doub, 'as.factor'), df$doub |> as.factor())

  # to double
  expect_identical(convert(df$doub, 'as.double'), c(1, 2, 3))

  # to complex
  expect_identical(convert(df$doub, 'as.complex'), c(1+0i, 2+0i, 3+0i))
})

# test convert function - complex ---------------
test_that('convert function - complex', {
  # to numeric
  expect_identical(convert(df$comp, 'as.numeric'), c(1, 2, 3))

  # to integer
  expect_identical(convert(df$comp, 'as.integer'), c(1L, 2L, 3L))

  # to char
  expect_identical(convert(df$comp, 'as.character'), c('1+2i', '2+3i', '3+4i'))

  # to date
  expect_identical(convert(df$comp, 'as.Date'), as.Date(c('1970-01-02', '1970-01-03', '1970-01-04')))
  expect_s3_class(convert(df$comp, 'as.Date'), 'Date')

  # to factor
  expect_identical(convert(df$comp, 'as.factor'), df$comp |> as.factor())

  # to double
  expect_identical(convert(df$comp, 'as.double'), c(1, 2, 3))

  # to complex
  expect_identical(convert(df$comp, 'as.complex'), c(1+2i, 2+3i, 3+4i))
})


# test filter_rows function ---------------------------------------------------
test_that('filter_rows function', {
  df <- data.frame(x = c(1, 2, 3, NA, NA)) |> as.data.table()

  expect_equal(nrow(filter_rows(df, 'x', '==', 2)), 1)
  expect_equal(nrow(filter_rows(df, 'x', 'is_na', NULL)), 2)
})

# test filter_rows 2 vars function --------------------------------------------
test_that('filter_rows 2 vars function', {
  df <- data.frame(x = c(1, 2, 3, NA, NA),
                   y = c(5:1)) |> as.data.table()

  df_filtered <- filter_rows_2vars(df, 'x', 'y', '==')

  expect_equal(nrow(df_filtered), 1)
  expect_equal(df_filtered[1, ]$y, 3)
  expect_equal(df_filtered[1, ]$x, 3)

  df_filtered <- filter_rows_2vars(df, 'x', 'y', '<=')

  expect_equal(nrow(df_filtered), 3)
  expect_equal(df_filtered$x, c(1, 2, 3))
  expect_equal(df_filtered$y, c(5, 4, 3))
})

test_that('filter_rows 2 vars function - wrong operator', {
  df <- data.frame(x = 1, y = 2) |> as.data.table()
  expect_error(filter_rows_2vars(df, 'x', 'y', 'xxx'))
})

# test f_dec function ---------------------------------------------------------
test_that('f_dec formats numbers', {
  expect_equal(f_dec(1.2345, dig = 0), '1')
  expect_equal(f_dec(1.2345, dig = 2), '1.23')
  expect_equal(f_dec(1.2345, dig = 5), '1.23450')
  expect_equal(f_dec(1.2345, dig = 7), '1.2345000')
})

# test make var names ---------------------------------------------------------
test_that('make var names', {

  df <- data.frame(
    `v 1` = 1:3,
    `v 1` = 4:6,
    `v 2` = 7:9
  )

  expect_error(make_var_names(1:3))
  expect_identical(make_var_names(df) |> names(), c('v.1', 'v.1.1', 'v.2'))
})

# test all equal --------------------------------------------------------------
test_that('all equal', {
  expect_false(test_all_equal(1:3))
  expect_true(test_all_equal(c(1, 1, 1)))

  expect_equal(test_all_equal(c(NA, NA, NA)), NA)
  expect_equal(test_all_equal(c(NA, 1, 2)), NA)
})

# test obj type function ------------------------------------------------------
test_that('obj_type returns correct type', {
  expect_equal(obj_type(1.2), 'double')
  expect_equal(obj_type(1L), 'integer')
  expect_equal(obj_type(as.Date('2020-01-01')), 'date')
  expect_equal(obj_type(factor('a')), 'factor')
  expect_equal(obj_type(as.raw(10)), 'other')
})

# test make valid cols function -----------------------------------------------
test_that('make_valid_cols handles raw values', {
  expect_type(make_valid_cols(as.raw(1)), 'character')
})

# test data frame -------------------------------------------------------------
test_that('is_spada_df returns TRUE for pure data.frames', {
  df <- data.frame(a = 1:3, b = letters[1:3])
  expect_true(is_spada_df(df))
})

test_that('is_spada_df returns FALSE for non-data.frame objects', {
  expect_false(is_spada_df(1:5))
  expect_false(is_spada_df(list(a = 1:3, b = 4:6)))
})

test_that('is_spada_df returns TRUE for data.frames with factors and dates', {
  df <- data.frame(
    a = factor(c('x', 'y', 'z')),
    b = as.Date('2020-01-01') + 0:2
  )
  expect_true(is_spada_df(df))
})

test_that('is_spada_df returns FALSE for nested data.table', {

  dt <- data.table::data.table(
    group = c('A', 'B'),
    data = list(
      data.table::data.table(x = 1:2),
      data.table::data.table(x = 3:4)
    )
  )

  expect_false(is_spada_df(dt))
})

# test make names append list function ----------------------------------------
test_that('make_names_append_list appends suffix to conflicting names', {
  new_list <- list(a = 1, b = 2)
  actual_list <- c('a1', 'a', 'b1', 'b')
  out <- make_names_append_list(new_list, actual_list, suffix = '_test')
  expect_true(all(names(out) %in% c('a_test', 'b_test')))
})

# test is hex color function --------------------------------------------------
test_that('is_hex_color validates hex codes correctly', {
  expect_true(is_hex_color('#FFFFFF'))
  expect_false(is_hex_color('123456'))
})


# test data format function ---------------------------------------------------
test_that('test data format function', {
  expect_false(test_data_format(list()))
  expect_true(test_data_format(list(iris)))
  expect_false(test_data_format(list(data.frame())))
})

# test check dir function -----------------------------------------------------
test_that('check_dir creates directory if it does not exist', {
  tmp <- file.path(tempdir(), 'new_dir')
  unlink(tmp, recursive = TRUE)

  expect_false(dir.exists(tmp))
  check_dir(tmp)
  expect_true(dir.exists(tmp))
  unlink(tmp, recursive = TRUE)
})

# test linear model df output -------------------------------------------------
test_that('linear model df output', {

  s1 <- lm(1:10 ~ rnorm(10)) |> summary()
  expect_error(linear_model_df_output(iris))
  expect_error(linear_model_df_output(1:3))
  expect_s3_class(linear_model_df_output(s1), 'data.frame')
})

# test linear model df metrics ------------------------------------------------
test_that('linear model df metrics', {

  s1 <- lm(1:10 ~ rnorm(10)) |> summary()
  expect_error(linear_model_df_metrics(iris))
  expect_error(linear_model_df_metrics(1:3))
  expect_s3_class(linear_model_df_metrics(s1), 'data.frame')
})

# test desc stats -------------------------------------------------------------
test_that('descriptive stats function', {

  df <- data.frame(
    a = c(1, 2, 3, NA, 5),
    b = c(10, 20, 30, 40, NA),
    c = factor(c('x', 'x', 'y', 'y', 'z'))
  )

  res <- desc_stats(df, fmt_digits = 2)

  expect_type(res, 'list')
  expect_true(all(c('Mean', 'Gmean', 'Hmean', 'Median', 'Mode',
                    'Min', 'Max', 'IQR', 'Range', 'Variance', 'Standard Deviation',
                    'Skewness', 'Kurtosis') %in% names(res)))

  # Mean
  expect_equal(res$Mean['a'][[1]], mean(df$a, na.rm = TRUE) |> f_num(dig = 2))

  # Gmean
  expect_equal(res$Gmean['a'][[1]], Gmean(df$a, na.rm = TRUE) |> f_num(dig = 2))

  # Hmean
  expect_equal(res$Hmean['a'][[1]], Hmean(df$a, na.rm = TRUE) |> f_num(dig = 2))

  # Median
  expect_equal(as.numeric(res$Median['a']), median(df$a, na.rm = TRUE))

  # Mode (coluna fator deve retornar um valor não vazio)
  expect_equal(res$Mode['c'][[1]], 'x')

  # Min e Max
  expect_equal(res$Min['a'][[1]], min(df$a, na.rm = TRUE) |> f_num(dig = 2))
  expect_equal(res$Max['b'][[1]], max(df$b, na.rm = TRUE) |> f_num(dig = 2))

  # IQR
  expect_equal(res[['IQR']]['b'][[1]],
               df$b |> IQR(na.rm = T) |> f_num(dig = 2))

  # Range
  expect_equal(res[['Range']]['b'][[1]],
               paste('[', df$b |> range(na.rm = T) |> f_num(dig = 2) , ']', collapse = '--->'))

  # Variance
  expect_equal(res[['Variance']]['b'][[1]],
               df$b |> var(na.rm = T) |> f_num(dig = 2))

  # SD
  expect_equal(res[['Standard Deviation']]['b'][[1]],
               df$b |> sd(na.rm = T) |> f_num(dig = 2))

  # Skewness
  expect_equal(res[['Skewness']]['b'][[1]],
               df$b |> Skew(na.rm = T) |> f_num(dig = 2))

  # Kurtosis
  expect_equal(res[['Kurtosis']]['b'][[1]],
               df$b |> Kurt(na.rm = T) |> f_num(dig = 2))

})

# test spada plot -------------------------------------------------------------
test_that('spada plot function', {
  p <- spada_plot(
    type = 'hist',
    df = data.frame(x = rnorm(100), y = rnorm(100)),
    xvar = 'x'
  )

  expect_s3_class(p, "ggplot")
})

# test output format ----------------------------------------------------------
test_that('test empty list as output', {
  expect_true(test_output_format(list()))
})

test_that('test valid output', {
  valid_output <- list(
    'id1' = list(
      id = '123',
      title = 'Title 1',
      annotation = 'Annotation 1',
      element = shiny::tags$div('content 1'),
      card = shiny::tags$div('content 1'),
      btn_x = actionButton('b1', 'l1'),
      btn_e = actionButton('b2', 'l2')
    ),
    'id2' = list(
      id = '456',
      title = 'title 2',
      annotation = 'Annotation 1',
      element = shiny::tags$div('content 2'),
      card = shiny::tags$div('content 2'),
      btn_x = actionButton('b1', 'l1'),
      btn_e = actionButton('b2', 'l2')
    )
  )

  expect_true(test_output_format(valid_output))
})

test_that('test other class than list', {
  expect_false(test_output_format('texto'))
  expect_false(test_output_format(123))
})

test_that('test inside elements - not lists', {
  not_list <- list(a = 'test')
  expect_false(test_output_format(not_list))
})

test_that('test missing element in inside list', {
  missing_element <- list(
    'id1' = list(
      id = '123',
      title = 'ok'
    )
  )
  expect_false(test_output_format(missing_element))
})

test_that('test class of inside elements', {
  invalid_output <- list(
    'id1' = list(
      id = 123,  # should be char
      title = 'ok',
      card = shiny::tags$div('content')
    )
  )
  expect_false(test_output_format(invalid_output))
})

# test load conf --------------------------------------------------------------
test_that('no previous conf file', {
  tmpdir <- tempdir()
  conf_path <- file.path(tmpdir, 'conf.qs2')

  if (file.exists(conf_path)) file.remove(conf_path)
  expect_false(file.exists(conf_path))

  start_conf <- c(
    'empty_datasets' = 1,
    'conf_dir' = normalizePath(R_user_dir('spada', 'config'), winslash = '/', mustWork = F),
    'data_dir' = normalizePath(R_user_dir('spada', 'data'), winslash = '/', mustWork = F),
    default_conf
  )

  res <- load_conf(start_conf, tmpdir, themes_names)

  # expect return of the start conf since there is no previous file
  expect_equal(res, start_conf)
  expect_true(file.exists(conf_path))
})

test_that('invalid previous conf file', {
  tmpdir <- tempdir()
  conf_path <- file.path(tmpdir, 'conf.qs2')

  valid_conf <- c(
    'empty_datasets' = 1,
    'conf_dir' = normalizePath(R_user_dir('spada', 'config'), winslash = '/', mustWork = F),
    'data_dir' = normalizePath(R_user_dir('spada', 'data'), winslash = '/', mustWork = F),
    default_conf
  )

  invalid_conf <- valid_conf
  invalid_conf$theme <- 'new_theme'
  # save invalid for test
  qs_save(invalid_conf, conf_path)

  res <- load_conf(valid_conf, tmpdir, themes_names)

  # expect return of the valid conf since the saved is invalid
  expect_equal(res, valid_conf)
})

# test summarise_dt function --------------------------------------------------
dt <- iris |> as.data.table()

test_that('summarise dt funciton - distinct', {

  dt_s0 <- dt[, .SD, .SDcols = 'Species'] |> unique()
  dt_s <- summarise_dt(dt, 'distinct', 'Species')

  expect_equal(dt_s0, dt_s)
  expect_s3_class(dt_s, c('data.table', 'data.frame'))
})

test_that('summarise dt funciton - count', {

  dt_s0 <- dt[, .N, by = 'Species']
  dt_s <- summarise_dt(dt, 'count', 'Species')

  expect_equal(dt_s0, dt_s)
  expect_s3_class(dt_s, c('data.table', 'data.frame'))
})

test_that('summarise dt funciton - wrong fun', {
  expect_error(summarise_dt(dt, 'count2', 'Species'))
})

test_that('summarise dt funciton - not data.table', {
  expect_error(summarise_dt(iris, 'count', 'Species'))

})

test_that('summarise dt funciton - not variable', {
  expect_error(summarise_dt(dt, 'count', c('Species', 'Sepal.Length2')))
})

# test update metadata function -----------------------------------------------
dt <- iris |> as.data.table()

test_that('update metadata funciton - wrogn change type', {
  expect_error(update_meta(previous_meta = data.table(), change_type = 'xx'))
})

test_that('update metadata funciton - rename cols', {

  metadata <- list('dt' = dt |> df_info())
  meta0 <- metadata[['dt']]

  dt2 <- copy(dt)
  names(dt2)[names(dt2) == 'Species'] <- 'Species_2'

  new_meta <- update_meta(
    previous_meta = meta0,
    col_names = dt2 |> names(),
    change_type = 'rename_cols'
  )

  meta0 <- dt2 |> df_info()

  expect_equal(meta0, new_meta)
})

test_that('update metadata funciton - select cols', {

  metadata <- list('dt' = dt |> df_info())

  meta0 <- metadata[['dt']]
  dt2 <- subset(dt, select = c('Petal.Width', 'Species'))

  new_meta <- update_meta(
    previous_meta = meta0,
    col_names = dt2 |> names(),
    ncols = ncol(dt2),
    change_type = 'select_cols'
  )

  meta0 <- dt2 |> df_info()

  expect_equal(meta0, new_meta)
})

test_that('update metadata funciton - order cols', {

  metadata <- list('dt' = dt |> df_info())

  meta0 <- metadata[['dt']]
  dt2 <- dt[, c(1, 3, 2, 4, 5)]

  new_meta <- update_meta(
    previous_meta = meta0,
    col_names = dt2 |> names(),
    change_type = 'order_cols'
  )

  meta0 <- dt2 |> df_info()

  expect_equal(meta0, new_meta)
})

test_that('update metadata funciton - convert cols', {

  metadata <- list('dt' = dt |> df_info())

  meta0 <- metadata[['dt']]
  dt2 <- copy(dt)
  dt2$Sepal.Length <- dt2$Sepal.Length |> as.integer()

  new_meta <- update_meta(
    dt = subset(dt2, select = 'Sepal.Length'),
    previous_meta = meta0,
    col_names = dt2 |> names(),
    updated_cols = 'Sepal.Length',
    ncols = dt2 |> ncol(),
    change_type = 'convert_cols'
  )

  meta0 <- dt2 |> df_info()

  expect_equal(meta0, new_meta)
})

# test is_date function -------------------------------------------------------
test_that("is_date - test Sys.Date()", {
  expect_equal(is_date(Sys.Date()), TRUE)
})

test_that("is_date - test Sys.time()", {
  expect_equal(is_date(Sys.time()), TRUE)
})

test_that("is_date - test numeric", {
  expect_equal(is_date(10), FALSE)
})

test_that("is_date - test char", {
  expect_equal(is_date('20'), FALSE)
})

test_that("is_date - test char in Date format", {
  expect_equal(is_date('2024-12-01'), FALSE)
})

test_that('is_date - test only first element of vector', {
  expect_equal(is_date(c(Sys.Date(), 10)), TRUE)
})

test_that('is_date - test only first element of vector - 2', {
  expect_equal(is_date(c(10, Sys.Date())), FALSE)
})

# test is_categorical function ------------------------------------------------
test_that("is_categorical", {
  expect_equal(is_categorical('A'), TRUE)
  expect_equal(is_categorical(TRUE), TRUE)
  expect_equal(is_categorical(factor(1:2)), TRUE)
  expect_equal(is_categorical(NA), TRUE)
  expect_equal(is_categorical(1), FALSE)
  expect_equal(is_categorical(Sys.Date()), FALSE)
  expect_equal(is_categorical(NA_real_), FALSE)
})

# test is_valid_name function -------------------------------------------------
test_that("is_valid_name - test a valid name", {
  expect_equal(is_valid_name('name'), TRUE)
})

test_that("is_valid_name - test 2 size vector with valid", {
  expect_equal(is_valid_name(c('a', 'b')), c(TRUE, TRUE))
})

test_that("is_valid_name - test name with space", {
  expect_equal(is_valid_name('a a'), FALSE)
})

test_that("is_valid_name - test name starting with number", {
  expect_equal(is_valid_name('1a'), FALSE)
})

test_that("is_valid_name - test name starting with blank space", {
  expect_equal(is_valid_name(' 1a'), FALSE)
})

# test wrapper function from collapse -----------------------------------------
test_that('max_nona - all valid values', {
  expect_equal(max_nona(c(1, 2, 3)), 3)
})

test_that('max_nona - all NA values', {
  expect_equal(max_nona(c(NA, NA, NA)), NA_integer_)
})

test_that('max_nona - all NaN values', {
  expect_equal(max_nona(c(NaN, NaN, NaN)), NA_real_)
})

test_that('max_nona - only n 1 valid out of 3', {
  expect_equal(max_nona(c(1, NA, NA)), 1)
})

test_that('max_nona - only n 2 valid out of 3', {
  expect_equal(max_nona(c(NA, 2, NA)), 2)
})

test_that('max_nona - only n 3 valid out of 3', {
  expect_equal(max_nona(c(NA, NA, 3)), 3)
})

test_that('max_nona - 2 and 3 valid out of 3', {
  expect_equal(max_nona(c(NA, 2, 3)), 3)
})

test_that('max_nona - NA, 3, 2', {
  expect_equal(max_nona(c(NA, 3, 2)), 3)
})

test_that('min_nona - all valid values', {
  expect_equal(min_nona(c(1, 2, 3)), 1)
})

test_that('min_na returns NA', {
  expect_true(is.na(min_na(c(NA, 2, 3))))
})

test_that('min_nona - all NA values', {
  expect_equal(min_nona(c(NA, NA, NA)), NA_integer_)
})

test_that('min_nona - all NaN values', {
  expect_equal(min_nona(c(NaN, NaN, NaN)), NA_real_)
})

test_that('min_nona - only n 1 valid out of 3', {
  expect_equal(min_nona(c(1, NA, NA)), 1)
})

test_that('min_nona - only n 2 valid out of 3', {
  expect_equal(min_nona(c(NA, 2, NA)), 2)
})

test_that('min_nona - only n 3 valid out of 3', {
  expect_equal(min_nona(c(NA, NA, 3)), 3)
})

test_that('min_nona - 2 and 3 valid out of 3', {
  expect_equal(min_nona(c(NA, 2, 3)), 2)
})

test_that('min_nona - NA, 3, 2', {
  expect_equal(min_nona(c(NA, 3, 2)), 2)
})

test_that('min_nona - test is.na all NA', {
  expect_true(is.na(min_nona(c(NA, NA, NA))))
})

test_that('min_nona - works with logical vectors', {
  expect_equal(min_nona(c(TRUE, FALSE, TRUE)), 0)
})

test_that('last_nona - all valid values', {
  expect_equal(last_nona(c(1, 2, 3)), 3)
})

test_that('last_nona - all NA values', {
  expect_equal(last_nona(c(NA, NA, NA)), NA)
})

test_that('last_nona - all NaN values', {
  expect_equal(last_nona(c(NaN, NaN, NaN)), NA_real_)
})

test_that('last_nona - only n 1 valid out of 3', {
  expect_equal(last_nona(c(1, NA, NA)), 1)
})

test_that('last_nona - only n 2 valid out of 3', {
  expect_equal(last_nona(c(NA, 2, NA)), 2)
})

test_that('last_nona - only n 3 valid out of 3', {
  expect_equal(last_nona(c(NA, NA, 3)), 3)
})

test_that('last_nona - 2 and 3 valid out of 3', {
  expect_equal(last_nona(c(NA, 2, 3)), 3)
})

test_that('first_nona - all valid values', {
  expect_equal(first_nona(c(1, 2, 3)), 1)
})

test_that('first_nona - all NA values', {
  expect_equal(first_nona(c(NA, NA, NA)), NA)
})

test_that('first_nona - all NaN values', {
  expect_equal(first_nona(c(NaN, NaN, NaN)), NA_real_)
})

test_that('first_nona - only n 1 valid out of 3', {
  expect_equal(first_nona(c(1, NA, NA)), 1)
})

test_that('first_nona - only n 2 valid out of 3', {
  expect_equal(first_nona(c(NA, 2, NA)), 2)
})

test_that('first_nona - only n 3 valid out of 3', {
  expect_equal(first_nona(c(NA, NA, 3)), 3)
})

test_that('first_nona - 2 and 3 valid out of 3', {
  expect_equal(first_nona(c(NA, 2, 3)), 2)
})

test_that('sum_nona - works with numeric vector without NA values', {
  expect_equal(sum_nona(c(1, 2, 3)), 6)
})

test_that('sum_nona - ignores NA values when na_rm = TRUE (default)', {
  expect_equal(sum_nona(c(NA, 2, 3)), 5)
})

test_that('sum_nona - returns the correct value for a vector with a single element', {
  expect_equal(sum_nona(5), 5)
})

test_that('sum_nona - calculates the sum for a vector with negative values', {
  expect_equal(sum_nona(c(-1, -2, -3)), -6)
})

test_that('sum_nona - calculates the sum for a vector with positive, negative, and zero values', {
  expect_equal(sum_nona(c(-1, 0, 1)), 0)
})

test_that('sum_nona - works with logical vectors (TRUE and FALSE treated as 1 and 0)', {
  expect_equal(sum_nona(c(TRUE, FALSE, TRUE)), 2)
})

test_that('sum_nona - throws an error when given a character vector', {
  expect_error(sum_nona(c('a', 'b', 'c')))
})

# test df_info ----------------------------------------------------------------

test_that('df_info works with a simple data frame', {
  df <- data.frame(
    a = c(1, 2, 3),
    b = c('x', 'y', 'z'),
    c = c(TRUE, FALSE, TRUE)
  )

  info <- df_info(df)

  expect_equal(info$var, c('a', 'b', 'c'))
  expect_equal(info$type, c('double', 'character', 'logical'))
  expect_equal(info$class, c('numeric', 'character', 'logical'))
  expect_equal(info$rows[1], 3)
  expect_equal(info$cols[1], 3)
})

test_that('df_info calculates valid counts and percentages correctly', {
  df <- data.frame(
    a = c(1, NA, 3),
    b = c('x', 'y', NA),
    c = c(TRUE, FALSE, NA)
  )

  info <- df_info(df)

  expect_equal(info$n_valid, c(2, 2, 2))
  expect_equal(info$perc_valid, c(2/3, 2/3, 2/3))
})

test_that('df_info calculates unique counts and percentages correctly', {
  df <- data.frame(
    a = c(1, 2, 2),
    b = c('x', 'y', 'x'),
    c = c(TRUE, FALSE, TRUE)
  )

  info <- df_info(df, type = 'complete')

  expect_equal(info$n_unique, c(2, 2, 2))
  expect_equal(info$perc_unique, c(2/3, 2/3, 2/3))
})

test_that('df_info calculates zero counts and percentages correctly', {
  df <- data.frame(
    a = c(0, 0, 1),
    b = c(0, 1, 2),
    c = c(TRUE, FALSE, FALSE)
  )

  info <- df_info(df)

  expect_equal(info$n_zero, c(2, 1, 0))
  expect_equal(info$perc_zero, c(2/3, 1/3, 0))
})

test_that('df_info calculates NA counts and percentages correctly', {
  df <- data.frame(
    a = c(1, NA, 3),
    b = c(NA, NA, 'z'),
    c = c(TRUE, FALSE, NA)
  )

  info <- df_info(df)

  expect_equal(info$n_nas, c(1, 2, 1))
  expect_equal(info$perc_nas, c(1/3, 2/3, 1/3))
})

test_that('df_info handles numeric min and max correctly', {
  df <- data.frame(
    a = c(1, 2, 3),
    b = c(10, 20, 30),
    c = c(TRUE, FALSE, TRUE)
  )

  info <- df_info(df)

  expect_equal(info$min, c(1, 10, NA))
  expect_equal(info$max, c(3, 30, NA))
})

test_that('df_info works with an empty data frame', {
  df <- data.frame()

  info <- df_info(df, type = 'complete')

  df0 <- data.table(
    var = 'v1',
    type = NA,
    class = NA,
    size = 0,
    min = NA,
    max = NA,
    n_valid = NA,
    perc_valid = NA,
    n_unique = NA,
    perc_unique = NA,
    n_zero = NA,
    perc_zero = NA,
    n_nas = NA,
    perc_nas = NA,
    rows = NA,
    cols = NA
  )

  expect_equal(nrow(info), 1)
  expect_equal(ncol(info), 16)
  expect_equal(info, df0)
})

test_that('df_info throws an error for non-data-frame inputs', {
  expect_error(df_info(1:10))
})

# df_info compare with r base function ----------------------------------------

r_base_df_info <- function(df) {
  stopifnot(is.data.frame(df))

  if (ncol(df) == 0) {
    return(data.table(
      var = 'v1',
      type = NA,
      class = NA,
      size = 0,
      min = NA,
      max = NA,
      n_valid = NA,
      perc_valid = NA,
      n_unique = NA,
      perc_unique = NA,
      n_zero = NA,
      perc_zero = NA,
      n_nas = NA,
      perc_nas = NA,
      rows = NA,
      cols = NA
    ))
  }

  rows <- nrow(df)
  cols <- ncol(df)

  res <- lapply(seq_len(cols), function(j) {
    x <- df[[j]]

    nas <- sum(is.na(x))
    valid <- rows - nas
    uniq <- length(unique(x))
    zeros <- if (is.numeric(x)) sum_nona(x == 0) else 0
    minv <- if (is.numeric(x)) min_nona(x) else NA
    maxv <- if (is.numeric(x)) max_nona(x) else NA

    list(
      var = names(df)[j],
      type = typeof(x),
      class = paste(class(x), collapse = "/"),
      size = as.numeric(object.size(x)),
      min = minv,
      max = maxv,
      n_valid = valid,
      perc_valid = valid / rows,
      n_unique = uniq,
      perc_unique = uniq / rows,
      n_zero = zeros,
      perc_zero = zeros / rows,
      n_nas = nas,
      perc_nas = nas / rows,
      rows = rows,
      cols = cols
    )
  })

  do.call(rbind.data.frame, res) |> as.data.table()
}

test_that('df_info compare with r base function', {

  df <- data.frame(
    a = c(1, NA, 3),
    b = c(NA, NA, 'z'),
    c = c(TRUE, FALSE, NA)
  )

  info <- df_info(df, type = 'complete')
  info_rbase <- r_base_df_info(df)

  expect_equal(info, info_rbase)
})

# test f_num function ---------------------------------------------------------

test_that("f_num - default billion test", {
  expect_equal(f_num(12345678956), '12 B')
})

test_that("f_num - default million test", {
  expect_equal(f_num(123456789), '123 M')
})

test_that("f_num - default thousand test", {
  expect_equal(f_num(1234), '1 K')
})

test_that("f_num - default Inf test", {
  expect_equal(f_num(Inf), 'Inf')
})

test_that("f_num - numeric Inf test", {
  expect_equal(f_num(100^308), 'Inf')
})

test_that("f_num - numeric -Inf test", {
  expect_equal(f_num(-100^308), '-Inf')
})

# test gt info ----------------------------------------------------------------
test_that('Test gt info', {

  gt_info_test <- iris |> df_info() |> gt_info(df_name = 'iris')
  expect_equal(gt_info_test |> class(), c('gt_tbl', 'list'))
  expect_equal(gt_info_test |> length(), 21)

  expect_equal(gt_info_test$`_data`$var, names(iris))
  expect_equal(gt_info_test$`_data`$type, sapply(iris, typeof) |> unname())
  expect_equal(gt_info_test$`_data`$class, sapply(iris, class) |> unname())
  expect_equal(gt_info_test$`_data`$rows, rep(150, 5))
  expect_equal(gt_info_test$`_data`$cols, rep(5, 5))
})

# test dt_join ----------------------------------------------------------------
test_that('dt_join checks variable types correctly', {

  dt1 <- data.table::data.table(
    id = c(1, 2, 3) |> as.Date(),
    x = letters[1:3]
  )

  dt2 <- data.table::data.table(
    id = c(2, 3, 4) |> as.Date(),
    x = LETTERS[1:3]
  )

  expect_no_error(dt_join(dt1, dt2, 'id', 'id'))

  dt2$id <- as.double(dt2$id)

  expect_error(dt_join(dt1, dt2, 'id', 'id'), 'The variables types must match')
})

# tests/testthat/test-one_t_test_module.R

# test one t test module - variable selection -------------------------------------
test_that('One-sample t Test module - variable selection', {
  testServer(one_t_test_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$setInputs(sel_var = 'Petal.Length')

    expect_equal(var_analysis(), setdiff(names(iris), 'Species'))
    expect_equal(var(), iris$Petal.Length)
    expect_equal(sample_mean(), mean(iris$Petal.Length))
  })
})

# test one t test - results ---------------------------------------------------
test_that('One-sample t Test module - test results', {
  testServer(one_t_test_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$setInputs(
      sel_var = 'Petal.Length',
      confidence = 90,
      radio_alternative = 'two.sided',
      mu = 5,
      btn_run_test = 1
    )

    t1 <- t.test(iris$Petal.Length,
                alternative = 'two.sided',
                mu = 5,
                conf.level = 0.9)

    expect_s3_class(ttest_results_gt(), 'gt_tbl')
    expect_equal(var(), iris$Petal.Length)
    expect_equal(ttest$results |> nrow(), 11)
    expect_equal(ttest$results |> ncol(), 2)
    expect_equal(ttest$results |> names(), c('values', 'results'))
    expect_equal(ttest$results$values[1], t1$statistic |> as.character())
    expect_equal(ttest$results$values[2], t1$parameter |> as.character())
    expect_equal(ttest$results$values[3], t1$p.value |> as.character())
    expect_equal(ttest$results$values[4], t1$conf.int[1] |> as.character())
    expect_equal(ttest$results$values[5], t1$conf.int[2] |> as.character())
    expect_equal(ttest$results$values[6], t1$estimate |> as.character())
    expect_equal(ttest$results$values[7], t1$null.value |> as.character())
    expect_equal(ttest$results$values[8], t1$stderr |> as.character())
    expect_equal(ttest$results$values[9], t1$alternative |> as.character())
    expect_equal(ttest$results$values[10], t1$method |> as.character())
    expect_equal(ttest$results$values[11], 'Petal.Length')
  })
})

# test one t test - histogram -------------------------------------------------
test_that('One-sample t Test module - histogram', {
  testServer(one_t_test_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$conf <- reactiveValues(
      plot_fill_color = '#229999',
      plot_line_color = '#44aa44',
      plot_limit = 1e5
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$setInputs(
      sel_var = 'Petal.Length',
      btn_hist = 1
    )

    # wait for the extended_task histogram
    while (task_hist$status() == 'running') {
      session$flushReact()
    }

    expect_equal(task_hist$status(), 'success')
    expect_s3_class(ttest_hist(), c('gg', 'ggplot'))
  })
})

# tests/testthat/test-filter_rows_module.R

df_test <- iris
df_test[150, 1:5] <- NA

# test filter rows - empty inputs ---------------------------------------------
test_that('Test filter rows - empty inputs - should not crash', {
  testServer(filter_rows_server, {

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test |> as.data.table()),
      act_name = 'df_test'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(btn_filter = 1)
    expect_equal(get_act_dt(session), as.data.table(df_test))

  })
})

# test filter rows - one variable ---------------------------------------------
test_that('Test filter rows - one variable - equal operator', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg = function(text, ...) { last_msg <<- text },
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test |> as.data.table()),
      act_name = 'df_test'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'one',
                      one_var_sel = 'Species',
                      one_var_operator = '==',
                      one_var_value = 'setosa')

    session$setInputs(btn_filter = 1)
    expect_equal(get_act_dt(session),
                 subset(df_test, subset = Species == 'setosa') |>
                   as.data.table())
    expect_equal(last_msg, 'Remove modal')

    # test no selection after first operation -----
    session$setInputs(filter_type = character(0),
                      one_var_sel = character(0),
                      btn_filter = 2)

    expect_equal(last_msg, 'Select the Filter type')

    session$setInputs(filter_type = 'one',
                      one_var_sel = character(0),
                      btn_filter = 3)

    expect_equal(last_msg, 'Select a variable')
  })
})

test_that('Test filter rows - one variable - not equal operator', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test |> as.data.table()),
      act_name = 'df_test'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'one',
                      one_var_sel = 'Species',
                      one_var_operator = '!=',
                      one_var_value = 'virginica')

    session$setInputs(btn_filter = 1)
    expect_equal(get_act_dt(session),
                 subset(df_test, subset = Species != 'virginica') |>
                   as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test filter rows - one variable - in operator', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test |> as.data.table()),
      act_name = 'df_test'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'one',
                      one_var_sel = 'Species',
                      one_var_operator = 'in',
                      one_var_value = c('virginica', 'setosa'))

    session$setInputs(btn_filter = 1)
    expect_equal(get_act_dt(session),
                 subset(df_test, subset = Species %in% c('virginica', 'setosa')) |>
                   as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test filter rows - one variable - not in operator', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test |> as.data.table()),
      act_name = 'df_test'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'one',
                      one_var_sel = 'Species',
                      one_var_operator = 'not_in',
                      one_var_value = c('virginica', 'setosa'))

    session$setInputs(btn_filter = 1)
    expect_equal(get_act_dt(session),
                 subset(df_test, subset = Species %notin% c('virginica', 'setosa')) |>
                   as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test filter rows - one variable - is na', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test |> as.data.table()),
      act_name = 'df_test'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'one',
                      one_var_sel = 'Species',
                      one_var_operator = 'is_na')

    session$setInputs(btn_filter = 1)
    expect_equal(get_act_dt(session),
                 subset(df_test, subset = is.na(Species)) |>
                   as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test filter rows - one variable - not na', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test |> as.data.table()),
      act_name = 'df_test'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'one',
                      one_var_sel = 'Species',
                      one_var_operator = 'not_na')

    session$setInputs(btn_filter = 1)
    expect_equal(get_act_dt(session),
                 subset(df_test, subset = !is.na(Species)) |>
                   as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test filter rows - one variable - grater', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test |> as.data.table()),
      act_name = 'df_test'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'one',
                      one_var_sel = 'Sepal.Length',
                      one_var_operator = '>',
                      one_var_value = 5)

    session$setInputs(btn_filter = 1)
    expect_equal(get_act_dt(session),
                 subset(df_test, subset = Sepal.Length > 5) |>
                   as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test filter rows - one variable - grater or equal', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test |> as.data.table()),
      act_name = 'df_test'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'one',
                      one_var_sel = 'Sepal.Length',
                      one_var_operator = '>=',
                      one_var_value = 6)

    session$setInputs(btn_filter = 1)
    expect_equal(get_act_dt(session),
                 subset(df_test, subset = Sepal.Length >= 6) |>
                   as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test filter rows - one variable - less', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test |> as.data.table()),
      act_name = 'df_test'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'one',
                      one_var_sel = 'Sepal.Width',
                      one_var_operator = '<',
                      one_var_value = 3)

    session$setInputs(btn_filter = 1)
    expect_equal(get_act_dt(session),
                 subset(df_test, subset = Sepal.Width < 3) |>
                   as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test filter rows - one variable - less or equal', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test |> as.data.table()),
      act_name = 'df_test'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'one',
                      one_var_sel = 'Sepal.Width',
                      one_var_operator = '<=',
                      one_var_value = 3.3)

    session$setInputs(btn_filter = 1)
    expect_equal(get_act_dt(session),
                 subset(df_test, subset = Sepal.Width <= 3.3) |>
                   as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test filter rows - one variable - outlier', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'one',
                      one_var_sel = 'hp',
                      one_var_operator = 'outlier')

    session$setInputs(btn_filter = 1)
    expect_equal(get_act_dt(session),
                 mtcars[Outlier(mtcars$hp, value = F), ] |>
                   as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test filter rows - one variable - not outlier', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'one',
                      one_var_sel = 'hp',
                      one_var_operator = 'not_outlier')

    session$setInputs(btn_filter = 1)

    mtcars_dt <- mtcars |> as.data.table()
    expect_equal(get_act_dt(session), mtcars_dt[!Outlier(mtcars$hp, value = F), ])
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test filter rows - one variable - between', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'one',
                      one_var_sel = 'hp',
                      one_var_operator = 'between',
                      one_var_value = 90,
                      one_var_value2 = 100)

    session$setInputs(btn_filter = 1)

    mtcars_dt <- mtcars |> as.data.table()
    expect_equal(get_act_dt(session), mtcars_dt[hp %between% c(90, 100), ])
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test filter rows - one variable - not between', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'one',
                      one_var_sel = 'hp',
                      one_var_operator = 'not_between',
                      one_var_value = 100,
                      one_var_value2 = 110)

    session$setInputs(btn_filter = 1)

    mtcars_dt <- mtcars |> as.data.table()
    expect_equal(get_act_dt(session), mtcars_dt[!(hp %between% c(100, 110)), ])
    expect_equal(last_msg, 'Remove modal')
  })
})

# test filter rows - two variables --------------------------------------------
test_that('Test filter rows - two variables - bigger', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'two',
                      two_var_sel1 = 'Sepal.Width',
                      two_var_operator = '>',
                      two_var_sel2 = 'Petal.Length')

    session$setInputs(btn_filter = 1)
    expect_equal(
      get_act_dt(session),
      iris |> as.data.table() |> subset(Sepal.Width > Petal.Length)
    )

  })
})

# test filter rows - sample ---------------------------------------------------
test_that('Test filter rows - sample - nrows', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'sample',
                      sample_type = 'rows',
                      n_rows = '15',
                      x_sample_replace = T)

    session$setInputs(btn_filter = 1)
    expect_equal(
      get_act_dt(session) |> nrow(),
      as.data.table(iris)[sample(1:nrow(iris), 15, replace = T), ] |> nrow()
    )
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test filter rows - sample - percent', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'sample',
                      sample_type = 'percent',
                      sample_size = 10,
                      x_sample_replace = T)

    session$setInputs(btn_filter = 1)
    expect_equal(
      get_act_dt(session) |> nrow(),
      as.data.table(iris)[sample(1:nrow(iris), 10/100 * nrow(iris), replace = T), ] |> nrow()
    )
    expect_equal(last_msg, 'Remove modal')
  })
})

# test filter rows - free -----------------------------------------------------
test_that('Test filter rows - free - allowed code', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'free',
                      txt_code_input = 'Sepal.Width <= Petal.Length'
                      )

    session$setInputs(btn_filter = 1)
    expect_equal(
      get_act_dt(session),
      iris |> as.data.table() |> subset(Sepal.Width <= Petal.Length)
    )
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test filter rows - free - not allowed code - data stay unchanged', {
  testServer(filter_rows_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'free',
                      txt_code_input = 'do.call(sum, list(1, 2))'
    )

    session$setInputs(btn_filter = 1)
    expect_equal(
      get_act_dt(session),
      iris |> as.data.table()
    )

  })
})

# test filter rows - from other dataset ---------------------------------------
test_that('Test filter rows - from other dataset - in', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table(),
                'df_test' = iris |> as.data.table() |> subset(Species == 'setosa')),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'dataset',
                      dt_1_var_sel = 'Species',
                      dt_var_operator = 'in',
                      dt_dt_sel = 'df_test',
                      dt_2_var_sel = 'Species')

    session$setInputs(btn_filter = 1)
    expect_equal(get_act_dt(session), iris |> as.data.table() |> subset(Species == 'setosa'))
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test filter rows - from other dataset - not in', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table(),
                'df_test' = iris |> as.data.table() |> subset(Species == 'setosa')),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(filter_type = 'dataset',
                      dt_1_var_sel = 'Species',
                      dt_var_operator = 'not_in',
                      dt_dt_sel = 'df_test',
                      dt_2_var_sel = 'Species')

    session$setInputs(btn_filter = 1)
    expect_equal(get_act_dt(session), iris |> as.data.table() |> subset(Species != 'setosa'))
    expect_equal(last_msg, 'Remove modal')
  })
})

# test input check messages ---------------------------------------------------
iris_dt <- iris |> as.data.table()
iris_dt$date <- Sys.Date()

test_that('Test filter rows - check inputs - filter type on', {
  testServer(filter_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg_error = function(text, ...) { last_msg <<- text },
      msg = function(text, ...) { last_msg <<- text }
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris_dt,
                'df_test' = iris_dt |> subset(Species == 'setosa')),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(filter_type = NULL, btn_filter = 1)

    expect_equal(last_msg, 'Select the Filter type')

    session$setInputs(
      filter_type = 'one',
      one_var_sel = NULL,
      btn_filter = 2
    )

    expect_equal(last_msg, 'Select a variable')

    session$setInputs(
      one_var_sel = 'Species',
      one_var_operator = NULL,
      btn_filter = 3
    )

    expect_equal(last_msg, 'Select an operator')

    session$setInputs(
      one_var_sel = 'Species',
      one_var_operator = 'in',
      one_var_value = NULL,
      btn_filter = 4
    )

    expect_equal(last_msg, 'Insert a value')

    session$setInputs(
      one_var_sel = 'Sepal.Length',
      one_var_operator = 'between',
      one_var_value = 3,
      one_var_value2 = NULL,
      btn_filter = 5
    )

    expect_equal(last_msg, 'Inform inicial and final values')

    session$setInputs(
      one_var_sel = 'date',
      one_var_operator = 'between',
      one_var_value = c('1970-01-01', NA),
      btn_filter = 6
    )

    expect_equal(last_msg, 'Inform inicial and final dates')

    session$setInputs(
      one_var_sel = 'date',
      one_var_operator = 'in',
      one_var_value = c('1970-01-01', NA),
      btn_filter = 6
    )

    expect_equal(last_msg, 'Insert values')

    session$setInputs(
      one_var_sel = 'date',
      one_var_operator = '==',
      one_var_value = c('1970-01-01', '1971-01-01'),
      btn_filter = 7
    )

    expect_equal(last_msg, 'Operator requires value of length 1')
  })
})

# tests/testthat/test-stats_table_module.R

v1 <- c(iris$Petal.Length, NA, NA)
v2 <- c(iris$Sepal.Length, 0, NA)

var <- reactive(v1)
var_name <- reactive('Petal.Length')
input_percentile <- reactive(90)
percentile <- reactive(quantile(v1, probs = 0.9, na.rm = TRUE))
var_sd <- reactive(sd(v1, na.rm = TRUE))
pearson_correlation <- reactive(cor(v1, v2, method = 'p', use = 'na.or.complete'))

# test stats table - calculations ---------------------------------------------
test_that('stats table server - test calculations', {
  testServer(
    stats_table_server,
    args = list(
      var1 = var,
      var1_name = var_name,
      input_percentile = input_percentile,
      percentile = percentile,
      var1_sd = var_sd,
      pearson_correlation = pearson_correlation
    ),
    {
      session$flushReact()

      while(task_stats$status() == 'running'){
        session$flushReact()
      }

      expect_equal(stats_result()$n, 152)
      expect_equal(stats_result()$n_nas, 2)
      expect_equal(stats_result()$min, min_nona(v1))
      expect_equal(stats_result()$q1, quantile(v1, 0.25, na.rm = T))
      expect_equal(stats_result()$median, median(v1, na.rm = T))
      expect_equal(stats_result()$mean, mean(v1, na.rm = T))
      expect_equal(stats_result()$mode, fmode(v1, na.rm = T))
      expect_equal(stats_result()$q3, quantile(v1, 0.75, na.rm = T))
      expect_equal(stats_result()$max, max_nona(v1))
  })
})

# test stats table - data.frame  and gt table ---------------------------------
test_that('stats table server - test data.frame and gt table', {
  testServer(
    stats_table_server,
    args = list(
      var1 = var,
      var1_name = var_name,
      input_percentile = input_percentile,
      percentile = percentile,
      var1_sd = var_sd,
      pearson_correlation = pearson_correlation
    ),
    {
      session$flushReact()

      while(task_stats$status() == 'running'){
        session$flushReact()
      }

      expect_s3_class(stats_table(), 'data.frame')
      expect_s3_class(stats_table_fmt(), 'gt_tbl')

      expect_equal(nrow(stats_table()), 12)
      expect_equal(ncol(stats_table()), 2)
      expect_equal(names(stats_table()), c('measure', 'value'))
      expect_equal(stats_table()[1, 2], var1_name())
    })
})

# tests/testthat/test-spada_server.R

dfs <- list('df_iris' = iris |> as.data.table(),
            'df_mtcars' = mtcars |> as.data.table())

temp_dir <- tempdir()

start_conf <- list(
  'conf_dir' = file.path(temp_dir, 'conf'),
  'data_dir' = file.path(temp_dir, 'data'),
  'theme' = 'spada_theme',
  'file_size' = 1000,
  'restore_session' = 'never',
  'save_session' = 'ask',
  'restore_data_status' = 0,
  'restore_output_status' = 0,
  'restore_status' = NULL,
  'plot_fill_color' = plot_fill_color,
  'plot_line_color' = plot_line_color
)

# test inputed datasets -------------------------------------------------------
test_that('Test inputed datasets', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {
    expect_equal(session$userData$dt$dt[[1]], iris |> as.data.table())
    expect_equal(session$userData$dt$dt[[2]], mtcars |> as.data.table())
    expect_equal(names(session$userData$dt$dt), c('df_iris', 'df_mtcars'))
    expect_equal(session$userData$dt_names(), c('df_iris', 'df_mtcars'))
  })
})

# test check datasets classes -------------------------------------------------
test_that('Test inputed datasets - classes', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {
    expect_equal(session$userData$dt$dt[[1]] |> class(), c('data.table', 'data.frame'))
    expect_equal(session$userData$dt$dt[[2]] |> class(), c('data.table', 'data.frame'))
  })
})

# test active dt --------------------------------------------------------------
test_that('Test active dt', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {
    expect_equal(session$userData$dt$act_name, 'df_iris')
    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 dfs[['df_iris']] |> as.data.table())
    expect_equal(session$userData$dt$bkp, NULL)
  })
})

# test active dt metadata -----------------------------------------------------
test_that('Test active dt metadata', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {
    expect_equal(session$userData$dt$act_meta(), df_info(iris))
  })
})

# test metadata - dt info -----------------------------------------------------
test_that('Test metadata - df_info', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    expect_equal(df_info(session$userData$dt$dt[[session$userData$dt$act_name]]) |> nrow(), 5)
    expect_equal(df_info(session$userData$dt$dt[[session$userData$dt$act_name]]), df_info(iris))
  })
})

# test metadata - dt info -----------------------------------------------------
test_that('Test metadata - gt class', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    session$userData$dt$act_name <- 'df_mtcars'

    expect_true('gt_tbl' %in% class(session$userData$dt$act_meta() |>
                                      gt_info(df_name = names(dfs)[2])))

  })
})

# test metadata - gt metadata -------------------------------------------------
test_that('Test metadata - gt metadata', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    mtcars_meta <- mtcars |> df_info() |> gt_info(df_name = 'df_mtcars')
    session$userData$dt$act_name <- 'df_mtcars'

    gt_temp <- session$userData$dt$act_meta() |> gt_info(df_name = names(dfs)[2])

    expect_equal(gt_temp$`_data`, mtcars_meta$`_data`)
    expect_equal(gt_temp$`_data` |> nrow(), 11)
    expect_equal(gt_temp$`_data`$var, names(mtcars))
    expect_equal(gt_temp$`_data`$type, sapply(mtcars, typeof) |> unname())
    expect_equal(gt_temp$`_data`$class, sapply(mtcars, class) |> unname())
  })
})

# test metadata - calculate cols ----------------------------------------------
test_that('Test metadata - calculate cols', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    session$userData$dt$act_name <- 'df_iris'

    calculate_cols_server('calculate_cols')

    session$setInputs(
      `calculate_cols-vars_sel` = 'Sepal.Length',
      `calculate_cols-txt_new_name_fun` = 'mean_Sepal.Length',
      `calculate_cols-fun` = 'mean_na',
      `calculate_cols-btn_apply_fun` = 1
    )

    iris2 <- iris |> as.data.table()
    iris2[, mean_Sepal.Length := mean(Sepal.Length)]

    expect_equal(get_act_dt(session), iris2)
    expect_equal(session$userData$dt$act_meta()[['var']], names(iris2))

    expect_equal(session$userData$dt$act_meta(), iris2 |> df_info())

    session$userData$dt$act_name <- 'df_mtcars'

    expect_equal(get_act_dt(session), mtcars |> as.data.table())
    expect_equal(session$userData$dt$act_meta(), mtcars |> df_info())
  })
})

# test metadata - convert cols ------------------------------------------------
test_that('Test metadata - convert cols', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    session$userData$dt$act_name <- 'df_iris'

    convert_cols_server('convert_cols')

    session$setInputs(
      'convert_cols-vars_sel' = 'Sepal.Length',
      'convert_cols-sel_format' = 'as.integer',
      'convert_cols-btn_apply' = 1
    )

    iris2 <- iris |> as.data.table()
    iris2[, Sepal.Length := as.integer(Sepal.Length)]

    expect_equal(get_act_dt(session), iris2)
    expect_equal(session$userData$dt$act_meta()[['var']], names(iris2))

    expect_equal(session$userData$dt$act_meta(), iris2 |> df_info())

    session$userData$dt$act_name <- 'df_mtcars'

    expect_equal(get_act_dt(session), mtcars |> as.data.table())
    expect_equal(session$userData$dt$act_meta(), mtcars |> df_info())
  })
})

# test metadata - filter rows -------------------------------------------------
test_that('Test metadata - filter rows', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    session$userData$dt$act_name <- 'df_mtcars'

    filter_rows_server('filter_rows')

    session$setInputs(
      'filter_rows-one_var_sel' = 'hp',
      'filter_rows-filter_type' = 'one',
      'filter_rows-one_var_operator' = '>',
      'filter_rows-one_var_value' = 100,
      'filter_rows-btn_filter' = 1
    )

    mtcars2 <- mtcars |> as.data.table() |> subset(hp > 100)

    expect_equal(get_act_dt(session), mtcars2)
    expect_equal(session$userData$dt$act_meta()[['var']], names(mtcars2))
    expect_equal(session$userData$dt$act_meta(), mtcars2 |> df_info())

    session$userData$dt$act_name <- 'df_iris'

    expect_equal(get_act_dt(session), iris |> as.data.table())
    expect_equal(session$userData$dt$act_meta(), iris |> df_info())
  })
})

# test metadata - select cols -------------------------------------------------
test_that('Test metadata - selet cols - keep', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    session$userData$dt$act_name <- 'df_mtcars'

    select_cols_server('select_cols')

    session$setInputs(
      'select_cols-vars_sel' = 'mpg',
      'select_cols-radio_var_sel' = 'keep',
      'select_cols-btn_sel' = 1
    )

    mtcars2 <- mtcars |> as.data.table() |> subset(select = mpg)
    expect_equal(get_act_dt(session), mtcars2)
    expect_equal(session$userData$dt$act_meta()[['var']], 'mpg')
    expect_equal(session$userData$dt$act_meta(), mtcars2 |> df_info())

    session$userData$dt$act_name <- 'df_iris'
    expect_equal(get_act_dt(session), iris |> as.data.table())
    expect_equal(session$userData$dt$act_meta(), iris |> df_info())
  })
})

# test metadata - order rows --------------------------------------------------
test_that('Test metadata - order rows', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    session$userData$dt$act_name <- 'df_mtcars'

    order_rows_server('order_rows')

    session$setInputs(
      'order_rows-vars_rows' = 'hp',
      'order_rows-radio_nas' = 'last',
      'order_rows-btn_order_rows' = 1
    )
    expect_equal(get_act_dt(session), mtcars[order(mtcars$hp), ] |> as.data.table())
    expect_equal(session$userData$dt$act_meta()[['var']], names(mtcars))
    expect_equal(session$userData$dt$act_meta(), mtcars |> df_info())

  })
})

# test metadata - reorder only ------------------------------------------------
test_that('Test metadata - order_cols after', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    session$userData$dt$act_name <- 'df_mtcars'

    order_cols_server('order_cols')

    session$setInputs(
      'order_cols-vars_cols' = 'cyl',
      'order_cols-vars_rest' = 'gear',
      'order_cols-radio_cols' = 'after',
      'order_cols-btn_order_cols' = 1
    )

    vars <- c('cyl')
    var1 <- names(mtcars)[10]

    mtcars_reordered <- mtcars |>
      dplyr::relocate(tidyselect::all_of(vars), .after = tidyselect::all_of(var1))

    expect_equal(get_act_dt(session), mtcars_reordered |> as.data.table())
    expect_equal(session$userData$dt$act_meta()[['var']], names(mtcars_reordered))
    expect_equal(session$userData$dt$act_meta(), mtcars_reordered |> df_info())

  })
})

# test metadata - rename only -------------------------------------------------
test_that('Test metadata - rename 1 col', {
  testServer(spada_server(datasets = dfs, conf = start_conf), {

    session$userData$dt$act_name <- 'df_mtcars'

    rename_cols_server('rename_cols')

    session$setInputs(
      'rename_cols-vars_sel' = 'cyl',
      'rename_cols-txt_new_name' = 'cyl2',
      'rename_cols-btn_rename' = 1
    )

    df <- mtcars
    names(df)[2] <- 'cyl2'

    expect_equal(session$userData$dt$act_meta()[['var']], names(df))

    session$userData$dt$act_name <- 'df_iris'

    session$setInputs(
      'rename_cols-vars_sel' = 'Species',
      'rename_cols-txt_new_name' = 'Species2',
      'rename_cols-btn_rename' = 2
    )

    df <- iris
    names(df)[5] <- 'Species2'

    expect_equal(get_act_dt(session), df |> as.data.table())
    expect_equal(session$userData$dt$act_meta()[['var']], names(df))
    expect_equal(session$userData$dt$act_meta(), df |> df_info())

  })
})

# tests/testthat/test-descriptive_stats_module.R

var1 <- 'hp'
df_mtcars <- mtcars |> as.data.table()

# test descriptive stats - desc_stats mean 1 var ------------------------------
test_that('Test desc stats - desc_stats mean 1 var', {

  testServer(descriptive_stats_server, {

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = df_mtcars),
      act_name = 'mtcars'
    )

    session$setInputs(
      sel_var = var1,
      table_digits = 9,
      xg_central_tendency = 'mean',
      xg_dispersion = '',
      xg_shape = '',
      btn_stats = 1
    )

    expect_true(calculated_stats |> is.reactive())

    while(task_desc_stats$status() == 'running'){
      session$flushReact()
    }

    expect_true(calculated_stats() |> is.list())
    expect_true((calculated_stats() |> length()) == 1)

    mean_hp <- mean(df_mtcars[[var1]]) |> f_num(dig = 9)
    names(mean_hp) = var1
    expect_equal(calculated_stats(), list('Mean' = mean_hp))
  })
})

# test descriptive stats - gt stats -------------------------------------------
test_that('Test descriptive stats - gt stats', {

  testServer(descriptive_stats_server, {

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = df_mtcars),
      act_name = 'mtcars'
    )

    session$setInputs(
      sel_var = var1,
      table_digits = 9,
      xg_central_tendency = 'mean',
      xg_dispersion = '',
      xg_shape = '',
      btn_stats = 1
    )

    while(task_desc_stats$status() == 'running'){
      session$flushReact()
    }

    expect_true(gt_stats |> is.reactive())
    expect_equal(gt_stats() |> class(), c('gt_tbl', 'list'))

  })
})

# tests/testthat/test-missing_module.R

dt <- data.table(
  Var1 = c(1, 1, NA, 10)
)

# test missing - identify - check messages ------------------------------------
test_that('Test missing - identify - check inputs', {
  testServer(missing_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg = function(text, ...) { last_msg <<- text },
      msg_error = function(text, ...) { last_msg <<- text }
    )

    session$userData$dt <- reactiveValues(
      dt = list('dt' = dt),
      act_name = 'dt'
    )

    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      var_sel_missing = NULL,
      txt_missing_name = 'Missing',
      btn_identify = 1
    )

    expect_equal(last_msg, 'Select at least one variable')

    session$setInputs(
      var_sel_missing = 'Var1',
      txt_missing_name = 'Var1',
      btn_identify = 2
    )

    expect_equal(last_msg, 'New names are not valid or already in use')

  })
})

# test missing - identify - check result --------------------------------------
test_that('Test missing - identify - check result', {
  testServer(missing_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg = function(text, ...) { last_msg <<- text },
      msg_error = function(text, ...) { last_msg <<- text }
    )

    session$userData$dt <- reactiveValues(
      dt = list('dt' = dt),
      act_name = 'dt'
    )

    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      var_sel_missing = 'Var1',
      txt_missing_name = 'Missing',
      btn_identify = 1
    )

    expect_equal(
      get_act_dt(session)$Missing,
      c('Not Missing', 'Not Missing', 'Missing', 'Not Missing') |> as.factor()
    )

  })
})






# test missing - replace - check inputs ---------------------------------------
test_that('Test missing - replace - check inputs', {
  testServer(missing_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg = function(text, ...) { last_msg <<- text },
      msg_error = function(text, ...) { last_msg <<- text }
    )

    session$userData$dt <- reactiveValues(
      dt = list('dt' = dt),
      act_name = 'dt'
    )

    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      var_sel_replace = 'Var1',
      sel_replace_method = 'constant',
      num_value = NULL,
      btn_replace = 1
    )

    expect_equal(last_msg, 'Inform a value')

  })
})


# test missing - replace - check result ---------------------------------------
test_that('Test missing - replace - check result', {
  testServer(missing_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg = function(text, ...) { last_msg <<- text },
      msg_error = function(text, ...) { last_msg <<- text }
    )

    session$userData$dt <- reactiveValues(
      dt = list('dt' = dt),
      act_name = 'dt'
    )

    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      var_sel_replace = 'Var1',
      sel_replace_method = 'constant',
      num_value = 2,
      btn_replace = 1
    )

    expect_equal(get_act_dt(session)$Var1, c(1, 1, 2, 10))

    session$userData$dt <- reactiveValues(
      dt = list('dt' = dt),
      act_name = 'dt'
    )

    session$setInputs(
      var_sel_replace = 'Var1',
      sel_replace_method = 'mean',
      btn_replace = 2
    )

    expect_equal(get_act_dt(session)$Var1, c(1, 1, 4, 10))

    session$userData$dt <- reactiveValues(
      dt = list('dt' = dt),
      act_name = 'dt'
    )

    session$setInputs(
      var_sel_replace = 'Var1',
      sel_replace_method = 'median',
      btn_replace = 3
    )

    expect_equal(get_act_dt(session)$Var1, c(1, 1, 1, 10))

  })
})

# tests/testthat/test-normality_test_module.R

df_test <- data.frame(
  var1 = rnorm(100)
)

# test normality test module - variable selection -------------------------------------
test_that('Normality Test module - variable selection', {
  testServer(normality_test_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$setInputs(sel_var = 'Setal.Width')

    expect_equal(var_analysis(), setdiff(names(iris), 'Species'))
    expect_equal(var(), iris$Setal.Width)
    expect_equal(var_len(), iris$Setal.Width |> length())

  })
})

# test normality test - test plots --------------------------------------------
test_that('Normality Test module - test plots', {
  testServer(normality_test_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$userData$conf <- reactiveValues(
      plot_fill_color = '#229999',
      plot_line_color = '#44aa44',
      plot_limit = 1e5
    )

    session$setInputs(
      sel_var = 'Petal.Length',
      bins = 5,
      btn_qq = 1,
      btn_hist = 1
    )

    # wait for the extended_task histogram
    while (task_hist$status() == 'running') {
      session$flushReact()
    }

    expect_equal(task_hist$status(), 'success')
    expect_s3_class(norm_hist(), c('gg', 'ggplot'))

    # wait for the extended_task qq plot
    while (task_qq$status() == 'running') {
      session$flushReact()
    }

    expect_equal(task_qq$status(), 'success')
    expect_s3_class(norm_qq_plot(), c('gg', 'ggplot'))
  })
})

# test normality test - ks test -------------------------------------------------------
test_that('Normality Test module - ks test results', {
  testServer(normality_test_server, {
    session$userData$dt <- reactiveValues(
      dt = list('df' = df_test |> as.data.table()),
      act_name = 'df'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$userData$conf <- reactiveValues(
      plot_fill_color = '#229999',
      plot_line_color = '#44aa44',
      plot_limit = 1e5
    )

    session$setInputs(
      sel_var = 'var1',
      btn_ks = 1
    )

    ks1 <- ks.test(df_test$var1, 'pnorm')

    expect_s3_class(ks_results_gt(), 'gt_tbl')
    expect_equal(var(), df_test$var1)
    expect_equal(ks_results()$results |> nrow(), 6)
    expect_equal(ks_results()$results |> ncol(), 2)
    expect_equal(ks_results()$results |> names(), c('values', 'results'))

    expect_equal(ks_results()$results$values[1], ks1$statistic |> as.character())
    expect_equal(ks_results()$results$values[2], ks1$p.value |> as.character())
    expect_equal(ks_results()$results$values[3], ks1$alternative |> as.character())
    expect_equal(ks_results()$results$values[4], ks1$method |> as.character())
    expect_equal(ks_results()$results$values[5], 'var1')
    expect_equal(ks_results()$results$values[6], ks1$exact |> as.character())
  })
})

# test normality test - shapiro wilk test -------------------------------------
test_that('Normality Test module - shapiro wilk test results', {
  testServer(normality_test_server, {
    session$userData$dt <- reactiveValues(
      dt = list('df' = df_test |> as.data.table()),
      act_name = 'df'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$userData$conf <- reactiveValues(
      plot_fill_color = '#229999',
      plot_line_color = '#44aa44',
      plot_limit = 1e5
    )

    session$setInputs(
      sel_var = 'var1',
      btn_sw = 1
    )

    sw1 <- shapiro.test(df_test$var1)

    expect_s3_class(sw_results_gt(), 'gt_tbl')
    expect_equal(var(), df_test$var1)
    expect_equal(sw_results() |> nrow(), 4)
    expect_equal(sw_results() |> ncol(), 2)
    expect_equal(sw_results() |> names(), c('values', 'results'))

    expect_equal(sw_results()$values[1], sw1$statistic |> as.character())
    expect_equal(sw_results()$values[2], sw1$p.value |> as.character())
    expect_equal(sw_results()$values[3], sw1$method |> as.character())
    expect_equal(sw_results()$values[4], 'var1')
  })
})

# test normality test - shapiro francia test -------------------------------------
test_that('Normality Test module - shapiro francia test results', {
  testServer(normality_test_server, {
    session$userData$dt <- reactiveValues(
      dt = list('df' = df_test |> as.data.table()),
      act_name = 'df'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$userData$conf <- reactiveValues(
      plot_fill_color = '#229999',
      plot_line_color = '#44aa44',
      plot_limit = 1e5
    )

    session$setInputs(
      sel_var = 'var1',
      btn_sf = 1
    )

    sf1 <- ShapiroFranciaTest(df_test$var1)

    expect_s3_class(sf_results_gt(), 'gt_tbl')
    expect_equal(var(), df_test$var1)
    expect_equal(sf_results() |> nrow(), 4)
    expect_equal(sf_results() |> ncol(), 2)
    expect_equal(sf_results() |> names(), c('values', 'results'))

    expect_equal(sf_results()$values[1], sf1$statistic |> as.character())
    expect_equal(sf_results()$values[2], sf1$p.value |> as.character())
    expect_equal(sf_results()$values[3], sf1$method |> as.character())
    expect_equal(sf_results()$values[4], 'var1')
  })
})

# tests/testthat/test-rename_cols_module.R

# test rename one column ------------------------------------------------------
test_that("Test rename one column", {
  testServer(rename_cols_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris',
      updated_cols = NULL,
      data_changed_rename = FALSE,
      data_changed_reorder = FALSE
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_sel = 'Species')
    session$setInputs(txt_new_name = 'Species_new_name')
    session$setInputs(btn_rename = 1)

    expect_equal(names(get_act_dt(session))[5], 'Species_new_name')
  })
})

# test rename prefix ----------------------------------------------------------
test_that("Test rename prefix - two columns", {
  testServer(rename_cols_server, {

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars[c('hp', 'mpg')] |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(
      vars_sel_multi = c('hp', 'mpg'),
      rename_method = 'prefix_suffix',
      part_position = 'prefix',
      name_separator = '_',
      txt_new_name_multi = 'new')
    session$setInputs(btn_rename_multi = 1)

    expect_equal(names(get_act_dt(session)), c('new_hp', 'new_mpg'))
  })
})

# test rename suffix ----------------------------------------------------------
test_that("Test rename suffix - three columns", {
  testServer(rename_cols_server, {

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars[c('hp', 'mpg', 'cyl')] |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(
      vars_sel_multi = c('hp', 'mpg', 'cyl'),
      rename_method = 'prefix_suffix',
      part_position = 'suffix',
      name_separator = '..',
      txt_new_name_multi = 'new_suffix')
    session$setInputs(btn_rename_multi = 1)

    expect_equal(names(get_act_dt(session)),
                 c('hp..new_suffix', 'mpg..new_suffix', 'cyl..new_suffix'))
  })
})

# test rename function ---------------------------------------------------------
test_that("Test rename function - one column", {
  testServer(rename_cols_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(
      vars_sel_multi = c('Species'),
      rename_method = 'function',
      name_function = 'toupper'
    )
    session$setInputs(btn_rename_multi = 1)

    expect_equal(names(get_act_dt(session))[5], 'SPECIES')
  })
})

# test rename replace ---------------------------------------------------------
test_that("Test rename replace - four columns", {
  testServer(rename_cols_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(
      vars_sel_multi = c('Sepal.Length', 'Sepal.Width', 'Petal.Length',
                         'Petal.Width'),
      rename_method = 'replace',
      txt_replace_part = '.',
      txt_replace_new_part = '_'
      )
    session$setInputs(btn_rename_multi = 1)

    expect_equal(names(get_act_dt(session)),
                 c('Sepal_Length', 'Sepal_Width', 'Petal_Length', 'Petal_Width',
                   'Species'))
  })
})

# test rename remove ----------------------------------------------------------
test_that("Test rename remove - two columns", {
  testServer(rename_cols_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(
      vars_sel_multi = c('Sepal.Length', 'Sepal.Width'),
      rename_method = 'remove',
      txt_remove_part = '.'
    )
    session$setInputs(btn_rename_multi = 1)

    expect_equal(names(get_act_dt(session))[1:2], c('SepalLength', 'SepalWidth'))
  })
})

# test input check messages ---------------------------------------------------
test_that('Test rename - check inputs', {
  testServer(rename_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg_error = function(text, ...) { last_msg <<- text },
      msg = function(text, ...) { last_msg <<- text },
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )
    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(vars_sel = NULL, btn_rename = 1)

    expect_equal(last_msg, 'Select at least one variable')

    session$setInputs(vars_sel = 'Species', txt_new_name = 123, btn_rename = 2)

    expect_equal(last_msg, 'New name is not valid or already in use')

    session$setInputs(vars_sel = 'Species', txt_new_name = 'v123', btn_rename = 3)

    expect_equal(last_msg, 'Remove modal')
  })
})


test_that('Test rename multi - check inputs', {
  testServer(rename_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg_error = function(text, ...) { last_msg <<- text },
      msg = function(text, ...) { last_msg <<- text }
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )
    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(vars_sel_multi = NULL, btn_rename_multi = 1)

    expect_equal(last_msg, 'Select at least one variable')

    session$setInputs(vars_sel_multi = 'Species',
                      rename_method = 'prefix_suffix',
                      txt_new_name_multi = NULL,
                      btn_rename_multi = 2
    )

    expect_equal(last_msg, 'Inform a new part to insert')

    session$setInputs(vars_sel_multi = 'Species',
                      rename_method = 'replace',
                      txt_replace_part = NULL,
                      btn_rename_multi = 3
    )

    expect_equal(last_msg, 'Inform the text to replace')

    session$setInputs(vars_sel_multi = 'Species',
                      rename_method = 'remove',
                      txt_remove_part = NULL,
                      btn_rename_multi = 4
    )

    expect_equal(last_msg, 'Inform the text to be removed')

    session$setInputs(vars_sel_multi = 'Species',
                      rename_method = 'replace',
                      txt_replace_part = 'Sp',
                      txt_replace_new_part = 'Sp',
                      btn_rename_multi = 5
    )

    expect_equal(last_msg, 'No changes to apply')

    session$setInputs(vars_sel_multi = 'Petal.Length',
                      txt_replace_part = 'Petal',
                      txt_replace_new_part = 'Sepal',
                      btn_rename_multi = 6
    )

    expect_equal(last_msg, 'Some new names already in use')

    session$setInputs(vars_sel_multi = 'Petal.Length',
                      txt_replace_part = 'Petal.Length',
                      txt_replace_new_part = 123,
                      btn_rename_multi = 7
    )

    expect_equal(last_msg, 'New names are not valid')
  })
})

# tests/testthat/test-data_highlights_module.R

# test data highlights - basic test -------------------------------------------
test_that('Test data highlights - basic test', {
  testServer(data_highlights_server, {

    df_test <- data.frame(
      num = c(1, 2, 3, NA, 5),
      char = c('A', 'B', 'A', 'B', NA),
      fact = factor(c('x', 'y', 'x', 'y', 'z')),
      date = as.Date(c('2020-01-01', '2020-01-02', NA, '2020-01-04', '2020-01-05')),
      all_na = rep(NA, 5),
      all_zeros = rep(0, 5),
      max_value = rep(Inf, 5),
      min_value = rep(-Inf, 5)
    )

    df_test_size <- (object.size(df_test) / 2^20) |> as.numeric() |> round(2) |>
      format(big.mark = ',', decimal = '.')

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test |> as.data.table()),
      act_name = 'df_test'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    expect_equal(output$var_num_vars, '4')
    expect_equal(output$var_char_vars, '1')
    expect_equal(output$var_factor_vars, '1')
    expect_equal(output$var_date_vars, '1')
    expect_equal(output$n_rows, '5')

    expect_equal(output$var_most_valid, 'fact')
    expect_equal(output$var_most_valid_n_valid, '5')

    expect_equal(output$var_most_zeros, 'all_zeros')
    expect_equal(output$var_most_zeros_n_zeros, '5')

    expect_equal(output$var_most_nas, 'all_na')
    expect_equal(output$var_most_nas_n, '5')

    expect_equal(output$var_max_value, 'max_value')
    expect_equal(output$max_value, 'Inf')

    expect_equal(output$var_min_value, 'min_value')
    expect_equal(output$min_value, '-Inf')

    expect_equal(act_dt_size(), df_test_size)
    expect_equal(output$act_dt_size, as.character(df_test_size))

  })
})

# tests/testthat/test-correlation_module.R

# test correlation - pearson --------------------------------------------------
test_that('test correlation - pearson method', {
  testServer(correlation_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    r1 <- cor.test(
      iris$Petal.Length,
      iris$Sepal.Length,
      alternative = 'two.sided',
      method = 'p',
      conf.level = 0.95,
      exact = F
    )

    session$setInputs(
      sel_var1 = 'Petal.Length',
      sel_var2 = 'Sepal.Length',
      radio_method = 'pearson',
      radio_alternative = 'two.sided',
      confidence = 95,
      btn_run_test = 1
    )

    expect_s3_class(cor_results_gt(), 'gt_tbl')
    expect_equal(cor_test$results |> nrow(), 10)
    expect_equal(cor_test$results |> ncol(), 2)
    expect_equal(cor_test$results |> names(), c('values', 'results'))
    expect_equal(cor_test$results$values[1], r1$statistic |> as.character())
    expect_equal(cor_test$results$values[2], r1$parameter |> as.character())
    expect_equal(cor_test$results$values[3], r1$p.value |> as.character())
    expect_equal(cor_test$results$values[4], r1$estimate |> as.character())
    expect_equal(cor_test$results$values[5], r1$null.value |> as.character())
    expect_equal(cor_test$results$values[6], r1$alternative |> as.character())
    expect_equal(cor_test$results$values[7], r1$method |> as.character())
    expect_equal('Petal.Length / Sepal.Length', cor_test$results$values[8])

  })
})

# test correlation - kendall --------------------------------------------------
test_that('test correlation - kendall method', {
  testServer(correlation_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    r1 <- cor.test(
      iris$Petal.Length,
      iris$Sepal.Length,
      alternative = 'two.sided',
      method = 'kendall',
      conf.level = 0.95,
      exact = F
    )

    session$setInputs(
      sel_var1 = 'Petal.Length',
      sel_var2 = 'Sepal.Length',
      radio_method = 'kendall',
      radio_alternative = 'two.sided',
      confidence = 95,
      btn_run_test = 1
    )

    expect_s3_class(cor_results_gt(), 'gt_tbl')
    expect_equal(cor_test$results |> nrow(), 7)
    expect_equal(cor_test$results |> ncol(), 2)
    expect_equal(cor_test$results |> names(), c('values', 'results'))
    expect_equal(cor_test$results$values[1], r1$statistic |> as.character())
    expect_equal(cor_test$results$values[2], r1$p.value |> as.character())
    expect_equal(cor_test$results$values[3], r1$estimate |> as.character())
    expect_equal(cor_test$results$values[4], r1$null.value |> as.character())
    expect_equal(cor_test$results$values[5], r1$alternative |> as.character())
    expect_equal(cor_test$results$values[6], r1$method |> as.character())
    expect_equal('Petal.Length / Sepal.Length', cor_test$results$values[7])

  })
})

# test correlation - spearman -------------------------------------------------
test_that('test correlation - spearman method', {
  testServer(correlation_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    r1 <- cor.test(
      iris$Petal.Length,
      iris$Sepal.Length,
      alternative = 'two.sided',
      method = 'spearman',
      conf.level = 0.95,
      exact = F
    )

    session$setInputs(
      sel_var1 = 'Petal.Length',
      sel_var2 = 'Sepal.Length',
      radio_method = 'spearman',
      radio_alternative = 'two.sided',
      confidence = 95,
      btn_run_test = 1
    )

    expect_s3_class(cor_results_gt(), 'gt_tbl')
    expect_equal(cor_test$results |> nrow(), 7)
    expect_equal(cor_test$results |> ncol(), 2)
    expect_equal(cor_test$results |> names(), c('values', 'results'))
    expect_equal(cor_test$results$values[1], r1$statistic |> as.character())
    expect_equal(cor_test$results$values[2], r1$p.value |> as.character())
    expect_equal(cor_test$results$values[3], r1$estimate |> as.character())
    expect_equal(cor_test$results$values[4], r1$null.value |> as.character())
    expect_equal(cor_test$results$values[5], r1$alternative |> as.character())
    expect_equal(cor_test$results$values[6], r1$method |> as.character())
    expect_equal('Petal.Length / Sepal.Length', cor_test$results$values[7])

  })
})

# test correlation - scatter plot -------------------------------------------------
test_that('test correlation - scatter plot', {
  testServer(correlation_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$userData$conf <- reactiveValues(
      plot_fill_color = '#229999',
      plot_line_color = '#44aa44',
      plot_limit = 1e5
    )

    session$setInputs(
      sel_var1 = 'Petal.Length',
      sel_var2 = 'Sepal.Length',
      btn_scatter = 1
    )

    # wait for the extended_task
    while (task_scatter$status() == 'running') {
      session$flushReact()
    }

    expect_equal(task_scatter$status(), 'success')
    expect_s3_class(scatter_plot(), c('gg', 'ggplot'))

  })
})

# tests/testthat/test-duplicates_module.R

iris_dt <- iris |> as.data.table()

# test duplicates - check messages --------------------------------------------
test_that('Test duplicates - check inputs', {
  testServer(duplicates_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg = function(text, ...) { last_msg <<- text },
      msg_error = function(text, ...) { last_msg <<- text }
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris_dt),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      vars_sel_dup = NULL,
      txt_new_dup_name = 'Duplicated',
      radio_last = 'FALSE',
      radio_primary = 'FALSE',
      txt_new_primary_name = 'Primary',
      btn_identify = 1
    )

    expect_equal(last_msg, 'Select at least one variable')

    session$setInputs(
      vars_sel_dup = 'Species',
      txt_new_dup_name = 'Species',
      btn_identify = 2
    )

    expect_equal(last_msg, 'New names are not valid or already in use')

    session$setInputs(
      txt_new_dup_name = 'Species2',
      txt_new_primary_name = 'Species',
      btn_identify = 3
    )

    expect_equal(last_msg, 'New names are not valid or already in use')

    # remove duplicates
    session$setInputs(
      vars_sel_drop = 'Species',
      btn_remove = 2
    )

    expect_equal(last_msg, 'The variable is not a factor of Duplicated and Not Duplicated Levels')

  })
})

# test duplicates - identify --------------------------------------------------
test_that('Test duplicates', {
  testServer(duplicates_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris_dt),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      vars_sel_dup = 'Species',
      txt_new_dup_name = 'Duplicated',
      radio_last = 'FALSE',
      radio_primary = 'TRUE',
      txt_new_primary_name = 'Primary',
      btn_identify = 1
    )

    expect_equal(
      get_act_dt(session)$Duplicated,
      factor(
        duplicated(iris_dt, by = 'Species'),
        levels = c(T, F),
        labels = c('Duplicated', 'Not Duplicated')
      )
    )

    expect_equal(
      get_act_dt(session)$Primary,
      factor(
        !duplicated(iris_dt, by = 'Species'),
        levels = c(T, F),
        labels = c('Primary', 'Not Primary')
      )
    )

    session$setInputs(
      radio_last = 'TRUE',
      txt_new_dup_name = 'Duplicated2',
      txt_new_primary_name = 'Primary2',
      btn_identify = 2
    )

    expect_equal(
      get_act_dt(session)$Duplicated2,
      factor(
        duplicated(iris_dt, fromLast = TRUE, by = 'Species'),
        levels = c(T, F),
        labels = c('Duplicated', 'Not Duplicated')
      )
    )

    expect_equal(
      get_act_dt(session)$Primary2,
      factor(
        !duplicated(iris_dt, fromLast = TRUE, by = 'Species'),
        levels = c(T, F),
        labels = c('Primary', 'Not Primary')
      )
    )
  })
})

# test duplicates - remove ----------------------------------------------------
test_that('Test duplicates', {
  testServer(duplicates_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris_dt),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      vars_sel_dup = 'Species',
      txt_new_dup_name = 'Duplicated',
      radio_last = 'FALSE',
      radio_primary = 'FALSE',
      btn_identify = 1
    )

    session$setInputs(
      vars_sel_drop = 'Duplicated',
      btn_remove = 2
    )

    expect_equal(get_act_dt(session) |> nrow(), 3)
  })
})

# tests/testthat/test-metadata_module.R

# test gt ---------------------------------------------------------------------
test_that("Test gt object", {
  testServer(metadata_server, {

    # create elements from outside of module
    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = as.data.table(mtcars), 'iris' = as.data.table(iris)),
      act_name = 'mtcars'
    )

    session$userData$dt_names <- reactive({
      names(session$userData$dt$dt)
    })

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$gt_info <- reactive({
      req(session$userData$dt$df_info())

      Map(gt_info, session$userData$dt$df_info(),
          df_name = names(session$userData$dt$df_info()))
    })

    # start test -----------------------------
    session$setInputs(dataset_sel = 'mtcars')

    expect_equal(names(session$userData$dt$gt_info()), c('mtcars', 'iris'))
    expect_equal(class(session$userData$dt$gt_info()), 'list')
    expect_equal(class(session$userData$dt$gt_info()[[1]]), c('gt_tbl', 'list'))
    expect_equal(class(session$userData$dt$gt_info()[[2]]), c('gt_tbl', 'list'))

    expect_equal(session$userData$dt$gt_info()[[1]][[1]][['var']],
                 names(mtcars))

    expect_equal(session$userData$dt$gt_info()[[2]][[1]],
                 gt_info(df_info(iris), df_name = 'mtcars')[[1]]
    )

    expect_equal(session$userData$dt$gt_info()[[2]][[1]][['var']],
                 names(iris))

    expect_equal(session$userData$dt$gt_info()[[2]][[1]],
                 gt_info(df_info(iris), df_name = 'iris')[[1]]
    )

  })
})

# tests/testthat/test-convert_cols_module.R

# test textInput - type/class -------------------------------------------------
test_that('Test currernt type/class', {
  testServer(convert_cols_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_sel = 'Sepal.Length')

    expect_equal(
      paste(
        'Type: [', iris$Sepal.Length |> typeof(), '] | Class: [',
        paste(iris$Sepal.Length |> class(), collapse = '/'),
        ']'
      ),
      paste(
        'Type: [', get_act_dt(session)[[input$vars_sel]] |> typeof(),
        '] | Class: [',
        paste(get_act_dt(session)[[input$vars_sel]] |> class(), collapse = '/'),
        ']'
      )
    )
  })
})

# test sample trigger reactiveVal ---------------------------------------------
test_that('Test Sample trigger - updates on button click', {
  testServer(convert_cols_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)
    trigger_before <- preview_sample_trigger()

    session$setInputs(btn_preview_sample = 1)
    trigger_after <- preview_sample_trigger()

    expect_true(trigger_after > trigger_before)
  })
})

# test preview df -------------------------------------------------------------
test_that('Test Preview dataframe after variable and format selection', {
  testServer(convert_cols_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$act_meta <- reactive({ iris |> as.data.table() |> df_info()})
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_sel = 'Species', sel_format = 'as.character')

    expect_true(is.data.frame(preview_df()))
    expect_true(ncol(preview_df()) == 2)
    expect_true(nrow(preview_df()) == 8)

    expect_equal(
      preview_df(),
      data.table(
        Species = iris[preview_sample(), ]$Species,
        preview = iris[preview_sample(), ]$Species |> as.character()
      )
    )
  })
})

# test conversion -------------------------------------------------------------
test_that('Test Conversion applies when button clicked', {
  testServer(convert_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg = function(text, ...) { last_msg <<- text },
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_sel = '', sel_format = '', btn_apply = 1)

    expect_equal(last_msg, 'Select a variable and a new format')

    original_class <- class(get_act_dt(session)[['Species']])

    session$setInputs(vars_sel = 'Species',
                      sel_format = 'as.character',
                      btn_apply = 2)

    converted_class <- class(get_act_dt(session)[['Species']])

    expect_false(identical(original_class, converted_class))
    expect_equal(converted_class, 'character')
    expect_equal(get_act_dt(session)[['Species']] |> class(), 'character')
    expect_equal(get_act_dt(session) |> class(), c('data.table', 'data.frame'))
    expect_equal(last_msg, 'Remove modal')

    # test no selection after first operation -----
    last_msg <- NULL
    session$setInputs(vars_sel = character(0),
                      sel_format = 'as.Date',
                      btn_apply = 3)

    expect_equal(last_msg, 'Select a variable and a new format')

    last_msg <- NULL
    session$setInputs(vars_sel = 'Species',
                      sel_format = character(0),
                      btn_apply = 4)

    expect_equal(last_msg, 'Select a variable and a new format')
  })
})

# tests/testthat/test-sidebar_module.R

# test change active dataset --------------------------------------------------
test_that('Test change active dataset', {
  testServer(sidebar_server, args = list(app_session = NULL), {

    session$userData$dt <- reactiveValues(
      dt = list('df_iris' = iris |> as.data.table(),
                'df_mtcars' = mtcars |> as.data.table()),
      act_name = 'df_iris'
    )

    session$userData$dt_names <- reactive({
      req(session$userData$dt$dt)
      names(session$userData$dt$dt)
    })

    session$setInputs(sel_datasets_names = 'iris')

    session$userData$dt_names <- reactive({
      req(session$userData$dt$dt)
      names(session$userData$dt$dt)
    })

    session$setInputs(sel_act_dt = 'df_mtcars')

    expect_true(session$userData$dt$act_name == 'df_mtcars')
    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 mtcars |> as.data.table())

    session$setInputs(sel_act_dt = 'df_iris')

    expect_true(session$userData$dt$act_name == 'df_iris')
    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 iris |> as.data.table())

    session$setInputs(sel_act_dt = 'df_mtcars')

    expect_true(session$userData$dt$act_name == 'df_mtcars')
    expect_equal(session$userData$dt$dt |> names(), c('df_iris', 'df_mtcars'))
    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 mtcars |> as.data.table()
    )
  })
})

test_that('Test active dataset - after deletion', {
  testServer(sidebar_server, args = list(app_session = NULL), {


    session$userData$dt <- reactiveValues(
      dt = list('df_iris' = iris |> as.data.table(),
                'df_mtcars' = mtcars |> as.data.table(),
                'df_cars' = cars |> as.data.table()
                ),
      act_name = 'df_iris'
    )

    session$userData$dt_names <- reactive({
      req(session$userData$dt$dt)
      names(session$userData$dt$dt)
    })

    session$setInputs(sel_datasets_names = 'iris')

    session$userData$dt_names <- reactive({
      req(session$userData$dt$dt)
      names(session$userData$dt$dt)
    })

    session$setInputs(sel_act_dt = 'df_mtcars')

    expect_true(session$userData$dt$act_name == 'df_mtcars')
    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 mtcars |> as.data.table())

    session$setInputs(sel_act_dt = 'df_iris')

    expect_true(session$userData$dt$act_name == 'df_iris')
    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 iris |> as.data.table())

    session$setInputs(sel_act_dt = 'df_mtcars')

    expect_true(session$userData$dt$act_name == 'df_mtcars')
    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 mtcars |> as.data.table()
    )

    # delete dataset
    session$userData$dt$dt[['df_cars']] <- NULL

    expect_true(session$userData$dt$act_name == 'df_mtcars')
    expect_equal(session$userData$dt$dt |> names(), c('df_iris', 'df_mtcars'))
    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 mtcars |> as.data.table())
  })
})

# test df info UI -------------------------------------------------------------
test_that('Test render df_info UI', {
  testServer(sidebar_server, args = list(app_session = NULL), {

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = data.frame(var1 = 1:5, var2 = rep(NA, 5) |>
          as.data.table())),
      act_name = 'df_test'
    )

    session$userData$dt_names <- reactive({
      req(session$userData$dt$dt)
      names(session$userData$dt$dt)
    })

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$userData$dt$act_mini_meta <- reactive({
      req(session$userData$dt$act_meta())

      list(
        'row_col' = paste(
          f_int(session$userData$dt$act_meta()[['rows']][1]),
          '/',
          f_int(session$userData$dt$act_meta()[['cols']][1])
        ),
        'col_nas' = session$userData$dt$act_meta()[n_nas > 0, ] |> nrow(),
        'size' = (object.size(get_act_dt(session)) / 2^20) |>
          as.numeric() |> round(2)
      )
    })

    expect_true(is.list(output$df_info))
    expect_true(grepl('Rows/Columns:', output$df_info$html))
    expect_true(grepl("Columns with NA's:", output$df_info$html))

    expect_equal(output$row_col, '5 / 2')
    expect_equal(output$col_nas, '1')
  })
})

# test render preview ---------------------------------------------------------
test_that('Test renders dataset preview first dataset', {
  testServer(sidebar_server, args = list(app_session = NULL), {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt_names <- reactive({
      req(session$userData$dt$dt)
      names(session$userData$dt$dt)
    })

    session$setInputs(sel_datasets_names = 'iris')

    expect_true(is.list(output$df_preview))
    expect_true(names(output$df_preview)[1] == 'html')
    expect_true(grepl('iris', output$df_preview$html))
    expect_true(grepl('Sepal.Length', output$df_preview$html))
    expect_true(grepl('Sepal.Width', output$df_preview$html))
    expect_true(grepl('Petal.Length', output$df_preview$html))
    expect_true(grepl('Petal.Width', output$df_preview$html))
    expect_true(grepl('Species', output$df_preview$html))

  })
})

test_that('Test renders dataset preview second dataset', {
  testServer(sidebar_server, args = list(app_session = NULL), {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table(),
                'cars' = cars |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt_names <- reactive({
      req(session$userData$dt$dt)
      names(session$userData$dt$dt)
    })

    session$setInputs(sel_datasets_names = 'cars')

    expect_true(is.list(output$df_preview))
    expect_true(names(output$df_preview)[1] == 'html')
    expect_true(grepl('cars', output$df_preview$html))
    expect_true(grepl('speed', output$df_preview$html))
    expect_true(grepl('dist', output$df_preview$html))

  })
})

# test dark mode --------------------------------------------------------------
test_that('Test dark mode', {
  testServer(sidebar_server, args = list(app_session = NULL), {
    session$userData$conf <- reactiveValues(theme = 'spada_theme')
    session$userData$dt <- reactiveValues(
      dt = list(iris = iris),
      act_name = 'iris'
    )
    session$userData$dt_names <- reactive( names(session$userData$dt$dt) )

    session$setInputs(dark_mode = FALSE)

    expect_equal(session$userData$conf$theme, 'spada_theme')

    session$setInputs(dark_mode = TRUE)

    expect_equal(session$userData$conf$theme, 'spada_dark_theme')
  })
})

# tests/testthat/test-exploratory_module.R

df_test <- data.frame(
  cyl = as.factor(mtcars$cyl),
  am = as.factor(mtcars$am)
)

# test exploratory test module - variable selection ---------------------------
test_that('Exploratory module - variable selection', {
  testServer(exploratory_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$setInputs(
      sel_vars = 'Petal.Length',
      sel_vars2 = 'Petal.Width',
      outliers = FALSE
    )

    expect_equal(var_analysis(), names(iris))
    expect_equal(var(), iris$Petal.Length)
    expect_equal(var2(), iris$Petal.Width)

  })
})

# test exploratory test module - dist plot ------------------------------------
test_that('Exploratory module - dist plot', {
  testServer(exploratory_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$userData$conf <- reactiveValues(
      plot_fill_color = '#229999',
      plot_line_color = '#44aa44',
      plot_limit = 1e5
    )

    session$setInputs(
      sel_vars = 'Petal.Length',
      sel_vars2 = 'Petal.Width',
      outliers = FALSE,
      radio_dist_plot = 'hist',
      var_percentile = 90,
      bins = 10
    )

    expect_equal(var_analysis(), names(iris))
    expect_equal(var(), iris$Petal.Length)
    expect_equal(var2(), iris$Petal.Width)
    expect_s3_class(dist_plot(), c('gg', 'ggplot'))

  })
})

# test exploratory test module - scatter plot ---------------------------------
test_that('Exploratory module - scatter plot', {
  testServer(exploratory_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$userData$conf <- reactiveValues(
      plot_fill_color = '#229999',
      plot_line_color = '#44aa44',
      plot_limit = 1e5
    )

    session$setInputs(
      sel_vars = 'Petal.Length',
      sel_vars2 = 'Petal.Width',
      outliers = FALSE,
      radio_dist_plot = 'hist',
      var_percentile = 90,
      bins = 10,
      btn_scatter = 1,
      scatter_lm = FALSE
    )

    expect_equal(var_analysis(), names(iris))
    expect_equal(var(), iris$Petal.Length)
    expect_equal(var2(), iris$Petal.Width)
    expect_s3_class(scatter_plot(), c('gg', 'ggplot'))

  })
})

# test exploratory test module - linear model ---------------------------------
test_that('Exploratory module - linear model', {
  testServer(exploratory_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$userData$conf <- reactiveValues(
      plot_fill_color = '#229999',
      plot_line_color = '#44aa44',
      plot_limit = 1e5
    )

    session$setInputs(
      sel_vars = 'Petal.Width',
      sel_vars2 = 'Petal.Length',
      outliers = FALSE,
      btn_scatter_lm_run = 1
    )

    m1 <- lm(Petal.Width ~ Petal.Length, data = iris, model = F)

    expect_true(!is.null(linear_model$model))
    expect_equal(linear_model$y_name, 'Petal.Width')
    expect_equal(linear_model$x_name, 'Petal.Length')
    expect_equal(linear_model$y, m1$fitted.values)
    expect_equal(summary(linear_model$model)$r.squared,
                 summary(m1)$r.squared)

    expect_equal(summary(linear_model$model)$r.squared,
                 summary(m1)$r.squared)

    expect_equal(summary(linear_model$model)$residuals,
                 summary(m1)$residuals)

    expect_equal(summary(linear_model$model)$fstatistic,
                 summary(m1)$fstatistic)

    expect_s3_class(lm_var_table(), 'gt_tbl')
    expect_s3_class(lm_metrics(), 'gt_tbl')


    session$setInputs(radio_lm_resid = 'hist')

    expect_equal(update_lm_resid_plot(), 0)

    session$setInputs(btn_lm_resid = 1)

    expect_equal(update_lm_resid_plot(), 1)

    expect_s3_class(lm_resid_plot(), 'ggplot')

    # clear model
    session$setInputs(btn_scatter_lm_clear = 1)
    expect_null(linear_model$model)
    expect_equal(update_lm_resid_plot(), 2)

    expect_null(linear_model$model$residuals)
    expect_null(linear_model$x)
    expect_null(linear_model$y)
    expect_equal(linear_model$x_name, '')
    expect_equal(linear_model$y_name, '')
  })
})

# test exploratory test module - stats ----------------------------------------
test_that('Exploratory module - stats', {
  testServer(exploratory_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$userData$conf <- reactiveValues(
      plot_fill_color = '#229999',
      plot_line_color = '#44aa44',
      plot_limit = 1e5
    )

    session$setInputs(
      sel_vars = 'Petal.Width',
      sel_vars2 = 'Petal.Length',
      outliers = FALSE
    )

    expect_equal(stats_sd(), sd(iris$Petal.Width))
    expect_equal(stats_correlation(), cor(iris$Petal.Width, iris$Petal.Length))

  })
})

# tests/testthat/test-z_test_module.R

# test z test module - variable selection -------------------------------------
test_that('Z Test module - variable selection', {
  testServer(z_test_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$setInputs(sel_var = 'Petal.Length')

    expect_equal(var_analysis(), setdiff(names(iris), 'Species'))
    expect_equal(var(), iris$Petal.Length)
    expect_equal(sample_mean(), mean(iris$Petal.Length))
    expect_equal(sample_sd(), sd(iris$Petal.Length))
  })
})

# test z test - results -------------------------------------------------------
test_that('Z Test module - test results', {
  testServer(z_test_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$setInputs(
      sel_var = 'Petal.Length',
      confidence = 90,
      radio_alternative = 'two.sided',
      mu = 5,
      sd = 2,
      btn_run_test = 1
    )

    z1 <- ZTest(iris$Petal.Length,
                alternative = 'two.sided',
                mu = 5,
                sd_pop = 2,
                conf.level = 0.9)

    expect_s3_class(ztest_results_gt(), 'gt_tbl')
    expect_equal(var(), iris$Petal.Length)
    expect_equal(ztest$results |> nrow(), 11)
    expect_equal(ztest$results |> ncol(), 2)
    expect_equal(ztest$results |> names(), c('values', 'results'))
    expect_equal(ztest$results$values[1], z1$statistic |> as.character())
    expect_equal(ztest$results$values[2], z1$parameter |> as.character())
    expect_equal(ztest$results$values[3], z1$p.value |> as.character())
    expect_equal(ztest$results$values[4], z1$conf.int[1] |> as.character())
    expect_equal(ztest$results$values[5], z1$conf.int[2] |> as.character())
    expect_equal(ztest$results$values[6], z1$estimate |> as.character())
    expect_equal(ztest$results$values[7], z1$null.value |> as.character())
    expect_equal(ztest$results$values[8], z1$stderr |> as.character())
    expect_equal(ztest$results$values[9], z1$alternative |> as.character())
    expect_equal(ztest$results$values[10], z1$method |> as.character())
    expect_equal('Petal.Length', ztest$results$values[11])
  })
})

# test z test - histogram -----------------------------------------------------
test_that('Z Test module - histogram', {
  testServer(z_test_server, {
    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$conf <- reactiveValues(
      plot_fill_color = '#229999',
      plot_line_color = '#44aa44',
      plot_limit = 1e5
    )

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$setInputs(
      sel_var = 'Petal.Length',
      btn_hist = 1
    )

    # wait for the extended_task histogram
    while (task_hist$status() == 'running') {
      session$flushReact()
    }

    expect_equal(task_hist$status(), 'success')
    expect_s3_class(ztest_hist(), c('gg', 'ggplot'))
  })
})

# tests/testthat/test-join_module.R

dt1 <- data.table(char_x = LETTERS[1:3], num = 1:3)
dt2 <- data.table(char_y = LETTERS[2:4], num = 10:12)

datasets <- list('dt1' = dt1, 'dt2' = dt2)

# test joins - new ------------------------------------------------------------
test_that('Test joins - new', {
  testServer(join_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt <- reactiveValues(
      dt = datasets,
      act_name = 'dt1'
    )

    session$userData$dt_names <- reactive({ names(session$userData$dt$dt) })

    session$setInputs(
      dt1_sel = 'dt1',
      vars_sel1 = 'char_x',
      dt2_sel = 'dt2',
      vars_sel2 = 'char_y',
      join_type = 'left',
      radio_overwrite = 'new',
      txt_new_dt_name = 'left_dt',
      btn_apply = 1
    )

    session$setInputs(
      join_type = 'right',
      txt_new_dt_name = 'right_dt',
      btn_apply = 2
    )

    session$setInputs(
      join_type = 'full',
      txt_new_dt_name = 'full_dt',
      btn_apply = 3
    )

    session$setInputs(
      join_type = 'inner',
      txt_new_dt_name = 'inner_dt',
      btn_apply = 4
    )

    # left join
    expect_s3_class(session$userData$dt$dt [[ 'left_dt' ]], 'data.table')
    expect_equal(session$userData$dt$dt [['left_dt']],
                 dt_join(dt1, dt2, 'char_x', 'char_y', 'left'),
                 ignore_attr = TRUE)
    # right join
    expect_s3_class(session$userData$dt$dt [[ 'right_dt' ]], 'data.table')
    expect_equal(session$userData$dt$dt [['right_dt']],
                 dt_join(dt1, dt2, 'char_x', 'char_y', 'right'),
                 ignore_attr = TRUE)

    # full join
    expect_s3_class(session$userData$dt$dt [[ 'full_dt' ]], 'data.table')
    expect_equal(session$userData$dt$dt [['full_dt']],
                 dt_join(dt1, dt2, 'char_x', 'char_y', 'full'),
                 ignore_attr = TRUE)

    # inner join
    expect_s3_class(session$userData$dt$dt [[ 'inner_dt' ]], 'data.table')
    expect_equal(session$userData$dt$dt [['inner_dt']],
                 dt_join(dt1, dt2, 'char_x', 'char_y', 'inner'),
                 ignore_attr = TRUE)
    expect_equal(last_msg, 'Remove modal')
  })
})

# test joins - overwrite ------------------------------------------------------
test_that('Test joins - overwrite - left', {
  testServer(join_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt <- reactiveValues(
      dt = datasets,
      act_name = 'dt1'
    )

    session$userData$data_changed <- reactiveVal(0)
    session$userData$dt_names <- reactive({ names(session$userData$dt$dt) })

    session$setInputs(
      dt1_sel = 'dt1',
      vars_sel1 = 'char_x',
      dt2_sel = 'dt2',
      vars_sel2 = 'char_y',
      join_type = 'left',
      radio_overwrite = 'overwrite',
      btn_apply = 1
    )

    expect_s3_class(session$userData$dt$dt [[ 'dt1' ]], 'data.table')
    expect_equal(session$userData$dt$dt [['dt1']],
                 dt_join(dt1, dt2, 'char_x', 'char_y', 'left'),
                 ignore_attr = TRUE)
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test joins - overwrite - right', {
  testServer(join_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt <- reactiveValues(
      dt = datasets,
      act_name = 'dt1'
    )

    session$userData$data_changed <- reactiveVal(0)
    session$userData$dt_names <- reactive({ names(session$userData$dt$dt) })

    session$setInputs(
      dt1_sel = 'dt1',
      vars_sel1 = 'char_x',
      dt2_sel = 'dt2',
      vars_sel2 = 'char_y',
      join_type = 'right',
      radio_overwrite = 'overwrite',
      btn_apply = 1
    )

    expect_s3_class(session$userData$dt$dt [[ 'dt1' ]], 'data.table')
    expect_equal(session$userData$dt$dt [['dt1']],
                 dt_join(dt1, dt2, 'char_x', 'char_y', 'right'),
                 ignore_attr = TRUE)
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test joins - overwrite - full', {
  testServer(join_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt <- reactiveValues(
      dt = datasets,
      act_name = 'dt1'
    )

    session$userData$data_changed <- reactiveVal(0)
    session$userData$dt_names <- reactive({ names(session$userData$dt$dt) })

    session$setInputs(
      dt1_sel = 'dt1',
      vars_sel1 = 'char_x',
      dt2_sel = 'dt2',
      vars_sel2 = 'char_y',
      join_type = 'full',
      radio_overwrite = 'overwrite',
      btn_apply = 1
    )

    expect_s3_class(session$userData$dt$dt [[ 'dt1' ]], 'data.table')
    expect_equal(session$userData$dt$dt [['dt1']],
                 dt_join(dt1, dt2, 'char_x', 'char_y', 'full'),
                 ignore_attr = TRUE)
    expect_equal(last_msg, 'Remove modal')
  })
})


test_that('Test joins - overwrite - inner', {
  testServer(join_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt <- reactiveValues(
      dt = datasets,
      act_name = 'dt1'
    )

    session$userData$data_changed <- reactiveVal(0)
    session$userData$dt_names <- reactive({ names(session$userData$dt$dt) })

    session$setInputs(
      dt1_sel = 'dt1',
      vars_sel1 = 'char_x',
      dt2_sel = 'dt2',
      vars_sel2 = 'char_y',
      join_type = 'inner',
      radio_overwrite = 'overwrite',
      btn_apply = 1
    )

    expect_s3_class(session$userData$dt$dt [[ 'dt1' ]], 'data.table')
    expect_equal(session$userData$dt$dt [['dt1']],
                 dt_join(dt1, dt2, 'char_x', 'char_y', 'inner'),
                 ignore_attr = TRUE)
    expect_equal(last_msg, 'Remove modal')
  })
})

# test sample trigger reactiveVal ---------------------------------------------
test_that("Test Sample trigger - updates on button click", {
  testServer(join_server, {
    session$userData$data_changed <- reactiveVal(0)
    trigger_before <- preview_sample_trigger()

    session$setInputs(btn_preview_sample = 1)
    trigger_after <- preview_sample_trigger()

    expect_true(trigger_after > trigger_before)
  })
})

# test joins - preview df -----------------------------------------------------
test_that('Test joins - preview', {
  testServer(join_server, {

    session$userData$dt <- reactiveValues(
      dt = datasets,
      act_name = 'dt1'
    )

    preview_sample_trigger <- reactiveVal(0)

    session$setInputs(
      dt1_sel = 'dt1',
      vars_sel1 = 'char_x',
      dt2_sel = 'dt2',
      vars_sel2 = 'char_y',
      join_type = 'left',
      btn_preview_sample = 1
    )

    # left join
    expect_s3_class(preview_df(), 'data.table')
    expect_equal(preview_df(),
                 dt_join(dt1, dt2, 'char_x', 'char_y', 'left'),
                 ignore_attr = TRUE)

    # right join
    session$setInputs(join_type = 'right', btn_preview_sample = 2)

    expect_s3_class(preview_df(), 'data.table')
    expect_equal(preview_df(),
                 dt_join(dt1, dt2, 'char_x', 'char_y', 'right'),
                 ignore_attr = TRUE)

    # full join
    session$setInputs(join_type = 'full', btn_preview_sample = 3)

    expect_s3_class(preview_df(), 'data.table')
    expect_equal(preview_df(),
                 dt_join(dt1, dt2, 'char_x', 'char_y', 'full'),
                 ignore_attr = TRUE)
  })
})

# test errors -----------------------------------------------------------------
test_that('Preview_df returns validation error', {
  testServer(join_server, {

    session$userData$dt <- reactiveValues(
      dt = datasets,
      act_name = 'dt1'
    )
    session$setInputs(
      dt1_sel = 'dt1',
      dt2_sel = 'dt2',
      vars_sel1 = 'a',
      vars_sel2 = 'b',
      join_type = 'left'
    )

    expect_error(preview_df(), class = 'shiny.silent.error')

  })
})

# test check messages ---------------------------------------------------------
test_that('Apply join - check inputs', {
  testServer(join_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg_error = function(text, ...) { last_msg <<- text },
      msg = function(text, ...) { last_msg <<- text },
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = datasets,
      act_name = 'dt1'
    )

    session$userData$dt_names <- reactive({ names(session$userData$dt$dt) })

    # equal datasets
    session$setInputs(
      dt1_sel = 'dt1',
      dt2_sel = 'dt1',
      btn_apply = 1
    )
    expect_equal(last_msg, 'Select two diferent datasets')

    # selected variables
    session$setInputs(
      dt1_sel = 'dt1',
      dt2_sel = 'dt2',
      btn_apply = 2
    )
    expect_equal(last_msg, 'Select variables for both datasets')

    # lenght of selected variables
    session$setInputs(
      vars_sel1 = c('char_x', 'num'),
      vars_sel2 = c('num'),
      btn_apply = 3
    )
    expect_equal(last_msg, 'The number of selected variables must be the same for both datasets')

    # new name not valid
    session$setInputs(
      vars_sel1 = 'char_x',
      vars_sel2 = 'char_y',
      radio_overwrite = 'new',
      txt_new_dt_name = 123,
      btn_apply = 4
    )
    expect_equal(last_msg, 'New name is not valid or already in use')

    # new name already in use
    session$setInputs(
      txt_new_dt_name = 'dt1',
      btn_apply = 5
    )
    expect_equal(last_msg, 'New name is not valid or already in use')

    # invalid vars
    session$setInputs(
      vars_sel2 = 'char_y2',
      txt_new_dt_name = 'new_dt',
      btn_apply = 6
    )
    expect_equal(last_msg, 'Select valid variables')

    # type of vars must match
    session$setInputs(
      vars_sel1 = 'char_x',
      vars_sel2 = 'num',
      btn_apply = 7
    )
    expect_equal(last_msg, 'The variables types must match')

    # join applied
    session$setInputs(
      vars_sel1 = 'char_x',
      vars_sel2 = 'char_y',
      btn_apply = 8
    )
    expect_equal(last_msg, 'Remove modal')

    # error
    session$setInputs(
      vars_sel1 = 'char_x',
      vars_sel2 = 'char_y',
      txt_new_dt_name = 'dt_error',
      join_type = 'error',
      btn_apply = 9
    )
    expect_equal(last_msg, 'Error: join aborted')

  })
})

# tests/testthat/test-navbar_df_info_module.R

# test df info UI -------------------------------------------------------------
test_that("Test render df_info UI", {
  testServer(navbar_df_info_server, args = list(app_session = NULL), {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |>  as.data.table()),
      act_name = 'iris'
    )

    session$userData$dt_names <- reactive({
      req(session$userData$dt$dt)
      names(session$userData$dt$dt)
    })

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$userData$dt$act_meta <- reactive({
      req(session$userData$dt$df_info())
      session$userData$dt$df_info()[[session$userData$dt$act_name]]
    })

    session$userData$dt$act_mini_meta <- reactive({
      req(session$userData$dt$act_meta())

      list(
        'row_col' = paste(
          f_int(session$userData$dt$act_meta()[['rows']][1]),
          '/',
          f_int(session$userData$dt$act_meta()[['cols']][1])
        ),
        'col_nas' = session$userData$dt$act_meta()[n_nas > 0, ] |> nrow()
      )
    })

    expect_true(is.list(output$navbar_df_info))
    expect_true(grepl('0', output$navbar_df_info$html))
    expect_true(grepl('Rows/Columns:\n  150 / 5\n', output$navbar_df_info$html))
    expect_true(grepl("Columns with NA's:\n  0", output$navbar_df_info$html))
  })
})

# ui --------------------------------------------------------------------------
stats_table_ui <- function(id) {
  ns <- NS(id)
  tagList(
    card(
      card_body(gt_output(ns('gt_stats'))),
      card_footer(
        layout_columns(
          col_widths = c(6, 6),
          numericInput(ns('table_digits'), 'Digits', 2, 0, 9, 1),
          div(style = "margin-top: 28px !important;",
              save_gt_ui(ns('pA_stats_table_save_gt'))
          )
        ),
        div(style = "margin-bottom: -18px !important;")
      )
    ),
    card_footer(insert_output_ui(ns('insert_stats_table')))
  )
}

# server ----------------------------------------------------------------------
stats_table_server <- function(id, var1, var1_name, input_percentile, percentile,
                               var1_sd, pearson_correlation) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns

    task_stats <- ExtendedTask$new(function(x, x_name) {
      mirai({
        list(
          var_name = x_name,

          n = length(x),

          n_nas = collapse::whichNA(x) |> NROW(),

          min = if(is.numeric(x)) collapse::fmin(x, na.rm = TRUE) else NA,

          q1 = if(is.numeric(x)) collapse::fquantile(x, 0.25) else NA,

          median = if(is.numeric(x)) collapse::fmedian(x, na.rm = TRUE) else NA,

          mean = if(is.numeric(x)) collapse::fmean(x, na.rm = TRUE) else NA,

          mode = if(is.numeric(x) || is.character(x) || is.factor(x)) {
            collapse::fmode(x, na.rm = TRUE)
          } else {
            NA
          },

          q3 = if(is.numeric(x)) collapse::fquantile(x, 0.75) else NA,

          max = if(is.numeric(x)) collapse::fmax(x, na.rm = TRUE) else NA
        )

      }, x = x, x_name = x_name)
    })

    observe({
      req(var1(), var1_name())
      task_stats$invoke(var1(), var1_name())
    }) |> bindEvent(var1())

    stats_result <- reactive({
      task_stats$result()
    })

    # table -------------------------------------------------------------------
    stats_table <- reactive({
      req(stats_result(), input_percentile())

      fmt_digits <- min(max(0, input$table_digits), 9)

      data.frame(
        measure = c(
          'Variable',
          paste0("% NA's (", f_int(stats_result()$n_nas), '/', f_int(stats_result()$n), ')'),
          'Minimum',
          'Percentile 25',
          'Median',
          'Mean',
          'Mode',
          'Percentile 75',
          'Maximum',
          paste('Percentile', input_percentile()),
          'Standard Deviation',
          'Pearson Correlation'
        ),
        value = c(
          stats_result()$var_name,
          (stats_result()$n_nas / stats_result()$n * 100) |> f_num(dig = fmt_digits),
          stats_result()$min |> f_num(dig = fmt_digits),
          stats_result()$q1 |> f_num(dig = fmt_digits),
          stats_result()$median |> f_num(dig = fmt_digits),
          stats_result()$mean |> f_num(dig = fmt_digits),
          if(stats_result()$mode |> allNA()) NA else paste(
            stats_result()$mode|> f_num(dig = fmt_digits), collapse = ' | '),
          stats_result()$q3 |> f_num(dig = fmt_digits),
          stats_result()$max |> f_num(dig = fmt_digits),
          percentile() |> f_num(dig = fmt_digits),
          var1_sd() |> f_num(dig = fmt_digits),
          pearson_correlation() |> f_num(dig = fmt_digits)
        )
      )
    })

    # format table ------------------------------------------------------------
    stats_table_fmt <- reactive({
      req(stats_table())

      var_name <- stats_table()[1, 2]
      stats_table()[-1, ] |>
        gt() |>
        tab_header(var_name) |>
        sub_missing(missing_text = '-') |>
        sub_values(values = 'NA', replacement = '-') |>
        cols_label(measure = 'Measure', value = 'Value') |>
        cols_align('left', measure) |>
        cols_align('right', value)
    })

    # render table ------------------------------------------------------------
    output$gt_stats <- render_gt({
      req(stats_table_fmt())
      validate(
        need(
          isTruthy(input_percentile()) &&
            between(input_percentile(), 0, 100),
          'Percentile must be between 0 and 100'),
        need(
          isTruthy(input$table_digits) &&
            between(input$table_digits, 0, 9),
          'Percentile must be between 0 and 9')
        )

      stats_table_fmt() |>
        opt_interactive(use_pagination = F,
                        use_highlight = T,
                        use_compact_mode = T,
                        use_sorting = F) |>
        tab_options(table.background.color = '#ffffff')
    })

    save_gt_server('pA_stats_table_save_gt', stats_table_fmt)

    # insert stats table ------------------------------------------------------
    insert_output_server('insert_stats_table', stats_table_fmt, 'Stats Table')

  })
}

# ui --------------------------------------------------------------------------
restore_session_ui <- function(id) {
  ns <- NS(id)
}

# server ----------------------------------------------------------------------
restore_session_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  # restore saved session data ----------------------------------------------
	  observe({
	    req(session$userData$conf$data_dir,
	        session$userData$conf$restore_session,
	        session$userData$conf$restore_data_status,
	        session$userData$conf$restore_output_status,
	        session$userData$dt$dt
	    )

	    if(session$userData$run_local && session$userData$conf$restore_session == 'always'){

	      if(!file.exists(paste0(session$userData$conf$data_dir, '/data.qs2'))){
	        session$userData$conf$restore_data_status <- 2

	      } else {
	        update_waiter_text(session, 'spada_startup_text', 'Loading data')
	        previous_data <- qs_read(paste0(session$userData$conf$data_dir, '/data.qs2'))

	        # check data format
	        if(!test_data_format(previous_data)){
	          session$userData$conf$restore_data_status <- 3
	        } else {

	          previous_data <- lapply(previous_data, as.data.table)

	          # if empty entry only keep loaded data
	          if(session$userData$conf$no_input_data){
	            session$userData$dt$dt <- previous_data
	            # update meta
	            session$userData$dt$meta <- lapply(previous_data, df_info)
	          } else {
	            previous_data <- make_names_append_list(
	              previous_data,
	              names(session$userData$dt$dt)
	            )

	            session$userData$dt$dt <- c(previous_data, session$userData$dt$dt)

	            # append meta
	            new_meta <- lapply(previous_data, df_info)

	            session$userData$dt$meta <- c(
	              new_meta,
	              session$userData$dt$meta
	            )
	          }

	          session$userData$dt$act_name <- names(session$userData$dt$dt)[1]
	          session$userData$conf$restore_data_status <- 1
	        }
	      }

	      # import output
	      if(!file.exists(paste0(session$userData$conf$data_dir, '/output.qs2'))){
	        session$userData$conf$restore_output_status <- 2

	      } else {

	        update_waiter_text(session, 'spada_startup_text', 'Loading output')
	        previous_output <- qs_read(paste0(session$userData$conf$data_dir, '/output.qs2'))

	        # check output format
	        if(!test_output_format(previous_output)){
	          session$userData$conf$restore_output_status <- 3
	        } else {
	          previous_output <- lapply(previous_output, function(x) {

	            id <- gen_element_id()
	            x$id <- id
	            btn_xid <- paste0('btn_xout_', id)
	            btn_eid <- paste0('btn_eout_', id)

	            register_output_events(id, session, input, btn_xid, btn_eid)

	            add_output_element(
	              id,
	              x$title,
	              x$annotation,
	              x$element,
	              ns(btn_xid),
	              ns(btn_eid)
	            )
	          })
            # rename items of list with new ids
	          names(previous_output) <- vapply(previous_output, function(x) x$id, character(1))

	          session$userData$out$elements <- previous_output
	          session$userData$conf$restore_output_status <- 1
	        }
	      }
	    }

	    session$userData$conf$restore_status <- paste0(
	      session$userData$conf$restore_data_status, '.',
	      session$userData$conf$restore_output_status
	    )
	  }) |> bindEvent(session$userData$conf$restore_session, once = T)

	  # show modal with restored status
	  observe({
	    req(session$userData$conf$restore_status)

	    if(session$userData$run_local && session$userData$conf$restore_session == 'always'){

	      display_restore_status(
	        session$userData$conf$restore_status,
	        actionButton(ns('btn_dismiss_restore_status'), 'OK', class = 'btn-task')
	      )
	    }
	  }) |> bindEvent(session$userData$conf$restore_status, once = T)

	  observe({
	    removeModal()
	  }) |> bindEvent(input$btn_dismiss_restore_status)

  })
}

# ui --------------------------------------------------------------------------
metadata_ui <- function(id) {
  ns <- NS(id)
  card(
    card_body(
      selectInput(
        inputId = ns('dataset_sel'),
        label = 'Dataset',
        choices = NULL
      ),
      gt_output(ns('gt'))
    ),
    card_footer(insert_output_ui(ns('metadata_insert_output')))
  )
}

# server ----------------------------------------------------------------------
metadata_server <- function(id) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns

    # metadata ----------------------------------------------------------------
    observe({
      req(session$userData$dt$gt_info())

      updateSelectInput(
        session,
        inputId = 'dataset_sel',
        choices = c(
          session$userData$dt$act_name,
          setdiff(
            names(session$userData$dt$meta),
            session$userData$dt$act_name
          )
        )
      )
    })

    output$gt <- render_gt({
      req(act_meta_gt())

      act_meta_gt() |>
        opt_interactive(
          use_filters = T,
          use_resizers = T,
          use_highlight = T,
          use_compact_mode = T,
          use_text_wrapping = F,
          use_page_size_select = T
        )
    })

    act_meta_gt <- reactive({
      req(session$userData$dt$gt_info(), input$dataset_sel)
      session$userData$dt$gt_info()[[input$dataset_sel]]
    })

    # insert to output module -------------------------------------------------
    insert_output_server(
      'metadata_insert_output',
      reactive(act_meta_gt() |> tab_options(table.width = pct(90))),
      'Metadata'
    )

  })
}

# ui --------------------------------------------------------------------------
order_cols_ui <- function(id) {
  ns <- NS(id)
  card(
    card_header('Order Columns', class = 'mini-header'),
    card_body(
      selectizeInput(
        ns('vars_cols'),
        'Variables to move',
        NULL,
        multiple = T,
        options = list(plugins = list('remove_button', 'clear_button')),
        width = '80%'
      ),
      radioButtons(ns('radio_cols'),
                   NULL,
                   c('Before' = 'before', 'After' = 'after'),
                   inline = T),
      selectInput(
        ns('vars_rest'),
        list('Other Variables', bs_icon('info-circle')) |>
          ttip(
            'If not informed, the Other Variables will be placed at the end',
            PLACE = 'right'
          ),
        '',
        width = '80%'
      )
    ),
    card_footer(
      btn_task(ns('btn_order_cols'), 'Order Columns', icon('arrow-right-arrow-left')),
    )
  )
}

# server ----------------------------------------------------------------------
order_cols_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

    df_names <- reactive(get_act_dt(session) |> names())

    observe({
      req(df_names())
      updateSelectizeInput(
        session,
        'vars_cols',
        choices = c('', df_names())
      )
    }) |> bindEvent(df_names())

    # rest of vars to order
    observe({
      updateSelectInput(
        session,
        'vars_rest',
        label = 'Other Variables',
        choices = df_names()[df_names() %notin% input$vars_cols]
      )
    }) |> bindEvent(input$vars_cols)

    # btn order cols ---------------------------
    observe({
      if(!isTruthy(input$vars_cols)){
        msg('Select at least one variable')
      } else {
        running_modal()
        temp <- copy(get_act_dt(session))

        if(all(df_names() %in% input$vars_cols) || input$vars_rest == ''){
          setcolorder(temp, input$vars_cols)
        } else if(input$radio_cols == 'before'){
          setcolorder(temp, input$vars_cols, before = input$vars_rest)
        } else if (input$radio_cols == 'after') {
          setcolorder(temp, input$vars_cols, after = input$vars_rest)
        }

        update_act_dt(session, copy(temp), updated_cols = temp |> names(),
                      change_type = 'order_cols')
        rm(temp)
        remove_running_modal()
      }
    }) |> bindEvent(input$btn_order_cols)

  })
}

# ui --------------------------------------------------------------------------
z_test_ui <- function(id) {
  ns <- NS(id)
  card(
    full_screen = T,
    card_header('Z Test', class = 'mini-header'),
    layout_sidebar(
      class = 'card-sidebar',
      sidebar = sidebar(uiOutput(ns('parameters'))),
      navset_card_pill(
        nav_panel(
          'Test',
          card(
            layout_sidebar(
              sidebar = sidebar(
                width = 380,
                fluidRow(
                  h5('Sampe Values'),
                  column(6, p(textOutput(ns('sample_mean')))),
                  column(6, p(textOutput(ns('sample_sd'))))
                ),
                h5('Parameters', style = 'margin-bottom: -18px;'),
                layout_columns(
                  numericInput(
                    ns('mu'),
                    list('Mean', bs_icon('info-circle') |>
                           ttip('Hypothesized Mean')), 0),
                  numericInput(
                    ns('sd'),
                    list('Std Deviation', bs_icon('info-circle') |>
                           ttip('Standard Deviation of Population')),
                    value = 1, min = 0)
                ),
                radioButtons(ns('radio_alternative'), 'Alternative',
                             c('Two sided' = 'two.sided',
                               'Less' = 'less',
                               'Greater' = 'greater'), inline = T),
                numericInput(ns('confidence'), 'Confidence Interval - %',
                             value = 95, 0, 100, 5, width = '200px'),
                layout_columns(
                  col_widths = c(6, 6),
                  btn_task(ns('btn_run_test'), 'Run Test', icon('gear')),
                  btn_task(ns('btn_help_ztest'), 'Help', icon('question'))
                )
              ),
              card_body(
                layout_columns(
                  col_widths = c(2, 7, 3),
                  fluidRow(
                    div(
                      style = 'margin-bottom: 10px;',
                      uiOutput(ns('conditional_staticard_ztest')),
                    )
                  ),
                  uiOutput(ns('conditional_plot')),
                  div(
                    gt_output(ns('ztest_gt')),
                    br(), br(),
                    uiOutput(ns('conditional_save_gt'))
                  )
                )
              ),
              card_footer(
                uiOutput(ns('conditional_add_output'))
              )
            )
          )
        ),
        nav_panel(
          'Histogram',
          card(
            full_screen = T,
            card_body(plotOutput(ns('hist'))),
            card_footer(
              div(style = 'margin-bottom: -8px !important;'),
              layout_columns(
                col_widths = c(1, 3, 3),
                numericInput(ns('bins'), 'Bins', 25, 5, step = 5),
                btn_task(ns('btn_hist'), 'Show Histogram', icon('chart-simple'),
                         style = 'margin-top: 28px'),
                div(insert_output_ui(ns('ztest_insert_output_hist')),
                    style = 'margin-top: 28px')
              ),
              div(style = 'margin-bottom: -24px !important;'),
            )
          )
        )
      )
    )
  )
}

# server ----------------------------------------------------------------------
z_test_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  df <- reactive(get_act_dt(session))

    ztest <- reactiveValues(results = NULL)

    df_active <- reactive(df()[, lapply(df(), is.numeric) == T, with = F])

    var_analysis <- reactive({
      df_names <- df_active() |> names()

      var_analysis <- session$userData$dt$act_meta() |> filter(perc_nas != 1) |> pull(var)

      df_names[df_names %in% var_analysis]
    })

    var <- reactive({
      req(input$sel_var)
      df_active()[[input$sel_var]]
    })

    # calculate values --------------------------------------------------------
    sample_mean <- reactive(var() |> fmean(na.rm = T))

    output$sample_mean <- renderText({
      paste('Mean:', sample_mean() |> f_num(dig = 2))
    })

    sample_sd <- reactive(var() |> fsd(na.rm = T))

    output$sample_sd <- renderText({
      paste('Std Deviation:', sample_sd() |> f_num(dig = 2))
    })

    # input vars --------------------------------------------------------------
    output$parameters <- renderUI({
      tagList(
        selectInput(ns('sel_var'), 'Variable', var_analysis()),
        p('* Showing only numeric variables')
      )
    })

    # run z test --------------------------------------------------------------
    observe({
      req(input$sel_var)
      req(input$radio_alternative)

      if(!isTruthy(input$mu)){
        msg_error('Inform a value for the Mean')
      } else if(!isTruthy(input$sd)){
        msg_error('Inform a value for the Std Deviation')
      } else if(input$sd <= 0){
        msg_error('Standard Deviation must be positive ( > 0)', 2)
      } else if(!isTruthy(input$confidence) ||
         !between(input$confidence, 0, 100)) {
        msg_error('Confidence interval must be between 0 and 100%', 2)
      } else if (sum(!is.na(var())) < 2) {
        msg_error('Inform at least 2 valid values2')
      } else {
        df <- ZTest(var(),
                    alternative = input$radio_alternative,
                    mu = input$mu,
                    sd_pop = input$sd,
                    conf.level = input$confidence/100)

        df <- df |>
          unlist() |> as.data.frame()

        df$results <- rownames(df)
        names(df) <- c('values', 'results')

        df[df$results == 'data.name', ]$values <- input$sel_var

        ztest$results <- df

        ztest$confidence <- input$confidence/100
      }
    }) |> bindEvent(input$btn_run_test)

    # results gt table --------------------------------------------------------
    ztest_results_gt <- reactive({
      req(ztest$results)
      ztest$results |>
        gt() |>
        cols_move(columns = 'values', after = 'results') |>
        cols_label('values' = 'Values', 'results' = 'Results') |>
        tab_header('Z Test')
    })

    output$ztest_gt <- render_gt({
      req(ztest_results_gt())
      ztest_results_gt()
    })

    # save gt module ----------------------------------------------------------
    save_gt_server('ztest_save_gt', ztest_results_gt)

    output$conditional_save_gt <- renderUI({
      req(ztest_results_gt())
      save_gt_ui(ns('ztest_save_gt'))
    })

    # insert to output --------------------------------------------------------
    insert_output_server(
      'ztest_insert_output',
      reactive(
        gen_table2(
          plotTag(ztest_plot()(), '', width = 1000, height = 500),
          ztest_results_gt()
        )
      ),
      'Z Test'
    )

    output$conditional_add_output <- renderUI({
      req(ztest_results_gt())
      insert_output_ui(ns('ztest_insert_output'))
    })

    # test plot ---------------------------------------------------------------
    output$conditional_plot <- renderUI({
      req(ztest_plot())
      plotOutput(ns('ztest_plot'))
    })

    ztest_plot <- reactive({
      req(ztest$results)
      req(ztest$confidence)

      function(){
        plot_z_test(
          confidence = ztest$confidence,
          test_type = ztest$results |>
            filter(results == 'alternative') |>
            pull(values),
          z_value = ztest$results |>
            filter(results == 'statistic.z') |>
            pull(values) |>
            as.numeric()
        )
      }
    })

    output$ztest_plot <- renderPlot({
      req(ztest$results)
      req(ztest$confidence)
      req(ztest_plot())

      ztest_plot()()
    })

    # staticards --------------------------------------------------------------
    output$conditional_staticard_ztest <- renderUI({
      req(ztest_results_gt())
      tagList(
        stati_card(ztest$results |>
                   filter(results %in% c('statistic.z')) |>
                   pull(values) |>
                   as.numeric() |>
                   f_num(dig = 3),
                   'Z value'),
        stati_card(ztest$results |>
                   filter(results %in% c('p.value')) |>
                   pull(values) |>
                   as.numeric() |>
                   f_num(dig = 3),
                   'p value')
      )
    })

    # help modal --------------------------------------------------------------
    observe({
      showModal(modalDialog(
        HTML(get_help_file('DescTools', 'ZTest')),
        easyClose = TRUE, size = 'xl'
      ))
    }) |> bindEvent(input$btn_help_ztest)

    # histogram ---------------------------------------------------------------
    output$hist <- renderPlot({
      req(ztest_hist())
      validate(need(input$bins > 0, 'Bins must be 1 or higher'))

      ztest_hist()
    }, res = 96)

    task_hist <- ExtendedTask$new(function(spada_plot_fun,
                                           df,
                                           title,
                                           bins,
                                           plot_fill_color,
                                           plot_line_color,
                                           sample_limit,
                                           mean_value,
                                           sd_value) {
      mirai({
        spada_plot_fun(
          type = 'hist_density',
          df = df,
          xvar = 'x',
          xlab = 'Values',
          ylab = 'Density',
          title = title,
          bins = bins,
          fill_color = plot_fill_color,
          line_color = plot_line_color,
          point_shape = if (sample_limit > 1e4 && nrow(df) > 1e4) '.' else 20,
          sample_limit = sample_limit,
          mean_value = mean_value,
          sd_value = sd_value
        )
      },
      spada_plot_fun = spada_plot_fun,
      df = df,
      title = title,
      bins = bins,
      plot_fill_color = plot_fill_color,
      plot_line_color = plot_line_color,
      sample_limit = sample_limit,
      mean_value = mean_value,
      sd_value = sd_value)
    }) |> bind_task_button('btn_hist')

    observe({
      req(input$sel_var, var())
      task_hist$invoke(
        spada_plot_fun = spada_plot,
        df = data.frame(x = var()),
        bins = input$bins,
        title = paste('Histogram -', input$sel_var),
        plot_fill_color = session$userData$conf$plot_fill_color,
        plot_line_color = session$userData$conf$plot_line_color,
        sample_limit = session$userData$conf$plot_limit,
        mean_value = fmean(var(), na.rm = TRUE),
        sd_value = fsd(var(), na.rm = TRUE)
      )
    }) |> bindEvent(input$btn_hist)

    ztest_hist <- reactive({ task_hist$result() })

    # insert histogram to output ----------------------------------------------
    insert_output_server(
      'ztest_insert_output_hist',
      reactive(
        plotTag(ztest_hist(), '', width = 1000, height = 500)
      ),
      'Histogram'
    )

  })
}

# ui --------------------------------------------------------------------------
save_gt_ui <- function(id) {
  ns <- NS(id)

  dropdownButton(
    inputId = ns('drop_btn'),
    label = 'Save',
    tagList(
      h3('Save table'),
      textInput(ns('file_name'), 'File name', value = 'table'),
      radioGroupButtons(ns('radio_format'), 'Format',
                        c('html', 'rtf'), size = 'sm', individual = T),
      downloadButton(ns('down_handler'), 'Save table', icon('download')),
      br()),
    circle = F, size = 'sm', icon = icon('download')
  )
}

# server ----------------------------------------------------------------------
save_gt_server <- function(id, input_table) {
  moduleServer(id, function(input, output, session) {

    output$down_handler <- downloadHandler(

      filename = function() {
        req(input_table())
        paste0(input$file_name, switch(input$radio_format, html = '.html', rtf = '.rtf'))
      },
      content = function(file) {
        req(input_table())
        gtsave(input_table(), file)
      }
    )
  })
}

# ui --------------------------------------------------------------------------
descriptive_stats_ui <- function(id) {
  ns <- NS(id)
  card(
    full_screen = T,
    card_header('Descriptive Statistics', class = 'mini-header'),
    layout_sidebar(
      class = 'card-sidebar',
      sidebar = sidebar(
        uiOutput(ns('parameters')),
        checkboxGroupInput(
          ns('xg_central_tendency'),
          h6('Central Tendency'), inline = T,
          c('Mean' = 'mean', 'Geometric Mean' = 'gmean', 'Harmonic Mean' = 'hmean',
            'Median' = 'median', 'Mode' = 'mode'),
          c('mean', 'gmean', 'hmean', 'median', 'mode')
        ),
        checkboxGroupInput(
          ns('xg_dispersion'), h6('Dispersion'), inline = T,
          c('Minimum' = 'min', 'Maximum' = 'max', 'IQR',
            'Range' = 'range', 'Variance' = 'var',
            'Standard Deviation' = 'sd'),
          c('min', 'max', 'IQR', 'range', 'var', 'sd')
        ),
        checkboxGroupInput(
          ns('xg_shape'), h6('Shape'), inline = T,
          c('Skewness' = 'skew', 'Kurtosis' = 'kurt'), c('skew', 'kurt')
        ),
        numericInput(ns('table_digits'), 'Digits', 2, 0, 9, 1),
        btn_task(ns('btn_stats'), 'Generate Table', icon('gear'))
      ),
      navset_card_pill(
        nav_panel(
          'Stats',
          card(full_screen = T,
            card_body(gt_output(ns('gt_stats'))),
            card_footer(
              layout_columns(
                col_widths = c(3, 3),
                uiOutput(ns('conditional_add_output')),
                uiOutput(ns('conditional_save_gt'))
              )
            )
          )
        )
      )
    )
  )
}

# server ----------------------------------------------------------------------
descriptive_stats_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  dt <- reactive(get_act_dt(session))

    var_analysis <- reactive({ dt() |> names() })

    output$parameters <- renderUI({
      tagList(
        selectizeInput(
          ns('sel_var'), 'Variables', var_analysis(), var_analysis()[1],
          multiple = T,
          options = list(plugins = list('remove_button', 'clear_button'))
        ),
      )
    })

    # calculate stats ---------------------------------------------------------
    task_desc_stats <- ExtendedTask$new(
      function(desc_stats_fun, df, fmt_digits, central_tendency, dispersion, shape){
      mirai({

        desc_stats_fun(
          df = df,
          fmt_digits = fmt_digits,
          central_tendency = central_tendency,
          dispersion = dispersion,
          shape = shape
        )
      },
      desc_stats_fun = desc_stats,
      df = df,
      fmt_digits = fmt_digits,
      central_tendency = central_tendency,
      dispersion = dispersion,
      shape = shape
      )
    }) |> bind_task_button('btn_freq_table')

    observe({
      if(!isTruthy(input$sel_var)) {
        msg('Select at least 1 variable')
        return()
      } else if(!isTruthy(input$table_digits) || input$table_digits < 1) {
        msg('The precision digits must be between 1 and 9')
        return()
      } else if(all(!isTruthy(input$xg_central_tendency),
                    !isTruthy(input$xg_dispersion),
                    !isTruthy(input$xg_shape))
                ) {
        msg('Select at least one measure')
        return()
      }

      task_desc_stats$invoke(
        desc_stats_fun = desc_stats,
        df = dt()[, .SD, .SDcols = input$sel_var],
        fmt_digits = min(max(0, input$table_digits), 9),
        central_tendency = input$xg_central_tendency,
        dispersion = input$xg_dispersion,
        shape = input$xg_shape
      )
    }) |> bindEvent(input$btn_stats)

    calculated_stats <- reactive({ task_desc_stats$result() })

    # gt table ----------------------------------------------------------------
    gt_stats <- reactive({
      req(calculated_stats())
      data.frame(
        Measures = names(calculated_stats()),
        do.call(rbind, calculated_stats())
      ) |>
        gt() |>
        cols_align(align = 'right') |>
        cols_align('left', Measures) |>
        sub_missing(missing_text = '-') |>
        sub_values(values = 'NA', replacement = '-') |>
        sub_values(values = 'Gmean', replacement = 'Geometric Mean') |>
        sub_values(values = 'Hmean', replacement = 'Harmonic Mean') |>
        tab_header('Descriptive Statistics')
    })

    output$gt_stats <- render_gt({
      req(gt_stats())
      gt_stats() |>
        opt_interactive(page_size_default = 25)
    })

    # save table module -------------------------------------------------------
    save_gt_server('pA_desciptive_stats_save_gt', gt_stats)

    output$conditional_save_gt <- renderUI({
      req(gt_stats())
      save_gt_ui(ns('pA_desciptive_stats_save_gt'))
    })

    # insert to output --------------------------------------------------------
    insert_output_server('desc_stats_insert_output', gt_stats, 'Descriptive Statistics')

    output$conditional_add_output <- renderUI({
      req(gt_stats())
      insert_output_ui(ns('desc_stats_insert_output'))
    })
  })
}

# ui --------------------------------------------------------------------------
about_spada_ui <- function(id) {
  ns <- NS(id)
  nav_panel(
    value = 'about',
    title = 'About',
    icon = icon('info'),
    card(
      class = 'big-card',
      layout_columns(
        col_widths = c(6, 6),
        card(
          card_body(
            div(
              style = 'display:flex; align-items:center; gap:12px;',
              tags$img(src = 'spada/favicon-32x32.png'),
              h4('Package Description', style = 'margin:0;')
            ),
            verbatimTextOutput(ns('info_about_spada'))
          )
        ),
        card(
          card_body(
            h4(icon('r-project'), 'Session Info'),
            verbatimTextOutput(ns('session_info'))
          )
        )
      )
    )
  )
}

# server ----------------------------------------------------------------------
about_spada_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

    output$session_info <- renderPrint(
      sessionInfo()
    )

    output$info_about_spada <- renderPrint(
      packageDescription('spada')
    )
  })
}

# ui --------------------------------------------------------------------------
join_ui <- function(id) {
  ns <- NS(id)

  layout_column_wrap(
    card(
      card_header('Join', class = 'mini-header'),
      card_body(
        fluidRow(
          column(6, selectInput(ns('dt1_sel'), 'Active Dataset', NULL)),
          column(6, selectInput(ns('dt2_sel'), 'Dataset 2', NULL))
        ),
        radioGroupButtons(
          ns('join_type'),
          'Select a Join',
          c(
            'Left Join' = 'left',
            'Right Join' = 'right',
            'Full Join' = 'full',
            'Inner Join' = 'inner'
            # ,
            # 'Anti Join' = 'anti',
            # 'Semi Join' = 'semi'
          ),
          selected = 'left',
          size = 'sm',
          individual = T
        ),
        fluidRow(
          column(
            6,
            selectizeInput(
              ns('vars_sel1'),
              'Variables 1',
              NULL,
              multiple = T,
              options = list(plugins = list('remove_button', 'clear_button'))
            )
          ),
          column(
            6 ,
            selectizeInput(
              ns('vars_sel2'),
              'Variables 2',
              NULL,
              multiple = T,
              options = list(plugins = list('remove_button', 'clear_button'))
            )
          )
        ),
        radioButtons(
          ns('radio_overwrite'),
          NULL,
          c('New' = 'new', 'Overwrite' = 'overwrite'),
          inline = T
        ),
        conditionalPanel(
          condition = "input.radio_overwrite == 'new'",
          ns = ns,
          textInput(ns('txt_new_dt_name'), 'New Dataset Name', placeholder = 'new_dataset')
        )
      ),
      card_footer(btn_task(ns('btn_apply'), 'Apply', icon('check')))
    ),
    card(
      card_header('Preview', class = 'mini-header'),
      card_body(gt_output(ns('preview_gt'))),
      card_footer(btn_task(
        ns('btn_preview_sample'),
        'Show new sample',
        icon('rotate-right')
      ))
    )
  )
}

# server ----------------------------------------------------------------------
join_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  n_dt <- reactive({
	    req(session$userData$dt$dt)
	    session$userData$dt$dt |> length()
	  })

	  dt_names <- reactive({
	    req(session$userData$dt$dt)
	    session$userData$dt$dt |> names()
	  })

	  act_name <- reactive({
	    req(session$userData$dt$act_name)
	    session$userData$dt$act_name
	  })

	  var1_names <- reactive({
	    req(session$userData$dt$dt)
	    req(input$dt1_sel)
      session$userData$dt$dt[[ input$dt1_sel ]] |> names()
    })

	  var2_names <- reactive({
	    req(session$userData$dt$dt)
	    req(input$dt2_sel)
	    session$userData$dt$dt[[ input$dt2_sel ]] |> names()
	  })

    # update select inputs -------------------
    observe({
      req(act_name())
      updateSelectizeInput(
        session,
        'dt1_sel',
        choices = c(act_name())
      )
    }) |> bindEvent(act_name())

    observe({
      req(var1_names())
      updateSelectizeInput(
        session,
        'vars_sel1',
        choices = c('', var1_names())
      )
    }) |> bindEvent(var1_names())

    observe({
      req(dt_names())

      choices <- dt_names()

      updateSelectizeInput(
        session,
        'dt2_sel',
        choices = choices
      )
    }) |> bindEvent(dt_names())

    observe({
      choices <- if(n_dt() <= 1) character(0) else var2_names()
      updateSelectizeInput(
        session,
        'vars_sel2',
        choices = choices
      )
    }) |> bindEvent(var2_names(), n_dt())

    # # sample to preview conversion -----------
    preview_sample_trigger <- reactiveVal(0)
    preview_sample <- reactive({
      req(input$dt2_sel)

      nrow1 <- get_act_dt(session) |> nrow()
      nrow2 <- session$userData$dt$dt[[ input$dt2_sel ]] |> nrow()

      nrow1 <- if(nrow1 < 8) rep(TRUE, nrow1) else sample(nrow1, 8, replace = F)
      nrow2 <- if(nrow2 < 8) rep(TRUE, nrow2) else sample(nrow2, 8, replace = F)

      list('s_dt1' = get_act_dt(session)[nrow1, ],
           's_dt2' = session$userData$dt$dt[[ input$dt2_sel ]][nrow2, ])
    }) |> bindEvent(preview_sample_trigger(),
                    input$dt1_sel,
                    input$dt2_sel,
                    input$vars_sel1,
                    input$vars_sel2
                    )

    # # update sample in button click ----------
    observe({
      preview_sample_trigger(preview_sample_trigger() + 1)
    }) |> bindEvent(input$btn_preview_sample)

    preview_df <- reactive({
      req(preview_sample())
      req(input$join_type)
      req(input$dt1_sel)
      req(input$dt2_sel)
      req(input$vars_sel1)
      req(input$vars_sel2)

      join_type <- input$join_type
      dt1_name <- input$dt1_sel
      dt2_name <- input$dt2_sel
      vars_dt1 <- input$vars_sel1
      vars_dt2 <- input$vars_sel2

      validate(
        need(dt1_name != dt2_name, 'Select two diferent datasets'),
        need(length(vars_dt1) == length(vars_dt2),
             'The number of selected variables must has the same for both datasets')
      )

      dt1 <- preview_sample()[[ 's_dt1' ]]
      dt2 <- preview_sample()[[ 's_dt2' ]]

      req(vars_dt1 %in% names(dt1))
      req(vars_dt2 %in% names(dt2))

      types_x <- sapply(dt1[, .SD, .SDcols = vars_dt1], obj_type)
      types_y <- sapply(dt2[, .SD, .SDcols = vars_dt2], obj_type)

      validate(
        need(all(types_x == types_y), 'All variables type must match')
      )

      temp <- try(
        dt_join(dt1, dt2, vars_dt1, vars_dt2, join_type), silent = T
      )

      validate(
        need(!inherits(temp, 'try-error'), 'Error')
      )

      temp

    })

    # render preview gt ----------------------
    output$preview_gt <- render_gt({

        preview_df() |>
          gt() |>
          cols_align(align = 'right') |>
          opt_interactive(
            page_size_default = 10,
            use_sorting = F,
            use_pagination = F,
            use_highlight = T,
            use_compact_mode = T) |>
          tab_options(table.background.color = bg_color)
    })

    # apply join --------------------------------------------------------------
    observe({

      join_type <- input$join_type
      dt1_name <- input$dt1_sel
      dt2_name <- input$dt2_sel
      vars_dt1 <- input$vars_sel1
      vars_dt2 <- input$vars_sel2
      radio <- input$radio_overwrite
      new_dt_name <- input$txt_new_dt_name

      # check diferrent datasets
      if((!isTruthy(dt1_name) || !isTruthy(dt2_name)) || dt1_name == dt2_name){
        msg('Select two diferent datasets')
        return()

      # check selected variables
      } else if (!isTruthy(vars_dt1) || !isTruthy(vars_dt2)){
        msg('Select variables for both datasets')
        return()

      # check len of selected variables
      } else if(length(vars_dt1) != length(vars_dt2)){
        msg('The number of selected variables must be the same for both datasets', 4)
        return()

      # check new dt name
      } else if (radio == 'new' &&
                 (!is_name_available(new_dt_name, session$userData$dt_names())))
      {
        msg_error('New name is not valid or already in use')
        return()

      # check vars in datasets
      } else if(
        !all(
          vars_dt1 %in% names(session$userData$dt$dt[[ dt1_name ]]),
          vars_dt2 %in% names(session$userData$dt$dt[[ dt2_name ]])
        ))
        {
        msg_error('Select valid variables')
        return()

      # passed all ui checks
      } else {

        dt1 <- copy(session$userData$dt$dt[[ dt1_name ]])
        dt2 <- copy(session$userData$dt$dt[[ dt2_name ]])

        types_x <- sapply(dt1[, .SD, .SDcols = vars_dt1], obj_type)
        types_y <- sapply(dt2[, .SD, .SDcols = vars_dt2], obj_type)

        if(!all(types_x == types_y)){
          msg_error('The variables types must match')
          return()
        } else {
          running_modal()

          temp <- try(
            dt_join(dt1, dt2, vars_dt1, vars_dt2, join_type), silent = T
          )

          # return error msg
          if(inherits(temp, 'try-error')){
            remove_running_modal()
            msg_error('Error: join aborted')
            return()
          # apply join
          } else {
            if(radio == 'overwrite'){
              update_act_dt(session, temp)
            } else if (radio == 'new'){
              append_dt(session, temp, new_dt_name)
              append_meta(session, temp |> df_info(), new_dt_name)
            }
            remove_running_modal()
          }
        }
      }
    }) |> bindEvent(input$btn_apply)
  })
}

# ui --------------------------------------------------------------------------
data_highlights_ui <- function(id) {
  ns <- NS(id)

  card_body(
    layout_column_wrap(
      value_box(
        title = 'Numeric Vars',
        value = textOutput(ns('var_num_vars')),
        showcase = bs_icon('123'),
        theme = 'bg-gradient-yellow-orange'
      ),
      value_box(
        title = 'Character Vars',
        value = textOutput(ns('var_char_vars')),
        showcase = bs_icon('alphabet'),
        theme = 'bg-gradient-blue-indigo'
      ),
      value_box(
        title = 'Factor Vars',
        value = textOutput(ns('var_factor_vars')),
        showcase = bs_icon('diagram-3'),
        theme = 'bg-gradient-green-blue'
      ),
      value_box(
        title = 'Date Vars',
        value = textOutput(ns('var_date_vars')),
        showcase = bs_icon('calendar3'),
        theme = 'bg-gradient-purple-indigo'
      )
    ),
    layout_column_wrap(
      value_box(
        title = 'Rows',
        value = textOutput(ns('n_rows')),
        showcase = bs_icon('list'),
        theme = 'bg-gradient-blue-purple'
      ),
      value_box(
        title = 'Columns',
        value = textOutput(ns('n_cols')),
        showcase = bs_icon('grid-3x3-gap-fill'),
        theme = 'bg-gradient-indigo-green'
      ),
      value_box(
        title = 'Most valid values',
        value = textOutput(ns('var_most_valid')),
        showcase = bs_icon('list-check'),
        theme = 'bg-gradient-indigo-yellow',
        p('Number of valid:', textOutput(ns('var_most_valid_n_valid'), inline = T))
      ) |> tooltip('Showing 1, there may be ties', placement = 'top'),
      value_box(
        title = "Most NA's",
        value = textOutput(ns('var_most_nas')),
        showcase = bs_icon('database-x'),
        theme = 'bg-gradient-red-indigo',
        p("Number of NA's:", textOutput(ns('var_most_nas_n'), inline = T), ' rows')
      ) |> tooltip('Showing 1, there may be ties', placement = 'top'),
    ),
    layout_column_wrap(
      value_box(
        title = 'Max value',
        value = textOutput(ns('var_max_value'), inline = T),
        showcase = bs_icon('graph-up-arrow', placement = 'top'),
        theme = 'bg-gradient-blue-green',
        p('Max value:', textOutput(ns('max_value'), inline = T))
      ) |> tooltip('Showing 1, there may be ties', placement = 'top'),
      value_box(
        title = 'Min value',
        value = textOutput(ns('var_min_value')),
        showcase = bs_icon('graph-down-arrow'),
        theme = 'bg-gradient-pink-indigo',
        p('Min value:', textOutput(ns('min_value'), inline = T))
      ) |> tooltip('Showing 1, there may be ties', placement = 'top'),
      value_box(
        title = 'Most zeros',
        value = textOutput(ns('var_most_zeros')),
        showcase = bs_icon('0-circle'),
        theme = 'bg-gradient-orange-indigo',
        p('Number of zeros:', textOutput(ns('var_most_zeros_n_zeros'), inline = T))
      ) |> tooltip('Showing 1, there may be ties', placement = 'top'),
      value_box(
        title = 'Size',
        value = p(textOutput(ns('act_dt_size'), inline = T), 'MB'),
        showcase = bs_icon('sd-card'),
        theme = 'bg-gradient-teal-indigo'
      )
    )
  )

}

# server ----------------------------------------------------------------------
data_highlights_server <- function(id) {
  moduleServer(id, function(input, output, session) {

    dt <- reactive({
      req(session$userData$dt$dt)
      get_act_dt(session)
    })

    dt_meta <- reactive({
      req(session$userData$dt$act_meta())
      copy(session$userData$dt$act_meta())
    })

    output$var_num_vars <- renderText(sapply(dt(), is.numeric) |> sum_nona())
    output$var_char_vars <- renderText(sapply(dt(), is.character) |> sum_nona())
    output$var_factor_vars <- renderText(sapply(dt(), is.factor) |> sum_nona())
    output$var_date_vars <- renderText(sapply(dt(), is_date) |> sum_nona())

    output$n_rows <- renderText( dt_meta()[1, ]$rows |> f_num() )

    output$n_cols <- renderText( dt_meta()[1, ]$cols |> f_num() )

    output$var_most_valid <- renderText(
      if (dt_meta()[n_valid > 0, ] |> nrow() < 1) {
        '---'
      } else {
        setorderv(dt_meta(), 'n_valid', -1, na.last = T)[1, var]
      }
    )

    output$var_most_valid_n_valid <- renderText(
      if (dt_meta()[n_valid > 0, ] |> nrow() < 1) {
        '0'
      } else {
        setorderv(dt_meta(), 'n_valid', -1, na.last = T)[1, f_num(n_valid)]
      }
    )

    output$var_most_zeros <- renderText(
      if (dt_meta()[n_zero > 0, ] |> nrow() < 1) {
        '---'
      } else {
        setorderv(dt_meta(), 'n_zero', -1, na.last = T)[1, var]
      }
    )

    output$var_most_zeros_n_zeros <- renderText(
      if (dt_meta()[n_zero > 0, ] |> nrow() < 1) {
        '0'
      } else {
        setorderv(dt_meta(), 'n_zero', -1, na.last = T)[1, f_num(n_zero)]
      }
    )

    output$var_most_nas <- renderText({
      if (dt_meta()[n_nas > 0, ] |> nrow() < 1) {
        '---'
      } else {
        setorderv(dt_meta()[n_nas > 0, ],
                  c('n_nas', 'perc_nas'), c(-1, -1), na.last = T)[1, var]
      }
    })

    output$var_most_nas_n <- renderText({
      if (dt_meta()[n_nas > 0, ] |> nrow() < 1) {
        '0'
      } else {
        setorderv(dt_meta()[n_nas > 0, ],
                  c('n_nas', 'perc_nas'), c(-1, -1), na.last = T)[1, f_num(n_nas)]
      }
    })

    output$var_max_value <- renderText(
      if (dt_meta()[!is.na(max), ] |> nrow() < 1) {
        '---'
      } else {
        setorderv(dt_meta(), 'max', -1, na.last = T)[1, var]
      }
    )

    output$max_value <- renderText(
      if (dt_meta()[!is.na(max), ] |> nrow() < 1) {
        '0'
      } else {
        setorderv(dt_meta(), 'max', -1, na.last = T)[1, f_num(max, dig = 3)]
      }
    )

    output$var_min_value <- renderText(
      if (dt_meta()[!is.na(min), ] |> nrow() < 1) {
        '---'
      } else {
        setorderv(dt_meta(), 'min', na.last = T)[1, var]
      }
    )

    output$min_value <- renderText(
      if (dt_meta()[!is.na(min), ] |> nrow() < 1) {
        '0'
      } else {
        setorderv(dt_meta(), 'min', na.last = T)[1, f_num(min, dig = 3)]
      }
    )

    act_dt_size <- reactive({
      req(dt())
      (object.size(dt()) / 2^20) |> as.numeric() |> round(2) |>
        format(big.mark = ',', decimal.mark = '.')
    })

    output$act_dt_size <- renderText({
      req(act_dt_size())
      act_dt_size()
    })
  })
}

# ui --------------------------------------------------------------------------
frequencies_ui <- function(id) {
  ns <- NS(id)

  tagList(
    card_body(
      gt_output(ns('table')),
    ),
    card_footer(
      fluidRow(
        column(3, btn_task(ns('btn_freq_table'), 'Generate Table', icon('gear'))),
        column(3, insert_output_ui(ns('insert_frequencies')))
      )
    )
  )
}

# server ----------------------------------------------------------------------
frequencies_server <- function(id, var, var_name) {
  moduleServer(id, function(input, output, session) {

    ns <- session$ns

    task_freq_table <- ExtendedTask$new(function(is_date_fun, var, var_name){
      mirai({
        if(is_date_fun(var)){

          if(unique(var) |> length() < 9){
            freq_df <- DescTools::Freq(var |> as.factor(), useNA = 'always')
          } else {
            freq_df <- DescTools::Freq(var, breaks = 9, useNA = 'always')
          }
        } else {
          freq_df <- DescTools::Freq(var, useNA = 'always')
        }
        freq_df <- freq_df |> as.data.frame()

        attr(freq_df, 'title') <- var_name
        freq_df

      },
      is_date_fun = is_date_fun,
      var = var,
      var_name = var_name)
    }) |> bind_task_button('btn_freq_table')

    observe({
      req(var(), var_name())
      task_freq_table$invoke(
        is_date_fun = is_date,
        var = var(),
        var_name = var_name()
      )
    }) |> bindEvent(input$btn_freq_table)

    freq_table <- reactive({ task_freq_table$result() })

    freq_table_gt <- reactive({

      req(freq_table())

      freq_table() |>
        gt() |>
        tab_header(attr(freq_table(), 'title')) |>
        fmt_percent(
          columns = c('perc', 'cumperc'),
          decimals = 4
        ) |>
        cols_label(
          level = 'Level',
          freq = 'Frequency',
          perc = 'Percent',
          cumfreq = 'Cum Frequency',
          cumperc = 'Cum Percent'
        ) |>
        fmt_number(
          columns = c('freq', 'cumfreq'),
          decimals = 0,
          sep_mark = ',',
          dec_mark = '.'
        ) |> sub_missing()
    })

    output$table <- render_gt({
      req(freq_table_gt())
      freq_table_gt() |>
        opt_interactive(
          page_size_default = 10,
          use_filters = T,
          use_resizers = T,
          use_highlight = T,
          use_compact_mode = T,
          use_text_wrapping = F,
          use_page_size_select = T
        )
    })

    # insert table of values to output ----------------------------------------
    insert_output_server('insert_frequencies', freq_table_gt, 'Frequencies')

  })
}

# ui --------------------------------------------------------------------------
summarise_ui <- function(id) {
  ns <- NS(id)

  card(
    card_header('Summarise', class = 'mini-header'),
    card_body(
      selectizeInput(
        ns('vars_sel'), 'Variables', '',
        multiple = T,
        options = list(plugins = list('remove_button', 'clear_button')),
        width = '80%'
      ),
      selectInput(ns('fun_sel'), 'Function', summarise_functions),
      radioButtons(ns('radio_overwrite'), 'Dataset',
                   c('New' = 'new', 'Overwrite' = 'overwrite'), inline = T),
      conditionalPanel(
        condition = "input.radio_overwrite == 'new'", ns = ns,
        textInput(ns('txt_new_dt_name'), 'New Dataset Name', placeholder = 'new_dataset')
      )
    ),
    card_footer(btn_task(ns('btn_summarise'), 'Summarise', icon('check')))
  )
}

# server ----------------------------------------------------------------------
summarise_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  vars_names <- reactive(get_act_dt(session) |> names())

    # variables ------------------------------
    observe({
      req(vars_names())
      updateSelectInput(
        session,
        'vars_sel',
        choices = c('', vars_names())
      )
    }) |> bindEvent(vars_names())

	  # summarise event ---------
    observe({
      if(!isTruthy(input$vars_sel)){
        msg('Insert at least one variable')
        return()
      } else if (input$radio_overwrite == 'new' &&
                 !is_name_available(input$txt_new_dt_name, session$userData$dt_names()))
      {
        msg_error('New name is not valid or already in use')
        return()
      } else {

        running_modal()
        temp <- copy(get_act_dt(session))

        temp <- summarise_dt(temp, input$fun_sel, input$vars_sel)

        # overwrite -------
        if(input$radio_overwrite == 'overwrite'){

          update_act_dt(session, temp)

        } else if(input$radio_overwrite == 'new'){

          append_dt(session, temp, input$txt_new_dt_name)
          append_meta(session, temp |> df_info(), input$txt_new_dt_name)

        }
        rm(temp)
        remove_running_modal()
      }

    }) |> bindEvent(input$btn_summarise)

  })
}

# themes names ----------------------------------------------------------------
themes_names <- c('spada_theme', 'spada_dark_theme')

# app colors ------------------------------------------------------------------
main_color <- '#02517d'
navbar_bg <- '#0A5A88'
sidebar_color <- '#E3E3E4'
inputs_border_color <- '#C8C8C8'
inputs_border_color2 <- '#D1D1D1'
bg_color <- '#F9F9F9'
secondary <- '#0072B2'
sucess <- '#009E73'
danger <- '#B60020'
plot_fill_color <- '#0099F8'
plot_line_color <- '#EE7942'
plot_title_color <- '#02517d'

spada_white <- '#FFFFFF'
spada_black <- '#000000'
spada_navpills_hover <- '#F0F0F0'
spada_navpills_hover_dark <- '#5A5A5A'
spada_blue5 <- '#0A6CA4'
spada_blue6 <- '#0A74B8'
spada_gray5 <- '#424242'
spada_red5 <- '#DC3545'

# palettes --------------------------------------------------------------------
gray_palette <- c('#CFCFCF', '#585858', '#232323')
blue_palette <- c('#229BD4', '#096691', '#134359')
yl_palette   <- c('#F5CB4E', '#FFC107', '#F7A305')
dg_palette   <- c('#378C88', '#1C6561', '#284E4C')
lg_palette   <- c('#0CB0A8', '#228F8A', '#09918B')
pk_palette   <- c('#A35D8C', '#BF007F', '#8F0360')
red_palette  <- c('#8F3646', '#B60020', '#750217')

# themes function -------------------------------------------------------------
spada_bs_theme <- function(theme) {

  basic_rules <- list(
    paste(
      " $main_color:", main_color, ";",
      " $secondary:", secondary, ";",
      " $bg_color:", bg_color, ";",
      " $sidebar_bg:", sidebar_color, ";",
      " $inputs_border_color:", inputs_border_color, ";",
      " $inputs_border_color2:", inputs_border_color2, ";",
      " $navbar_bg:", navbar_bg, ";",
      " $spada_navpills_hover:", spada_navpills_hover, ";",
      " $spada_navpills_hover_dark:", spada_navpills_hover_dark, ";",
      " $spada_white:", spada_white, ";",
      " $spada_blue5:", spada_blue5, ";",
      " $spada_blue6:", spada_blue6, ";",
      " $spada_red5:", spada_red5, ";",
      " $btn_neutral:", spada_gray5, ";",

      "

      $stati_card_text: $spada_white;

      .navbar {
        min-height: 45px !important;
        padding-top: 4px !important;
        padding-bottom: 4px !important;

        background: linear-gradient(
          to right,
          $main_color,
          $spada_blue6
        ) !important;

      }

      .main{
        padding-right: 16px !important;
        padding-top: 16px !important;
        padding-bottom: 8px !important;
      }

      .spada-main-sidebar{
        background: $sidebar_bg
      }

      .spada-main-sidebar .accordion-item{
        border: none !important;
        border-color: transparent !important;
      }

      .nav-link { font-size: 18px; }

      .big-card{ background-color: $main_color; }

      .mini-header {
        color: $spada_white;

        background:
          linear-gradient(
            180deg,
            $spada_blue5 0%,
            $navbar_bg 100%
          );

        border-bottom: 1px solid $spada_white;
        border-radius: 0.2rem;
      }

      .btn-task:active,
      .btn-default:active {
        background-color: darken($bg_color, 10%) !important;
        transform: scale(0.99);
        box-shadow: inset 0 2px 4px rgba(0, 0, 0, 0.2);
      }

      .btn-task:hover,
      .btn-default:hover {
        background-color: $secondary !important;
        border-color: $secondary !important;
        color: white !important;
      }

      .btn-task-cancel {
        color: $spada_red5 !important;
        background-color: $bg_color !important;
        border-color: $spada_red5 !important;
      }

      .btn-task-cancel:hover {
        background-color: $danger !important;
        border-color: $danger !important;
        color: $bg_color !important;
      }

      .btn-task.btn-task-neutral {
        color: $btn_neutral !important;
        background-color: white !important;
        border-color: $btn_neutral !important;
      }

      .btn-task.btn-task-neutral:hover {
        background-color: $btn_neutral !important;
        border-color: $btn_neutral !important;
        color: $bg_color !important;
      }

      .mini-btn {
        color: $secondary !important;
        background-color: $bg_color !important;
        border-color: $secondary !important;
        border-radius: 0.2rem;
        font-weight: 400;

        width: 48px;
        height: 32px;
        padding: 0 !important;
      }

      .micro-btn-cancel {
        width: 18px !important;
        height: 18px !important;
        padding: 0 !important;
        font-size: 0.65rem !important;
        border: none !important;
        background: transparent !important;
      }

      .micro-btn-cancel:hover,
      .micro-btn-cancel:focus,
      .micro-btn-cancel:active {
        color: $secondary !important;
        background-color: transparent !important;
        border-color: transparent !important;
      }

      .card, .well { --bs-card-inner-border-radius: '0.2rem'; }

      .card-body { border-radius: '0.2rem'; }

      .card-sidebar {
        background:
        linear-gradient(
            180deg,
            $navbar_bg 0%,
            $spada_blue5 100%
        );
      }

      .value-box-title { font-size: 1rem !important; }

      .value-box-value { font-size: 1.5rem !important; }

      .nav-pills .nav-link {
        font-size: 14px !important;
        padding-top: 6px !important;
        padding-bottom: 6px !important;
      }

      .popover.preview-dt-popup {
        max-width: 120vw;
        width: 600px;
        border-radius: 0.2rem !important;
      }

      .popover.preview-dt-popup-mini {
        max-width: 120vw;
        width: 300px;
        border-radius: 0.2rem !important;
      }

      .popover.preview-dt-popup .popover-body {
        max-height: 100vh;
        overflow: auto;
      }

      .modal-content {
        border: none !important;
        border-radius: 0.6rem !important;
        box-shadow:
          0 1.5rem 4rem rgba(0, 0, 0, .20),
          0 .5rem 1rem rgba(0, 0, 0, .08);
      }

      .modal-header {
        background: linear-gradient(
          135deg,
          $main_color 0%,
          $spada_blue6 100%
        );

        color: $spada_white;
        border: none !important;
        padding: 1rem 1.5rem;
      }

      .modal-footer {
        border-top: 1px solid rgba(0, 0, 0, .06);
        padding: 1rem 1.5rem;
      }

      .navbar-nav .nav-link {
        transition: background-color 0.2s ease;
        border-radius: 0.6rem;
        margin: 1px;
        padding-left: 12px !important;
        padding-right: 12px !important;
      }

      .navbar-nav .nav-link:hover {
        background-color: rgba(255, 255, 255, 0.12);
        color: white !important;
      }

      .navbar-nav .nav-link.active {
        background-color: rgba(255, 255, 255, 0.18);
        color: white !important;
      }

      .selectize-input {
        word-wrap: break-word;
        word-break: break-word;
      }

      .selectize-dropdown {  }

      .navbar .dropdown-menu {
        border: none;
        border-radius: 0.875rem;
        padding: 6px;
        overflow: hidden;
        box-shadow: 0 10px 30px rgba(0, 0, 0, 0.15);
      }

      .navbar .dropdown-item {
        border-radius: 0.5rem;
        transition: background-color .15s ease, color .15s ease;
      }

      .selectize-dropdown {
        word-wrap: break-word;
        border: none !important;
        border-radius: 0.4rem !important;
        overflow: hidden;
        box-shadow:
          0 0.75rem 2rem rgba(0, 0, 0, .15);
      }

      .selectize-dropdown .option {
        margin: 0.2rem;
        border-radius: 0.2rem;
        transition: background-color .15s ease;
      }

      .selectize-dropdown .active {
        background-color: rgba(10, 108, 164, .08);
      }

      .spada-navbar-df-info-header {
        display: inline-block;
        background: linear-gradient(135deg, $main_color, $spada_blue5);
        color: white;
        padding: 0.6rem 1.2rem;
        border-radius: 0.6rem;
        margin-bottom: 1rem;
      }

      .spada-navbar-df-info-header h2{
        margin: 0;
        font-size: 1.25rem;
        font-weight: 600;
        color: white;
        overflow-wrap: anywhere;
        word-break: break-word;
      }
    "
    )
  )

  if (theme == 'spada_theme' || !(theme %in% themes_names)) {
    bs_theme(
      version = 5,
      bg = bg_color,
      fg = spada_black,
      primary = main_color,
      secondary = secondary,
      success = sucess,
      danger = danger,
      font_size_base = '1rem',
      'nav-pills-border-radius' = '0.2rem',
      'nav-pills-link-active-color' = main_color,
      'nav-pills-link-active-bg' = sidebar_color,
      'border-radius-sm' = '0.2rem',
      'border-radius' = '0.2rem',
      'navbar-brand-font-size' = '1.5rem',
      'navbar-brand-padding-y' = '0.250rem',
      'btn-font-weight' = 400,
      'dropdown-bg' = bg_color,
      'dropdown-color' = main_color,
      'dropdown-link-color' = spada_black,
      'dropdown-link-hover-bg' = sidebar_color,
      base_font = font_collection('Segoe UI', 'Ubuntu', 'system-ui')
    ) |>
      bs_add_rules(basic_rules) |>
      bs_add_rules(
        list(
          "
          .accordion-sidebar{
            background-color: $main_color !important;
            color: $spada_white;
          }

          .card {
            border-radius: 0.2rem;
            margin: -8px;
          }

          .big-card-footer{
            background-color: $main_color;
            margin-top: -12px !important;
            padding-bottom: 6px !important;
            height: 60px;
          }

          .btn-task,
          .btn-default {
            color: $secondary !important;
            background-color: $bg_color !important;
            border-color: $secondary !important;
            padding-top: 6px !important;
            padding-bottom: 6px !important;
            border-radius: 0.2rem
          }

          .btn-task-cancel {
            color: $spada_red5 !important;
            background-color: white !important;
            border-color: $spada_red5 !important;
            border-radius: 0.2rem
          }

          .control-label {
            margin-bottom: 3px !important;
            padding-top: 3px !important;
          }

          .bslib-sidebar-layout>.sidebar {
            background-color: $sidebar_bg !important;
            color: black !important;
          }

          .bslib-sidebar-layout>.collapse-toggle {
            color: black !important;
            background-color: unset;
          }

          .nav-pills .nav-link:hover {
            background-color: $spada_navpills_hover !important;
          }

          .shiny-input-text,
          .shiny-input-number,
          .selectize-input,
          .shiny-input-textarea textarea {
            border: 1px solid $inputs_border_color2 !important;
            border-radius: 0.2rem !important;
            box-shadow: none !important;

            transition:
              border-color 0.2s ease,
              box-shadow 0.2s ease;
          }

          .selectize-input:hover,
          .shiny-input-text:hover,
          .shiny-input-number:hover,
          .shiny-input-textarea textarea:hover {
            border-color: $inputs_border_color !important;
            border-bottom: 1px solid $main_color !important;
          }

          .selectize-input:focus,
          .shiny-input-text:focus,
          .shiny-input-number:focus,
          .shiny-input-textarea textarea:focus {
            border-color: $inputs_border_color !important;
            border-bottom: 1px solid $main_color !important;
          }

          .navbar .dropdown-item.active,
          .navbar .dropdown-item:active {
            background: linear-gradient(
              135deg,
              $main_color 0%,
              $spada_blue5 100%
            ) !important;
            color: $spada_white !important;
          }

        "
        )
      )
  } else if (theme == 'spada_dark_theme') {
    bs_theme(
      version = 5,
      bootswatch = 'darkly',
      'border-radius-sm' = '0.2rem',
      'border-radius' = '0.2rem',
      'navbar-brand-font-size' = '1.5rem',
      'btn-font-weight' = 400,
      base_font = font_collection('Open Sans', 'Ubuntu', 'system-ui')
    ) |>
      bs_add_rules(basic_rules) |>
      bs_add_rules(
        list(
          "
          .card {
            border-radius: '0.2rem';
            margin: -4px;
          }

          .big-card-footer{
            margin-top: -12px !important;
            padding-bottom: 0px !important;
            height: 60px;
          }

          .btn-task,
          .btn-default {
            color: $secondary;
            background-color: $bg_color;
            border-color: $secondary;
            border-radius: '0.2rem'
          }

          .nav-pills .nav-link:hover {
            background-color: $spada_navpills_hover_dark !important;
            color: $spada_white !important;
          }

        "
        )
      )
  }
}

# ui --------------------------------------------------------------------------
cor_matrix_ui <- function(id) {
  ns <- NS(id)
  card(
    full_screen = T,
    card_header('Correlation Matrix', class = 'mini-header'),
    layout_sidebar(
      class = 'card-sidebar',
      sidebar = sidebar(
        uiOutput(ns('parameters')),
        radioButtons(ns('radio_method'), 'Method',
                     c('Pearson' = 'pearson',
                       'Kendall' = 'kendall',
                       'Spearman' = 'spearman')),
        btn_task(ns('btn_cor_matrix'), 'Run', icon('gear'))
      ),
      navset_card_pill(
        nav_panel(
          'Correlation Matrix',
          card(
            card_body(plotOutput(ns('cor_plot'))),
            card_footer(insert_output_ui(ns('insert_cor_matrix')))
          )
        )
      )
    )
  )
}

# server ----------------------------------------------------------------------
cor_matrix_server <- function(id) {
  moduleServer(id, function(input, output, session) {
	  ns <- session$ns

	  df <- reactive(get_act_dt(session))

    # outupt objects ----------------------------------------------------------
    df_active <- reactive({
      req(df())
      df()[, lapply(df(), is.numeric) == T, with = F]
    })

    var_analysis <- reactive({
      df_names <- df_active() |> names()

      var_analysis <- session$userData$dt$act_meta()[perc_nas != 1, var]

      df_names[df_names %in% var_analysis]
    })

    output$parameters <- renderUI({
      tagList(
        selectizeInput(
          ns('sel_vars'),
          'Variables',
          var_analysis(),
          var_analysis(),
          multiple = T,
          options = list(plugins = list('remove_button', 'clear_button'))
        ),
        p('* Showing only numeric variables')
      )
    })

    # run cor matrix ----------------------------------------------------------
    task_cor_matrix <- ExtendedTask$new(function(df, vars, method){
      mirai({

        df_cor <- cor(
            subset(df, select = vars), method = method, use = 'na.or.complete'
          ) |>
          as.data.frame()

        df_cor <- cbind(Var1 = rownames(df_cor), df_cor) |>
          data.table::as.data.table()

        df_cor <- data.table::melt(
          data = df_cor,
          id.vars = 'Var1',
          variable.name = 'Var2',
          value.name = 'value'
        )

        df_cor[, label := round(value, digits = 3)]
        df_cor[, method := method]
        df_cor

      },
      df = df,
      vars = vars,
      method = method
      )
    }) |> bind_task_button('btn_cor_matrix')

    observe({
      req(df(), input$sel_vars, input$radio_method)
      task_cor_matrix$invoke(
        df = df(),
        vars = input$sel_vars,
        method = input$radio_method
      )
    }) |> bindEvent(input$btn_cor_matrix)

    df_cor <- reactive({ task_cor_matrix$result() })

    # cor matrix plot ---------------------------------------------------------
    cor_matrix <- reactive({
      req(df_cor())

      method <- df_cor()$method[1]

      text_size <- max(
        6,
        10 / sqrt(length(df_cor()$Var1))
      )

      ggplot(
        data = df_cor(),
        mapping = aes(x = Var1, y = Var2, fill = value)
      ) + geom_tile(color = '#ffffff') +
        labs(title = paste(StrCap(method), 'Correlation'),
             x = NULL, y = NULL, fill = '') +
        theme(
          panel.grid = element_blank(),
          panel.background = element_blank(),
          plot.background = element_blank(),
          axis.text.x = element_text(size = 16),
          axis.text.y = element_text(size = 16),
          plot.title = element_text(
            color = session$userData$conf$plot_title_color, size = 16, face = 'bold'
          )
        ) +
        geom_text(aes(label = label), size = text_size, color = '#000000') +
        scale_fill_gradient2(
          low = danger,
          mid = '#ffffff',
          high = secondary,
          midpoint = 0,
          limits = c(-1, 1)
        )
    })

    output$cor_plot <- renderPlot({
      cor_matrix()
    })

    # insert cor matrix to output ---------------------------------------------
    insert_output_server(
      'insert_cor_matrix',
      reactive(plot_tag(cor_matrix())),
      'Correlation Matrix'
    )

  })
}

# ui --------------------------------------------------------------------------
export_file_ui <- function(id) {
  ns <- NS(id)
    layout_column_wrap(
      card(
        fluidRow(
          column(3, textInput(ns('file_name'), 'File name', value = 'dataset')),
          column(3, radioButtons(ns('radio_format'), 'File format',
                                 c('csv', 'qs2', 'RDS', 'sav', 'xlsx'), inline = T))
        ),
        conditionalPanel(
          condition = "input.radio_format == 'csv'", ns = ns,
          card(
            card_header('Csv Parameters', class = 'mini-header'),
            card_body(
              checkboxInput(ns('x_rownames'), 'Save row names'),
              fluidRow(
                column(3, radioButtons(ns('radio_separator'), 'Separator',
                             c('Semicolon' = ';', 'Comma' = ','), inline = T)),
                column(3, radioButtons(ns('radio_decimal'), 'Decimal mark',
                       c('Dot' = '.', 'Comma' = ','), inline = T))
              ),
              fluidRow(
                column(3, textInput(ns('txt_na'), 'Missing (NA) substitute', value = '')),
                column(3, radioButtons(ns('radio_scientific'), 'Scientific notation',
                             c('No' = 999999999, 'Allow' = 0), inline = T))
              )
            )
          )
        ),
        conditionalPanel(
          condition = "input.radio_format == 'RDS'", ns = ns,
          card(
            card_header('RDS Parameters', class = 'mini-header'),
            card_body(
              checkboxInput(ns('checkbox_rds_compress'), 'Compress')
            ),
          )
        ),
        conditionalPanel(
          condition = "input.radio_format == 'sav'", ns = ns,
          card(
            card_header('Sav Parameters', class = 'mini-header'),
            card_body(
              radioButtons(ns('radio_sav_compress'), 'Compress',
                           c('Byte' = 'byte', 'None' = 'none', 'Zsav' = 'zsav'),
                           inline = T)
            ),
          )
        ),
        card_footer(downloadButton(
          ns('down_handler'),
          'Export Active dataset',
          class = 'btn-task',
          icon = icon('download')
        ))
      )
    )
}

# server ----------------------------------------------------------------------
export_file_server <- function(id) {
  moduleServer(id, function(input, output, session) {

    act_dt <- reactive({ get_act_dt(session) })

    output$down_handler <- downloadHandler(

      filename = function() {

        paste0(
          input$file_name,
          switch(
            input$radio_format,
            'csv' = '.csv',
            'qs2' = '.qs2',
            'RDS' = '.RDS',
            'RDS Compressed' = '.RDS',
            'sav' = '.sav',
            'xlsx' = '.xlsx',
            ''
          )
        )
      },
      content = function(file) {

        dt <- act_dt()

        if(input$radio_format == 'csv'){
          fwrite(dt, file,
                 row.names = input$x_rownames,
                 sep = input$radio_separator,
                 dec = input$radio_decimal,
                 na = input$txt_na,
                 scipen = as.integer(input$radio_scientific)
          )
        } else if (input$radio_format == 'qs2'){
          qs_save(dt, file)
        } else if (input$radio_format == 'RDS'){
          saveRDS(dt, file, compress = input$checkbox_rds_compress)
        } else if (input$radio_format == 'sav') {
          write_sav(dt, file, compress = input$radio_sav_compress)
        } else if (input$radio_format == 'xlsx') {

          if (nrow(dt) <= 1048575 && ncol(dt) <= 16384) {
            write_xlsx(dt, file, format_headers = F)
          } else {

            rows <- min(1048575, nrow(dt))
            cols <- min(16384, ncol(dt))
            showModal(
              modalDialog(
                title = 'xlsx Limits',
                paste(
                  'xlsx format supports 1,048,576 rows and 16,384 columns.',
                  'Exporting', f_int(rows), 'rows (and the headers) and',
                  f_int(cols), 'columns.'
                ),
                easyClose = TRUE,
                footer = modalButton('Dismiss')
              )
            )

            write_xlsx(dt[1:rows, 1:cols], file, format_headers = F)
          }
        }
      }
    )
  })
}

# ui --------------------------------------------------------------------------
insert_output_ui <- function(id) {
  ns <- NS(id)

  btn_task(ns('btn_add_output'), 'Add to output', icon('plus'))
}

# server ----------------------------------------------------------------------
insert_output_server <- function(id, input_element, element_title = 'Title') {
  moduleServer(id, function(input, output, session) {

    ns <- session$ns

    is_large_data_frame <- reactive({
      if('gt_tbl' %in% (input_element() |> class()) &&
         'data.frame' %in% (input_element()$`_data` |> class()) &&
         ((input_element()$`_data` |> nrow()) > 250 ||
          (input_element()$`_data` |> ncol()) > 250)) TRUE else FALSE
    })

    # convert gt table to raw HTML for speed
    new_element <- reactive({
      req(input_element())

      el <- input_element()

      if(inherits(el, 'gt_tbl')) {
        HTML(as_raw_html(el))
      } else {
        el
      }
    })

    # add output -----------------------------
    observe({
      req(new_element())

      showModal(
        modalDialog(
          title = div(icon('plus'), 'Add element to Output'),
          size = 'l',
          tagList(
            if (is_large_data_frame()) {
              div(
                class = 'alert alert-warning',
                icon('triangle-exclamation'),
                'Inserting an extensive data.frame in the Output may cause performance degradation'
              )
            },
            textInput(ns('output_title'), 'Title', value = element_title),
            textAreaInput(
              ns('output_annot'),
              'Annotation',
              width = '90%',
              rows = 7,
              resize = 'both'
            )
          ),
          footer = tagList(
            actionButton(ns('btn_cancel_add_output'), 'Cancel',
                         icon = icon('xmark'), class = 'btn-task btn-task-cancel'),
            actionButton(ns('btn_confirm_add_output'), 'Submit',
                         icon = icon('check'), class = 'btn-task')
          )
        )
      )

    }) |> bindEvent(input$btn_add_output)

    # cancel_add_output ----------------------
    observe({
      removeModal()
    }) |> bindEvent(input$btn_cancel_add_output)

    # confirm add output ---------------------
    observe({
      if(!isTruthy(input$output_title) || trimws(input$output_title) == ''){
        msg('Insert a Title', DURATION = 1)
      } else {
        removeModal()

        id <- gen_element_id()
        btn_xid <- paste0('btn_xout_', id)
        btn_eid <- paste0('btn_eout_', id)

        register_output_events(id, session, input, btn_xid, btn_eid)

        session$userData$out$elements[[id]] <- add_output_element(
          id,
          input$output_title,
          input$output_annot,
          new_element(),
          ns(btn_xid),
          ns(btn_eid)
        )
        msg('Added to output', DURATION = 1)
      }

    }) |> bindEvent(input$btn_confirm_add_output)

  })
}

# ui --------------------------------------------------------------------------
import_file_ui <- function(id) {
  file_extensions <- c('csv', 'qs2', 'RDS', 'sav')
  ns <- NS(id)
    card(
      card_body(
        fluidRow(
          column(3, textInput(ns('dataset_name'), 'Dataset name', 'dataset')),
          column(3, radioButtons(ns('radio_file_ext'), 'File format',
                                        file_extensions, inline = T)),
          column(6, fileInput(ns('file'), 'Select a File', width = '80%')),
        ),
        conditionalPanel(
          condition = "input.radio_file_ext == 'csv'", ns = ns,
          card(
            card_header('Csv Parameters', class = 'mini-header'),
            card_body(
              layout_column_wrap(
                checkboxInput(ns('x_csv_header'), 'Header', T),
                radioButtons(ns('radio_separator'), 'Separator',
                             c('Semicolon' = ';', 'Comma' = ','), inline = T),
                radioButtons(ns('radio_decimal'), 'Decimal Mark',
                             c('Dot' = '.', 'Comma' = ','), inline = T),
                radioButtons(ns('csv_lines'), 'Lines to read',
                             choices = c('All' = 'all', 'Specific (N)' = 'some'), inline = TRUE),
                conditionalPanel(
                  condition = "input.csv_lines == 'some'", ns = ns,
                  numericInput(ns('csv_n_lines'), NULL, width = '150px',
                               value = 100, min = 1, step = 100) |>
                    ttip('You may use Scientific Notation, e.g. 1e6')
                ),
                btn_task(ns('btn_preview_raw'), 'Preview raw file', icon('magnifying-glass')),
              )
            )
          )
        ),
        conditionalPanel(
          condition = "input.radio_file_ext == 'sav'", ns = ns,
          card(
            card_header('Sav Parameters', class = 'mini-header'),
            card_body(
              layout_columns(
                col_widths = c(2, 2),
                radioButtons(ns('sav_lines'), 'Lines to read',
                             choices = c('All' = 'all', 'Specific (N)' = 'some'), inline = TRUE),
                conditionalPanel(
                  condition = "input.sav_lines == 'some'", ns = ns,
                  numericInput(ns('sav_n_lines'), NULL, width = '150px',
                               value = 100, min = 1, step = 100) |>
                    ttip('You may use Scientific Notation, e.g. 1e6')
                )
              )
            )
          )
        )
      ),
      card_footer(
        btn_task(ns('btn_import'), 'Import file', icon('upload')),
      )
    )
}

# server ----------------------------------------------------------------------
import_file_server <- function(id) {
  moduleServer(id, function(input, output, session) {

    data <- reactiveValues(data = NULL)

    observe({
      if (is.null(input$file)) {
        msg_error('Insert a file')
        return()
      }

      if (!is_name_available(input$dataset_name, session$userData$dt_names())){
        msg_error('Name invalid or already in use')
        return()
      }

      # import file
      ext <- file_ext(input$file$datapath) |> tolower()

      if (ext == 'csv' && input$radio_file_ext == 'csv'){
        n <- if (input$csv_lines == 'all') Inf else as.integer(input$csv_n_lines)
        if (n < 1) {
          msg_error('Number of lines must be at least 1')
          return()
        }
        data$data <- tryCatch({
          fread(input$file$datapath,
                sep = input$radio_separator,
                dec = input$radio_decimal,
                header = input$x_csv_header,
                nrows = n,
                check.names = TRUE)
          },
          warning = \(w) msg_error('Check parameters'),
          error = \(e) msg_error('Check parameters')
        )
      } else if (ext == 'qs2' && input$radio_file_ext == 'qs2'){
        data$data <- qs_read(input$file$datapath)

        if (!is_spada_df(data$data)) {
          msg_error('Object must be a valid dataset (nrow and ncol > 0)', 5)
          return()
        }
      } else if (ext == 'rds' && input$radio_file_ext == 'RDS'){
        data$data <- readRDS(input$file$datapath)
        if (!is_spada_df(data$data)) {
          msg_error('Object must be a valid dataset (nrow and ncol > 0)', 5)
          return()
        }
      } else if (ext == 'sav' && input$radio_file_ext == 'sav'){
        n <- if (input$sav_lines == 'all') Inf else as.integer(input$sav_n_lines)
        if (n < 1) {
          msg_error('Number of lines must be at least 1')
          return()
        }
        data$data <- read_sav(input$file$datapath, n_max = n,
                                  .name_repair = make.names) |>
          as.data.table()
        data$data[, names(.SD) := lapply(.SD, as_factor), .SDcols = is.labelled]

      } else {
        msg_error(paste('Insert a', input$radio_file_ext, 'file'))
        return()
      }

      # update dt ----------------------------
      append_dt(session, data$data, input$dataset_name)

      data$data <- NULL
      gc()

      # update metadata ----------------------
      append_meta(session,
                  session$userData$dt$dt[[input$dataset_name]] |> df_info(),
                  input$dataset_name)

      msg('File imported successfully')

    }) |> bindEvent(input$btn_import)

    # preview csv file --------------------------------------------------------
    observe({
      req(input$file)

        ext <- file_ext(input$file$datapath) |> tolower()

        if (ext == 'csv' && input$radio_file_ext == 'csv') {
          preview_data <- readLines(input$file$datapath, n = 8)

          showModal(modalDialog(
            size = 'xl',
            title = div(icon('magnifying-glass'), 'Preview Raw File'),
            HTML(paste(preview_data, collapse = '<br>'))
          ))

        } else {
          msg_error(paste('Insert a', input$radio_file_ext, 'file'))
        }
      }) |> bindEvent(input$btn_preview_raw)

  })
}

# tests/testthat/test-select_cols_module.R

# test drop columns -----------------------------------------------------------
test_that('Test drop one column', {
  testServer(select_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_sel = 'Species',
                      radio_var_sel = 'drop',
                      btn_sel = 1)

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 iris |> as.data.table() |> subset(select = -Species))

    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test drop two column', {
  testServer(select_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_sel = c('hp', 'mpg'),
                      radio_var_sel = 'drop',
                      btn_sel = 1)

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 mtcars |> as.data.table() |> subset(select = -c(hp, mpg)))
    expect_equal(last_msg, 'Remove modal')
  })
})

# test keep columns -----------------------------------------------------------
test_that('Test keep one column', {
  testServer(select_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_sel = 'hp',
                      radio_var_sel = 'keep',
                      btn_sel = 1)

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 mtcars |> as.data.table() |> subset(select = hp))
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test keep two column', {
  testServer(select_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_sel = c('Species', 'Petal.Length'),
                      radio_var_sel = 'keep',
                      btn_sel = 1)

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 iris |> as.data.table() |> subset(select = c(Species, Petal.Length)))
    expect_equal(last_msg, 'Remove modal')
  })
})

# test check messages ---------------------------------------------------------
test_that('Select cols - check inputs', {
  testServer(select_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg = function(text, ...) { last_msg <<- text }
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris |> as.data.table()),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(btn_sel = 1)

    expect_equal(last_msg, 'Select at least one variable')

    session$setInputs(radio_var_sel = 'drop', vars_sel = names(iris), btn_sel = 2)

    expect_equal(last_msg, 'Leave at least 1 variable')

  })
})

# tests/testthat/test-insert_output_module.R

# test insert output ----------------------------------------------------------
test_that('Test insert outuput', {
  testServer(insert_output_server, args = list(reactive(div('Content'))), {

    # Values inserted in modal
    session$setInputs(output_title = 'Title',
                      output_annot = 'Annotation',
                      btn_confirm_add_output = 1)

    # check names
    expect_true(grepl('id_', session$userData$out$elements |> names()))
    # check id
    expect_type(session$userData$out$elements[[1]]$id, 'character')
    expect_true(grepl('id_', session$userData$out$elements[[1]]$id))
    # check title
    expect_type(session$userData$out$elements[[1]]$title, 'character')
    expect_true(session$userData$out$elements[[1]]$title == 'Title')

    # test annotation
    expect_true(inherits(session$userData$out$elements[[1]]$annotation, 'character'))

    # test element
    expect_true(inherits(session$userData$out$elements[[1]]$element, c('shiny.tag', 'gt_tbl')))

    # test card
    expect_true(inherits(session$userData$out$elements[[1]]$card, 'shiny.tag'))

    # test btn_x
    expect_true(inherits(session$userData$out$elements[[1]]$btn_x, 'shiny.tag'))

    # test btn_e
    expect_true(inherits(session$userData$out$elements[[1]]$btn_e, 'shiny.tag'))

  })
})

# tests/testthat/test-about_spada_module.R

# test info about spada -------------------------------------------------------
test_that('Test info about spada', {

  testServer(about_spada_server, {

    package_output <- output$info_about_spada
    expect_type(package_output, 'character')
    expect_true(grepl('Package: spada', package_output))
    expect_true(grepl("A 'shiny' Package for Data Analysis", package_output))
    expect_true(grepl('Depends: R (>= 4.4.0)', package_output, fixed = T))
  })
})

# test dependencies -----------------------------------------------------------
test_that('Test dependencies', {

  testServer(about_spada_server, {

    package_output <- output$info_about_spada
    expect_true(grepl('bsicons', package_output, fixed = T))
    expect_true(grepl('bslib', package_output, fixed = T))
    expect_true(grepl('collapse', package_output, fixed = T))
    expect_true(grepl('data.table', package_output, fixed = T))
    expect_true(grepl('DescTools', package_output, fixed = T))
    expect_true(grepl('ggplot2', package_output, fixed = T))
    expect_true(grepl('gt', package_output, fixed = T))
    expect_true(grepl('haven', package_output, fixed = T))
    expect_true(grepl('htmltools', package_output, fixed = T))
    expect_true(grepl('mirai', package_output, fixed = T))
    expect_true(grepl('qs2', package_output, fixed = T))
    expect_true(grepl('rlang', package_output, fixed = T))
    expect_true(grepl('shiny', package_output, fixed = T))
    expect_true(grepl('shinyWidgets', package_output, fixed = T))
    expect_true(grepl('waiter', package_output, fixed = T))
  })
})

# tests/testthat/test-data_module.R

dfs <- list('df_iris' = iris |> as.data.table(),
            'df_mtcars' = mtcars |> as.data.table())

temp_dir <- tempdir()

start_conf <- list(
  'conf_dir' = file.path(temp_dir, 'conf'),
  'data_dir' = file.path(temp_dir, 'data'),
  'theme' = 'spada_theme',
  'file_size' = 1000,
  'restore_session' = 'never',
  'save_session' = 'ask',
  'restore_data_status' = 0,
  'restore_output_status' = 0,
  'restore_status' = NULL,
  'plot_fill_color' = plot_fill_color,
  'plot_line_color' = plot_line_color
)

# test make active dt ---------------------------------------------------------
test_that('Test data - make active dt', {
  testServer(data_server, {

    session$userData$dt_names <- reactive(dfs |> names())
    session$userData$dt$dt <- dfs
    session$userData$dt$act_name <- 'df_iris'

    session$setInputs(
      sel_dt = 'df_mtcars',
      btn_active = 1
    )

    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      mtcars |> as.data.table()
    )
  })
})

# test copy dt ----------------------------------------------------------------
test_that('Test data - copy dt', {
  testServer(data_server, {

    session$userData$dt_names <- reactive(dfs |> names())
    session$userData$dt$dt <- dfs
    session$userData$dt$act_name <- 'df_iris'

    session$setInputs(
      sel_dt = 'df_mtcars',
      txt_new_name = 'copy',
      btn_copy_dt = 1
    )

    expect_equal(
      session$userData$dt$dt[['copy']],
      mtcars |> as.data.table()
    )
  })
})

# test del dt -----------------------------------------------------------------
test_that('Test data - del dt', {
  testServer(data_server, {

    session$userData$dt_names <- reactive(dfs |> names())
    session$userData$dt$dt <- dfs
    session$userData$dt$act_name <- 'df_iris'

    session$setInputs(
      sel_dt = 'df_mtcars',
      btn_del_dt = 1
    )

    expect_equal(session$userData$dt$dt |> names(), 'df_iris')
    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      iris |> as.data.table()
    )
  })
})

test_that('Test try delete active df', {
  testServer(data_server, {

    session$userData$dt_names <- reactive(dfs |> names())
    session$userData$dt$dt <- dfs
    session$userData$dt$act_name <- 'df_iris'

    session$setInputs(
      sel_dt = 'df_iris',
      btn_del_dt = 1
    )

    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      iris |> as.data.table()
    )
  })
})

# test rename dt --------------------------------------------------------------
test_that('Test rename dataset', {
  testServer(data_server, {

    session$userData$dt_names <- reactive(dfs |> names())
    session$userData$dt$dt <- dfs
    session$userData$dt$act_name <- 'df_iris'
    session$userData$dt$meta = lapply(dfs, df_info)

    session$setInputs(
      txt_new_name = 'dataset',
      sel_dt = 'df_mtcars',
      btn_new_name = 1
    )

    expect_equal(session$userData$dt$dt |> names(), c('df_iris', 'dataset'))
    expect_equal(names(session$userData$dt$meta), c('df_iris', 'dataset'))
    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      iris |> as.data.table()
    )
    expect_equal(
      session$userData$dt$dt[['dataset']],
      mtcars |> as.data.table()
    )
  })
})

test_that('Test rename active df - invalid name', {
  testServer(data_server,{

    session$userData$dt_names <- reactive(dfs |> names())
    session$userData$dt$dt <- dfs
    session$userData$dt$act_name <- 'df_iris'
    session$userData$dt$meta = lapply(dfs, df_info)

    session$setInputs(
      txt_new_name = 'new dataset',
      sel_dt = 'df_mtcars',
      btn_new_name = 1
    )

    expect_equal(session$userData$dt$act_name, 'df_iris')
    expect_equal(names(session$userData$dt$meta), c('df_iris', 'df_mtcars'))
  })
})

test_that('Test rename non active df', {
  testServer(data_server,{

    session$userData$dt_names <- reactive(dfs |> names())
    session$userData$dt$dt <- dfs
    session$userData$dt$act_name <- 'df_iris'
    session$userData$dt$meta = lapply(dfs, df_info)

    session$setInputs(
      txt_new_name = 'df_mtcars2',
      sel_dt = 'df_mtcars',
      btn_new_name = 1
    )

    expect_equal(session$userData$dt$act_name, 'df_iris')
    expect_equal(names(session$userData$dt$meta), c('df_iris', 'df_mtcars2'))

  })
})

# tests/testthat/test-summarise_module.R

dt <- iris |> as.data.table()

# test summarise - distinct ---------------------------------------------------
test_that('Test summarise - new dt - distinct function', {
  testServer(summarise_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt_names <- reactive('iris')
    session$userData$dt$dt[['iris']] <- dt
    session$userData$dt$act_name <- 'iris'

    session$setInputs(
      vars_sel = 'Petal.Length',
      fun_sel = 'distinct',
      radio_overwrite = 'new',
      txt_new_dt_name = 'new_dt',
      btn_summarise = 1
    )

    expect_equal(
      session$userData$dt$dt[['new_dt']],
      dt[, .SD, .SDcols = 'Petal.Length'] |> unique()
    )
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test summarise - overwrite dt - distinct function', {
  testServer(summarise_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt_names <- reactive('iris')
    session$userData$dt$dt[['iris']] <- dt
    session$userData$dt$act_name <- 'iris'
    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      vars_sel = 'Petal.Length',
      fun_sel = 'distinct',
      radio_overwrite = 'overwrite',
      btn_summarise = 1
    )

    expect_equal(
      session$userData$dt$dt[['iris']],
      dt[, .SD, .SDcols = 'Petal.Length'] |> unique()
    )
    expect_equal(last_msg, 'Remove modal')
  })
})

# test input check messages ---------------------------------------------------
test_that('Test sumamrise - check inputs', {
  testServer(summarise_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg_error = function(text, ...) { last_msg <<- text },
      msg = function(text, ...) { last_msg <<- text }
    )

    session$userData$dt_names <- reactive('iris')
    session$userData$dt$dt[['iris']] <- dt
    session$userData$dt$act_name <- 'iris'

    session$setInputs(vars_sel = NULL,btn_summarise = 1)

    expect_equal(last_msg, 'Insert at least one variable')

    session$setInputs(
      vars_sel = 'Species',
      radio_overwrite = 'new',
      txt_new_dt_name = 123,
      btn_summarise = 2
    )

    expect_equal(last_msg, 'New name is not valid or already in use')
  })
})

# tests/testthat/test-data_bkp_module.R

dfs <- list('df_iris' = iris |> as.data.table(),
            'df_mtcars' = mtcars |> as.data.table())

# test create bkp -------------------------------------------------------------
test_that('Test data bkp - create bkp', {
  testServer(data_bkp_server, {

    session$userData$dt_names <- reactive('iris')

    session$userData$dt <- reactiveValues(
      dt = list('iris' = dfs[['df_iris']]),
      act_name = 'iris'
    )

    expect_null(session$userData$dt$bkp)
    session$setInputs(btn_bkp = 1)

    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      session$userData$dt$bkp
    )
  })
})

# test clear bkp --------------------------------------------------------------
test_that('Test data bkp - clear bkp', {
  testServer(data_bkp_server, {

    session$userData$dt_names <- reactive('iris')
    session$userData$dt <- reactiveValues(
      dt = list('iris' = dfs[['df_iris']]),
      act_name = 'iris'
    )

    session$setInputs(btn_bkp = 1)
    expect_true(!is.null(session$userData$dt$bkp))

    session$setInputs(btn_clear_bkp = 1)

    expect_null(session$userData$dt$bkp)
  })
})

# test restore bkp --------------------------------------------------------------
test_that('Test data bkp - restore bkp', {
  testServer(data_bkp_server, {

    session$userData$data_changed <- reactiveVal(0)
    session$userData$dt_names <- reactive('iris')
    session$userData$dt <- reactiveValues(
      dt = list('iris' = dfs[['df_iris']]),
      act_name = 'iris'
    )

    session$setInputs(btn_bkp = 1)

    session$userData$dt$dt[['iris']] <- iris |> subset(select = 'Species')

    session$setInputs(btn_restore = 1)

    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      iris |> as.data.table()
    )
  })
})

# test reset bkp --------------------------------------------------------------
test_that('Test data bkp - reset bkp', {
  testServer(data_bkp_server, {

    session$userData$data_changed <- reactiveVal(0)
    session$userData$dt_names <- reactive('iris')

    session$userData$dt <- reactiveValues(
      dt = list('iris' = dfs[['df_iris']]),
      act_name = 'iris',
      bkp0 = dfs[['df_iris']]
    )

    session$setInputs(btn_bkp = 1)

    session$userData$dt$dt[['iris']] <- iris |> subset(select = 'Species')

    session$setInputs(btn_restore = 1)

    session$userData$dt$dt[['iris']] <- iris |> subset(select = 'Petal.Length')

    session$setInputs(btn_reset = 1)

    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      iris |> as.data.table()
    )
  })
})

# test messages ---------------------------------------------------------------
test_that('Test data bkp - check messages', {
  testServer(data_bkp_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg_error = function(text, ...) { last_msg <<- text },
      msg = function(text, ...) { last_msg <<- text }
    )

    session$userData$dt <- reactiveValues(
      bkp = NULL
    )

    session$setInputs(btn_restore = 1)
    expect_equal(last_msg, 'No backup to restore')

    session$setInputs(btn_clear_bkp = 1)
    expect_equal(last_msg, 'No backup to clear')
  })
})

# tests/testthat/test-frequencies_module.R

# test frequencies - table ----------------------------------------------------
test_that('Frequencies', {
  testServer(frequencies_server,
             args = list(reactive(iris$Species), reactive('Species')), {

    session$setInputs(btn_freq_table = 1)

    expect_equal(var(), iris$Species)

    tab <- Freq(iris$Species, useNA = 'always') |> as.data.frame()
    attr(tab, 'title') <- 'Species'

    while(task_freq_table$status() == 'running'){
      session$flushReact()
    }

    expect_equal(freq_table(), tab)
    expect_s3_class(freq_table(), 'data.frame')
    expect_s3_class(freq_table_gt(), 'gt_tbl')
  })
})

# tests/testthat/test-config_module.R

# test colors -----------------------------------------------------------------
test_that('Test chose colors', {
  testServer(config_server, args = list(app_session = NULL), {

    session$setInputs(sel_fill = '#aaccff')
    session$setInputs(sel_line = '#ccee55')

    expect_equal(palette()[[1]], '#aaccff')
    expect_equal(palette()[[2]], '#ccee55')
  })
})

# test input file size --------------------------------------------------------
test_that('Test input file size', {
  testServer(config_server, args = list(app_session = NULL),  {

    last_msg <- NULL

    local_mocked_bindings(
      msg_error = function(text, ...) { last_msg <<- text },
      msg = function(text, ...) { last_msg <<- text }
    )

    session$userData$conf$conf_dir <- reactiveValues(conf_dir = tempdir())

    session$setInputs(input_file_size = 1, btn_file_size = 1)

    expect_equal(max_request_size(), 1 * 1024 ^ 2)
    expect_equal(last_msg, 'New limit applied')
  })
})

# test theme choice -----------------------------------------------------------
test_that('Test theme choice', {
  testServer(config_server, args = list(app_session = MockShinySession$new()), {
    session$userData$conf <- reactiveValues(theme = 'spada_theme')

    session$setInputs(theme_choice = 'spada_dark_theme', btn_theme = 1)

    expect_equal(session$userData$conf$theme, 'spada_dark_theme')

    session$setInputs(theme_choice = 'spada_theme', btn_theme = 2)

    expect_equal(session$userData$conf$theme, 'spada_theme')
  })
})

# test restore session settings -----------------------------------------------
test_that('Test restore session settings', {
  testServer(config_server, args = list(app_session = NULL),  {

    last_msg <- NULL

    local_mocked_bindings(
      msg_error = function(text, ...) { last_msg <<- text },
      msg = function(text, ...) { last_msg <<- text }
    )

    session$setInputs(radio_restore_session = 'always',
                      radio_save_session = 'ask',
                      btn_save_session_conf = 1)

    expect_equal(session$userData$conf$restore_session, 'always')
    expect_equal(session$userData$conf$save_session, 'ask')
    expect_equal(last_msg, 'New settings applied')

  })
})

# test plot limit -------------------------------------------------------------
test_that('Test plot limit', {
  testServer(config_server, args = list(app_session = NULL),  {

    last_msg <- NULL

    local_mocked_bindings(
      msg_error = function(text, ...) { last_msg <<- text },
      msg = function(text, ...) { last_msg <<- text }
    )

    session$setInputs(input_plot_limit = 100, btn_plot_limit = 1)

    expect_equal(session$userData$conf$plot_limit, 100e3)
    expect_equal(last_msg, 'New limit applied')

    session$setInputs(input_plot_limit = -100, btn_plot_limit = 2)
    expect_equal(last_msg, 'Value must be > 0')
  })
})

# tests/testthat/test-save_gt_module.R

# test save gt - save ---------------------------------------------------------
test_that('Test save gt - html', {

  testServer(save_gt_server, args = list(input_table = reactive(iris |> gt())), {

    session$setInputs(
      file_name = 'gt_iris',
      radio_format = 'html'
    )

    expect_equal(output$down_handler |> basename(), 'gt_iris.html')

  })
})

test_that('Test save gt - rtf', {

  testServer(save_gt_server, args = list(input_table = reactive(iris |> gt())), {

    session$setInputs(
      file_name = 'gt_iris',
      radio_format = 'rtf'
    )

    expect_equal(output$down_handler |> basename(), 'gt_iris.rtf')

  })
})

# test save gt - check input_table --------------------------------------------
test_that('Test save gt - html', {

  testServer(save_gt_server, args = list(input_table = reactive(iris |> gt())), {

    session$setInputs(
      file_name = 'gt_iris',
      radio_format = 'html'
    )

    expect_true(input_table |> is.reactive())
    expect_equal(input_table() |> class(), c('gt_tbl', 'list'))

  })
})

# tests/testthat/test-calculate_cols_module.R

iris_dt <- iris |> as.data.table()

# test calculate cols - apply function ----------------------------------------
test_that('Test calculate cols - sum numeric', {
  testServer(calculate_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg = function(text, ...) { last_msg <<- text },
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris_dt),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      vars_sel = 'Sepal.Width',
      fun = 'sum_na',
      txt_new_name_fun = 'new_var',
      vars_groupby_fun = NULL
      )

    session$setInputs(btn_apply_fun = 1)

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]]$new_var[1],
                 iris$Sepal.Width |> sum())
    expect_equal(last_msg, 'Remove modal')

    # test no selection after first operation -----
    session$setInputs(vars_sel = character(0),
                      btn_apply_fun = 2)

    expect_equal(last_msg, 'Select a variable')
  })
})

test_that('Test calculate cols - mean numeric - groupby', {
  testServer(calculate_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris_dt),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      vars_sel = 'Sepal.Width',
      fun = 'mean_na',
      txt_new_name_fun = 'new_var',
      vars_groupby_fun = 'Species'
    )

    session$setInputs(btn_apply_fun = 1)

    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      copy(iris_dt)[, new_var := mean(Sepal.Width), by = 'Species']
    )
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test calculate cols - factor', {
  testServer(calculate_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris_dt),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      vars_sel = 'Species',
      fun = 'is.factor',
      txt_new_name_fun = 'new_var',
      vars_groupby_fun = NULL
    )

    session$setInputs(btn_apply_fun = 1)

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 copy(iris_dt)[, new_var := is.factor(Species)])
    expect_equal(last_msg, 'Remove modal')
  })
})

# test calculate cols - freehand ----------------------------------------------
test_that('Test calculate cols - freehand', {
  testServer(calculate_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris_dt),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      txt_new_name_free = 'new_var',
      vars_groupby_free = NULL,
      txt_code_input = '10'
    )

    session$setInputs(btn_apply_free = 1)

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 copy(iris_dt)[, new_var := 10 ])
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test calculate cols - freehand - groupby', {
  testServer(calculate_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris_dt),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      txt_new_name_free = 'new_var',
      vars_groupby_free = 'Species',
      txt_code_input = 'mean(Sepal.Length)'
    )

    session$setInputs(btn_apply_free = 1)

    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      copy(iris_dt)[, new_var := mean(Sepal.Length), by = 'Species' ]
    )
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test calculate cols - freehand - fifelse', {
  testServer(calculate_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris_dt),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      txt_new_name_free = 'new_var',
      vars_groupby_free = NULL,
      txt_code_input = "fifelse(Species == 'setosa', 1, 2)"
    )

    session$setInputs(btn_apply_free = 1)

    expect_equal(
      session$userData$dt$dt[[session$userData$dt$act_name]],
      copy(iris_dt)[, new_var := fifelse(Species == 'setosa', 1, 2) ]
    )
    expect_equal(last_msg, 'Remove modal')
  })
})

# test calculate cols - not allowed function ----------------------------------
test_that('Test calculate cols - not allowed function', {
  testServer(calculate_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg = function(text, ...) { last_msg <<- text }
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris_dt),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)

    session$setInputs(
      vars_sel = 'Sepal.Width',
      fun = 'not_allowed',
      txt_new_name_fun = 'new_var',
      vars_groupby_fun = NULL,
      btn_apply_fun = 1)

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]]$Sepal.Width,
                 iris$Sepal.Width)
    expect_equal(last_msg, 'Function is not allowed')

    # dangerous function
    session$setInputs(fun = 'assign', btn_apply_fun = 2)

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]]$Sepal.Width,
                 iris$Sepal.Width)
    expect_equal(last_msg, 'Function is not allowed')

  })
})

# test check messages ---------------------------------------------------------
test_that('Calculate cols - apply fun - check inputs', {
  testServer(calculate_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg_error = function(text, ...) { last_msg <<- text },
      msg = function(text, ...) { last_msg <<- text },
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris_dt),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)

    # vars_sel NULL
    session$setInputs(vars_sel = NULL, btn_apply_fun = 1)
    expect_equal(last_msg, 'Select a variable')

    # fun NULL
    session$setInputs(vars_sel = 'Species', fun = NULL, btn_apply_fun = 2)
    expect_equal(last_msg, 'Select a function')

    # dangerous function
    session$setInputs(vars_sel = 'Species', fun = 'unlink', btn_apply_fun = 3)
    expect_equal(last_msg, 'Function is not allowed')

    # new var name invalid
    session$setInputs(vars_sel = 'Species', fun = 'mean_na',
                      txt_new_name_fun = 123, btn_apply_fun = 4)
    expect_equal(last_msg, 'New name is not valid or already in use')

    # new var name already in use
    session$setInputs(vars_sel = 'Species', fun = 'mean_na',
                      txt_new_name_fun = 'Species', btn_apply_fun = 5)
    expect_equal(last_msg, 'New name is not valid or already in use')

    # apply function ok
    session$setInputs(vars_sel = 'Species', fun = 'is.factor',
                      txt_new_name_fun = 'Species2', btn_apply_fun = 6)
    expect_equal(last_msg, 'Remove modal')

  })
})

test_that('Calculate cols - free hand - check inputs', {
  testServer(calculate_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg_error = function(text, ...) { last_msg <<- text },
      msg = function(text, ...) { last_msg <<- text },
      remove_running_modal = function(){last_msg <<- 'Remove modal'}
    )

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris_dt),
      act_name = 'iris'
    )

    session$userData$data_changed <- reactiveVal(0)

    # error parse code
    session$setInputs(txt_code_input = '///', txt_new_name_free = 'Species2',
                      btn_apply_free = 1)
    expect_equal(last_msg, 'Error to validate expression. Check code')

    # not allowed
    session$setInputs(txt_code_input = 'unlink(x)', txt_new_name_free = 'Species2',
                      btn_apply_free = 2)
    expect_equal(last_msg, 'Some operations are not allowed')


    # varible not present
    session$setInputs(txt_code_input = 'mean(var1)',
                      txt_new_name_free = 'Species2',
                      btn_apply_free = 3)
    expect_equal(last_msg, 'Some variables are not present in the dataset')

    # new var name already in use
    session$setInputs(txt_code_input = 'mean(Sepal.Length)',
                      txt_new_name_free = 'Species_mean 123',
                      vars_groupby_free = 'Species',
                      btn_apply_free = 4)
    expect_equal(last_msg, 'New name is not valid or already in use')

    # error - log of factor
    session$setInputs(txt_code_input = 'log(Species)',
                      txt_new_name_free = 'Species_log',
                      btn_apply_free = 5)
    expect_equal(last_msg, 'Error in expression. Check code')

    # calculate ok
    session$setInputs(txt_code_input = 'mean(Sepal.Length)',
                      txt_new_name_free = 'Species_mean',
                      vars_groupby_free = 'Species',
                      btn_apply_free = 6)
    expect_equal(last_msg, 'Remove modal')

  })
})

# tests/testthat/test-data_overview_module.R

# test validate number of rows ------------------------------------------------
test_that("Test sample rows", {
  testServer(data_overview_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = as.data.table(iris), 'mtcars' = as.data.table(mtcars)),
      act_name = 'iris'
    )

    session$userData$dt_names <- reactive({
      names(session$userData$dt$dt)
    })

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$setInputs(sel_dataset = 'iris', size_sample = 0)

    expect_error(idx())

  })
})

# test first rows -------------------------------------------------------------
test_that("Test first rows", {
  testServer(data_overview_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = as.data.table(iris), 'mtcars' = as.data.table(mtcars)),
      act_name = 'iris'
    )

    session$userData$dt_names <- reactive({
      names(session$userData$dt$dt)
    })

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$setInputs(size_sample = 5,
                      sel_dataset = 'iris',
                      radio_sample = 'first')

    expect_equal(data_filtered(), iris[1:5, ] |> as.data.table())

  })
})

# test last rows -------------------------------------------------------------
test_that("Test last rows", {
  testServer(data_overview_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = as.data.table(iris), 'mtcars' = as.data.table(mtcars)),
      act_name = 'iris'
    )

    session$userData$dt_names <- reactive({
      names(session$userData$dt$dt)
    })

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$setInputs(size_sample = 3,
                      sel_dataset = 'mtcars',
                      radio_sample = 'last')

    expect_equal(data_filtered(), mtcars[30:32, ], ignore_attr = TRUE)

  })
})

# test sample of rows ---------------------------------------------------------
test_that("Test sample rows", {
  testServer(data_overview_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = as.data.table(iris), 'mtcars' = as.data.table(mtcars)),
      act_name = 'iris'
    )

    session$userData$dt_names <- reactive({
      names(session$userData$dt$dt)
    })

    session$userData$dt$df_info <- reactive({
      req(session$userData$dt)

      lapply(session$userData$dt$dt, df_info)
    })

    session$setInputs(size_sample = 5,
                      sel_dataset = 'iris',
                      radio_sample = 'first')

    # test for equality (unlist to avoid error caused by row.names)
    expect_equal(data_filtered() |> unlist(), iris[idx(),] |> unlist())
  })
})

# tests/testthat/test-export_file_module.R

# test export csv -------------------------------------------------------------
test_that('Test export csv - test name', {

  testServer(export_file_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris),
      act_name = 'iris'
    )

    session$setInputs(
      file_name = 'iris_export',
      radio_format = 'csv',
      x_rownames = FALSE,
      radio_separator = ',',
      radio_decimal = '.',
      txt_na = 'NA',
      radio_scientific = 0
    )

    expect_equal(output$down_handler |> basename(), 'iris_export.csv')

  })
})

# test export RDS -------------------------------------------------------------
test_that('Test export RDS - test name', {

  testServer(export_file_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris),
      act_name = 'iris'
    )

    session$setInputs(
      file_name = 'iris_export',
      radio_format = 'RDS',
      checkbox_rds_compress = F
    )

    expect_equal(output$down_handler |> basename(), 'iris_export.RDS')

  })
})

# test export sav -------------------------------------------------------------
test_that('Test export sav - test name', {

  testServer(export_file_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris),
      act_name = 'iris'
    )

    session$setInputs(
      file_name = 'iris_export',
      radio_format = 'sav',
      radio_sav_compress = 'none'
    )

    expect_equal(output$down_handler |> basename(), 'iris_export.sav')

  })
})

# test export qs2 -------------------------------------------------------------
test_that('Test export qs2 - test name', {

  testServer(export_file_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris),
      act_name = 'iris'
    )

    session$setInputs(
      file_name = 'iris_export',
      radio_format = 'qs2',
      radio_sav_compress = 'none'
    )

    expect_equal(output$down_handler |> basename(), 'iris_export.qs2')

  })
})

# test export xlsx ------------------------------------------------------------
test_that('Test export xlsx - test name', {
  testServer(export_file_server, {

    session$userData$dt <- reactiveValues(
      dt = list('iris' = iris),
      act_name = 'iris'
    )

    session$setInputs(
      file_name = 'iris_export',
      radio_format = 'xlsx',
      radio_sav_compress = 'none'
    )

    expect_equal(output$down_handler |> basename(), 'iris_export.xlsx')

  })
})

test_that('Test export xlsx - test limit', {

  testServer(export_file_server, {

    session$userData$dt <- reactiveValues(
      dt = list('df' = rep(1, 16385) |> as.list() |> as.data.table()),
      act_name = 'df'
    )

    session$setInputs(
      file_name = 'df_export',
      radio_format = 'xlsx',
      radio_sav_compress = 'none'
    )

    expect_equal(output$down_handler |> basename(), 'df_export.xlsx')

  })
})

# tests/testthat/test-import_file_module.R

# test import RDS -------------------------------------------------------------
test_that('Test import RDS', {
  testServer(import_file_server, {

    tmp <- tempfile(fileext = ".RDS")
    saveRDS(iris, tmp)

    session$userData$dt_names <- reactive('mtcars')
    session$userData$dt$dt[['mtcars']] <- mtcars
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(
      dataset_name = 'df_test',
      radio_file_ext = 'RDS',
      file = list(datapath = tmp)
    )

    session$setInputs(btn_import = 1)

    expect_null(data$data)
    expect_equal(session$userData$dt$dt[['df_test']], iris |> as.data.table())
  })
})

# test import sav -------------------------------------------------------------
test_that('Test import Sav - all lines', {
  testServer(import_file_server, {

    tmp <- tempfile(fileext = ".sav")
    haven::write_sav(iris, tmp)

    session$userData$dt_names <- reactive('mtcars')
    session$userData$dt$dt[['mtcars']] <- mtcars
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(
      dataset_name = 'df_test',
      radio_file_ext = 'sav',
      file = list(datapath = tmp),
      sav_lines = 'all'
    )

    session$setInputs(btn_import = 1)

    # import data to test
    data_temp_test <- read_sav(tmp, .name_repair = make.names) |>
      as.data.table()
    data_temp_test[ , names(.SD) := lapply(.SD, as_factor), .SDcols = is.labelled]

    expect_null(data$data)
    expect_equal(session$userData$dt$dt[[input$dataset_name]], data_temp_test)
  })
})

test_that('Test import Sav - 10 lines', {
  testServer(import_file_server, {

    tmp <- tempfile(fileext = ".sav")
    haven::write_sav(iris, tmp)

    session$userData$dt_names <- reactive('mtcars')
    session$userData$dt$dt[['mtcars']] <- mtcars
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(
      dataset_name = 'df_test',
      radio_file_ext = 'sav',
      file = list(datapath = tmp),
      sav_lines = 'some',
      sav_n_lines = '10'
    )

    session$setInputs(btn_import = 1)

    # import data to test
    data_temp_test <- read_sav(tmp, n_max = 10, .name_repair = make.names) |>
      as.data.table()
    data_temp_test[ , names(.SD) := lapply(.SD, as_factor), .SDcols = is.labelled]

    expect_null(data$data)
    expect_equal(session$userData$dt$dt[[input$dataset_name]], data_temp_test)
  })
})

# test import csv -------------------------------------------------------------
test_that('Test import all lines - csv standard', {
  testServer(import_file_server, {

    tmp <- tempfile(fileext = ".csv")
    write.table(iris, tmp, sep = ',', dec = '.', row.names = F)

    session$userData$dt_names <- reactive('mtcars')
    session$userData$dt$dt[['mtcars']] <- mtcars
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(
      dataset_name = 'df_test',
      radio_file_ext = 'csv',
      file = list(datapath = tmp),
      radio_separator = ',',
      radio_decimal = '.',
      csv_lines = 'all',
      x_csv_header = TRUE
    )

    session$setInputs(btn_import = 1)

    data_temp <- fread(
        file = tmp,
        sep = ',',
        dec = '.',
        check.names = T,
        nrows = Inf,
        skip = 0,
        header = TRUE)

    expect_null(data$data)
    expect_equal(session$userData$dt$dt[[input$dataset_name]], data_temp)
  })
})

test_that('Test import all lines - csv 2 standard', {
  testServer(import_file_server, {

    tmp <- tempfile(fileext = ".csv")
    write.table(iris, tmp, sep = ';', dec = ',', row.names = F)

    session$userData$dt_names <- reactive('mtcars')
    session$userData$dt$dt[['mtcars']] <- mtcars
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(
      dataset_name = 'df_test',
      radio_file_ext = 'csv',
      file = list(datapath = tmp),
      radio_separator = ';',
      radio_decimal = ',',
      csv_lines = 'all',
      x_csv_header = TRUE
    )

    session$setInputs(btn_import = 1)

    data_temp <- fread(
      file = tmp,
      sep = ';',
      dec = ',',
      check.names = T,
      nrows = Inf,
      skip = 0,
      header = TRUE)

    expect_null(data$data)
    expect_equal(session$userData$dt$dt[[input$dataset_name]], data_temp)
  })
})

# tests/testthat/test-order_cols_module.R

# test order columns - before -------------------------------------------------
test_that('Test order one column - before all', {
  testServer(order_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('wt')
    var1 <- names(mtcars)[1]
    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_cols = vars)
    session$setInputs(vars_rest = var1)
    session$setInputs(radio_cols = 'before')
    session$setInputs(btn_order_cols = 1)

    mtcars_reordered <- mtcars |>
      dplyr::relocate(tidyselect::all_of(vars), .before = tidyselect::all_of(var1))

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 mtcars_reordered |> as.data.table())

    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test order two column - before all', {
  testServer(order_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('wt', 'hp')
    var1 <- names(mtcars)[1]
    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_cols = vars)
    session$setInputs(vars_rest = var1)
    session$setInputs(radio_cols = 'before')
    session$setInputs(btn_order_cols = 1)

    mtcars_reordered <- mtcars |>
      dplyr::relocate(tidyselect::all_of(vars), .before = tidyselect::all_of(var1))

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 mtcars_reordered |> as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test order one column - before especific', {
  testServer(order_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('wt')
    var1 <- names(mtcars)[4]
    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_cols = vars)
    session$setInputs(vars_rest = var1)
    session$setInputs(radio_cols = 'before')
    session$setInputs(btn_order_cols = 1)

    mtcars_reordered <- mtcars |>
      dplyr::relocate(tidyselect::all_of(vars), .before = tidyselect::all_of(var1))

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 mtcars_reordered |> as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test order two columns - before especific', {
  testServer(order_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('wt','mpg')
    var1 <- names(mtcars)[4]
    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_cols = vars)
    session$setInputs(vars_rest = var1)
    session$setInputs(radio_cols = 'before')
    session$setInputs(btn_order_cols = 1)

    mtcars_reordered <- mtcars |>
      dplyr::relocate(tidyselect::all_of(vars), .before = tidyselect::all_of(var1))

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 mtcars_reordered |> as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

# test order columns - after -------------------------------------------------
test_that('Test order one column - after all', {
  testServer(order_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('wt')
    var1 <- names(mtcars)[length(mtcars)]
    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_cols = vars)
    session$setInputs(vars_rest = var1)
    session$setInputs(radio_cols = 'after')
    session$setInputs(btn_order_cols = 1)

    mtcars_reordered <- mtcars |>
      dplyr::relocate(tidyselect::all_of(vars), .after = tidyselect::all_of(var1))

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 mtcars_reordered |> as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test order two columns - after all', {
  testServer(order_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('wt', 'drat')
    var1 <- names(mtcars)[length(mtcars)]
    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_cols = vars)
    session$setInputs(vars_rest = var1)
    session$setInputs(radio_cols = 'after')
    session$setInputs(btn_order_cols = 1)

    mtcars_reordered <- mtcars |>
      dplyr::relocate(tidyselect::all_of(vars), .after = tidyselect::all_of(var1))

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 mtcars_reordered |> as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test order one column - after specific', {
  testServer(order_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('am')
    var1 <- names(mtcars)[8]
    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_cols = vars)
    session$setInputs(vars_rest = var1)
    session$setInputs(radio_cols = 'after')
    session$setInputs(btn_order_cols = 1)

    mtcars_reordered <- mtcars |>
      dplyr::relocate(tidyselect::all_of(vars), .after = tidyselect::all_of(var1))

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 mtcars_reordered |> as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test order two columns - after specific', {
  testServer(order_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('am', 'carb')
    var1 <- names(mtcars)[2]
    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars |> as.data.table()),
      act_name = 'mtcars'
    )
    session$userData$data_changed <- reactiveVal(0)
    session$setInputs(vars_cols = vars)
    session$setInputs(vars_rest = var1)
    session$setInputs(radio_cols = 'after')
    session$setInputs(btn_order_cols = 1)

    mtcars_reordered <- mtcars |>
      dplyr::relocate(tidyselect::all_of(vars), .after = tidyselect::all_of(var1))

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]],
                 mtcars_reordered |> as.data.table())
    expect_equal(last_msg, 'Remove modal')
  })
})

# test input check messages ---------------------------------------------------
test_that('Test order cols - check inputs', {
  testServer(order_cols_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg = function(text, ...) { last_msg <<- text }
    )

    session$userData$dt <- reactiveValues(
      dt = list('mtcars' = mtcars |> as.data.table()),
      act_name = 'mtcars'
    )

    session$setInputs(vars_cols = NULL, btn_order_cols = 1)

    expect_equal(last_msg, 'Select at least one variable')

  })
})

# tests/testthat/test-table_values_module.R

# test table values module - tables -------------------------------------------
test_that('Tables values - 1 variable', {
  testServer(table_values_server,
             args = list(reactive(iris$Species),
                         reactive(iris$Petal.Width),
                         reactive('Species'),
                         reactive('Petal.Width')
                         ), {

    session$setInputs(
      table_var = '1v',
      table_type = 'abs_table'
    )

    expect_equal(var(), iris$Species)
    expect_equal(table_values(), table(iris$Species) |> as.data.frame())
    expect_s3_class(table_values(), 'data.frame')
    expect_s3_class(table_values_gt(), 'gt_tbl')
  })
})

test_that('Exploratory module - tables - 2 variables', {
  testServer(table_values_server,
             args = list(reactive(mtcars$cyl |> as.factor()),
                         reactive(mtcars$am |> as.factor()),
                         reactive('cyl'),
                         reactive('am')
             ), {

    session$setInputs(
      table_var = '2v',
      table_type = 'abs_table'
    )

    tab1 <- table(mtcars$cyl |> as.factor(),
                  mtcars$am |> as.factor())
    tab1 <- tab1 |> as.data.frame.matrix()
    tab1 <- cbind(Var1 = rownames(tab1), tab1)

    expect_equal(var(), mtcars$cyl |> as.integer() |> as.factor())
    expect_equal(table_values(), tab1)
    expect_s3_class(table_values(), 'data.frame')
    expect_s3_class(table_values_gt(), 'gt_tbl')
  })
})

# tests/testthat/test-order_rows_module.R

df_test <- iris |> as.data.table()
df_test[150, 1:5] <- NA

# test order rows - na last ---------------------------------------------------
test_that('Test order rows - na last - one variable ascending', {
  testServer(order_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('Species')

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test),
      act_name = 'df_test'
    )

    session$setInputs(vars_rows = vars)
    session$setInputs(radio_nas = 'last')

    session$setInputs(btn_order_rows = 1)

    df_reordered <- df_test[order(df_test[[vars]]), ]

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]], df_reordered)
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test order rows - na last - one variable descending', {
  testServer(order_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('Species')

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test),
      act_name = 'df_test'
    )

    session$setInputs(vars_rows = vars)
    session$setInputs(radio_nas = 'last')
    session$setInputs(vars_descending = vars)
    session$setInputs(btn_order_rows = 1)

    df_reordered <- df_test[order(df_test[[vars]], decreasing = TRUE), ]

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]], df_reordered)
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test order rows - na last - two variables ascending', {
  testServer(order_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('Sepal.Length', 'Sepal.Width')

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test),
      act_name = 'df_test'
    )

    session$setInputs(vars_rows = vars)
    session$setInputs(radio_nas = 'last')

    session$setInputs(btn_order_rows = 1)

    df_reordered <- setorderv(df_test, vars, na.last = TRUE)

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]], df_reordered)
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test order rows - na last - two variables descending', {
  testServer(order_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('Sepal.Length', 'Sepal.Width')

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test),
      act_name = 'df_test'
    )

    session$setInputs(vars_rows = vars)
    session$setInputs(radio_nas = 'last')

    session$setInputs(btn_order_rows = 1)

    df_reordered <- setorderv(df_test, vars, c(-1, -1), na.last = TRUE)

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]], df_reordered)
    expect_equal(last_msg, 'Remove modal')
  })
})

# test order rows - na first --------------------------------------------------
test_that('Test order rows - na first - one variable ascending', {
  testServer(order_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('Species')

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test),
      act_name = 'df_test'
    )

    session$setInputs(vars_rows = vars)
    session$setInputs(radio_nas = 'first')

    session$setInputs(btn_order_rows = 1)

    df_reordered <- setorderv(df_test, vars, na.last = FALSE)

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]], df_reordered)
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test order rows - na first - one variable descending', {
  testServer(order_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('Species')

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test),
      act_name = 'df_test'
    )

    session$setInputs(vars_rows = vars)
    session$setInputs(radio_nas = 'first')
    session$setInputs(vars_descending = vars)
    session$setInputs(btn_order_rows = 1)

    df_reordered <- setorderv(df_test, vars, -1, na.last = FALSE)

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]], df_reordered)
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test order rows - na first - two variables ascending', {
  testServer(order_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('Sepal.Length', 'Petal.Width')

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test),
      act_name = 'df_test'
    )

    session$setInputs(vars_rows = vars)
    session$setInputs(radio_nas = 'first')

    session$setInputs(btn_order_rows = 1)

    df_reordered <- setorderv(df_test, vars, na.last = FALSE)

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]], df_reordered)
    expect_equal(last_msg, 'Remove modal')
  })
})

test_that('Test order rows - na first - two variables descending', {
  testServer(order_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      remove_running_modal = function() { last_msg <<- 'Remove modal' }
    )

    vars <- c('Sepal.Length', 'Sepal.Width')

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test),
      act_name = 'df_test'
    )

    session$setInputs(vars_rows = vars)
    session$setInputs(radio_nas = 'first')
    session$setInputs(vars_descending = vars)
    session$setInputs(btn_order_rows = 1)

    df_reordered <- setorderv(df_test, vars, c(-1, -1), na.last = FALSE)

    expect_equal(session$userData$dt$dt[[session$userData$dt$act_name]], df_reordered)
    expect_equal(last_msg, 'Remove modal')
  })
})

# test input check messages ---------------------------------------------------
test_that('Test order rows - check inputs', {
  testServer(order_rows_server, {

    last_msg <- NULL

    local_mocked_bindings(
      msg_error = function(text, ...) { last_msg <<- text },
      msg = function(text, ...) { last_msg <<- text }
    )

    session$userData$dt <- reactiveValues(
      dt = list('df_test' = df_test),
      act_name = 'df_test'
    )

    session$setInputs(vars_rows = NULL, btn_order_rows = 1)

    expect_equal(last_msg, 'Select at least one variable')
  })
})

---
title: "Intro"
date: "`r format(Sys.Date(), '%d/%m/%Y')`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Intro}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Datasets

The basic object in Spada is a dataset. A dataset is an object of the data.frame class. Internally, for speed and programmability purposes, Spada uses the data.table class from the [**data.table**](https://rdatatable.gitlab.io/data.table/) package.

Spada demands at least one dataset. If none is provided in the funcion call then Iris and Mtcars, from datasets package, are used as sample data.

```{r, eval = FALSE}
library(spada)
spada()
```

For multiples datasets as input:

```{r, eval = FALSE}
spada(iris, mtcars, cars)
```

### Dataset Names

If the user desires to give names to datasets, then the Spada function should be called like this:

```{r, eval = FALSE}
spada('dataset' = iris, 'test_data' = mtcars)
```

Spada will make valid and unique names with **make.names** function. So:

```{r, eval = FALSE}
spada('dataset 1' = iris, 'test data' = mtcars, 'dataset 1' = cars)
```

Will result in names: dataset.1, test.data and dataset.1.1.

If the user inform at least one name, Spada will assume that the datasets should receive a name (instead the name of the original object) and **will create valid and unique names** using dataset_n, being n a sequencial number:

```{r, eval = FALSE}
spada('dataset 1' = iris, mtcars, cars)
```

Will result in names: dataset.1, dataset_2 and dataset_3.

### Active Dataset

The dataset in which the operations will be applied are the **Active dataset**. One can think of this dataset as the current object. As Spada requires at least one dataset, won't be possible to delete the Active dataset.

## Navbar

Spada uses a page_navbar from bslib package. In the left side are the menus that lead to the app pages. In the right side are the **Active dataset** name.

![](../man/figures/articles/spada/navbar_active_dataset.png)

## Navbar Menus

### Data

This is the default selected item.

#### Highlights

![](../man/figures/articles/spada/data_menu/highlights.png)

#### Metadata

![](../man/figures/articles/spada/data_menu/metadata.png)

#### Overview

![](../man/figures/articles/spada/data_menu/overview.png)

#### Data

![](../man/figures/articles/spada/data_menu/data.png)

#### Import

![](../man/figures/articles/spada/data_menu/import.png)

#### Export

![](../man/figures/articles/spada/data_menu/export.png)

### Edit

![](../man/figures/articles/spada/spada_edit.png)

### Analysis

![](../man/figures/articles/spada/spada_analysis_exploratory.png)

### Output

![](../man/figures/articles/spada/spada_output.png)

### Config

![](../man/figures/articles/spada/spada_config.png)

---
title: "Intro"
date: "`r format(Sys.Date(), '%d/%m/%Y')`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Intro}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Datasets

The basic object in Spada is a dataset. A dataset is an object of the data.frame class. Internally, for speed and programmability purposes, Spada uses the data.table class from the [**data.table**](https://rdatatable.gitlab.io/data.table/) package.

Spada demands at least one dataset. If none is provided in the funcion call then Iris and Mtcars, from datasets package, are used as sample data.

```{r, eval = FALSE}
library(spada)
spada()
```

For multiples datasets as input:

```{r, eval = FALSE}
spada(iris, mtcars, cars)
```

### Dataset Names

If the user desires to give names to datasets, then the Spada function should be called like this:

```{r, eval = FALSE}
spada('dataset' = iris, 'test_data' = mtcars)
```

Spada will make valid and unique names with **make.names** function. So:

```{r, eval = FALSE}
spada('dataset 1' = iris, 'test data' = mtcars, 'dataset 1' = cars)
```

Will result in names: dataset.1, test.data and dataset.1.1.

If the user inform at least one name, Spada will assume that the datasets should receive a name (instead the name of the original object) and **will create valid and unique names** using dataset_n, being n a sequencial number:

```{r, eval = FALSE}
spada('dataset 1' = iris, mtcars, cars)
```

Will result in names: dataset.1, dataset_2 and dataset_3.

### Active Dataset

The dataset in which the operations will be applied are the **Active dataset**. One can think of this dataset as the current object. As Spada requires at least one dataset, won't be possible to delete the Active dataset.

## Navbar

Spada uses a page_navbar from bslib package. In the left side are the menus that lead to the app pages. In the right side are the **Active dataset** name.

![Navbar](../man/figures/navbar_active_dataset.png)

## Navbar Menus

### Data

This is the default selected item.

#### Highlights

![Highlights](../man/figures/highlights.png)

#### Metadata

![Metadata](../man/figures/metadata.png)

#### Overview

![Overview](../man/figures/overview.png)

#### Data

![Data](../man/figures/data.png)

#### Import

![Import](../man/figures/import.png)

#### Export

![Export](../man/figures/export.png)

### Edit

![Edit](../man/figures/spada_edit.png)

### Analysis

![Analysis](../man/figures/spada_analysis_exploratory.png)

### Output

![Output](../man/figures/spada_output.png)

### Config

![Config](../man/figures/spada_config.png)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spada.R
\name{spada}
\alias{spada}
\title{Launch a 'shiny' application for data analysis}
\usage{
spada(..., run_local = TRUE)
}
\arguments{
\item{...}{Objects of data.frame class}

\item{run_local}{Logical. Whether to run the application locally. If TRUE the app will close at session end.}
}
\value{
An object of class 'shiny.appobj' representing the 'shiny' application.
  Printing the object launches the interactive app in a web browser.
}
\description{
Generates a 'shiny' application for interactive data analysis.
}
\examples{
if(interactive()) spada(datasets::mtcars)

}
